/* 8051 emulation module
   Copyright (C) 2002, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>

#include <cereal.h>
#include <disasm8051.h>
#include <lib.h>
#include <module.h>

#include "8051.h"

/* Default 8051 frequency */
enum { I8051_DEFAULT_FREQUENCY = 11059200 };

/* Create a new module instance;
   returns: module instance pointer if OK, NULL on error */
static void *
i8051_mi_new (void)
{
  struct mi *mi;
  const struct sfr_fn_skel *src;
  size_t prev;

  { /* Reference i8051_disasm so that KC8051ViewWin can load and work */
    static const uint8_t nul; /* = 0; */
    const char *opcode;
    char buf[64];
    
    i8051_disasm (0, &nul, &opcode, buf);
  }
  mi = xmalloc (sizeof (*mi));
  memset (mi->sfr, 0, sizeof (mi->sfr));
  memset (mi->memory, 0, sizeof (mi->memory));
  mi->program = xmalloc (I8051_PROG_MEM_SIZE_DEFAULT);
  memset (mi->program, I8051_PROG_MEM_DEFAULT_VALUE,
	  I8051_PROG_MEM_SIZE_DEFAULT);
  mi->xmem = xmalloc (I8051_XMEM_MEM_SIZE_DEFAULT);
  memset (mi->xmem, I8051_XMEM_MEM_DEFAULT_VALUE, I8051_XMEM_MEM_SIZE_DEFAULT);
  mi->prog_mem_size = I8051_PROG_MEM_SIZE_DEFAULT;
  mi->data_mem_size = I8051_MEM_SIZE_MIN;
  mi->xmem_mem_size = I8051_XMEM_MEM_SIZE_DEFAULT;
  mi->pc = 0;
  mi->flags = 0;
  mi->frequency = I8051_DEFAULT_FREQUENCY;
  memset (mi->insn, 0, sizeof (mi->insn));
  mi->insn_next = mi->insn;
  mi->insn_special = 0;
  mi->prev_pc = 0;
  memset (mi->prev_insn, 0, sizeof (mi->prev_insn));
  mi->tmp1 = 0;
  mi->tmp2 = 0;
  mi->pins = PIN_PSEN; /* & ~PIN_ALE */
  memset (mi->ports_in, 0xFF, sizeof (mi->ports_in));
  memset (mi->ports_out, 0xFF, sizeof (mi->ports_out));
  mi->port_3_mask = 0xFF;
  mi->intr_pending = 0;
  mi->intr_new = 0;
  mi->prev_p3 = 0;
  mi->uart_rx_buf = 0;
  mi->uart_tx_buf = 0;
  mi->uart_rx_status = URX_NONE;
  mi->uart_tx_status = UTX_NONE;
  mi->uart_rx_step = 0;
  mi->uart_tx_step = 0;
  mi->s1p1_event = schedule_new (mi);
  mi->s3p1_event = schedule_new (mi);
  mi->s3p2_event = schedule_new (mi);
  mi->s4p1_event = schedule_new (mi);
  mi->s5p1_event = schedule_new (mi);
  mi->s5p2_event = schedule_new (mi);
  mi->s6p2_event = schedule_new (mi);
  mi->s14p1_event = schedule_new (mi);
  mi->s14p2_event = schedule_new (mi);
  mi->s25p1_event = schedule_new (mi);
  mi->s25p2_event = schedule_new (mi);
  mi->s36p1_event = schedule_new (mi);
  mi->uart_rx_event = schedule_new (mi);
  mi->uart_tx_event = schedule_new (mi);
  mi->s5p1_fn = NULL;
  memset (mi->sfr_fns, 0, sizeof (mi->sfr_fns));
  prev = 0;
  for (src = i8051_sfr_fns; src->reg < I8051_SFR_SIZE; src++)
    {
      struct sfr_fn *dest;

      if (src != i8051_sfr_fns)
	assert (prev < src->reg); /* Make sure the array is sorted */
      prev = src->reg;
      dest = mi->sfr_fns + src->reg;
#define DO_FN(TYPE)					\
      if (src->TYPE##_fn != NULL)			\
	{						\
	  dest->TYPE.fn.TYPE##_8 = src->TYPE##_fn;	\
	  dest->TYPE.data = src->reg;			\
	  dest->TYPE.mi = mi;				\
	}
      DO_FN (read);
      DO_FN (write);
      DO_FN (display);
      DO_FN (modify);
    }
  i8051_sfr_fns_len = src - i8051_sfr_fns;
  memset (mi->pin_fns, 0, sizeof (mi->pin_fns));
  mi->regs = mi->memory;
  mi->phase_len = 1.0 / I8051_DEFAULT_FREQUENCY;
  /* We start in S4P1, so that we can properly fetch insn from external
     memory. */
  schedule_add (mi->s14p1_event, 0, i8051_emulate_s14p1);
  schedule_add (mi->s14p2_event, mi->phase_len, i8051_emulate_s14p2);
  schedule_add (mi->s25p1_event, 2 * mi->phase_len, i8051_emulate_s25p1);
  schedule_add (mi->s25p2_event, 3 * mi->phase_len, i8051_emulate_s25p2);
  schedule_add (mi->s5p1_event, 2 * mi->phase_len, i8051_emulate_s5p1);
  schedule_add (mi->s5p2_event, 3 * mi->phase_len, i8051_emulate_s5p2);
  schedule_add (mi->s1p1_event, 6 * mi->phase_len, i8051_emulate_insn);
  schedule_add (mi->s3p1_event, 10 * mi->phase_len, i8051_emulate_s3p1);
  i8051_emulate_reset (mi);
  return mi;
}

/* Delete MI_ (may be in any emulation state) */
static void 
i8051_mi_delete (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  schedule_delete (mi->s1p1_event);
  schedule_delete (mi->s3p1_event);
  schedule_delete (mi->s3p2_event);
  schedule_delete (mi->s4p1_event);
  schedule_delete (mi->s5p1_event);
  schedule_delete (mi->s5p2_event);
  schedule_delete (mi->s6p2_event);
  schedule_delete (mi->s14p1_event);
  schedule_delete (mi->s14p2_event);
  schedule_delete (mi->s25p1_event);
  schedule_delete (mi->s25p2_event);
  schedule_delete (mi->s36p1_event);
  schedule_delete (mi->uart_rx_event);
  schedule_delete (mi->uart_tx_event);
  free (mi->xmem);
  free (mi->program);
  free (mi);
}

/* Set internal data memory size of MI to VALUE
   returns: 0 if OK, -1 on error */
inline static int
i8051_set_data_mem_size (struct mi *mi, const char *value)
{
  unsigned long val;

  if (get_num (value, 0, &val) != 0)
    return -1;
  /* Any other size than 128 and 256 doesn't make much sense, but I've read
     about a chip (something called 8[037]44) with 192 bytes. */
  assert (I8051_MEM_SIZE_MIN == 128 && I8051_MEM_SIZE_MAX == 256);
  if (val != I8051_MEM_SIZE_MIN && val != 192 && val != I8051_MEM_SIZE_MAX)
    {
      error (0, "option `data_mem_size' argument must be %d, 192 or %d",
	     I8051_MEM_SIZE_MIN, I8051_MEM_SIZE_MAX);
      return -1;
    }
  mi->data_mem_size = val;
  return 0;
}

/* Set CPU frequency of MI to VALUE
   returns: 0 if OK, -1 on error */
inline static int
i8051_set_frequency (struct mi *mi, const char *value)
{
  unsigned long val;
  
  if (get_num (value, 0, &val) != 0)
    return -1;
  if (val == 0)
    {
      error (0, "frequency cannot be 0");
      return -1;
    }
  mi->frequency = val;
  mi->phase_len = 1.0 / val;
  return 0;
}

/* Set internal program memory size of MI to VALUE;
   returns: 0 if OK, -1 on error */
inline static int
i8051_set_prog_mem_size (struct mi *mi, const char *value)
{
  unsigned long val;

  if (get_num (value, 0, &val) != 0)
    return -1;
  if (val < I8051_PROG_MEM_SIZE_MIN || val > I8051_PROG_MEM_SIZE_MAX
      || (val & (val - 1)) != 0)
    {
      error (0, "option `prog_mem_size' argument must be a power of two in "
	     "range [%d, %d]", I8051_PROG_MEM_SIZE_MIN,
	     I8051_PROG_MEM_SIZE_MAX);
      return -1;
    }
  mi->program = xrealloc (mi->program, val);
  if (mi->prog_mem_size < val)
    memset (mi->program + mi->prog_mem_size, I8051_PROG_MEM_DEFAULT_VALUE,
	    val - mi->prog_mem_size);
  mi->prog_mem_size = val;
  return 0;
}

/* Set "external" program memory size of MI to VALUE;
   returns: 0 if OK, -1 on error */
inline static int
i8051_set_xmem_mem_size (struct mi *mi, const char *value)
{
  unsigned long val;

  if (get_num (value, 0, &val) != 0)
    return -1;
  if (val > I8051_XMEM_MEM_SIZE_MAX || (val & (val - 1)) != 0)
    {
      error (0, "option `xmem_mem_size' argument must be a power of two less "
	     "than or equal to  %d", I8051_XMEM_MEM_SIZE_MAX);
      return -1;
    }
  mi->xmem = xrealloc (mi->xmem, val);
  if (mi->xmem_mem_size < val)
    memset (mi->xmem + mi->xmem_mem_size, I8051_XMEM_MEM_DEFAULT_VALUE,
	    val - mi->xmem_mem_size);
  mi->xmem_mem_size = val;
  return 0;
}

/* Get value of option NAME of MI_;
   returns: string for free () or NULL on error */
static char *
i8051_get_option (void *mi_, const char *name)
{
  struct mi *mi;

  mi = mi_;
  if (strcmp (name, "data_mem_size") == 0)
    {
      char *res;

      assert (sizeof (mi->data_mem_size) * CHAR_BIT <= 64);
      res = xmalloc (32); /* Overkill */
      sprintf (res, "%zu", mi->data_mem_size);
      return res;
    }
  else if (strcmp (name, "frequency") == 0)
    {
      char *res;

      assert (sizeof (mi->frequency) * CHAR_BIT <= 64);
      res = xmalloc (32); /* Overkill */
      sprintf (res, "%lu", mi->frequency);
      return res;
    }
  else if (strcmp (name, "load_hex") == 0) /* FIXME: load_hex for xmem? */
    {
      error (0, "option `%s' can't be read", name);
      return NULL;
    }
  else if (strcmp (name, "prog_mem_size") == 0)
    {
      char *res;

      assert (sizeof (mi->prog_mem_size) * CHAR_BIT <= 64);
      res = xmalloc (32); /* Overkill */
      sprintf (res, "%zu", mi->prog_mem_size);
      return res;
    }
  else if (strcmp (name, "xmem_mem_size") == 0)
    {
      char *res;
      
      assert (sizeof (mi->xmem_mem_size) * CHAR_BIT <= 64);
      res = xmalloc (32); /* Overkill */
      sprintf (res, "%zu", mi->xmem_mem_size);
      return res;
    }
  error (0, "unknown option `%s'", name);
  return NULL;
}

/* Set option NAME of MI_ to VALUE;
   returns: 0 if OK, -1 on error */
static int
i8051_set_option (void *mi_, const char *name, const char *value)
{
  struct mi *mi;

  mi = mi_;
  if (strcmp (name, "data_mem_size") == 0)
    return i8051_set_data_mem_size (mi, value);
  if (strcmp (name, "frequency") == 0)
    return i8051_set_frequency (mi, value);
  if (strcmp (name, "load_hex") == 0)
    {
      int res;

      res = load_hex_file (value, mi->program, mi->prog_mem_size);
      if (res == 0)
	mi->flags |= MI_PROGRAM_LOADED;
      return res;
    }
  if (strcmp (name, "prog_mem_size") == 0)
    return i8051_set_prog_mem_size (mi, value);
  if (strcmp (name, "xmem_mem_size") == 0)
    return i8051_set_xmem_mem_size (mi, value);
  error (0, "unknown option `%s'", name);
  return -1;
}

/* Our XML namespace */
static const xmlChar i8051_xml_ns[]
= "http://www.fw.cz/mitr/cereal/modules/8051";

/* Save setup (configuration options) of MI_ to NODE */
static void
i8051_save_setup (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  char buf[32];

  mi = mi_;
  ns = xmlNewNs (node, i8051_xml_ns, (const xmlChar *)"i8051");
  assert (sizeof (mi->data_mem_size) * CHAR_BIT <= 64);
  sprintf (buf, "%zu", mi->data_mem_size);
  xmlNewTextChild (node, ns, (const xmlChar *)"data_mem_size", (xmlChar *)buf);
  assert (sizeof (mi->frequency) * CHAR_BIT <= 64);
  sprintf (buf, "%lu", mi->frequency);
  xmlNewTextChild (node, ns, (const xmlChar *)"frequency", (xmlChar *)buf);
  assert (sizeof (mi->prog_mem_size) * CHAR_BIT <= 64);
  sprintf (buf, "%zu", mi->prog_mem_size);
  xmlNewTextChild (node, ns, (const xmlChar *)"prog_mem_size", (xmlChar *)buf);
  assert (sizeof (mi->xmem_mem_size) * CHAR_BIT <= 64);
  sprintf (buf, "%zu", mi->xmem_mem_size);
  xmlNewTextChild (node, ns, (const xmlChar *)"xmem_mem_size", (xmlChar *)buf);
}

/* Load setup (configuration options) of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
i8051_load_setup (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  return generic_module_load_setup_options (mi_, doc, node, &i8051_module,
					    i8051_xml_ns);
}

/* If the code below isn't a good reason to make an automatic generator
   of XML save/restore code, then nothing is. */

/* Values saved for uart_rx_status */
static const char *const uart_rx_status_text[] =
{
  [URX_NONE] = "URX_NONE", [URX_0_S3P1] = "URX_0_S3P1",
  [URX_0_S5P2] = "URX_0_S5P2", [URX_0_S6P1] = "URX_0_S6P1",
  [URX_0_DONE_S6P2] = "URX_0_DONE_S6P2", [URX_0_DONE_S1P1] = "URX_0_DONE_S1P1",
  [URX_123_START_0] = "URX_123_START_0", [URX_123_START_1] = "URX_123_START_1",
  [URX_123_RECV_0] = "URX_123_RECV_0", [URX_123_RECV_1] = "URX_123_RECV_1",
  [URX_123_RECV_2] = "URX_123_RECV_2"
};

/* Values saved for uart_tx_status */
static const char *const uart_tx_status_text[] =
{
  [UTX_NONE] = "UTX_NONE", [UTX_0_S3P1] = "UTX_0_S3P1",
  [UTX_0_S6P1] = "UTX_0_S6P1", [UTX_0_S6P2] = "UTX_0_S6P2",
  [UTX_0_DONE] = "UTX_0_DONE", [UTX_123_SEND] = "UTX_123_SEND"
};

/* Save state of MI_ to NODE */
static void
i8051_save_state (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  char buf[32];

  mi = mi_;
  ns = xmlNewNs (node, i8051_xml_ns, (const xmlChar *)"i8051");
  save_hex_xml (node, ns, "sfr", mi->sfr, sizeof (mi->sfr));
  save_hex_xml (node, ns, "memory", mi->memory, sizeof (mi->memory));
  save_hex_xml (node, ns, "program", mi->program, mi->prog_mem_size);
  if (mi->xmem_mem_size != 0)
    save_hex_xml (node, ns, "xmem", mi->xmem, mi->xmem_mem_size);
  sprintf (buf, "0x%04X", mi->pc);
  xmlNewTextChild (node, ns, (const xmlChar *)"pc", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_PROGRAM_LOADED) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"program_loaded",
		   (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_FETCH_EXTERNAL) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"fetch_external",
		   (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_FETCH_EXTERNAL_NEW) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"fetch_external_new",
		   (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_MOVX_1) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"movx_1", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_MOVX_2) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"movx_2", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_INTR_LOW) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"intr_low", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_INTR_HIGH) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"intr_high", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_NO_INTR) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"no_intr", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_T0_DOWN) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"t0_down", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_T1_DOWN) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"t1_down", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_UART_DIV2) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"uart_div2", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_UART_TIMER) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"uart_timer", (xmlChar *)buf);
  save_hex_xml (node, ns, "insn", mi->insn, sizeof (mi->insn));
  assert (sizeof (mi->insn) * CHAR_BIT <= 64);
  sprintf (buf, "%d", mi->insn_next - mi->insn);
  xmlNewTextChild (node, ns, (const xmlChar *)"insn_next", (xmlChar *)buf);
  assert (sizeof (mi->insn_special) * CHAR_BIT <= 64);
  sprintf (buf, "%d", mi->insn_special);
  xmlNewTextChild (node, ns, (const xmlChar *)"insn_special", (xmlChar *)buf);
  sprintf (buf, "0x%04X", mi->prev_pc);
  xmlNewTextChild (node, ns, (const xmlChar *)"prev_pc", (xmlChar *)buf);
  save_hex_xml (node, ns, "prev_insn", mi->prev_insn, sizeof (mi->prev_insn));
  sprintf (buf, "0x%02X", mi->tmp1);
  xmlNewTextChild (node, ns, (const xmlChar *)"tmp1", (xmlChar *)buf);
  sprintf (buf, "0x%02X", mi->tmp2);
  xmlNewTextChild (node, ns, (const xmlChar *)"tmp2", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->pins & PIN_ALE) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"pin_ale", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->pins & PIN_PSEN) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"pin_psen", (xmlChar *)buf);
  save_hex_xml (node, ns, "ports_in", mi->ports_in, sizeof (mi->ports_in));
  save_hex_xml (node, ns, "ports_out", mi->ports_out, sizeof (mi->ports_out));
  sprintf (buf, "0x%02X", mi->port_3_mask);
  xmlNewTextChild (node, ns, (const xmlChar *)"port_3_mask", (xmlChar *)buf);
  sprintf (buf, "0x%02X", mi->intr_pending);
  xmlNewTextChild (node, ns, (const xmlChar *)"intr_pending", (xmlChar *)buf);
  sprintf (buf, "0x%02X", mi->intr_new);
  xmlNewTextChild (node, ns, (const xmlChar *)"intr_new", (xmlChar *)buf);
  sprintf (buf, "0x%02X", mi->prev_p3);
  xmlNewTextChild (node, ns, (const xmlChar *)"prev_p3", (xmlChar *)buf);
  sprintf (buf, "0x%04X", mi->uart_rx_buf);
  xmlNewTextChild (node, ns, (const xmlChar *)"uart_rx_buf", (xmlChar *)buf);
  sprintf (buf, "0x%04X", mi->uart_tx_buf);
  xmlNewTextChild (node, ns, (const xmlChar *)"uart_tx_buf", (xmlChar *)buf);
  xmlNewTextChild (node, ns, (const xmlChar *)"uart_rx_status",
		   (const xmlChar *)uart_rx_status_text[mi->uart_rx_status]);
  xmlNewTextChild (node, ns, (const xmlChar *)"uart_tx_status",
		   (const xmlChar *)uart_tx_status_text[mi->uart_tx_status]);
  sprintf (buf, "%u", mi->uart_rx_step);
  xmlNewTextChild (node, ns, (const xmlChar *)"uart_rx_step", (xmlChar *)buf);
  sprintf (buf, "%u", mi->uart_tx_step);
  xmlNewTextChild (node, ns, (const xmlChar *)"uart_tx_step", (xmlChar *)buf);
  schedule_save_event (mi->s1p1_event, node, ns, "s1p1_event",
		       i8051_event_list);
  schedule_save_event (mi->s3p1_event, node, ns, "s3p1_event",
		       i8051_event_list);
  schedule_save_event (mi->s3p2_event, node, ns, "s3p2_event",
		       i8051_event_list);
  schedule_save_event (mi->s4p1_event, node, ns, "s4p1_event",
		       i8051_event_list);
  schedule_save_event (mi->s5p1_event, node, ns, "s5p1_event",
		       i8051_event_list);
  schedule_save_event (mi->s5p2_event, node, ns, "s5p2_event",
		       i8051_event_list);
  schedule_save_event (mi->s6p2_event, node, ns, "s6p2_event",
		       i8051_event_list);
  schedule_save_event (mi->s14p1_event, node, ns, "s14p1_event",
		       i8051_event_list);
  schedule_save_event (mi->s14p2_event, node, ns, "s14p2_event",
		       i8051_event_list);
  schedule_save_event (mi->s25p1_event, node, ns, "s25p1_event",
		       i8051_event_list);
  schedule_save_event (mi->s25p2_event, node, ns, "s25p2_event",
		       i8051_event_list);
  schedule_save_event (mi->s36p1_event, node, ns, "s36p1_event",
		       i8051_event_list);
  schedule_save_event (mi->uart_rx_event, node, ns, "uart_rx_event",
		       i8051_uart_event_list);
  schedule_save_event (mi->uart_tx_event, node, ns, "uart_tx_event",
		       i8051_uart_event_list);
  if (mi->s5p1_fn == NULL)
    xmlNewTextChild (node, ns, (const xmlChar *)"s5p1_fn",
		     (const xmlChar *)"NULL");
  else
    {
      const struct i8051_fn_desc *d;
      
      for (d = i8051_s5p1_fn_list; d->fn != mi->s5p1_fn; d++)
	assert (d[1].fn != NULL);
      xmlNewTextChild (node, ns, (const xmlChar *)"s5p1_fn",
		       (const xmlChar *)d->name);
    }
}

/* Load state of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
i8051_load_state (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  int res;
  
  mi = mi_;
  ns = xmlSearchNsByHref (doc, node, i8051_xml_ns);
  res = 0;
  if (ns == NULL)
    goto err;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE || node->ns != ns)
	continue;
#define GET_VAL(NAME, LIMIT)					\
      if (strcmp ((const char *)node->name, #NAME) == 0)	\
	do							\
	  {							\
	    unsigned long val;					\
								\
	    res = get_xml_num (doc, node, LIMIT, &val);		\
	    if (res != 0)					\
	      goto err;						\
	    mi->NAME = val;					\
	  }							\
	while (0)
#define GET_FLAG(VAR, NAME, BIT)			\
      if (strcmp ((const char *)node->name, NAME) == 0)	\
	do						\
	  {						\
	    unsigned long val;				\
							\
	    res = get_xml_num (doc, node, 2, &val);	\
	    if (res != 0)				\
	      goto err;					\
	    if (val == 0)				\
	      mi->VAR &= ~(BIT);			\
	    else					\
	      mi->VAR |= (BIT);				\
	  }						\
	while (0)
#define GET_HEX(NAME, SIZE)					\
      if (strcmp ((const char *)node->name, #NAME) == 0)	\
	do							\
	  {							\
	    res = load_hex_xml (doc, node, mi->NAME, SIZE);	\
	    if (res != 0)					\
	      goto err;						\
	  }							\
	while (0)
#define GET_EVENT(NAME, LIST)					\
      if (strcmp ((const char *)node->name, #NAME) == 0)	\
	do							\
	  {							\
	    res = schedule_load_event (mi->NAME, node, LIST);	\
	    if (res != 0)					\
	      goto err;						\
	  }							\
	while (0)
      GET_HEX (sfr, sizeof (mi->sfr));
      else GET_HEX (memory, sizeof (mi->memory));
      else GET_HEX (program, mi->prog_mem_size);
      else GET_HEX (xmem, mi->xmem_mem_size);
      else GET_VAL (pc, 0x10000);
      else GET_FLAG (flags, "program_loaded", MI_PROGRAM_LOADED);
      else GET_FLAG (flags, "fetch_external", MI_FETCH_EXTERNAL);
      else GET_FLAG (flags, "fetch_external_new", MI_FETCH_EXTERNAL_NEW);
      else GET_FLAG (flags, "movx_1", MI_MOVX_1);
      else GET_FLAG (flags, "movx_2", MI_MOVX_2);
      else GET_FLAG (flags, "intr_low", MI_INTR_LOW);
      else GET_FLAG (flags, "intr_high", MI_INTR_HIGH);
      else GET_FLAG (flags, "no_intr", MI_NO_INTR);
      else GET_FLAG (flags, "t0_down", MI_T0_DOWN);
      else GET_FLAG (flags, "t1_down", MI_T1_DOWN);
      else GET_FLAG (flags, "uart_div2", MI_UART_DIV2);
      else GET_FLAG (flags, "uart_timer", MI_UART_TIMER);
      else GET_HEX (insn, sizeof (mi->insn));
      else if (memcmp (node->name, "insn_next", sizeof ("insn_next")) == 0)
	{
	  unsigned long val;
	  
	  res = get_xml_num (doc, node, sizeof (mi->insn) + 1, &val);
	  if (res != 0)
	    goto err;
	  mi->insn_next = mi->insn + val;
	}
      else GET_VAL (insn_special, sizeof (mi->insn) + 1);
      else GET_VAL (prev_pc, 0x10000);
      else GET_HEX (prev_insn, sizeof (mi->prev_insn));
      else GET_VAL (tmp1, 0x100);
      else GET_VAL (tmp2, 0x100);
      else GET_FLAG (pins, "pin_ale", PIN_ALE);
      else GET_FLAG (pins, "pin_psen", PIN_PSEN);
      else GET_HEX (ports_in, sizeof (mi->ports_in));
      else GET_HEX (ports_out, sizeof (mi->ports_out));
      else GET_VAL (port_3_mask, 0x100);
      else GET_VAL (intr_pending, 0x100);
      else GET_VAL (intr_new, 0x100);
      else GET_VAL (prev_p3, 0x100);
      else GET_VAL (uart_rx_buf, 0x10000);
      else GET_VAL (uart_tx_buf, 0x10000);
      else if (strcmp ((const char *)node->name, "uart_rx_status") == 0)
	{
	  res = get_xml_enum (doc, node, uart_rx_status_text,
			      ARRAY_SIZE (uart_rx_status_text),
			      &mi->uart_rx_status);
	  if (res != 0)
	    goto err;
	}
      else if (strcmp ((const char *)node->name, "uart_tx_status") == 0)
	{
	  res = get_xml_enum (doc, node, uart_tx_status_text,
			      ARRAY_SIZE (uart_tx_status_text),
			      &mi->uart_tx_status);
	  if (res != 0)
	    goto err;
	}
      else GET_VAL (uart_rx_step, 0x10);
      else GET_VAL (uart_tx_step, 0x10);
      else GET_EVENT (s1p1_event, i8051_event_list);
      else GET_EVENT (s3p1_event, i8051_event_list);
      else GET_EVENT (s3p2_event, i8051_event_list);
      else GET_EVENT (s4p1_event, i8051_event_list);
      else GET_EVENT (s5p1_event, i8051_event_list);
      else GET_EVENT (s5p2_event, i8051_event_list);
      else GET_EVENT (s6p2_event, i8051_event_list);
      else GET_EVENT (s14p1_event, i8051_event_list);
      else GET_EVENT (s14p2_event, i8051_event_list);
      else GET_EVENT (s25p1_event, i8051_event_list);
      else GET_EVENT (s25p2_event, i8051_event_list);
      else GET_EVENT (s36p1_event, i8051_event_list);
      else GET_EVENT (uart_rx_event, i8051_uart_event_list);
      else GET_EVENT (uart_tx_event, i8051_uart_event_list);
      else if (strcmp ((const char *)node->name, "s5p1_fn") == 0)
	{
	  xmlChar *str;

	  str = xmlNodeListGetString (doc, node->children, 1);
	  if (str == NULL)
	    {
	      res = -1;
	      goto err;
	    }
	  if (strcmp ((const char *)str, "NULL") == 0)
	    mi->s5p1_fn = NULL;
	  else
	    {
	      const struct i8051_fn_desc *d;

	      for (d = i8051_s5p1_fn_list; d->fn != NULL; d++)
		{
		  if (strcmp (d->name, (const char *)str) == 0)
		    goto found;
		}
	      xmlFree (str);
	      error (0, "invalid <i8051:s5p1_fn> value");
	      res = -1;
	      goto err;

	    found:
	      xmlFree (str);
	      mi->s5p1_fn = d->fn;
	    }
	  xmlFree (str);
	}
    }
  /* Reload mi->regs */
  mi->regs = mi->memory + (mi->sfr[SFR_PSW] & (SFR_PSW_RS0 | SFR_PSW_RS1));
 err:
  return res;
}

struct cereal_module i8051_module = 
{
  .next = NULL, .pprev = NULL, .name = "8051", .compat_names = NULL,
  .xml = NULL, .handle = NULL, .mi_new = i8051_mi_new,
  .mi_delete = i8051_mi_delete, .set_option = i8051_set_option,
  .get_option = i8051_get_option, .save_setup = i8051_save_setup,
  .load_setup = i8051_load_setup, .save_state = i8051_save_state,
  .load_state = i8051_load_state, .w =
  {
    [WIDTH_1] =
    {
      .spaces = i8051_spaces_1, .num_spaces = ARRAY_SIZE (i8051_spaces_1),
      .xml_spaces = NULL
    },
    [WIDTH_8] =
    {
      .spaces = i8051_spaces_8, .num_spaces = ARRAY_SIZE (i8051_spaces_8),
      .xml_spaces = NULL
    },
    [WIDTH_16] =
    {
      .spaces = i8051_spaces_16, .num_spaces = ARRAY_SIZE (i8051_spaces_16),
      .xml_spaces = NULL
    }
  }
};
