/* Internal interfaces of 8051 module
   Copyright (C) 2002, 2004 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#ifndef I8051_H__
#define I8051_H__

#include <config.h>

#include <stddef.h>
#include <stdint.h>

#include <cereal.h>
#include <module.h>

/* Size of 8051 internal program memory */
enum
{
  /* Minimal size (interrupt handlers don't fit in smaller power of 2) */
  I8051_PROG_MEM_SIZE_MIN	= 64,
  I8051_PROG_MEM_SIZE_DEFAULT 	= 4096,		/* Default size */
  I8051_PROG_MEM_SIZE_MAX	= 65536,	/* Maximal size */
};

/* Value set until program is loaded */
enum { I8051_PROG_MEM_DEFAULT_VALUE = 0x00 };

/* Size of 8051 internal data memory */
enum
{
  I8051_MEM_SIZE_MAX	= 256,	/* Maximal size */
  I8051_MEM_SIZE_MIN 	= 128	/* Minimal size */
};

/* Size of "external" data memory */
enum
{
  I8051_XMEM_MEM_SIZE_DEFAULT	= 0,
  I8051_XMEM_MEM_SIZE_MAX	= 65536
};

/* Default value */
enum { I8051_XMEM_MEM_DEFAULT_VALUE = 0x00 };

/* Size of special function registers area */
enum { I8051_SFR_SIZE = 128 };

/* Starting address of special function registers area */
enum { I8051_SFR_START = 0x80 };

/* Addresses in I8051_SPACE_1_PINS */
enum
{
  I8051_S1P_P0		= 0x00, /* P0 bits start here */
  I8051_S1P_P1		= 0x08, /* P1 bits start here */
  I8051_S1P_P2		= 0x10, /* P2 bits start here */
  I8051_S1P_P3		= 0x18, /* P3 bits start here */
  I8051_S1P_MISC	= 0x20, /* Start of "miscellaneous bits" area */
  I8051_S1P_ALE		= I8051_S1P_MISC,
  I8051_S1P_PSEN,
  I8051_S1P_EA,
  I8051_NUM_S1P
};

/* 8051 SFR port handlers */
struct sfr_fn
{
  struct port_fn read;
  struct port_fn write;
  struct port_fn modify;
  struct port_fn display;
};

/* 8051 pin port handlers **/
struct pin_fn
{
  struct port_fn read;
  struct port_fn write;
};

/* 8051 module instance */
struct mi
{
  /* Emulation state */
  /* FIXME: all state should be exported for user / plugins */
  uint8_t sfr[I8051_SFR_SIZE];
  uint8_t memory[I8051_MEM_SIZE_MAX];
  uint8_t *program;
  uint8_t *xmem;
  size_t prog_mem_size;
  size_t data_mem_size;
  size_t xmem_mem_size; /* 0 = use external memory */
  uint16_t pc;
  unsigned flags; /* See MI_* below */
  unsigned long frequency;
  uint8_t insn[3];
  uint8_t *insn_next;
  size_t insn_special; /* Number of reads to skip (data already in ->insn) */
  uint16_t prev_pc; /* pc and insn[] of last completed instruction */
  uint8_t prev_insn[3];
  /* ALU operands or other temporary data; if there are two, then generally
     tmp1 = dest, tmp2 = src */
  uint8_t tmp1, tmp2;
  uint8_t pins; /* "Other" pins - see PIN_* below */
  /* Values latched on ports at last S5P2 */
  uint8_t ports_in[4];
  /* Values really output on P[0123] (may not always match sfr[SFR_Px]) */
  uint8_t ports_out[4];
  uint8_t port_3_mask; /* "Alternate output function of P3" values */
  /* Currently handled / newly read interrupt pending flags */
  uint8_t intr_pending, intr_new; /* See INTR_* below */
  uint8_t prev_p3; /* For edge detection */
  uint16_t uart_rx_buf, uart_tx_buf; /* Both are 12 bits long */
  unsigned uart_rx_status, uart_tx_status; /* See UTX_*, URX_* below */
  unsigned char uart_rx_step, uart_tx_step;
  struct schedule_event *s1p1_event, *s3p1_event;
  /* s3p2_event also used for movx_c2s3p1 () */
  struct schedule_event *s3p2_event, *s4p1_event, *s5p1_event, *s5p2_event;
  struct schedule_event *s6p2_event;
  struct schedule_event *s14p1_event, *s14p2_event, *s25p1_event, *s25p2_event;
  struct schedule_event *s36p1_event;
  struct schedule_event *uart_rx_event, *uart_tx_event;
  void (*s5p1_fn) (struct mi *mi); /* Function done at next s5p1 */
  /* Data reinitialized on each load */
  struct sfr_fn sfr_fns[I8051_SFR_SIZE];
  struct pin_fn pin_fns[I8051_NUM_S1P];
  uint8_t *regs; /* Offset of registers in ->memory (cached from SFR_PSW) */
  double phase_len; /* 1/12 of cycle time = 1 / frequency */
};

enum
{
  MI_PROGRAM_LOADED	= 0x001, /* A program was loaded */
  MI_FETCH_EXTERNAL	= 0x002, /* Fetch next insn from external memory */
  MI_FETCH_EXTERNAL_NEW	= 0x004, /* Fetch 2nd next insn from external memory */
  MI_MOVX_1		= 0x008, /* C1S3P2 .. C2S3P1 in MOVX */
  MI_MOVX_2		= 0x010, /* C2S3P2 .. C2S6P2 in MOVX */
  MI_INTR_LOW		= 0x020, /* Low-priority interrupt being handled */
  MI_INTR_HIGH		= 0x040, /* High-priority interrupt being handled */
  MI_NO_INTR		= 0x080, /* Don't service interrupts this time */
  MI_T0_DOWN		= 0x100, /* 1 -> 0 transition occurred on SFR_P3_T0 */
  MI_T1_DOWN		= 0x200, /* 1 -> 0 transition occurred on SFR_P3_T1 */
  MI_UART_DIV2		= 0x400, /* Ignore next UART pulse */
  MI_UART_TIMER		= 0x800  /* Set when TL1 overflows */
};

/* Interrupt flags */
enum
{
  INTR_X0	= 0x01, /* External 0 */
  INTR_T0	= 0x02, /* Timer 0 */
  INTR_X1	= 0x04, /* External 1 */
  INTR_T1	= 0x08, /* Timer 1 */
  INTR_S	= 0x10, /* Serial */
};

/* UART RX status; update also uart_rx_status_text! */
enum
{
  URX_NONE,
  URX_0_S3P1,
  URX_0_S5P2,
  URX_0_S6P1,
  URX_0_DONE_S6P2, /* Write SBUF */
  URX_0_DONE_S1P1, /* Report receive completion */
  URX_123_START_0, /* Looking for start bit, read 0 last time */
  URX_123_START_1, /* Looking for start bit, read 1 last time */
  /* Keep these three values consecutive! */
  URX_123_RECV_0,  /* Receiving data, seen only zeroes for this bit */
  URX_123_RECV_1,  /* Receiving data, seen single one for this bit */
  URX_123_RECV_2   /* Receiving data, seen two ones for this bit */
};

/* UART TX status; update also uart_tx_status_text! */
enum
{
  UTX_NONE,
  UTX_0_S3P1,
  UTX_0_S6P1,
  UTX_0_S6P2,
  UTX_0_DONE,
  UTX_123_SEND
};

/* Offsets for known special function registers in mi->sfr */
enum
{
  SFR_P0	= 0x80 - I8051_SFR_START,
  SFR_SP	= 0x81 - I8051_SFR_START,
  SFR_DPL	= 0x82 - I8051_SFR_START,
  SFR_DPH	= 0x83 - I8051_SFR_START,
  SFR_PCON	= 0x87 - I8051_SFR_START,
  SFR_TCON	= 0x88 - I8051_SFR_START,
  SFR_TMOD	= 0x89 - I8051_SFR_START,
  SFR_TL0	= 0x8A - I8051_SFR_START,
  SFR_TL1	= 0x8B - I8051_SFR_START,
  SFR_TH0	= 0x8C - I8051_SFR_START,
  SFR_TH1	= 0x8D - I8051_SFR_START,
  SFR_P1	= 0x90 - I8051_SFR_START,
  SFR_SCON	= 0x98 - I8051_SFR_START,
  SFR_SBUF	= 0x99 - I8051_SFR_START,
  SFR_P2	= 0xA0 - I8051_SFR_START,
  SFR_IE	= 0xA8 - I8051_SFR_START,
  SFR_P3	= 0xB0 - I8051_SFR_START,
  SFR_IP	= 0xB8 - I8051_SFR_START,
  SFR_PSW	= 0xD0 - I8051_SFR_START,
  SFR_ACC	= 0xE0 - I8051_SFR_START,
  SFR_B		= 0xF0 - I8051_SFR_START
};

enum
{
  PIN_ALE	= 0x01,
  PIN_PSEN	= 0x02,
};

/* PCON bits */
enum
{
  SFR_PCON_IDL	= 0x01,
  SFR_PCON_PD	= 0x02,
  /* SFR_PCON_GF[01] aren't interesting enough */
  SFR_PCON_SMOD	= 0x80
};

/* TCON bits */
enum
{
  SFR_TCON_IT0	= 0x01,
  SFR_TCON_IE0	= 0x02,
  SFR_TCON_IT1	= 0x04,
  SFR_TCON_IE1	= 0x08,
  SFR_TCON_TR0	= 0x10,
  SFR_TCON_TF0	= 0x20,
  SFR_TCON_TR1	= 0x40,
  SFR_TCON_TF1	= 0x80
};

/* TMOD bits */
enum
{
  SFR_TMOD_0_M0		= 0x01,
  SFR_TMOD_0_M1		= 0x02,
  SFR_TMOD_0_CT		= 0x04,
  SFR_TMOD_0_GATE	= 0x08,
  SFR_TMOD_1_M0		= 0x10,
  SFR_TMOD_1_M1		= 0x20,
  SFR_TMOD_1_CT		= 0x40,
  SFR_TMOD_1_GATE	= 0x80
};

/* SCON bits */
enum
{
  SFR_SCON_RI	= 0x01,
  SFR_SCON_TI	= 0x02,
  SFR_SCON_RB8	= 0x04,
  SFR_SCON_TB8	= 0x08,
  SFR_SCON_REN	= 0x10,
  SFR_SCON_SM2	= 0x20, /* Ugly big-endian bit numbering */
  SFR_SCON_SM1	= 0x40,
  SFR_SCON_SM0	= 0x80
};

/* IE bits */
enum
{
  SFR_IE_EX0	= INTR_X0,
  SFR_IE_ET0	= INTR_T0,
  SFR_IE_EX1	= INTR_X1,
  SFR_IE_ET1	= INTR_T1,
  SFR_IE_ES	= INTR_S,
  SFR_IE_EA	= 0x80
};

/* P3 alternate function names */
enum
{
  SFR_P3_RXD	= 0x01,
  SFR_P3_TXD	= 0x02,
  SFR_P3_INT0	= 0x04,
  SFR_P3_INT1	= 0x08,
  SFR_P3_T0	= 0x10,
  SFR_P3_T1	= 0x20,
  SFR_P3_WR	= 0x40,
  SFR_P3_RD	= 0x80,
};

/* PSW bits */
enum
{
  SFR_PSW_P	= 0x01,
  SFR_PSW_OV	= 0x04,
  SFR_PSW_RS0	= 0x08,
  SFR_PSW_RS1	= 0x10,
  SFR_PSW_AC	= 0x40,
  SFR_PSW_C	= 0x80
};

/* SFR template structure */
struct sfr_fn_skel
{
  size_t reg; /* SFR_* */
  port_8_read_t read_fn;
  port_8_write_t write_fn;
  port_8_display_t display_fn;
  port_8_modify_t modify_fn;
};

/* mi->sfr_fns template.  Values given here can't be overriden using
   I8051_SPACE_8_SFR_EXT.  In ascending order of .reg values, terminated by
   .reg = I8051_SFR_SIZE. */
extern const struct sfr_fn_skel i8051_sfr_fns[];
/* Number of entries in i8051_sfr_fns, not including the terminating
   I8051_SFR_SIZE element.  Computed when first MI is created. */
extern size_t i8051_sfr_fns_len;

/* Bit address spaces */
enum
{
  I8051_SPACE_1_BITS,	/* 8051 bit addresses */
  I8051_SPACE_1_PINS,	/* 8051 chip pins */
  I8051_NUM_SPACES_1	/* Keep this last */
};

extern const struct space i8051_spaces_1[I8051_NUM_SPACES_1];

/* Byte address spaces */
enum
{
  I8051_SPACE_8_MEM,		/* Internal memory */
  I8051_SPACE_8_PROG,		/* Internal program memory */
  I8051_SPACE_8_SFR,		/* SFRs */
  I8051_SPACE_8_SFR_EXT,	/* For adding new SFRs from plugins */ 
  I8051_SPACE_8_PREV_INSN,	/* Last finished instruction */
  /* 3 bytes at mi->pc, if using internal program memory */
  I8051_SPACE_8_NEXT_INSN,
  I8051_NUM_SPACES_8		/* Keep this last */
};

extern const struct space i8051_spaces_8[I8051_NUM_SPACES_8];

/* 16b address spaces */
enum
{
  I8051_SPACE_16_MISC,	/* Miscellaneous 16b registers: see I8051_S16M_* */
  I8051_NUM_SPACES_16
};

/* Addresses in I8051_SPACE_16_MISC */
enum
{
  I8051_S16M_PC,	/* mi->pc */
  I8051_S16M_PREV_PC,	/* mi->prev_pc */
  I8051_NUM_S16M
};

extern const struct space i8051_spaces_16[I8051_NUM_SPACES_16];

/* Emulate one instruction in MI */
extern void i8051_emulate_insn (void *mi_);

/* Handle S3P1 */
extern void i8051_emulate_s3p1 (void *mi_);

/* Handle S5P1 */
extern void i8051_emulate_s5p1 (void *mi_);

/* Handle S5P2 */
extern void i8051_emulate_s5p2 (void *mi_);

/* Handle S1P1 or S4P1 */
extern void i8051_emulate_s14p1 (void *mi_);

/* Handle S1P2 or S4P2 */
extern void i8051_emulate_s14p2 (void *mi_);

/* Handle S2P1 or S5P1 */
extern void i8051_emulate_s25p1 (void *mi_);

/* Handle S2P2 or S5P2 */
extern void i8051_emulate_s25p2 (void *mi_);

/* Emulate reset of MI */
extern void i8051_emulate_reset (struct mi *mi);

/* Read current value of PORT (PORT is I8051_S1P_Pi) to *DEST;
   returns: number of connected bits */
extern unsigned i8051_port_read (struct mi *mi, size_t port, uint8_t *dest);

/* Set PORT (PORT is I8051_S1P_Pi) to VAL;
   returns: number of connected bits */
extern unsigned i8051_port_write (struct mi *mi, size_t port, uint8_t val);

/* Write VAL to byte port in MI_ */
extern void i8051_sfr_write_sbuf_fn (void *mi_, intptr_t data, uint8_t val);

/* Write VAL to byte port in MI_ */
extern void i8051_sfr_write_scon_fn (void *mi_, intptr_t data, uint8_t val);

/* Event handling function list (except for UART) */
extern const struct schedule_fn_desc i8051_event_list[];

/* Event handling function list for uart_[rt]x_event */
extern const struct schedule_fn_desc i8051_uart_event_list[];

/* 8051 function description */
struct i8051_fn_desc
{
  void (*fn) (struct mi *mi);
  const char *name;
};

/* List of functions that can be in mi->s5p1_fn */
extern const struct i8051_fn_desc i8051_s5p1_fn_list[];

#endif
