/* 8051 instruction emulator
   Copyright (C) 2002, 2004 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <inttypes.h>
#include <string.h>

#include <cereal.h>
#include <lib.h>

#include "8051.h"

/* Forward declaration */
static void s36p1 (void *mi_);

/* Handle S1P1 or S4P1 */
void
i8051_emulate_s14p1 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  if ((mi->flags & MI_MOVX_1) != 0)
    mi->flags |= MI_FETCH_EXTERNAL_NEW;
  else
    {
      const struct port_fn *pf;
      
      pf = &mi->pin_fns[I8051_S1P_EA].read;
      if (pf->fn.read_1 != NULL && pf->fn.read_1 (pf->mi, pf->data) == 0)
	mi->flags |= MI_FETCH_EXTERNAL_NEW;
      else
	mi->flags &= ~MI_FETCH_EXTERNAL_NEW;
      if ((mi->flags & MI_MOVX_2) != 0)
	/* P3 output is updated four lines later */
	mi->port_3_mask |= (mi->insn[0] & 0x10) == 0 ? SFR_P3_RD : SFR_P3_WR;
    }
  i8051_port_write (mi, I8051_S1P_P1, mi->sfr[SFR_P1]);
  i8051_port_write (mi, I8051_S1P_P3, mi->sfr[SFR_P3]);
  if ((mi->flags & (MI_FETCH_EXTERNAL | MI_MOVX_2)) == 0)
    {
      i8051_port_write (mi, I8051_S1P_P0, mi->sfr[SFR_P0]);
      i8051_port_write (mi, I8051_S1P_P2, mi->sfr[SFR_P2]);
    }
  schedule_add (mi->s14p1_event, 6 * mi->phase_len, i8051_emulate_s14p1);
}

/* Handle S1P2 or S4P2 */
void
i8051_emulate_s14p2 (void *mi_)
{
  struct mi *mi;
  const struct port_fn *pf;

  mi = mi_;
  if ((mi->flags & MI_FETCH_EXTERNAL_NEW) != 0 || mi->pc >= mi->prog_mem_size)
    mi->flags |= MI_FETCH_EXTERNAL;
  else
    mi->flags &= ~MI_FETCH_EXTERNAL;
  if ((mi->flags & (MI_FETCH_EXTERNAL | MI_MOVX_1)) == MI_FETCH_EXTERNAL)
    schedule_add (mi->s36p1_event, 3 * mi->phase_len, s36p1);
  if ((mi->pins & PIN_PSEN) == 0)
    {
      mi->pins |= PIN_PSEN;
      pf = &mi->pin_fns[I8051_S1P_PSEN].write;
      if (pf->fn.write_1 != NULL)
	pf->fn.write_1 (pf->mi, pf->data, 1);
    }
  if ((mi->flags & MI_MOVX_2) != 0 && (mi->insn[0] & 0x10) != 0)
    /* This is certainly a strange hardware wart. */
    i8051_port_write (mi, I8051_S1P_P0, mi->pc & 0xFF);
  mi->pins |= PIN_ALE;
  pf = &mi->pin_fns[I8051_S1P_ALE].write;
  if (pf->fn.write_1 != NULL)
    pf->fn.write_1 (pf->mi, pf->data, 1);
  schedule_add (mi->s14p2_event, 6 * mi->phase_len, i8051_emulate_s14p2);
}

/* Handle S2P1 or S5P1 */
void
i8051_emulate_s25p1 (void *mi_)
{
  struct mi *mi;
  uint16_t val;
  
  mi = mi_;
  if ((mi->flags & (MI_FETCH_EXTERNAL | MI_MOVX_1)) == 0)
    { /* Maybe we couldn't do it in S14P1, so do it now. */
      i8051_port_write (mi, I8051_S1P_P0, mi->sfr[SFR_P0]);
      i8051_port_write (mi, I8051_S1P_P2, mi->sfr[SFR_P2]);
    }
  else
    {
      if ((mi->flags & MI_MOVX_1) == 0)
	val = mi->pc;
      else
	{
	  switch (mi->insn[0] & 0x03)
	    {
	    case 0x00: /* 0xE0, 0xF0: MOVX A, @DPTR; MOVX @DPTR, A */
	      val = mi->sfr[SFR_DPL] | mi->sfr[SFR_DPH] << 8;
	      break;
	      
	    case 0x02: case 0x03: /* 0xE2, 0xE3, 0xF2, 0xF3 */
	      val = mi->regs[mi->insn[0] & 0x01];
	      break;
	      
	    default:
	      assert ("BUG" == NULL);
	      return;
	    }
	}
      i8051_port_write (mi, I8051_S1P_P0, val & 0xFF);
      if ((mi->flags & MI_MOVX_1) == 0 || (mi->insn[0] & 0x03) == 0x00)
	i8051_port_write (mi, I8051_S1P_P2, val >> 8);
    }
  schedule_add (mi->s25p1_event, 6 * mi->phase_len, i8051_emulate_s25p1);
}

/* Handle S2P2 or S5P2 */
void
i8051_emulate_s25p2 (void *mi_)
{
  struct mi *mi;
  const struct port_fn *pf;

  mi = mi_;
  mi->pins &= ~PIN_ALE;
  pf = &mi->pin_fns[I8051_S1P_ALE].write;
  if (pf->fn.write_1 != NULL)
    pf->fn.write_1 (pf->mi, pf->data, 0);
  schedule_add (mi->s25p2_event, 6 * mi->phase_len, i8051_emulate_s25p2);
}

/* Handle S3P1 or S6P1 */
static void
s36p1 (void *mi_)
{
  struct mi *mi;
  const struct pin_fn *pf;

  mi = mi_;
  mi->sfr[SFR_P0] = 0xFF;
  i8051_port_write (mi, I8051_S1P_P0, 0xFF);
  mi->pins &= ~PIN_PSEN;
  pf = mi->pin_fns + I8051_S1P_PSEN;
  if (pf->write.fn.write_1 != NULL)
    pf->write.fn.write_1 (pf->write.mi, pf->write.data, 0);
}

/* Read next instruction byte in MI, store it in mi->insn;
   Should be called only on S1P1 or S4P1 */
static void
read_insn (struct mi *mi)
{
  if (mi->insn_special != 0)
    mi->insn_special--;
  else
    {
      uint16_t pc;
      uint8_t val;
      
      pc = mi->pc++;
      if (pc == mi->prog_mem_size)
	{
	  error (0, "program counter overflows to external memory");
	  emulation_result |= ER_WARNING;
	}
      if ((mi->flags & MI_FETCH_EXTERNAL) == 0)
	val = mi->program[pc];
      else if (i8051_port_read (mi, I8051_S1P_P0, &val) != 8)
	{
	  error (0, "external program memory not connected");
	  emulation_result |= ER_WARNING;
	}
      *mi->insn_next = val;
    }
  mi->insn_next++;
}

/* Handle S3P1 */
void
i8051_emulate_s3p1 (void *mi_)
{
  struct mi *mi;
  int run, flag;

  mi = mi_;
  if ((mi->sfr[SFR_TCON] & SFR_TCON_TR0) != 0
      && ((mi->sfr[SFR_TMOD] & SFR_TMOD_0_GATE) == 0
	  || (mi->ports_in[3] & SFR_P3_INT0) != 0)
      && ((mi->sfr[SFR_TMOD] & SFR_TMOD_0_CT) == 0
	  || (mi->flags & MI_T0_DOWN) != 0))
    {
      switch (mi->sfr[SFR_TMOD] & (SFR_TMOD_0_M0 | SFR_TMOD_0_M1))
	{
	case 0:
	  {
	    uint16_t val;

	    val = mi->sfr[SFR_TH0] << 5 | (mi->sfr[SFR_TL0] & 0x1F);
	    if (val++ == 0x1FFF)
	      mi->sfr[SFR_TCON] |= SFR_TCON_TF0;
	    mi->sfr[SFR_TH0] = (val >> 5) & 0xFF;
	    mi->sfr[SFR_TL0] = (mi->sfr[SFR_TL0] & 0xE0) | (val & 0x1F);
	    break;
	  }

	case SFR_TMOD_0_M0:
	  {
	    uint16_t val;

	    val = mi->sfr[SFR_TH0] << 8 | mi->sfr[SFR_TL0];
	    if (val++ == 0xFFFF)
	      mi->sfr[SFR_TCON] |= SFR_TCON_TF0;
	    mi->sfr[SFR_TH0] = val >> 8;
	    mi->sfr[SFR_TL0] = val & 0xFF;
	    break;
	  }

	case SFR_TMOD_0_M1:
	  if (mi->sfr[SFR_TL0]++ == 0xFF)
	    {
	      mi->sfr[SFR_TCON] |= SFR_TCON_TF0;
	      mi->sfr[SFR_TL0] = mi->sfr[SFR_TH0];
	    }
	  break;

	case SFR_TMOD_0_M1 | SFR_TMOD_0_M0:
	  if (mi->sfr[SFR_TL0]++ == 0xFF)
	    mi->sfr[SFR_TCON] |= SFR_TCON_TF0;
	  break;

	default:
	  assert ("BUG" == NULL);
	  break;
	}
    }
  if ((mi->sfr[SFR_TMOD] & (SFR_TMOD_0_M0 | SFR_TMOD_0_M1))
      == (SFR_TMOD_0_M0 | SFR_TMOD_0_M1))
    {
      if ((mi->sfr[SFR_TCON] & SFR_TCON_TR1) != 0
	  && mi->sfr[SFR_TH0]++ == 0xFF)
	{
	  mi->sfr[SFR_TCON] |= SFR_TCON_TF1;
	  mi->flags |= MI_UART_TIMER;
	}
      run = 1;
    }
  else
    run = (mi->sfr[SFR_TCON] & SFR_TCON_TR1) != 0
      && ((mi->sfr[SFR_TMOD] & SFR_TMOD_1_GATE) == 0
	  || (mi->ports_in[3] & SFR_P3_INT1) != 0)
      && ((mi->sfr[SFR_TMOD] & SFR_TMOD_1_CT) == 0
	  || (mi->flags & MI_T1_DOWN) != 0);
  flag = 0;
  if (run != 0)
    {
      switch (mi->sfr[SFR_TMOD] & (SFR_TMOD_1_M0 | SFR_TMOD_1_M1))
	{
	case 0:
	  {
	    uint16_t val;
	    
	    val = mi->sfr[SFR_TH1] << 5 | (mi->sfr[SFR_TL1] & 0x1F);
	    if (val++ == 0x1FFF)
	      flag = 1;
	    mi->sfr[SFR_TH1] = (val >> 5) & 0xFF;
	    mi->sfr[SFR_TL1] = (mi->sfr[SFR_TL1] & 0xE0) | (val & 0x1F);
	    break;
	  }
	  
	case SFR_TMOD_1_M0:
	  {
	    uint16_t val;
	    
	    val = mi->sfr[SFR_TH1] << 8 | mi->sfr[SFR_TL1];
	    if (val++ == 0xFFFF)
	      flag = 1;
	    mi->sfr[SFR_TH1] = val >> 8;
	    mi->sfr[SFR_TL1] = val & 0xFF;
	    break;
	  }
	  
	case SFR_TMOD_1_M1:
	  if (mi->sfr[SFR_TL1]++ == 0xFF)
	    {
	      flag = 1;
	      mi->sfr[SFR_TL1] = mi->sfr[SFR_TH1];
	    }
	  break;

	case SFR_TMOD_1_M0 | SFR_TMOD_1_M1:
	  break;
	  
	default:
	  assert ("BUG" == NULL);
	  break;
	}  
    }
  if (flag != 0 && (mi->sfr[SFR_TMOD] & (SFR_TMOD_0_M0 | SFR_TMOD_0_M1))
      != (SFR_TMOD_0_M0 | SFR_TMOD_0_M1))
    {
      mi->sfr[SFR_TCON] |= SFR_TCON_TF1;
      mi->flags |= MI_UART_TIMER;
    }
  schedule_add (mi->s3p1_event, 12 * mi->phase_len, i8051_emulate_s3p1);
}

/* Handle S5P1 */
void
i8051_emulate_s5p1 (void *mi_)
{
  struct mi *mi;
  void (*fn) (struct mi *_mi);

  mi = mi_;
  i8051_port_read (mi, I8051_S1P_P0, mi->ports_in);
  i8051_port_read (mi, I8051_S1P_P1, mi->ports_in + 1);
  i8051_port_read (mi, I8051_S1P_P2, mi->ports_in + 2);
  i8051_port_read (mi, I8051_S1P_P3, mi->ports_in + 3);
  fn = mi->s5p1_fn;
  if (fn != NULL)
    {
      mi->s5p1_fn = NULL;
      fn (mi);
    }
  schedule_add (mi->s5p1_event, 12 * mi->phase_len, i8051_emulate_s5p1);
}

/* Handle S5P2 */
void
i8051_emulate_s5p2 (void *mi_)
{
  struct mi *mi;
  uint8_t down;

  mi = mi_;
  mi->intr_pending = mi->intr_new;
  mi->intr_new = 0;
  down = (mi->ports_in[3] ^ mi->prev_p3) & mi->prev_p3;
  if ((mi->sfr[SFR_TCON] & SFR_TCON_IT0) != 0)
    {
      if ((down & SFR_P3_INT0) != 0)
	mi->sfr[SFR_TCON] |= SFR_TCON_IE0;
      if ((mi->sfr[SFR_TCON] & SFR_TCON_IE0) != 0)
	mi->intr_new |= INTR_X0;
    }
  if ((mi->sfr[SFR_TCON] & SFR_TCON_IT1) != 0)
    {
      if ((down & SFR_P3_INT1) != 0)
	mi->sfr[SFR_TCON] |= SFR_TCON_IE1;
      if ((mi->sfr[SFR_TCON] & SFR_TCON_IE1) != 0)
	mi->intr_new |= INTR_X1;
    }
  if ((down & SFR_P3_T0) != 0)
    mi->flags |= MI_T0_DOWN;
  else
    mi->flags &= ~MI_T0_DOWN;
  if ((down & SFR_P3_T1) != 0)
    mi->flags |= MI_T1_DOWN;
  else
    mi->flags &= ~MI_T1_DOWN;
  mi->prev_p3 = mi->ports_in[3];
  if ((mi->sfr[SFR_TCON] & SFR_TCON_IT0) == 0
      && (mi->ports_in[3] & SFR_P3_INT0) == 0)
    mi->intr_new |= INTR_X0;
  if ((mi->sfr[SFR_TCON] & SFR_TCON_IT1) == 0
      && (mi->ports_in[3] & SFR_P3_INT1) == 0)
    mi->intr_new |= INTR_X1;
  if ((mi->sfr[SFR_TCON] & SFR_TCON_TF0) != 0)
    mi->intr_new |= INTR_T0;
  if ((mi->sfr[SFR_TCON] & SFR_TCON_TF1) != 0)
    mi->intr_new |= INTR_T1;
  if ((mi->sfr[SFR_SCON] & (SFR_SCON_RI | SFR_SCON_TI)) != 0)
    mi->intr_new |= INTR_S;
  schedule_add (mi->s5p2_event, 12 * mi->phase_len, i8051_emulate_s5p2);
}

/* Find highest-priority active interrupt in one of INTRS;
   returns: interrupt handler address or 0 if none */
static uint16_t
find_intr (struct mi *mi, uint8_t intrs)
{
  uint16_t addr, dest;
  uint8_t mask;

  intrs &= mi->intr_pending & mi->sfr[SFR_IE];
  dest = 0;
  addr = 0x03;
  for (mask = 0x01; mask < SFR_IE_EA; mask <<= 1)
    {
      if ((intrs & mask) != 0 && dest == 0)
	dest = addr;
      addr += 0x08;
    }
  return dest;
}

/* Handle pending interrupt to DEST of priority flag PRIO */
static void
handle_intr (struct mi *mi, uint16_t dest, unsigned prio)
{
  if ((mi->flags & (MI_INTR_HIGH | prio)) != 0)
    return;
  mi->insn[0] = 0x12; /* LCALL */
  mi->insn[1] = dest >> 8;
  mi->insn[2] = dest & 0xFF;
  mi->insn_special = 3;
  mi->flags |= prio;
  switch (dest)
    {
      /* This clears the flag even for level-triggered interrupts.  I'm not
	 really sure about it, but it's simpler and it also works reasonably
	 when SFR_TCON_ITi is cleared in the last instruction (after the flags
	 were polled). */
    case 0x03: /* External 0 */
      mi->sfr[SFR_TCON] &= ~SFR_TCON_IE0;
      break;

    case 0x0B: /* Timer 0 */
      mi->sfr[SFR_TCON] &= ~SFR_TCON_TF0;
      break;
      
    case 0x13: /* External 1 */
      mi->sfr[SFR_TCON] &= ~SFR_TCON_IE1;
      break;
      
    case 0x1B: /* Timer 1 */
      mi->sfr[SFR_TCON] &= ~SFR_TCON_TF1;
      break;
      
    case 0x23: /* Serial */
    case 0x2B: case 0x33: /* Possibly these two */
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Finish emulation of current instruction in MI, emulate next PHASES phases
   later */
static void
finish_insn (struct mi *mi, unsigned phases)
{
  assert (sizeof (mi->insn) == sizeof (mi->prev_insn));
  memcpy (mi->prev_insn, mi->insn, sizeof (mi->prev_insn));
  if ((mi->flags & MI_NO_INTR) == 0 && (mi->sfr[SFR_IE] & SFR_IE_EA) != 0)
    {
      uint16_t dest;
      
      dest = find_intr (mi, mi->sfr[SFR_IP]);
      if (dest != 0)
	handle_intr (mi, dest, MI_INTR_HIGH);
      else
	{
	  dest = find_intr (mi, ~mi->sfr[SFR_IP]);
	  if (dest != 0)
	    handle_intr (mi, dest, MI_INTR_LOW);
	}
    }
  mi->flags &= ~MI_NO_INTR;
  mi->insn_next = mi->insn;
  emulation_result |= ER_INSN;
  schedule_add (mi->s1p1_event, phases * mi->phase_len, i8051_emulate_insn);
}

/* Write ACC to SFR_ACC in MI, update SFR_PSW_P */
static void
write_acc (struct mi *mi, uint8_t acc)
{
  mi->sfr[SFR_ACC] = acc;
  acc ^= acc >> 4;
  acc ^= acc >> 2;
  acc ^= acc >> 1;
  if ((acc & 0x01) != 0)
    mi->sfr[SFR_PSW] |= SFR_PSW_P;
  else
    mi->sfr[SFR_PSW] &= ~SFR_PSW_P;
}

/* Read byte port value in MI_;
   returns: port value */
static uint8_t
sfr_read_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  mi = mi_;
  return mi->sfr[data];
}

/* Read byte port value in MI_;
   returns: port value */
static uint8_t
sfr_read_port_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  mi = mi_;
  switch (data)
    {
    case SFR_P0:
      return mi->ports_in[0];
      
    case SFR_P1:
      return mi->ports_in[1];
      
    case SFR_P2:
      return mi->ports_in[2];
      
    case SFR_P3:
      return mi->ports_in[3];
      
    default:
      assert ("BUG" == NULL);
      return 0;
    }
}

/* Write VAL to byte port in MI_ */
static void
sfr_write_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  mi = mi_;
  mi->sfr[data] = val;
}

/* Write VAL to byte port in MI_ */
static void
sfr_write_psw_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  (void)data;
  mi = mi_;
  mi->sfr[SFR_PSW] = val;
  mi->regs = mi->memory + (val & (SFR_PSW_RS0 | SFR_PSW_RS1));
  write_acc (mi, mi->sfr[SFR_ACC]); /* Recompute SFR_PSW_P */
}

/* Write VAL to byte port in MI_ */
static void
sfr_write_acc_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  (void)data;
  mi = mi_;
  write_acc (mi, val);
}

/* Read byte port value for display (unlike port_8_read_t may have no side
   effects);
   returns: port value */
static uint8_t
sfr_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  mi = mi_;
  return mi->sfr[data];
}

/* Modify byte port value by user */
static void
sfr_modify_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  mi = mi_;
  mi->sfr[data] = val;
}

/* Modify byte port value by user */
static void
sfr_modify_psw_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  (void)data;
  mi = mi_;
  mi->sfr[SFR_PSW] = val;
  mi->regs = mi->memory + (val & (SFR_PSW_RS0 | SFR_PSW_RS1));
  /* Recompute SFR_PSW_P.  If the user wants to change SFR_PSW_P, he'll have
     to change SFR_ACC. */
  write_acc (mi, mi->sfr[SFR_ACC]);
}

/* Modify byte port value by user */
static void
sfr_modify_acc_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  (void)data;
  mi = mi_;
  write_acc (mi, val);
}

/* mi->sfr_fns template.  Values given here can't be overriden using
   I8051_SPACE_8_SFR_EXT.  In ascending order of .reg values, terminated by
   .reg = I8051_SFR_SIZE. */
const struct sfr_fn_skel i8051_sfr_fns[] =
{
  {
    .reg = SFR_P0, .read_fn = sfr_read_port_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_SP, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_DPL, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_DPH, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  { /* FIXME: warn about writing 1 to reserved bits? */ /* FIXME: implement */
    .reg = SFR_PCON, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_TCON, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_TMOD, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_TL0, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_TL1, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_TH0, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_TH1, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_P1, .read_fn = sfr_read_port_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_SCON, .read_fn = sfr_read_fn,
    .write_fn = i8051_sfr_write_scon_fn, .display_fn = sfr_display_fn,
    .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_SBUF, .read_fn = sfr_read_fn,
    .write_fn = i8051_sfr_write_sbuf_fn, .display_fn = sfr_display_fn,
    .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_P2, .read_fn = sfr_read_port_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  { /* FIXME: warn about writing 1 to reserved bits? */
    .reg = SFR_IE, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_P3, .read_fn = sfr_read_port_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  { /* FIXME: warn about writing 1 to reserved bits? */
    .reg = SFR_IP, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = SFR_PSW, .read_fn = sfr_read_fn, .write_fn = sfr_write_psw_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_psw_fn
  },
  {
    .reg = SFR_ACC, .read_fn = sfr_read_fn, .write_fn = sfr_write_acc_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_acc_fn
  },
  {
    .reg = SFR_B, .read_fn = sfr_read_fn, .write_fn = sfr_write_fn,
    .display_fn = sfr_display_fn, .modify_fn = sfr_modify_fn
  },
  {
    .reg = I8051_SFR_SIZE, .read_fn = NULL, .write_fn = NULL,
    .display_fn = NULL, .modify_fn = NULL
  },
};

/* Number of entries in i8051_sfr_fns, not including the terminating
   I8051_SFR_SIZE element.  Computed when first MI is created. */
size_t i8051_sfr_fns_len;

/* Read direct-addressed data at ADDR in MI */
static uint8_t
read_direct (struct mi *mi, uint8_t addr)
{
  const struct port_fn *pf;
  
  if (addr < I8051_SFR_START)
    return mi->memory[addr];
  pf = &mi->sfr_fns[addr - I8051_SFR_START].read;
  if (pf->fn.read_8 != NULL)
    return pf->fn.read_8 (pf->mi, pf->data);
  else
    {
      error (0, "read from unknown special function register %02" PRIX8, addr);
      emulation_result |= ER_WARNING;
      return 0xFF; /* Just guessing a bogus value */
    }
}


/* Read direct-addressed data at ADDR in MI, read output latch instead of
   input value on Pi */
static uint8_t
read_direct_latch (struct mi *mi, uint8_t addr)
{
  switch (addr)
    {
    case I8051_SFR_START + SFR_P0:
      return mi->sfr[SFR_P0];

    case I8051_SFR_START + SFR_P1:
      return mi->sfr[SFR_P1];

    case I8051_SFR_START + SFR_P2:
      return mi->sfr[SFR_P2];

    case I8051_SFR_START + SFR_P3:
      return mi->sfr[SFR_P3];
    }
  return read_direct (mi, addr);
}

/* Write VAL to direct-addressed data at ADDR in MI */
static void
write_direct (struct mi *mi, uint8_t addr, uint8_t val)
{
  if (addr < I8051_SFR_START)
    mi->memory[addr] = val;
  else
    {
      const struct port_fn *pf;

      if (addr == I8051_SFR_START + SFR_IE || addr == I8051_SFR_START + SFR_IP)
	mi->flags |= MI_NO_INTR;
      pf = &mi->sfr_fns[addr - I8051_SFR_START].write;
      if (pf->fn.write_8 != NULL)
	pf->fn.write_8 (pf->mi, pf->data, val);
      else
	{
	  error (0, "write to unknown special function register %02" PRIX8,
		 addr);
	  emulation_result |= ER_WARNING;
	}
    }
}

/* Read indirect-addressed data at ADDR in MI */
static uint8_t
read_indirect (struct mi *mi, uint8_t addr)
{
  if (addr < mi->data_mem_size)
    return mi->memory[addr];
  error (0, "read from non-existent internal memory location %02" PRIX8, addr);
  emulation_result |= ER_WARNING;
  return 0xFF; /* Just guessing a bogus value */
}

/* Write VAL to indirect-addressed data at ADDR in MI */
static void
write_indirect (struct mi *mi, uint8_t addr, uint8_t val)
{
  if (addr < mi->data_mem_size)
    mi->memory[addr] = val;
  else
    {
      error (0, "write to non-existent internal memory location %02" PRIX8,
	     addr);
      emulation_result |= ER_WARNING;
    }
}

/* Read source operand of insn ending in 0x4 .. 0xF in MI;
   returns: source operand value */
static uint8_t
read_insn_source (struct mi *mi)
{
  uint8_t insn;

  insn = mi->insn[0];
  switch (insn & 0x0F)
    {
    case 0x04: /* #data */
      return mi->insn[1];
      
    case 0x05: /* Direct */
      return read_direct (mi, mi->insn[1]);

    case 0x06: case 0x07: /* @Ri */
      return read_indirect (mi, mi->regs[insn & 0x01]);
      
    case 0x08: case 0x09: case 0x0A: case 0x0B: case 0x0C: case 0x0D:
    case 0x0E: case 0x0F: /* Rn */
      return mi->regs[insn & 0x07];

    default:
      assert ("BUG" == NULL);
      return 0;
    }
}

/* Read destination operand of insn ending in 0x4 .. 0xF in MI;
   returns: destination operand value */
inline static uint8_t
read_insn_dest (struct mi *mi)
{
  uint8_t insn;

  insn = mi->insn[0];
  switch (insn & 0x0F)
    {
    case 0x04: /* A */
      return mi->sfr[SFR_ACC];
      
    case 0x05: /* Direct */
      return read_direct (mi, mi->insn[1]);

    case 0x06: case 0x07: /* @Ri */
      return read_indirect (mi, mi->regs[insn & 0x01]);
      
    case 0x08: case 0x09: case 0x0A: case 0x0B: case 0x0C: case 0x0D:
    case 0x0E: case 0x0F: /* Rn */
      return mi->regs[insn & 0x07];
      
    default:
      assert ("BUG" == NULL);
      return 0;
    }
}

/* Read destination operand of insn ending in 0x4 .. 0xF in MI, read output
   latch instead of input value on Pi;
   returns: destination operand value */
static uint8_t
read_insn_dest_latch (struct mi *mi)
{
  uint8_t insn;

  insn = mi->insn[0];
  switch (insn & 0x0F)
    {
    case 0x04: /* A */
      return mi->sfr[SFR_ACC];
      
    case 0x05: /* Direct */
      return read_direct_latch (mi, mi->insn[1]);

    case 0x06: case 0x07: /* @Ri */
      return read_indirect (mi, mi->regs[insn & 0x01]);
      
    case 0x08: case 0x09: case 0x0A: case 0x0B: case 0x0C: case 0x0D:
    case 0x0E: case 0x0F: /* Rn */
      return mi->regs[insn & 0x07];
      
    default:
      assert ("BUG" == NULL);
      return 0;
    }
}

/* Write VAL to destination operand of insn ending in 0x4 .. 0xF in MI */
static void
write_insn_dest (struct mi *mi, uint8_t val)
{
  uint8_t insn;

  insn = mi->insn[0];
  switch (insn & 0x0F)
    {
    case 0x04: /* A */
      write_acc (mi, val);
      break;
      
    case 0x05: /* Direct */
      write_direct (mi, mi->insn[1], val);
      break;

    case 0x06: case 0x07: /* @Ri */
      write_indirect (mi, mi->regs[insn & 0x01], val);
      break;
      
    case 0x08: case 0x09: case 0x0A: case 0x0B: case 0x0C: case 0x0D:
    case 0x0E: case 0x0F: /* Rn */
      mi->regs[insn & 0x07] = val;
      break;
      
    default:
      assert ("BUG" == NULL);
    }
}

/* Get byte address of BIT */
static uint8_t
bit_address (uint8_t bit)
{
  if (bit < 0x80)
    return 0x20 + bit / 8;
  else
    return bit & 0xF8;
}

/* Pop stack in MI;
   returns: popped value */
static uint8_t
stack_pop (struct mi *mi)
{
  uint8_t val;

  val = read_indirect (mi, mi->sfr[SFR_SP]);
  if (mi->sfr[SFR_SP]-- == 0)
    {
      error (0, "stack underflow");
      emulation_result |= ER_WARNING;
    }
  return val;
}

/* Push VAL on stack in MI */
static void
stack_push (struct mi *mi, uint8_t val)
{
  if (++mi->sfr[SFR_SP] == 0)
    {
      error (0, "stack overflow");
      emulation_result |= ER_WARNING;
    }
  write_indirect (mi, mi->sfr[SFR_SP], val);
}

/* Emulate S4P1 or S1P1 (just fetch instruction byte) in MI */
static void
s14p1 (void *mi_)
{
  struct mi *mi;
  
  mi = mi_;
  read_insn (mi);
}

/* Emulate generic S6P2 (finis instruction) in MI */
static void
s6p2 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  finish_insn (mi, 1);
}

/* Emulate S5P1 in one-cycle instruction in MI */
static void
oc_s5p1 (struct mi *mi)
{
  switch (mi->insn[0])
    {
    case 0x24: case 0x25: case 0x26: case 0x27: case 0x28: case 0x29:
    case 0x2A: case 0x2B: case 0x2C: case 0x2D: case 0x2E: case 0x2F:
      /* ADD A, source */
    case 0x34: case 0x35: case 0x36: case 0x37: case 0x38: case 0x39:
    case 0x3A: case 0x3B: case 0x3C: case 0x3D: case 0x3E: case 0x3F:
      /* ADDC A, source */
    case 0x44: case 0x45: case 0x46: case 0x47: case 0x48: case 0x49:
    case 0x4A: case 0x4B: case 0x4C: case 0x4D: case 0x4E: case 0x4F:
      /* ORL A, source */
    case 0x54: case 0x55: case 0x56: case 0x57: case 0x58: case 0x59:
    case 0x5A: case 0x5B: case 0x5C: case 0x5D: case 0x5E: case 0x5F:
      /* ANL A, source */
    case 0x64: case 0x65: case 0x66: case 0x67: case 0x68: case 0x69:
    case 0x6A: case 0x6B: case 0x6C: case 0x6D: case 0x6E: case 0x6F:
      /* XRL A, source */
    case 0x94: case 0x95: case 0x96: case 0x97: case 0x98: case 0x99:
    case 0x9A: case 0x9B: case 0x9C: case 0x9D: case 0x9E: case 0x9F:
      /* SUBB A, source */
      mi->tmp1 = mi->sfr[SFR_ACC];
      mi->tmp2 = read_insn_source (mi);
      break;

    case 0xE5: case 0xE6: case 0xE7: case 0xE8: case 0xE9: case 0xEA:
    case 0xEB: case 0xEC: case 0xED: case 0xEE: case 0xEF: /* MOV A, source */
      mi->tmp1 = read_insn_source (mi);
      break;

    case 0x42: /* ORL Direct, A */
    case 0x52: /* ANL Direct, A */
    case 0x62: /* XRL Direct, A */
      mi->tmp1 = read_direct_latch (mi, mi->insn[1]);
      mi->tmp2 = mi->sfr[SFR_ACC];
      break;

    case 0x03: /* RR A */
    case 0x13: /* RRC A */
    case 0x23: /* RL A */
    case 0x33: /* RLC A */
    case 0xC4: /* SWAP */
    case 0xD4: /* DA A */
    case 0xF4: /* CPL A */
    case 0xF5: case 0xF6: case 0xF7: case 0xF8: case 0xF9: case 0xFA:
    case 0xFB: case 0xFC: case 0xFD: case 0xFE: case 0xFF: /* MOV dest, A */
      mi->tmp1 = mi->sfr[SFR_ACC];
      break;

    case 0xE4: /* CLR A */
      mi->tmp1 = 0; /* Then like MOV A, source */
      break;

    case 0x04: case 0x05: case 0x06: case 0x07: case 0x08: case 0x09:
    case 0x0A: case 0x0B: case 0x0C: case 0x0D: case 0x0E: case 0x0F:
      /* INC dest */
    case 0x14: case 0x15: case 0x16: case 0x17: case 0x18: case 0x19:
    case 0x1A: case 0x1B: case 0x1C: case 0x1D: case 0x1E: case 0x1F:
      /* DEC dest */
      mi->tmp1 = read_insn_dest_latch (mi);
      break;

    case 0xC5: case 0xC6: case 0xC7: case 0xC8: case 0xC9: case 0xCA:
    case 0xCB: case 0xCC: case 0xCD: case 0xCE: case 0xCF: /* XCH A, dest */
    case 0xD6: case 0xD7: /* XCHD A, @Ri */
      mi->tmp1 = read_insn_dest (mi);
      mi->tmp2 = mi->sfr[SFR_ACC];
      break;
      
    case 0x74: case 0x76: case 0x77: case 0x78: case 0x79: case 0x7A:
    case 0x7B: case 0x7C: case 0x7D: case 0x7E: case 0x7F:
      /* MOV dest, #data */
      mi->tmp1 = mi->insn[1];
      break;

    case 0xA2: /* MOV C, bit */
      mi->tmp1 = read_direct (mi, bit_address (mi->insn[1]));
      break;
      
    case 0xB2: /* CPL bit */
    case 0xC2: /* CLR bit */
    case 0xD2: /* SETB bit */
      mi->tmp1 = read_direct_latch (mi, bit_address (mi->insn[1]));
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Emulate S6P2 in one-cycle instruction in MI;
   Most instruction are "finished" (and their results are visible) in S6P2,
   that's at least what we pretend. */
static void
oc_s6p2 (void *mi_)
{
  struct mi *mi;
  
  mi = mi_;
  switch (mi->insn[0])
    {
    case 0x00: /* NOP */
      break;
      
    case 0x24: case 0x25: case 0x26: case 0x27: case 0x28: case 0x29:
    case 0x2A: case 0x2B: case 0x2C: case 0x2D: case 0x2E: case 0x2F:
      /* ADD A, source */
    case 0x34: case 0x35: case 0x36: case 0x37: case 0x38: case 0x39:
    case 0x3A: case 0x3B: case 0x3C: case 0x3D: case 0x3E: case 0x3F:
      /* ADDC A, source */
      {
	unsigned val, old;
	
	old = mi->tmp1;
	val = mi->tmp1 + mi->tmp2;
	/* ADDC: add SFR_PSW_C */
	if ((mi->insn[0] & 0x10) != 0 && (mi->sfr[SFR_PSW] & SFR_PSW_C) != 0)
	  val++;
	mi->sfr[SFR_PSW] &= ~(SFR_PSW_C | SFR_PSW_AC | SFR_PSW_OV);
	if (val > 0xFF)
	  mi->sfr[SFR_PSW] |= SFR_PSW_C;
	if ((old & 0x08) != 0 && (val & 0x08) == 0)
	  mi->sfr[SFR_PSW] |= SFR_PSW_AC;
	if (((old & 0x40) != 0 && (val & 0x40) == 0)
	    ^ ((old & 0x80) != 0 && (val & 0x80) == 0))
	  mi->sfr[SFR_PSW] |= SFR_PSW_OV;
	write_acc (mi, val);
	break;
      }
      
    case 0x44: case 0x45: case 0x46: case 0x47: case 0x48: case 0x49:
    case 0x4A: case 0x4B: case 0x4C: case 0x4D: case 0x4E: case 0x4F:
      /* ORL A, source */
      write_acc (mi, mi->tmp1 | mi->tmp2);
      break;
      
    case 0x54: case 0x55: case 0x56: case 0x57: case 0x58: case 0x59:
    case 0x5A: case 0x5B: case 0x5C: case 0x5D: case 0x5E: case 0x5F:
      /* ANL A, source */
      write_acc (mi, mi->tmp1 & mi->tmp2);
      break;

    case 0x64: case 0x65: case 0x66: case 0x67: case 0x68: case 0x69:
    case 0x6A: case 0x6B: case 0x6C: case 0x6D: case 0x6E: case 0x6F:
      /* XRL A, source */
      write_acc (mi, mi->tmp1 ^ mi->tmp2);
      break;

    case 0x94: case 0x95: case 0x96: case 0x97: case 0x98: case 0x99:
    case 0x9A: case 0x9B: case 0x9C: case 0x9D: case 0x9E: case 0x9F:
      /* SUBB A, source */
      {
	unsigned val, old;

	val = mi->tmp1 - mi->tmp2;
	if ((mi->sfr[SFR_PSW] & SFR_PSW_C) != 0)
	  val--;
	mi->sfr[SFR_PSW] &= ~(SFR_PSW_C | SFR_PSW_AC | SFR_PSW_OV);
	old = mi->tmp1;
	/* Negative values wrap around to positive (regardless of sign
	   representation). */
	if (val > 0xFF)
	  mi->sfr[SFR_PSW] |= SFR_PSW_C;
	if ((old & 0x08) == 0 && (val & 0x08) != 0)
	  mi->sfr[SFR_PSW] |= SFR_PSW_AC;
	if (((old & 0x40) == 0 && (val & 0x40) != 0)
	    ^ ((old & 0x80) == 0 && (val & 0x80) != 0))
	  mi->sfr[SFR_PSW] |= SFR_PSW_OV;
	write_acc (mi, val);
	break;
      }

    case 0xE4: /* CLR A */
    case 0xE5: case 0xE6: case 0xE7: case 0xE8: case 0xE9: case 0xEA:
    case 0xEB: case 0xEC: case 0xED: case 0xEE: case 0xEF: /* MOV A, source */
      write_acc (mi, mi->tmp1);
      break;

    case 0x42: /* ORL Direct, A */
      write_direct (mi, mi->insn[1], mi->tmp1 | mi->tmp2);
      break;
      
    case 0x52: /* ANL Direct, A */
      write_direct (mi, mi->insn[1], mi->tmp1 & mi->tmp2);
      break;
      
    case 0x62: /* XRL Direct, A */
      write_direct (mi, mi->insn[1], mi->tmp1 ^ mi->tmp2);
      break;

    case 0xC4: /* SWAP */
      write_acc (mi, mi->tmp1 >> 4 | mi->tmp1 << 4);
      break;

    case 0xD4: /* DA A */
      {
	unsigned val;

	val = mi->tmp1;
	if ((val & 0x0F) > 0x09 || (mi->sfr[SFR_PSW] & SFR_PSW_AC) != 0)
	  val += 0x06;
	if ((val & 0xF0) > 0x90 || (mi->sfr[SFR_PSW] & SFR_PSW_C) != 0)
	  val += 0x60;
	if (val > 0x99)
	  mi->sfr[SFR_PSW] |= SFR_PSW_C;
	write_acc (mi, val);
	break;
      }

    case 0xF4: /* CPL A */
      write_acc (mi, ~mi->tmp1);
      break;

    case 0xB3: /* CPL C */
      mi->sfr[SFR_PSW] ^= SFR_PSW_C;
      break;

    case 0xC3: /* CLR C */
      mi->sfr[SFR_PSW] &= ~SFR_PSW_C;
      break;

    case 0xD3: /* SETB C */
      mi->sfr[SFR_PSW] |= SFR_PSW_C;
      break;

    case 0x04: case 0x05: case 0x06: case 0x07: case 0x08: case 0x09:
    case 0x0A: case 0x0B: case 0x0C: case 0x0D: case 0x0E: case 0x0F:
      /* INC dest */
      write_insn_dest (mi, mi->tmp1 + 1);
      break;

    case 0x14: case 0x15: case 0x16: case 0x17: case 0x18: case 0x19:
    case 0x1A: case 0x1B: case 0x1C: case 0x1D: case 0x1E: case 0x1F:
      /* DEC dest */
      write_insn_dest (mi, mi->tmp1 - 1);
      break;

    case 0x74: case 0x76: case 0x77: case 0x78: case 0x79: case 0x7A:
    case 0x7B: case 0x7C: case 0x7D: case 0x7E: case 0x7F:
      /* MOV dest, #data (mi->insn[1] doesn't matter) */
    case 0xF5: case 0xF6: case 0xF7: case 0xF8: case 0xF9: case 0xFA:
    case 0xFB: case 0xFC: case 0xFD: case 0xFE: case 0xFF: /* MOV dest, A */
      write_insn_dest (mi, mi->tmp1);
      break;

    case 0xA2: /* MOV C, bit */
      if ((mi->tmp1 & (1 << (mi->insn[1] % 8))) != 0)
	mi->sfr[SFR_PSW] |= SFR_PSW_C;
      else
	mi->sfr[SFR_PSW] &= ~SFR_PSW_C;
      break;

    case 0xB2: /* CPL bit */
      write_direct (mi, bit_address (mi->insn[1]),
		    mi->tmp1 ^ (1 << (mi->insn[1] % 8)));
      break;

    case 0xC2: /* CLR bit */
      write_direct (mi, bit_address (mi->insn[1]),
		    mi->tmp1 & ~(1 << (mi->insn[1] % 8)));
      break;

    case 0xD2: /* SETB bit */
      write_direct (mi, bit_address (mi->insn[1]),
		    mi->tmp1 | (1 << (mi->insn[1] % 8)));
      break;

    case 0x03: /* RR A */
      {
	uint8_t val;

	val = (mi->tmp1 & 0x01) != 0 ? 0x80: 0x00;
	val |= mi->tmp1 >> 1;
	write_acc (mi, val);
	break;
      }

    case 0x13: /* RRC A */
      {
	uint8_t val;

	val = (mi->sfr[SFR_PSW] & SFR_PSW_C) != 0 ? 0x80: 0x00;
	if ((mi->tmp1 & 0x01) != 0)
	  mi->sfr[SFR_PSW] |= SFR_PSW_C;
	else
	  mi->sfr[SFR_PSW] &= ~SFR_PSW_C;
	val |= mi->tmp1 >> 1;
	write_acc (mi, val);
	break;
      }

    case 0x23: /* RL A */
      {
	uint8_t val;

	val = (mi->tmp1 & 0x80) != 0 ? 0x01: 0x00;
	val |= mi->tmp1 << 1;
	write_acc (mi, val);
	break;
      }

    case 0x33: /* RLC A */
      {
	uint8_t val;

	val = (mi->sfr[SFR_PSW] & SFR_PSW_C) != 0 ? 0x01: 0x00;
	if ((mi->tmp1 & 0x80) != 0)
	  mi->sfr[SFR_PSW] |= SFR_PSW_C;
	else
	  mi->sfr[SFR_PSW] &= ~SFR_PSW_C;
	val |= mi->tmp1 << 1;
	write_acc (mi, val);
	break;
      }

    case 0xC5: case 0xC6: case 0xC7: case 0xC8: case 0xC9: case 0xCA:
    case 0xCB: case 0xCC: case 0xCD: case 0xCE: case 0xCF: /* XCH A, dest */
      write_insn_dest (mi, mi->tmp2);
      write_acc (mi, mi->tmp1);
      break;

    case 0xD6: case 0xD7: /* XCHD A, @Ri */
      write_insn_dest (mi, (mi->tmp1 & 0xF0) | (mi->tmp2 & 0x0F));
      write_acc (mi, (mi->tmp2 & 0xF0) | (mi->tmp1 & 0x0F));
      break;

    default:
      assert ("BUG" == NULL);
    }
  finish_insn (mi, 1);
}

/* Forward declaration */
static void tc_c2s6p2 (void *mi_);

/* Emulate C2S5P1 in two-cycle instruction in MI */
static void
tc_s5p1 (struct mi *mi)
{
  switch (mi->insn[0])
    {
    case 0x43: /* ORL Direct, #data */
    case 0x53: /* ANL Direct, #data */
    case 0x63: /* XRL Direct, #data */
      mi->tmp1 = read_direct_latch (mi, mi->insn[1]);
      mi->tmp2 = mi->insn[2];
      break;

    case 0x72: /* ORL C, bit */
    case 0x82: /* ANL C, bit */
    case 0xA0: /* ORL C, /bit */
    case 0xB0: /* ANL C, /bit */
      mi->tmp1 = read_direct (mi, bit_address (mi->insn[1]));
      break;

    case 0x92: /* MOV bit, C */
      mi->tmp1 = read_direct_latch (mi, bit_address (mi->insn[1]));
      break;

    case 0x85: /* MOV Direct, source */
    case 0x86: case 0x87: case 0x88: case 0x89: case 0x8A: case 0x8B:
    case 0x8C: case 0x8D: case 0x8E: case 0x8F: /* MOV Direct, source */
      mi->tmp1 = read_insn_source (mi);
      break;

    case 0xA6: case 0xA7: case 0xA8: case 0xA9: case 0xAA: case 0xAB:
    case 0xAC: case 0xAD: case 0xAF: /* MOV dest, Direct */
    case 0xC0: /* PUSH Direct */
      mi->tmp1 = read_direct (mi, mi->insn[1]);
      break;

    case 0x75: /* MOV dest, #data */
      mi->tmp1 = mi->insn[2];
      break;

    case 0xA3: /* INC DPTR */
      mi->tmp1 = mi->sfr[SFR_DPL];
      mi->tmp2 = mi->sfr[SFR_DPH];
      break;
      
    case 0xD0: /* POP Direct */
      mi->tmp1 = stack_pop (mi);
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Emulate C1S6P2 in two-cycle instruction in MI */
static void
tc_c1s6p2 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  if (mi->insn[0] != 0x90)
    mi->s5p1_fn = tc_s5p1;
  schedule_add (mi->s6p2_event, 12 * mi->phase_len, tc_c2s6p2);
}
     
/* Emulate C2S6P2 in two-cycle instruction in MI */
static void
tc_c2s6p2 (void *mi_)
{
  struct mi *mi;
  
  mi = mi_;
  switch (mi->insn[0])
    {
    case 0x43: /* ORL Direct, #data */
      write_direct (mi, mi->insn[1], mi->tmp1 | mi->tmp2);
      break;

    case 0x53: /* ANL Direct, #data */
      write_direct (mi, mi->insn[1], mi->tmp1 & mi->tmp2);
      break;
      
    case 0x63: /* XRL Direct, #data */
      write_direct (mi, mi->insn[1], mi->tmp1 ^ mi->tmp2);
      break;

    case 0x72: /* ORL C, bit */
      if ((mi->tmp1 & (1 << (mi->insn[1] % 8))) != 0)
	mi->sfr[SFR_PSW] |= SFR_PSW_C;
      break;

    case 0x82: /* ANL C, bit */
      if ((mi->tmp1 & (1 << (mi->insn[1] % 8))) == 0)
	mi->sfr[SFR_PSW] &= ~SFR_PSW_C;
      break;

    case 0xA0: /* ORL C, /bit */
      if ((mi->tmp1 & (1 << (mi->insn[1] % 8))) == 0)
	mi->sfr[SFR_PSW] |= SFR_PSW_C;
      break;

    case 0xB0: /* ANL C, /bit */
      if ((mi->tmp1 & (1 << (mi->insn[1] % 8))) != 0)
	mi->sfr[SFR_PSW] &= ~SFR_PSW_C;
      break;

    case 0x85: /* MOV Direct, source */
      write_direct (mi, mi->insn[2], mi->tmp1);
      break;
      
    case 0x86: case 0x87: case 0x88: case 0x89: case 0x8A: case 0x8B:
    case 0x8C: case 0x8D: case 0x8E: case 0x8F: /* MOV Direct, source */
    case 0xD0: /* POP Direct */
      write_direct (mi, mi->insn[1], mi->tmp1);
      break;
      
    case 0xA6: case 0xA7: case 0xA8: case 0xA9: case 0xAA: case 0xAB:
    case 0xAC: case 0xAD: case 0xAF: /* MOV dest, Direct */
      /* The BYTE param doesn't matter */
    case 0x75: /* MOV dest, #data */
      write_insn_dest (mi, mi->tmp1);
      break;

    case 0x92: /* MOV bit, C */
      {
	uint8_t tmp, mask;

	tmp = mi->tmp1;
	mask = 1 << (mi->insn[1] % 8);
	if ((mi->sfr[SFR_PSW] & SFR_PSW_C) != 0)
	  tmp |= mask;
	else
	  tmp &= ~mask;
	write_direct (mi, bit_address (mi->insn[1]), tmp);
	break;
      }
      
    case 0x90: /* MOV DPTR, #data16 */
      mi->sfr[SFR_DPH] = mi->insn[1];
      mi->sfr[SFR_DPL] = mi->insn[2];
      break;

    case 0xA3: /* INC DPTR */
      mi->sfr[SFR_DPL] = mi->tmp1 + 1;
      mi->sfr[SFR_DPH] = mi->tmp2;
      if (mi->sfr[SFR_DPL] == 0)
	mi->sfr[SFR_DPH]++;
      break;
      
    case 0xC0: /* PUSH Direct */
      stack_push (mi, mi->tmp1);
      break;

    default:
      assert ("BUG" == NULL);
    }
  finish_insn (mi, 1);
}

/* Forward declarations */
static void movc_s4p1 (void *mi_);
static void movc_c2s6p2 (void *mi_);

/* Emulate C1S5P1 in MOVC instruction in MI */
static void
movc_s5p1 (struct mi *mi)
{
  unsigned long base;

  if (mi->insn[0] == 0x93)
    base = mi->sfr[SFR_DPL] | mi->sfr[SFR_DPH] << 8;
  else
    base = mi->pc;
  base += mi->sfr[SFR_ACC];
  if (base > 0xFFFF)
    {
      error (0, "program memory adress wrapped around");
      emulation_result |= ER_WARNING;
      base &= 0xFFFF;
    }
  mi->tmp1 = base & 0xFF;
  mi->tmp2 = base >> 8;
}

/* Emulate C1S6P2 in MOVC instruction in MI */
static void
movc_c1s6p2 (void *mi_)
{
  struct mi *mi;
  uint16_t pc;

  mi = mi_;
  pc = mi->pc;
  mi->pc = mi->tmp1 | mi->tmp2 << 8;
  mi->tmp1 = pc & 0xFF;
  mi->tmp2 = pc >> 8;
  schedule_add (mi->s4p1_event, 9 * mi->phase_len, movc_s4p1);
  schedule_add (mi->s6p2_event, 12 * mi->phase_len, movc_c2s6p2);
}

/* Emulate C2S4P1 in MOVC instruction in MI */
static void
movc_s4p1 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  read_insn (mi);
  mi->pc = mi->tmp1 | mi->tmp2 << 8;
}

/* Emulate C2S6P2 in MOVC instruction in MI */
static void
movc_c2s6p2 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  write_acc (mi, mi->insn[1]);
  finish_insn (mi, 1);
}

/* Emulate C1S5P1 in MUL or DIV instruction in MI */
static void
muldiv_s5p1 (struct mi *mi)
{
  mi->tmp1 = mi->sfr[SFR_ACC];
  mi->tmp2 = mi->sfr[SFR_B];
}

/* Emulate C4S6P2 in MUL or DIV instruction in MI */
static void
muldiv_s6p2 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  if (mi->insn[0] == 0x84)
    {
      uint8_t a, b;
      
      a = mi->tmp1;
      b = mi->tmp2;
      if (b != 0)
	{
	  write_acc (mi, a / b);
	  mi->sfr[SFR_B] = a % b;
	  mi->sfr[SFR_PSW] &= ~(SFR_PSW_C | SFR_PSW_OV);
	}
      else
	{
	  /* FIXME: what values do A and B get? */
	  mi->sfr[SFR_PSW] = (mi->sfr[SFR_PSW] & ~SFR_PSW_C) | SFR_PSW_OV;
	  error (0, "divide by zero");
	  emulation_result |= ER_WARNING;
	}
    }
  else
    {
      unsigned val;
      
      val = mi->tmp1 * mi->tmp2;
      write_acc (mi, val);
      val >>= 8;
      mi->sfr[SFR_B] = val;
      if (val != 0)
	mi->sfr[SFR_PSW] = (mi->sfr[SFR_PSW] & ~SFR_PSW_C) | SFR_PSW_OV;
      else
	mi->sfr[SFR_PSW] &= ~(SFR_PSW_C | SFR_PSW_OV);
    }
  finish_insn (mi, 1);
}

/* Emulate C1S5P1 in relative jump instruction in MI */
static void
rj_s5p1 (struct mi *mi)
{
  switch (mi->insn[0])
    {
    case 0xB4: case 0xB5: /* CJNE A, source, rel */
      mi->tmp1 = mi->sfr[SFR_ACC];
      mi->tmp2 = read_insn_source (mi);
      break;

    case 0xB6: case 0xB7: case 0xB8: case 0xB9: case 0xBA: case 0xBB:
    case 0xBC: case 0xBD: case 0xBE: case 0xBF: /* CJNE source, #data, rel */
      mi->tmp1 = read_insn_source (mi);
      mi->tmp2 = mi->insn[1];
      break;

    case 0xD5: case 0xD8: case 0xD9: case 0xDA: case 0xDB: case 0xDC:
    case 0xDD: case 0xDE: case 0xDF: /* DJNZ dest, rel */
      mi->tmp1 = read_insn_dest_latch (mi);
      break;

    case 0x10: /* JBC bit, rel */
      mi->tmp1 = read_direct_latch (mi, bit_address (mi->insn[1]));
      break;

    case 0x20: /* JB bit, rel */
    case 0x30: /* JNB bit, rel */
      mi->tmp1 = read_direct (mi, bit_address (mi->insn[1]));
      break;

    case 0x40: /* JC rel */
    case 0x50: /* JNC rel */
    case 0x60: /* JZ rel */
    case 0x70: /* JNZ rel */
    case 0x80: /* SJMP rel */
      break;
      
    default:
      assert ("BUG" == NULL);
    }
}

/* Emulate C1S6P2 in relative jump instruction in MI */
static void
rj_c1s6p2 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  /* At output of this switch, mi->tmp1 != 0 if we are taking the jump */
  switch (mi->insn[0])
    {
    case 0xB4: case 0xB5: /* CJNE A, source, rel */
    case 0xB6: case 0xB7: case 0xB8: case 0xB9: case 0xBA: case 0xBB:
    case 0xBC: case 0xBD: case 0xBE: case 0xBF: /* CJNE source, #data, rel */
      if (mi->tmp1 < mi->tmp2)
	mi->sfr[SFR_PSW] |= SFR_PSW_C;
      else
	mi->sfr[SFR_PSW] &= ~SFR_PSW_C;
      mi->tmp1 ^= mi->tmp2; /* ((tmp1 ^ tmp2) != 0) == (tmp1 != tmp2) */
      break;

    case 0xD5: case 0xD8: case 0xD9: case 0xDA: case 0xDB: case 0xDC:
    case 0xDD: case 0xDE: case 0xDF: /* DJNZ dest, rel */
      mi->tmp1--;
      write_insn_dest (mi, mi->tmp1);
      break;

    case 0x10: /* JBC bit, rel */
      {
	uint8_t mask;

	mask = 1 << (mi->insn[1] % 8);
	if ((mi->tmp1 & mask) != 0)
	  write_direct (mi, bit_address (mi->insn[1]), mi->tmp1 & ~mask);
	mi->tmp1 &= mask;
	break;
      }

    case 0x20: /* JB bit, rel */
      mi->tmp1 &= 1 << (mi->insn[1] % 8);
      break;
      
    case 0x30: /* JNB bit, rel */
      mi->tmp1 = ~mi->tmp1 & (1 << (mi->insn[1] % 8));
      break;

    case 0x40: /* JC rel */
      mi->tmp1 = mi->sfr[SFR_PSW] & SFR_PSW_C;
      break;
      
    case 0x50: /* JNC rel */
      mi->tmp1 = ~mi->sfr[SFR_PSW] & SFR_PSW_C;
      break;

    case 0x60: /* JZ rel */
      mi->tmp1 = mi->sfr[SFR_ACC] == 0;
      break;
      
    case 0x70: /* JNZ rel */
      mi->tmp1 = mi->sfr[SFR_ACC];
      break;
      
    case 0x80: /* SJMP rel */
      mi->tmp1 = 1;
      break;
      
    default:
      assert ("BUG" == NULL);
    }
  schedule_add (mi->s6p2_event, 12 * mi->phase_len, s6p2);
}

/* Emulate C2S3P2 in relative jump instruction in MI */
static void
rj_s3p2 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  if (mi->tmp1 != 0)
    {
      int8_t disp; /* Signed! */
      
      disp = mi->insn_next[-1];
      if ((disp > 0 && (uint16_t)(mi->pc + disp) < mi->pc)
	  || (disp < 0 && (uint16_t)(mi->pc + disp) > mi->pc))
	{
	  error (0, "PC wrapped around");
	  emulation_result |= ER_WARNING;
	}
      mi->pc += disp;
    }
}

/* Emulate C2S3P2 in absolute jump / call instruction in MI;
   I don't suppose that all stack accesses etc. are done in this phase,
   I have no idea when they are done and they should have no external
   effects, so we might as well cluster them here. */
static void
aj_s3p2 (void *mi_)
{
  struct mi *mi;
  uint16_t new_pc;

  mi = mi_;
  switch (mi->insn[0])
    {
    case 0x11: case 0x31: case 0x51: case 0x71: case 0x91: case 0xB1:
    case 0xD1: case 0xF1: /* ACALL Addr */
      stack_push (mi, mi->pc & 0xFF);
      stack_push (mi, mi->pc >> 8);
      /* Fall through */
    case 0x01: case 0x21: case 0x41: case 0x61: case 0x81: case 0xA1:
    case 0xC1: case 0xE1: /* AJMP Addr */
      new_pc = (mi->pc & 0xF800) | (mi->insn[0] & 0xE0) << 3 | mi->insn[1];
      break;
      
    case 0x12: /* LCALL Addr */
      stack_push (mi, mi->pc & 0xFF);
      stack_push (mi, mi->pc >> 8);
      /* Fall through */
    case 0x02: /* LJMP Addr */
      new_pc = mi->insn[1] << 8 | mi->insn[2];
      break;

    case 0x32: /* RETI */
      mi->flags |= MI_NO_INTR;
      if ((mi->flags & MI_INTR_HIGH) != 0)
	mi->flags &= ~MI_INTR_HIGH;
      else if ((mi->flags & MI_INTR_LOW) != 0)
	mi->flags &= ~MI_INTR_LOW;
    case 0x22: /* RET */
      new_pc = stack_pop (mi) << 8;
      new_pc |= stack_pop (mi);
      break;
      
    case 0x73: /* JMP @A + DPTR */
      {
	unsigned long v;
	
	v = (mi->sfr[SFR_DPH] << 8 | mi->sfr[SFR_DPL]) + mi->sfr[SFR_ACC];
	if (v > 0xFFFF)
	  {
	    error (0, "PC wrapped arround");
	    emulation_result |= ER_WARNING;
	  }
	new_pc = v;
	break;
      }
      
    default:
      assert ("BUG" == NULL);
      return; /* To make GCC happy */
    }
  mi->pc = new_pc;
}

/* Forward declarations */
static void movx_c2s1p1 (void *mi_);
static void movx_c2s3p1 (void *mi_);
static void movx_c2s6p2 (void *mi_);

/* Emulate C1S3P2 in MOVX instruction in MI */
static void
movx_c1s3p2 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  mi->flags |= MI_MOVX_1;
}

/* Emulate C1S6P2 in MOVX instruction in MI */
static void
movx_c1s6p2 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  /* Skip one ALE cycle */
  schedule_cancel (mi->s14p1_event);
  schedule_cancel (mi->s14p2_event);
  schedule_cancel (mi->s25p1_event);
  schedule_cancel (mi->s25p2_event);
  schedule_add (mi->s14p1_event, 7 * mi->phase_len, i8051_emulate_s14p1);
  schedule_add (mi->s14p2_event, 8 * mi->phase_len, i8051_emulate_s14p2);
  schedule_add (mi->s25p1_event, 9 * mi->phase_len, i8051_emulate_s25p1);
  schedule_add (mi->s25p2_event, 10 * mi->phase_len, i8051_emulate_s25p2);
  if ((mi->insn[0] & 0x10) == 0)
    i8051_port_write (mi, I8051_S1P_P0, 0xFF);
  else if (i8051_port_write (mi, I8051_S1P_P0, mi->sfr[SFR_ACC]) == 0)
    {
      error (0, "external data memory not connected for writing");
      emulation_result |= ER_WARNING;
    }
  schedule_add (mi->s1p1_event, mi->phase_len, movx_c2s1p1);
  /* Using existing S3P2 event for S3P1 */
  schedule_add (mi->s3p2_event, 5 * mi->phase_len, movx_c2s3p1);
  schedule_add (mi->s6p2_event, 12 * mi->phase_len, movx_c2s6p2);
}

/* Emulate C2S3P1 in MOVX instruction in MI */
static void
movx_c2s1p1 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  mi->port_3_mask &= (mi->insn[0] & 0x10) == 0 ? ~SFR_P3_RD : ~SFR_P3_WR;
  /* Update for changed port_3_mask */
  i8051_port_write (mi, I8051_S1P_P3, mi->ports_out[3]);
}

/* Emulate C2S3P1 in MOVX instruction (writing to external memory) in MI */
static void
movx_c2s3p1 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  /* Allow reading from partially-connected external memory - it could make
     sense in some applications */
  if ((mi->insn[0] & 0x10) == 0
      && i8051_port_read (mi, I8051_S1P_P0, mi->sfr + SFR_ACC) == 0)
    {
      error (0, "external data memory not connected for reading");
      emulation_result |= ER_WARNING;
    }
  mi->flags = (mi->flags & ~MI_MOVX_1) | MI_MOVX_2;
}

/* Emulate C2S6P2 in MOVX instruction in MI */
static void
movx_c2s6p2 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  mi->flags &= ~MI_MOVX_2;
  finish_insn (mi, 1);
}

/* Emulate C1S5P1 in MOVX instruction in MI with xmem */
static void
movx_xmem_c1s5p1 (struct mi *mi)
{
  switch (mi->insn[0])
    {
    case 0xE0: /* MOVX A, @DPTR */
    case 0xE2: case 0xE3: /* MOVX A, @Ri */
      break;

    case 0xF0: /* MOVX @DPTR, A */
    case 0xF2: case 0xF3: /* MOVX @Ri, A */
      mi->tmp1 = mi->sfr[SFR_ACC];
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Emulate C2S6P2 in MOVX instruction in MI with xmem */
static void
movx_xmem_c2s6p2 (void *mi_)
{
  struct mi *mi;
  uint16_t addr;

  mi = mi_;
  if ((mi->insn[0] & 0x0F) == 0x00)
    addr = mi->sfr[SFR_DPH] << 8 | mi->sfr[SFR_DPL];
  else
    addr = mi->regs[mi->insn[0] & 0x01];
  if (addr >= mi->xmem_mem_size)
    {
      addr &= mi->xmem_mem_size - 1;
      error (0, "MOVX address too large");
      emulation_result |= ER_WARNING;
    }
  switch (mi->insn[0])
    {
    case 0xE0: /* MOVX A, @DPTR */
    case 0xE2: case 0xE3: /* MOVX A, @Ri */
      write_acc (mi, mi->xmem[addr]);
      break;
      
    case 0xF0: /* MOVX @DPTR, A */
    case 0xF2: case 0xF3: /* MOVX @Ri, A */
      mi->xmem[addr] = mi->sfr[SFR_ACC];
      break;

    default:
      assert ("BUG" == NULL);
    }
  finish_insn (mi, 1);
}

/* Emulate one instruction in MI */
void
i8051_emulate_insn (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  if ((mi->flags & MI_PROGRAM_LOADED) == 0
      && mi->pin_fns[I8051_S1P_PSEN].write.fn.write_1 == NULL)
    {
      error (0, "no program to emulate loaded");
      emulation_result |= ER_ERROR;
      return;
    }
  assert (mi->insn_next == mi->insn);
  mi->prev_pc = mi->pc;
  read_insn (mi);
  switch (mi->insn[0])
    {
      /* One-cycle (OC) instructions */
    case 0x05: case 0x15: case 0x24: case 0x25: case 0x34: case 0x35:
    case 0x42: case 0x44: case 0x45: case 0x52: case 0x54: case 0x55:
    case 0x62: case 0x64: case 0x65: case 0x74: case 0x76: case 0x77:
    case 0x78: case 0x79: case 0x7A: case 0x7B: case 0x7C: case 0x7D:
    case 0x7E: case 0x7F: case 0x94: case 0x95: case 0xA2: case 0xB2:
    case 0xC2: case 0xC5: case 0xD2: case 0xE5: case 0xF5:
      schedule_add (mi->s4p1_event, 6 * mi->phase_len, s14p1);
      /* Fall through */
    case 0x03: case 0x04: case 0x06: case 0x07: case 0x08: case 0x09:
    case 0x0A: case 0x0B: case 0x0C: case 0x0D: case 0x0E: case 0x0F:
    case 0x13: case 0x14: case 0x16: case 0x17: case 0x18: case 0x19:
    case 0x1A: case 0x1B: case 0x1C: case 0x1D: case 0x1E: case 0x1F:
    case 0x23: case 0x26: case 0x27: case 0x28: case 0x29: case 0x2A:
    case 0x2B: case 0x2C: case 0x2D: case 0x2E: case 0x2F: case 0x33:
    case 0x36: case 0x37: case 0x38: case 0x39: case 0x3A: case 0x3B:
    case 0x3C: case 0x3D: case 0x3E: case 0x3F: case 0x46: case 0x47:
    case 0x48: case 0x49: case 0x4A: case 0x4B: case 0x4C: case 0x4D:
    case 0x4E: case 0x4F: case 0x56: case 0x57: case 0x58: case 0x59:
    case 0x5A: case 0x5B: case 0x5C: case 0x5D: case 0x5E: case 0x5F:
    case 0x66: case 0x67: case 0x68: case 0x69: case 0x6A: case 0x6B:
    case 0x6C: case 0x6D: case 0x6E: case 0x6F: case 0x96: case 0x97:
    case 0x98: case 0x99: case 0x9A: case 0x9B: case 0x9C: case 0x9D:
    case 0x9E: case 0x9F: case 0xC4: case 0xC6: case 0xC7: case 0xC8:
    case 0xC9: case 0xCA: case 0xCB: case 0xCC: case 0xCD: case 0xCE:
    case 0xCF: case 0xD4: case 0xD6: case 0xD7: case 0xE4: case 0xE6:
    case 0xE7: case 0xE8: case 0xE9: case 0xEA: case 0xEB: case 0xEC:
    case 0xED: case 0xEE: case 0xEF: case 0xF4: case 0xF6: case 0xF7:
    case 0xF8: case 0xF9: case 0xFA: case 0xFB: case 0xFC: case 0xFD:
    case 0xFE: case 0xFF:
      mi->s5p1_fn = oc_s5p1;
      /* Fall through */
    case 0x00: case 0xB3: case 0xC3: case 0xD3:
      schedule_add (mi->s6p2_event, 11 * mi->phase_len, oc_s6p2);
      break;

      /* Two-cycle (TC) instructions */
    case 0x43: case 0x53: case 0x63: case 0x75: case 0x85: case 0x90:
      schedule_add (mi->s1p1_event, 12 * mi->phase_len, s14p1);
      /* Fall through */
    case 0x72: case 0x82: case 0x86: case 0x87: case 0x88: case 0x89:
    case 0x8A: case 0x8B: case 0x8C: case 0x8D: case 0x8E: case 0x8F:
    case 0x92: case 0xA0: case 0xA6: case 0xA7: case 0xA8: case 0xA9:
    case 0xAA: case 0xAB: case 0xAC: case 0xAD: case 0xAE: case 0xAF:
    case 0xB0: case 0xC0: case 0xD0:
      schedule_add (mi->s4p1_event, 6 * mi->phase_len, s14p1);
      /* Fall through */
    case 0xA3:
      schedule_add (mi->s6p2_event, 11 * mi->phase_len, tc_c1s6p2);
      break;
      
    case 0x83: case 0x93: /* MOVC A, @A + 16B */
      mi->s5p1_fn = movc_s5p1;
      schedule_add (mi->s6p2_event, 11 * mi->phase_len, movc_c1s6p2);
      /* Other events scheduled in movc_c1s6p2 */
      break;

    case 0x84: /* DIV AB */
    case 0xA4: /* MUL AB */
      mi->s5p1_fn = muldiv_s5p1;
      schedule_add (mi->s6p2_event, 47 * mi->phase_len, muldiv_s6p2);
      break;

      /* Absolute jump / call (AJ) instructions */
    case 0x02: case 0x12:
      schedule_add (mi->s1p1_event, 12 * mi->phase_len, s14p1);
      /* Fall through */
    case 0x01: case 0x11: case 0x21: case 0x31: case 0x41: case 0x51:
    case 0x61: case 0x71: case 0x81: case 0x91: case 0xA1: case 0xB1:
    case 0xC1: case 0xD1: case 0xE1: case 0xF1:
      schedule_add (mi->s4p1_event, 6 * mi->phase_len, s14p1);
      /* Fall through */
    case 0x22: case 0x32: case 0x73:
      schedule_add (mi->s3p2_event, 17 * mi->phase_len, aj_s3p2);
      schedule_add (mi->s6p2_event, 23 * mi->phase_len, s6p2);
      break;

      /* Relative jumps */
    case 0x10: case 0x20: case 0x30: case 0xB4: case 0xB5: case 0xB6:
    case 0xB7: case 0xB8: case 0xB9: case 0xBA: case 0xBB: case 0xBC:
    case 0xBD: case 0xBE: case 0xBF: case 0xD5: 
      schedule_add (mi->s1p1_event, 12 * mi->phase_len, s14p1);
      /* Fall through */
    case 0xD8: case 0xD9: case 0xDA: case 0xDB: case 0xDC: case 0xDD:
    case 0xDE: case 0xDF:
      mi->s5p1_fn = rj_s5p1;
      /* Fall through */
    case 0x40: case 0x50: case 0x60: case 0x70: case 0x80:
      schedule_add (mi->s4p1_event, 6 * mi->phase_len, s14p1);
      schedule_add (mi->s6p2_event, 11 * mi->phase_len, rj_c1s6p2);
      schedule_add (mi->s3p2_event, 17 * mi->phase_len, rj_s3p2);
      /* C2S6P2 added in rj_c1s6p2 () */
      break;

      /* MOVX instructions */
    case 0xE0: case 0xE2: case 0xE3: case 0xF0: case 0xF2: case 0xF3:
      /* FIXME: allow plugins */
      if (mi->xmem_mem_size == 0)
	{
	  schedule_add (mi->s3p2_event, 5 * mi->phase_len, movx_c1s3p2);
	  schedule_add (mi->s6p2_event, 11 * mi->phase_len, movx_c1s6p2);
	  /* C2S3P1, C2S6P2 added in movx_c1s6p2 () */
	}
      else
	{
	  mi->s5p1_fn = movx_xmem_c1s5p1;
	  schedule_add (mi->s6p2_event, 23 * mi->phase_len, movx_xmem_c2s6p2);
	}
      break;
      
    default:
      error (0, "unknown instruction");
      emulation_result |= ER_WARNING;
      /* Try to recover and continue */
      finish_insn (mi, 12);
      break;
    }
}

/* Emulate reset of in MI */
void
i8051_emulate_reset (struct mi *mi)
{
  mi->pc = 0;
  mi->sfr[SFR_P0] = 0xFF;
  mi->sfr[SFR_SP] = 0x07;
  mi->sfr[SFR_DPL] = 0x00;
  mi->sfr[SFR_DPH] = 0x00;
  mi->sfr[SFR_PCON] = 0x00;
  mi->sfr[SFR_TCON] = 0x00;
  mi->sfr[SFR_TMOD] = 0x00;
  mi->sfr[SFR_TL0] = 0x00;
  mi->sfr[SFR_TL1] = 0x00;
  mi->sfr[SFR_TH0] = 0x00;
  mi->sfr[SFR_TH1] = 0x00;
  mi->sfr[SFR_P1] = 0xFF;
  mi->sfr[SFR_SCON] = 0x00;
  mi->sfr[SFR_P2] = 0xFF;
  mi->sfr[SFR_IE] = 0x00;
  mi->sfr[SFR_P3] = 0xFF;
  mi->sfr[SFR_IP] = 0x00;
  mi->sfr[SFR_PSW] = 0x00;
  mi->sfr[SFR_ACC] = 0x00;
  mi->sfr[SFR_B] = 0x00;
}

#define FN(FN) { .fn = FN, .name = #FN }

/* Event handling function list (except for UART) */
const struct schedule_fn_desc i8051_event_list[] =
{
  FN (i8051_emulate_s14p1), FN (i8051_emulate_s14p2), FN (i8051_emulate_s25p1),
  FN (i8051_emulate_s25p2), FN (s36p1), FN (i8051_emulate_s3p1),
  FN (i8051_emulate_s5p1), FN (i8051_emulate_s5p2), FN (s14p1), FN (s6p2),
  FN (oc_s6p2), FN (tc_c1s6p2), FN (tc_c2s6p2), FN (movc_c1s6p2),
  FN (movc_s4p1), FN (movc_c2s6p2), FN (muldiv_s6p2), FN (rj_c1s6p2),
  FN (rj_s3p2), FN (aj_s3p2), FN (movx_c1s3p2), FN (movx_c1s6p2),
  FN (movx_c2s1p1), FN (movx_c2s3p1), FN (movx_c2s6p2), FN (movx_xmem_c2s6p2),
  FN (i8051_emulate_insn), 
  { .fn = NULL, .name = NULL }
};


/* List of functions that can be in mi->s5p1_fn */
const struct i8051_fn_desc i8051_s5p1_fn_list[] =
{
  FN (tc_s5p1), FN (oc_s5p1), FN (movc_s5p1), FN (muldiv_s5p1), FN (rj_s5p1),
  FN (movx_xmem_c1s5p1),
  { .fn = NULL, .name = NULL }
};
