/* Ports (in the cereal sense) handling for 8051
   Copyright (C) 2002, 2004 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <errno.h>
#include <stdlib.h>

#include <cereal.h>

#include "8051.h"

/* Get number of ports of this space in MI_ */
static size_t
memory_1_get_size (void *mi_)
{
  (void)mi_;
  return 0x100;
}

/* Read bit port value for display (unlike port_1_read_t may have no side
   effects);
   returns: port value */
static int
memory_1_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  mi = mi_;
  assert ((size_t)data < 0x100);
  if (data < 0x80)
    return (mi->memory[0x20 + data / 8] & 1 << data % 8) != 0;
  else
    {
      const struct port_fn *pf;

      pf = &mi->sfr_fns[data & 0x78].display;
      assert (pf->fn.display_1 != NULL);
      return (pf->fn.display_1 (pf->mi, pf->data) & 1 << data % 8) != 0;
    }
}

/* Save port_fn for PORT displaying to DEST */
static void
memory_1_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < 0x100);
  if (port >= 0x80 && mi->sfr_fns[port & 0x78].display.fn.display_1 == NULL)
    {
      dest->fn.display_1 = NULL;
      return;
    }
  dest->fn.display_1 = memory_1_display_fn;
  dest->mi = mi;
  dest->data = port;
}

/* Modify bit port value by user */
static void
memory_1_modify_fn (void *mi_, intptr_t data, int val)
{
  struct mi *mi;
  uint8_t mask;
      
  mi = mi_;
  assert ((size_t)data < 0x100);
  mask = 1 << data % 8;
  if (data < 0x80)
    {
      uint8_t *ptr;
      
      ptr = mi->memory + 0x20 + data / 8;
      if (val != 0)
	*ptr |= mask;
      else
	*ptr &= ~mask;
    }
  else
    {
      const struct sfr_fn *sf;
      uint8_t v;

      sf = mi->sfr_fns + (data & 0x78);
      assert (sf->display.fn.display_1 != NULL
	      && sf->modify.fn.modify_1 != NULL);
      v = sf->display.fn.display_1 (sf->display.mi, sf->display.data);
      if (val != 0)
	v |= mask;
      else
	v &= ~mask;
      sf->modify.fn.modify_1 (sf->modify.mi, sf->modify.data, v);
    }
}

/* Save port_fn for PORT modification to DEST */
static void
memory_1_get_modify_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < 0x100);
  if (port >= 0x80)
    {
      const struct sfr_fn *sf;

      sf = mi->sfr_fns + (port & 0x78);
      if (sf->display.fn.display_1 == NULL || sf->modify.fn.modify_1 == NULL)
	{
	  dest->fn.modify_1 = NULL;
	  return;
	}
    }
  dest->fn.modify_1 = memory_1_modify_fn;
  dest->mi = mi;
  dest->data = port;
}

/* Get number of ports of this space in MI_ */
static size_t
pins_1_get_size (void *mi_)
{
  (void)mi_;
  return I8051_NUM_S1P;
}

/* Read bit port value in MI_ (for emulation or display, so may have no side
   effects);
   returns: port value */
static int
pins_1_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;
  uint8_t flag;

  assert ((size_t)data < I8051_NUM_S1P);
  mi = mi_;
  if (data < I8051_S1P_MISC)
    {
      uint8_t val;
      
      val = mi->ports_out[data / 8];
      if (data >= I8051_S1P_P3 && data < I8051_S1P_P3 + 8)
	val &= mi->port_3_mask;
      return (val & (1 << (data % 8))) != 0;
    }
  switch (data)
    {
    case I8051_S1P_ALE:
      flag = PIN_ALE;
      break;

    case I8051_S1P_PSEN:
      flag = PIN_PSEN;
      break;

    default:
      assert ("BUG" == NULL);
      return 0; /* To make GCC happy */
    }
  return (mi->pins & flag) != 0;
}

/* Save port_fn for PORT reading to DEST */
static void
pins_1_get_read_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < I8051_NUM_S1P);
  if (port >= I8051_S1P_MISC)
    {
      switch (port)
	{
	case I8051_S1P_ALE: case I8051_S1P_PSEN:
	  break;
	  
	case I8051_S1P_EA:
	  dest->fn.read_1 = NULL;
	  return;

	default:
	  assert ("BUG" == NULL);
	}
    }
  dest->fn.read_1 = pins_1_display_fn;
  dest->mi = mi_;
  dest->data = port;
}

/* Write VAL to bit port in MI_ */
static void
pins_1_write_fn (void *mi_, intptr_t data, int val)
{
  struct mi *mi;

  (void)val;
  mi = mi_;
  assert ((size_t)data < I8051_NUM_S1P);
  if (data >= I8051_S1P_MISC) /* We will poll for Px bits at S5P2 */
    {
      switch (data)
	{
	case I8051_S1P_EA:
	  break; /* We will poll when needed */
	  
	default:
	  assert ("BUG" == NULL);
	}
    }
}

/* Save port_fn for PORT writing to DEST */
static void
pins_1_get_write_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < I8051_NUM_S1P);
  if (port >= I8051_S1P_MISC)
    {
      switch (port)
	{
	case I8051_S1P_ALE: case I8051_S1P_PSEN:
	  dest->fn.write_1 = NULL;
	  return;

	case I8051_S1P_EA:
	  break;
	  
	default:
	  assert ("BUG" == NULL);
	}
    }
  dest->fn.write_1 = pins_1_write_fn;
  dest->mi = mi_;
  dest->data = port;
}

/* Save port_fn for PORT displaying to DEST */
static void
pins_1_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < I8051_NUM_S1P);
  dest->fn.display_1 = pins_1_display_fn;
  dest->mi = mi_;
  dest->data = port;
}

/* Set peer read_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int
pins_1_set_read_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;

  mi = mi_;
  assert (port < I8051_NUM_S1P);
  if (port >= I8051_S1P_MISC)
    {
      switch (port)
	{
	case I8051_S1P_ALE: case I8051_S1P_PSEN:
	  return ENOTSUP;

	case I8051_S1P_EA:
	  break;
	  
	default:
	  assert ("BUG" == NULL);
	  break;
	}
    }
  pf = &mi->pin_fns[port].read;
  if (pf->fn.read_1 != NULL && src->fn.read_1 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

/* Set peer write_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int
pins_1_set_write_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;

  mi = mi_;
  assert (port < I8051_NUM_S1P);
  if (port >= I8051_S1P_MISC)
    {
      switch (port)
	{
	case I8051_S1P_ALE: case I8051_S1P_PSEN:
	  break;

	case I8051_S1P_EA:
	  return ENOTSUP;
	  
	default:
	  assert ("BUG" == NULL);
	  break;
	}
    }
  pf = &mi->pin_fns[port].write;
  if (pf->fn.write_1 != NULL && src->fn.write_1 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

/* Read current value of PORT (PORT is I8051_S1P_Pi) to *DEST;
   returns: number of connected bits */
unsigned
i8051_port_read (struct mi *mi, size_t port, uint8_t *dest)
{
  const struct pin_fn *pf;
  unsigned res;
  uint8_t val;

  res = 0;
  val = 0;
  pf = mi->pin_fns + port + 7;
  do
    {
      val <<= 1;
      if (pf->read.fn.read_1 != NULL)
	{
	  val |= pf->read.fn.read_1 (pf->read.mi, pf->read.data) != 0;
	  res++;
	}
    }
  while (pf-- != mi->pin_fns + port);
  switch (port)
    {
    case I8051_S1P_P0:
      break;

    case I8051_S1P_P1:
      val &= mi->sfr[SFR_P1];
      break;

    case I8051_S1P_P2:
      val &= mi->sfr[SFR_P2];
      break;

    case I8051_S1P_P3:
      val &= mi->sfr[SFR_P3];
      break;

    default:
      assert ("BUG" == NULL);
    }
  *dest = val;
  return res;
}

/* Set PORT (PORT is I8051_S1P_Pi) to VAL;
   returns: number of connected bits */
unsigned
i8051_port_write (struct mi *mi, size_t port, uint8_t val)
{
  size_t i;
  const struct pin_fn *pf;

  switch (port)
    {
    case I8051_S1P_P0:
      i = 0;
      break;

    case I8051_S1P_P1:
      i = 1;
      break;

    case I8051_S1P_P2:
      i = 2;
      break;

    case I8051_S1P_P3:
      i = 3;
      break;

    default:
      assert ("BUG" == NULL);
      return 0;
    }
  mi->ports_out[i] = val;
  if (port == I8051_S1P_P3)
    val &= mi->port_3_mask;
  i = 0;
  for (pf = mi->pin_fns + port; pf < mi->pin_fns + port + 8; pf++)
    {
      if (pf->write.fn.write_1 != NULL)
	{
	  pf->write.fn.write_1 (pf->write.mi, pf->write.data, val & 0x01);
	  i++;
	}
      val >>= 1;
    }
  return i;
}

/* Bit address spaces */
const struct space i8051_spaces_1[I8051_NUM_SPACES_1] =
{
  /* 8051 bit addresses */
  [I8051_SPACE_1_BITS] =
  {
    .get_size = memory_1_get_size, .get_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = memory_1_get_display_fn,
      [PORT_TYPE_MODIFY] = memory_1_get_modify_fn
    },
    .set_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
    }
  },
  /* 8051 chip pins */
  [I8051_SPACE_1_PINS] =
  {
    .get_size = pins_1_get_size, .get_fn =
    {
      [PORT_TYPE_READ] = pins_1_get_read_fn,
      [PORT_TYPE_WRITE] = pins_1_get_write_fn,
      [PORT_TYPE_DISPLAY] = pins_1_get_display_fn, [PORT_TYPE_MODIFY] = NULL
    },
    .set_fn =
    {
      [PORT_TYPE_READ] = pins_1_set_read_fn,
      [PORT_TYPE_WRITE] = pins_1_set_write_fn, [PORT_TYPE_DISPLAY] = NULL,
      [PORT_TYPE_MODIFY] = NULL
    }
  }
};

/* Get number of ports of this space in MI_ */
static size_t
memory_8_get_size (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  return mi->data_mem_size;
}

/* Read byte port value for display (unlike port_8_read_t may have no side
   effects);
   returns: port value */
static uint8_t
memory_8_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  mi = mi_;
  assert ((size_t)data < mi->data_mem_size);
  return mi->memory[data];
}

/* Save port_fn for PORT displaying to DEST */
static void
memory_8_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;
  
  mi = mi_;
  assert (port < mi->data_mem_size);
  dest->fn.display_8 = memory_8_display_fn;
  dest->mi = mi;
  dest->data = port;
}

/* Modify byte port value by user */
static void
memory_8_modify_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  mi = mi_;
  assert ((size_t)data < mi->data_mem_size);
  mi->memory[data] = val;
}

/* Save port_fn for PORT modification to DEST */
static void
memory_8_get_modify_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < mi->data_mem_size);
  dest->fn.modify_8 = memory_8_modify_fn;
  dest->mi = mi;
  dest->data = port;
}

/* Get number of ports of this space in MI_ */
static size_t
program_8_get_size (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  return mi->prog_mem_size;
}

/* Read byte port value for display (unlike port_8_read_t may have no side
   effects);
   returns: port value */
static uint8_t
program_8_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  mi = mi_;
  assert ((size_t)data < mi->prog_mem_size);
  return mi->program[data];
}

/* Save port_fn for PORT displaying to DEST */
static void
program_8_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < mi->prog_mem_size);
  dest->fn.display_8 = program_8_display_fn;
  dest->mi = mi;
  dest->data = port;
}

/* Modify byte port value by user */
static void
program_8_modify_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  mi = mi_;
  assert ((size_t)data < mi->prog_mem_size);
  mi->program[data] = val;
}

/* Save port_fn for PORT modification to DEST */
static void
program_8_get_modify_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < mi->prog_mem_size);
  dest->fn.modify_8 = program_8_modify_fn;
  dest->mi = mi;
  dest->data = port;
}

/* Get number of ports of this space in MI_ */
static size_t
sfr_8_get_size (void *mi_)
{
  (void)mi_;
  return I8051_SFR_SIZE;
}

/* Save port_fn for PORT displaying to DEST */
static void
sfr_8_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < I8051_SFR_SIZE);
  *dest = mi->sfr_fns[port].display;
}

/* Save port_fn for PORT modification to DEST */
static void
sfr_8_get_modify_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < I8051_SFR_SIZE);
  *dest = mi->sfr_fns[port].modify;
}

/* Get number of ports of this space in MI_ */
static size_t
sfr_ext_8_get_size (void *mi_)
{
  (void)mi_;
  return I8051_SFR_SIZE;
}

static int
cmp_sfr_fn (const void *xkey, const void *xskel)
{
  const size_t *key;
  const struct sfr_fn_skel *skel;

  key = xkey;
  skel = xskel;
  return (int)*key - (int)skel->reg;
}

/* Find i8051_sfr_fns entry for PORT;
   returns: i8051_sfr_fns pointer or NULL if not found */
static const struct sfr_fn_skel *
find_sfr_fn_skel (size_t port)
{
  return bsearch (&port, i8051_sfr_fns, i8051_sfr_fns_len,
		  sizeof (*i8051_sfr_fns), cmp_sfr_fn);
}

/* Set peer read_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int
sfr_ext_8_set_read_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;
  const struct sfr_fn_skel *skel;

  assert (port < I8051_SFR_SIZE);
  skel = find_sfr_fn_skel (port);
  if (skel != NULL && skel->read_fn != NULL)
    return ENOTSUP;
  mi = mi_;
  pf = &mi->sfr_fns[port].read;
  if (pf->fn.read_8 != NULL && src->fn.read_8 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

/* Set peer write_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int
sfr_ext_8_set_write_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;
  const struct sfr_fn_skel *skel;

  assert (port < I8051_SFR_SIZE);
  skel = find_sfr_fn_skel (port);
  if (skel != NULL && skel->write_fn != NULL)
    return ENOTSUP;
  mi = mi_;
  pf = &mi->sfr_fns[port].write;
  if (pf->fn.write_8 != NULL && src->fn.write_8 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

/* Set peer display_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int
sfr_ext_8_set_display_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;
  const struct sfr_fn_skel *skel;

  assert (port < I8051_SFR_SIZE);
  skel = find_sfr_fn_skel (port);
  if (skel != NULL && skel->display_fn != NULL)
    return ENOTSUP;
  mi = mi_;
  pf = &mi->sfr_fns[port].display;
  if (pf->fn.display_8 != NULL && src->fn.display_8 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

/* Set peer modify_fn for PORT in MI_ to SRC:
   returns: 0 if OK, errno otherwise */
static int
sfr_ext_8_set_modify_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;
  const struct sfr_fn_skel *skel;

  assert (port < I8051_SFR_SIZE);
  skel = find_sfr_fn_skel (port);
  if (skel != NULL && skel->modify_fn != NULL)
    return ENOTSUP;
  mi = mi_;
  pf = &mi->sfr_fns[port].modify;
  if (pf->fn.modify_8 != NULL && src->fn.modify_8 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

/* Get number of ports of this space in MI_ */
static size_t
prev_insn_8_get_size (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  return sizeof (mi->prev_insn);
}

/* Read byte port value for display (unlike port_8_read_t may have no side
   effects);
   returns: port value */
static uint8_t
prev_insn_8_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  mi = mi_;
  assert ((size_t)data < sizeof (mi->prev_insn));
  return mi->prev_insn[data];
}

/* Save port_fn for PORT displaying to DEST */
static void
prev_insn_8_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < sizeof (mi->prev_insn));
  dest->fn.display_8 = prev_insn_8_display_fn;
  dest->mi = mi;
  dest->data = port;
}

/* Get number of ports of this space in MI_ */
static size_t
next_insn_8_get_size (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  return sizeof (mi->insn);
}

/* Read byte port value for display (unlike port_8_read_t may have no side
   effects);
   returns: port value */
static uint8_t
next_insn_8_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  mi = mi_;
  assert ((size_t)data < sizeof (mi->insn));
  return (size_t)data < mi->insn_special
    ? mi->insn[data] : mi->program[mi->pc + data];
}

/* Save port_fn for PORT displaying to DEST */
static void
next_insn_8_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < sizeof (mi->insn));
  if (port >= mi->insn_special && ((mi->flags & MI_FETCH_EXTERNAL) != 0
				   || mi->pc + port >= mi->prog_mem_size))
    dest->fn.display_8 = NULL;
  else
    {
      dest->fn.display_8 = next_insn_8_display_fn;
      dest->mi = mi;
      dest->data = port;
    }
}

/* Byte address spaces */
const struct space i8051_spaces_8[I8051_NUM_SPACES_8] =
{
  /* Internal memory */
  [I8051_SPACE_8_MEM] =
  {
    .get_size = memory_8_get_size, .get_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = memory_8_get_display_fn,
      [PORT_TYPE_MODIFY] = memory_8_get_modify_fn
    },
    .set_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
    }
  },
  /* Internal program memory */
  [I8051_SPACE_8_PROG] =
  {
    .get_size = program_8_get_size, .get_fn =
    {
      [PORT_TYPE_READ] = NULL,
      [PORT_TYPE_WRITE] = NULL, [PORT_TYPE_DISPLAY] = program_8_get_display_fn,
      [PORT_TYPE_MODIFY] = program_8_get_modify_fn
    },
    .set_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
    }
  },
  /* SFRs */
  [I8051_SPACE_8_SFR] =
  {
    .get_size = sfr_8_get_size, .get_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = sfr_8_get_display_fn,
      [PORT_TYPE_MODIFY] = sfr_8_get_modify_fn
    },
    .set_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
    }
  },
  /* For adding new SFRs from plugins; FIXME: should we make separate SFR bit
     address space (e.g. for reserved bits)? */
  [I8051_SPACE_8_SFR_EXT] =
  {
    .get_size = sfr_ext_8_get_size, .get_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
    },
    .set_fn =
    {
      [PORT_TYPE_READ] = sfr_ext_8_set_read_fn,
      [PORT_TYPE_WRITE] = sfr_ext_8_set_write_fn,
      [PORT_TYPE_DISPLAY] = sfr_ext_8_set_display_fn,
      [PORT_TYPE_MODIFY] = sfr_ext_8_set_modify_fn
    }
  },
  /* Last finished instruction */
  [I8051_SPACE_8_PREV_INSN] =
  {
    .get_size = prev_insn_8_get_size, .get_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = prev_insn_8_get_display_fn,
      [PORT_TYPE_MODIFY] = NULL
    },
    .set_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
    }
  },
  /* Next 3 bytes at mi->pc, if we are using internal memory.  This means that
     the user needs to call get_fn[PORT_TYPE_DISPLAY] before each access to
     check whether the contents are valid (we return 0x00 in that case) */
  [I8051_SPACE_8_NEXT_INSN] =
  {
    .get_size = next_insn_8_get_size, .get_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = next_insn_8_get_display_fn,
      [PORT_TYPE_MODIFY] = NULL
    },
    .set_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
    }
  }
};

/* Get number of ports of this space in MI_ */
static size_t
misc_16_get_size (void *mi_)
{
  (void)mi_;
  return I8051_NUM_S16M;
}

/* Read 16b port value for display (unlike port_16_read_t may have no side
   effects);
   returns: port value */
static uint16_t
misc_16_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  mi = mi_;
  switch (data)
    {
    case I8051_S16M_PC:
      return mi->pc;

    case I8051_S16M_PREV_PC:
      return mi->prev_pc;
      
    default:
      assert ("BUG" == NULL);
      return 0; /* To make GCC happy */
    }
}

/* Save port_fn for PORT displaying to DEST */
static void
misc_16_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < I8051_NUM_S16M);
  dest->fn.display_16 = misc_16_display_fn;
  dest->mi = mi_;
  dest->data = port;
}

/* Modify 16b port value by user */
static void
misc_16_modify_fn (void *mi_, intptr_t data, uint16_t val)
{
  struct mi *mi;

  mi = mi_;
  switch (data)
    {
    case I8051_S16M_PC:
      mi->pc = val;
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Save port_fn for PORT modification to DEST */
static void
misc_16_get_modify_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < I8051_NUM_S16M);
  dest->fn.modify_16 = misc_16_modify_fn;
  dest->mi = mi_;
  dest->data = port;
}

/* 16b address spaces */
const struct space i8051_spaces_16[I8051_NUM_SPACES_16] =
{
  /* Miscellaneous 16b registers */
  [I8051_SPACE_16_MISC] =
  {
    .get_size = misc_16_get_size, .get_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = misc_16_get_display_fn,
      [PORT_TYPE_MODIFY] = misc_16_get_modify_fn
    },
    .set_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
    }
  }
};
