/* UART emulation
   Copyright (C) 2002, 2004 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>

#include <cereal.h>
#include <lib.h>

#include "8051.h"

/* Read current value of RXD in MI */
static int
read_rxd (struct mi *mi)
{
  uint8_t val;
  
  i8051_port_read (mi, I8051_S1P_P3, &val);
  return (val & SFR_P3_RXD) != 0;
}

/* Emulate next UART mode 0 receive event */
static void
emulate_uart_rx_0 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  /* We are cheating a bit.  Because mi->uart_rx_buf is not user-visible,
     we set it to 0x1FE early, shift it in S5P2 instead of S6P2 etc. */
  /* If you are reading the Atmel docs, note that they talk about input data
     coming to the right, and shifting to the left.  Unlike them we can't
     simply cross the wires, so if we want to read LSB first, we have to shift
     to the right.  In other words, swap right and left when reading the
     docs. */
  switch (mi->uart_rx_status)
    {
    case URX_0_S3P1:
      mi->port_3_mask &= ~SFR_P3_TXD;
      i8051_port_write (mi, I8051_S1P_P3, mi->ports_out[3]);
      schedule_add (mi->uart_rx_event, 5 * mi->phase_len, emulate_uart_rx_0);
      mi->uart_rx_status = URX_0_S5P2;
      break;

    case URX_0_S5P2:
      mi->uart_rx_buf >>= 1;
      if (read_rxd (mi) != 0)
	mi->uart_rx_buf |= 0x100;
      schedule_add (mi->uart_rx_event, mi->phase_len, emulate_uart_rx_0);
      mi->uart_rx_status = URX_0_S6P1;
      break;
      
    case URX_0_S6P1:
      mi->port_3_mask |= SFR_P3_TXD;
      i8051_port_write (mi, I8051_S1P_P3, mi->ports_out[3]);
      if ((mi->uart_rx_buf & 0x01) == 0)
	{
	  schedule_add (mi->uart_rx_event, mi->phase_len, emulate_uart_rx_0);
	  mi->uart_rx_status = URX_0_DONE_S6P2;
	  break;
	}
      schedule_add (mi->uart_rx_event, 6 * mi->phase_len, emulate_uart_rx_0);
      mi->uart_rx_status = URX_0_S3P1;
      break;

    case URX_0_DONE_S6P2:
      mi->sfr[SFR_SBUF] = mi->uart_rx_buf >> 1;
      schedule_add (mi->uart_rx_event, mi->phase_len, emulate_uart_rx_0);
      mi->uart_rx_status = URX_0_DONE_S1P1;
      break;

    case URX_0_DONE_S1P1:
      mi->sfr[SFR_SCON] |= SFR_SCON_RI;
      mi->uart_rx_status = URX_NONE;
      break;
    }
}

/* Emulate next UART mode 0 transmit event */
static void
emulate_uart_tx_0 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  switch (mi->uart_tx_status)
    {
    case UTX_0_S6P2:
      mi->uart_tx_buf >>= 1;
      if (mi->uart_tx_buf == 0x01)
	{
	  mi->port_3_mask |= SFR_P3_RXD;
	  schedule_add (mi->uart_tx_event, mi->phase_len, emulate_uart_tx_0);
	  mi->uart_tx_status = UTX_0_DONE;
	  break;
	}
      /* Yes, data goes to RxD.  Mode 0 is strange. */
      if ((mi->uart_tx_buf & 0x01) == 0)
	mi->port_3_mask &= ~SFR_P3_RXD;
      else
	mi->port_3_mask |= SFR_P3_RXD;
      i8051_port_write (mi, I8051_S1P_P3, mi->ports_out[3]);
      schedule_add (mi->uart_tx_event, 5 * mi->phase_len, emulate_uart_tx_0);
      mi->uart_tx_status = UTX_0_S3P1;
      break;

      /* In mode 0, RX, TX or both may be driving the TXD pin, but the result
	 is the same. */
    case UTX_0_S3P1:
      mi->port_3_mask &= ~SFR_P3_TXD;
      i8051_port_write (mi, I8051_S1P_P3, mi->ports_out[3]);
      schedule_add (mi->uart_tx_event, 6 * mi->phase_len, emulate_uart_tx_0);
      mi->uart_tx_status = UTX_0_S6P1;
      break;
      
    case UTX_0_S6P1:
      mi->port_3_mask |= SFR_P3_TXD;
      i8051_port_write (mi, I8051_S1P_P3, mi->ports_out[3]);
      schedule_add (mi->uart_tx_event, mi->phase_len, emulate_uart_tx_0);
      mi->uart_tx_status = UTX_0_S6P2;
      break;

    case UTX_0_DONE:
      mi->sfr[SFR_SCON] |= SFR_SCON_TI;
      mi->uart_tx_status = UTX_NONE;
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Emulate next UART mode 0 event */
static void
emulate_uart_123 (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  if ((mi->sfr[SFR_SCON] & (SFR_SCON_SM0 | SFR_SCON_SM1)) != SFR_SCON_SM0
      && (mi->flags & MI_UART_TIMER) == 0)
    goto sched_next;
  mi->flags &= ~MI_UART_TIMER;
  if ((mi->sfr[SFR_PCON] & SFR_PCON_SMOD) == 0)
    {
      mi->flags ^= MI_UART_DIV2;
      if ((mi->flags & MI_UART_DIV2) != 0)
	goto sched_next;
    }
  if (mi->uart_tx_status == UTX_123_SEND && mi->uart_tx_step == 0)
    {
      mi->uart_tx_buf >>= 1;
      if (mi->uart_tx_buf == 0x01)
	{
	  mi->port_3_mask |= SFR_P3_TXD;
	  mi->sfr[SFR_SCON] |= SFR_SCON_TI;
	  mi->uart_tx_status = UTX_NONE;
	}
      else
	{
	  if ((mi->uart_tx_buf & 0x01) == 0)
	    mi->port_3_mask &= ~SFR_P3_TXD;
	  else
	    mi->port_3_mask |= SFR_P3_TXD;
	}
      i8051_port_write (mi, I8051_S1P_P3, mi->ports_out[3]);
    }
  mi->uart_tx_step = (mi->uart_tx_step + 1) & 0x0F;
  /* The bit-swapping comment from emulate_uart_rx_0 () applies here as
     well. */
  switch (mi->uart_rx_status)
    {
    case URX_NONE:
      break;

    case URX_123_START_0:
      if (read_rxd (mi) != 0)
	mi->uart_rx_status = URX_123_START_1;
      break;

    case URX_123_START_1:
      if (read_rxd (mi) == 0)
	{
	  mi->uart_rx_buf = 0x1FF;
	  mi->uart_rx_step = 0;
	  mi->uart_rx_status = URX_123_RECV_0;
	}
      break;

    case URX_123_RECV_0: case URX_123_RECV_1: case URX_123_RECV_2:
      {
	int cnt, now;
	
	if (mi->uart_rx_step < 7 || mi->uart_rx_step > 9)
	  break;
	assert (URX_123_RECV_1 == URX_123_RECV_0 + 1);
	assert (URX_123_RECV_2 == URX_123_RECV_1 + 1);
	cnt = mi->uart_rx_status - URX_123_RECV_0;
	now = read_rxd (mi);
	if (now != 0)
	  cnt++;
	if (mi->uart_rx_step != 9)
	  {
	    mi->uart_rx_status = URX_123_RECV_0 + cnt;
	    break;
	  }
	if (cnt > 1) /* Received an one */
	  mi->uart_rx_buf |= 0x200;
	if (mi->uart_rx_buf == 0x3FF)
	  /* Start bit was received as 1 */
	  mi->uart_rx_status = now == 0 ? URX_123_START_0 : URX_123_START_1;
	else if ((mi->uart_rx_buf & 0x01) != 0)
	  {
	    mi->uart_rx_buf >>= 1;
	    mi->uart_rx_status = URX_123_RECV_0;
	  }
	else /* All 9 bits read */
	  {
	    if ((mi->sfr[SFR_SCON] & SFR_SCON_RI) == 0
		&& ((mi->sfr[SFR_SCON] & SFR_SCON_SM2) == 0
		    || (mi->uart_rx_buf & 0x200) != 0))
	      {
		mi->sfr[SFR_SBUF] = mi->uart_rx_buf >> 1;
		if ((mi->uart_rx_buf & 0x200) != 0)
		  mi->sfr[SFR_SCON] |= SFR_SCON_RB8;
		else
		  mi->sfr[SFR_SCON] &= ~SFR_SCON_RB8;
		mi->sfr[SFR_SCON] |= SFR_SCON_RI;
	      }
	    mi->uart_rx_status = now == 0 ? URX_123_START_0 : URX_123_START_1;
	  }
	break;
      }
	
    default:
      assert ("BUG" == NULL);
    }
  mi->uart_rx_step = (mi->uart_rx_step + 1) & 0x0F;
 sched_next:
  if ((mi->sfr[SFR_SCON] & (SFR_SCON_SM0 | SFR_SCON_SM1)) != SFR_SCON_SM0)
    schedule_add (mi->uart_tx_event, 12 * mi->phase_len, emulate_uart_123);
  else
    schedule_add (mi->uart_tx_event, 2 * mi->phase_len, emulate_uart_123);
}

/* Write VAL to byte port in MI_ */
void
i8051_sfr_write_sbuf_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  (void)data;
  mi = mi_;
  /* What to do when a transmit is in progress? We simply "reset the
     transmitter" and start a new transmit. */
  switch (mi->sfr[SFR_SCON] & (SFR_SCON_SM0 | SFR_SCON_SM1))
    {
    case 0:
      /* Leave bit 0 blank to eliminate a special case. */
      mi->uart_tx_buf = 0x200 | val << 1;
      if (mi->uart_tx_status != UTX_NONE)
	schedule_cancel (mi->uart_tx_event);
      schedule_add (mi->uart_tx_event, 12 * mi->phase_len, emulate_uart_tx_0);
      mi->uart_tx_status = UTX_0_S6P2;
      break;

    case SFR_SCON_SM1:
      /* Leave bit 0 blank to eliminate a special case, bit 1 = 0 is the start
	 bit. */
      mi->uart_tx_buf = 0x400 | val << 2;
      mi->uart_tx_status = UTX_123_SEND;
      break;
      
    case SFR_SCON_SM0:
    case SFR_SCON_SM0 | SFR_SCON_SM1:
      /* Leave bit 0 blank to eliminate a special case, bit 1 = 0 is the start
	 bit. */
      mi->uart_tx_buf = 0x800 | val << 2;
      if ((mi->sfr[SFR_SCON] & SFR_SCON_TB8) != 0)
	mi->uart_tx_buf |= 0x400;
      mi->uart_tx_status = UTX_123_SEND;
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Write VAL to byte port in MI_ */
void
i8051_sfr_write_scon_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  (void)data;
  mi = mi_;
  /* Reaction to mode change when I/O is in progress is not described.  We
     simply abort any pending I/O. */
  if (((mi->sfr[SFR_SCON] ^ val) & (SFR_SCON_SM0 | SFR_SCON_SM1)) != 0)
    { /* Mode change */
      if (mi->uart_tx_status != UTX_NONE
	  || (mi->sfr[SFR_SCON] & (SFR_SCON_SM0 | SFR_SCON_SM1)) != 0)
	{
	  schedule_cancel (mi->uart_tx_event);
	  mi->uart_tx_status = UTX_NONE;
	}
      if ((mi->sfr[SFR_SCON] & (SFR_SCON_SM0 | SFR_SCON_SM1)) == 0
	  && mi->uart_rx_status != URX_NONE)
	{
	  schedule_cancel (mi->uart_rx_event);
	  mi->uart_rx_status = URX_NONE;
	}
      mi->flags &= ~MI_UART_DIV2;
      switch (val & (SFR_SCON_SM0 | SFR_SCON_SM1))
	{
	case 0:
	  break;

	case SFR_SCON_SM1:
	case SFR_SCON_SM0 | SFR_SCON_SM1:
	  schedule_add (mi->uart_tx_event, mi->phase_len, emulate_uart_123);
	  break;

	case SFR_SCON_SM0:
	  schedule_add (mi->uart_tx_event, 2 * mi->phase_len,
			emulate_uart_123);
	  break;

	default:
	  assert ("BUG" == NULL);
	  break;
	}
    }
  if ((val & (SFR_SCON_SM0 | SFR_SCON_SM1)) == 0)
    {
      if ((val & SFR_SCON_REN) == 0 && mi->uart_rx_status != URX_NONE)
	{
	  schedule_cancel (mi->uart_rx_event);
	  mi->uart_rx_status = URX_NONE;
	}
      if ((val & (SFR_SCON_REN | SFR_SCON_RI)) == SFR_SCON_REN)
	{
	  mi->uart_rx_buf = 0xFF; /* See the comment in emulate_uart_rx_0 () */
	  schedule_add (mi->uart_rx_event, 17 * mi->phase_len,
			emulate_uart_rx_0);
	  mi->uart_rx_status = URX_0_S3P1;
	}
    }
  else
    {
      if ((val & SFR_SCON_REN) == 0)
	mi->uart_rx_status = URX_NONE;
      else
	mi->uart_rx_status = URX_123_START_0; /* Make sure we see 1 and 0 */
    }
  mi->sfr[SFR_SCON] = val;
}

#define FN(FN) { .fn = FN, .name = #FN }

/* Event handling function list for uart_[rt]x_event */
const struct schedule_fn_desc i8051_uart_event_list[] =
{
  FN (emulate_uart_rx_0), FN (emulate_uart_tx_0), FN (emulate_uart_123),
  { .fn = NULL, .name = NULL }
};
