/* Cereal module interface
   Copyright (C) 2002, 2004, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#ifndef MODULE_H__
#define MODULE_H__

#include <config.h>

#include <stddef.h>
#include <stdint.h>
#include <libxml/xmlstring.h>

#include <lib.h>

BEGIN_C_DECLS

struct _xmlDoc;
struct _xmlNode;

/* Port widths */
enum port_width
  {
    WIDTH_FIRST, WIDTH_1 = WIDTH_FIRST, WIDTH_8, WIDTH_16, NUM_WIDTHS
  };

/* Port type.  The values must be consecutive in range [0, NUM_PORT_TYPES) */
enum
{
  PORT_TYPE_READ,
  PORT_TYPE_WRITE,
  PORT_TYPE_DISPLAY,
  PORT_TYPE_MODIFY,
  NUM_PORT_TYPES
};

/* C++ is so type-safe that we can't do real work )-: */
#ifdef __cplusplus
extern "C++"
{
  inline enum port_width &operator ++ (enum port_width &w)
  {
    w = static_cast <enum port_width> (w + 1);
    return w;
  }
  
  inline enum port_width operator ++ (enum port_width &w, int)
  {
    enum port_width ret = w;
    ++w;
    return ret;
  }
};
#endif

/* Port width values */
extern const unsigned port_width_value[NUM_WIDTHS];

/* Get port width from port width value;
   returns: port width or NUM_WIDTHS if not found */
extern enum port_width get_port_width (unsigned value);

/* Per-width module data */
struct module_width
{
  const struct space *spaces;
  size_t num_spaces;
  /* Pointers to <space> element nodes from cereal_module.xml */
  struct _xmlNode **xml_spaces;
};

/* Every cereal module has this description.  cereal_module_register () should
   be called to make it accessible, cereal_module_unregister () must then be
   called before unloading it. */
struct cereal_module
{
  struct cereal_module *next, **pprev;
  const char *name;
  const char *const *compat_names; /* NULL-terminated, may be NULL */
  struct _xmlDoc *xml; /* Loaded cereal_module XML description file */
  void *handle; /* From dlopen (), or NULL */
  /* Create a new module instance;
     returns: module instance pointer if OK, NULL on error */
  void *(*mi_new) (void);
  /* Delete MI_ (may be in any emulation state) */
  void (*mi_delete) (void *mi_);
  /* Set option NAME of MI_ to VALUE;
     returns: 0 if OK, -1 on error */
  int (*set_option) (void *mi_, const char *name, const char *value);
  /* Get value of option NAME of MI_;
     returns: string for free () or NULL on error */
  char *(*get_option) (void *mi_, const char *name);
  /* Save setup (configuration options) of MI_ to NODE */
  void (*save_setup) (void *mi_, struct _xmlNode *node);
  /* Load setup (configuration options) of MI_ from NODE of DOC;
     returns: 0 if OK, -1 on error */
  int (*load_setup) (void *mi_, struct _xmlDoc *doc, struct _xmlNode *node);
  /* Save state of MI_ to NODE */
  void (*save_state) (void *mi_, struct _xmlNode *node);
  /* Load state of MI_ from NODE of DOC;
     returns: 0 if OK, -1 on error */
  int (*load_state) (void *mi_, struct _xmlDoc *doc, struct _xmlNode *node);
  struct module_width w[NUM_WIDTHS];
};

/* List of currently registered modules */
extern struct cereal_module *cereal_module_list;

/* Declare known modules */
#define MODULE(X) extern struct cereal_module (X);
#include <module_list.h>
#undef MODULE

/* Register MODULE to make it accessible by its name.
   returns: 0 if OK, -1 on error */
extern int cereal_module_register (struct cereal_module *module);

/* Unregister registered MODULE */
extern void cereal_module_unregister (const struct cereal_module *module);

/* Unregister all known modules */
extern void cereal_modules_unregister_all (void);

/* Save list of all known modules as children of NODE */
extern void cereal_modules_save (struct _xmlNode *node);

/* Load modules according to NODE of DOC;
   returns: 0 if OK, -1 on error */
extern int cereal_modules_load (struct _xmlDoc *doc, struct _xmlNode *node);

/* Find module NAME;
   returns: module or NULL if not found */
extern const struct cereal_module *cereal_module_find (const char *name);

/* Set option NAME of MI_ to VALUE;
   returns: 0 if OK, -1 on error;
   Just say that the option is unknown */
extern int generic_module_set_option (void *mi_, const char *name,
				      const char *value);

/* Get value of option NAME of MI_;
   returns: string for free () or NULL on error;
   Just say that the option is unknown */
extern char *generic_module_get_option (void *mi_, const char *name);

/* Save setup (configuration options) of MI_ to NODE;
   Do nothing */
extern void generic_module_save_setup (void *mi_, struct _xmlNode *node);

/* Load setup (configuration options) of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error;
   Just return 0 */
extern int generic_module_load_setup (void *mi_, struct _xmlDoc *doc,
				      struct _xmlNode *node);

/* Load setup (configuration options) of MI_ from NODE of DOC by setting
   options of MODULE according to elements of namespace NS;
   returns: 0 if OK, -1 on error */
extern int generic_module_load_setup_options (void *mi_, struct _xmlDoc *doc,
					      struct _xmlNode *node,
					      const struct cereal_module
					      *module, const xmlChar *ns);

/* Read bit port value in MI_;
   returns: port value */
typedef int (*port_1_read_t) (void *mi_, intptr_t data);

/* Write VAL to bit port in MI_ */
typedef void (*port_1_write_t) (void *mi_, intptr_t data, int val);

/* Read bit port value for display (unlike port_1_read_t may have no side
   effects);
   returns: port value */
typedef int (*port_1_display_t) (void *mi_, intptr_t data);

/* Modify bit port value by user */
typedef void (*port_1_modify_t) (void *mi_, intptr_t data, int val);

/* Read byte port value in MI_;
   returns: port value */
typedef uint8_t (*port_8_read_t) (void *mi_, intptr_t data);

/* Write VAL to byte port in MI_ */
typedef void (*port_8_write_t) (void *mi_, intptr_t data, uint8_t val);

/* Read byte port value for display (unlike port_8_read_t may have no side
   effects);
   returns: port value */
typedef uint8_t (*port_8_display_t) (void *mi_, intptr_t data);

/* Modify byte port value by user */
typedef void (*port_8_modify_t) (void *mi_, intptr_t data, uint8_t val);

/* Read 16b port value in MI_;
   returns: port value */
typedef uint16_t (*port_16_read_t) (void *mi_, intptr_t data);

/* Write VAL to 16b port in MI_ */
typedef void (*port_16_write_t) (void *mi_, intptr_t data, uint16_t val);

/* Read 16b port value for display (unlike port_16_read_t may have no side
   effects);
   returns: port value */
typedef uint16_t (*port_16_display_t) (void *mi_, intptr_t data);

/* Modify 16b port value by user */
typedef void (*port_16_modify_t) (void *mi_, intptr_t data, uint16_t val);

/* Description of a handling function and its arguments */
struct port_fn
{
  /* fn may be NULL to indicate unsupported function (in get_*)
     or disconect (in set_*) */
  union
  {
    port_1_read_t read_1;
    port_1_write_t write_1;
    port_1_display_t display_1;
    port_1_modify_t modify_1;
    port_8_read_t read_8;
    port_8_write_t write_8;
    port_8_display_t display_8;
    port_8_modify_t modify_8;
    port_16_read_t read_16;
    port_16_write_t write_16;
    port_16_display_t display_16;
    port_16_modify_t modify_16;
  } fn;
  void *mi;
  intptr_t data;
};

/* Address space */
struct space
{
  /* Get number of ports of this space in MI_ */
  size_t (*get_size) (void *mi_); 
  /* Save port_fn for PORT to DEST */
  void (*get_fn[NUM_PORT_TYPES]) (void *mi_, size_t port,
				  struct port_fn *dest);
  /* Set peer fn for PORT in MI_ to SRC;
     returns: 0 if OK, errno otherwise */
  int (*set_fn[NUM_PORT_TYPES]) (void *mi_, size_t port,
				 const struct port_fn *src);
};

/* Get XML node for SPACE of WIDTH in MODULE;
   returns: XML node pointer or NULL */
extern struct _xmlNode *space_get_xml (const struct cereal_module *module,
				       enum port_width width, size_t space);

/* Create name for SPACE of WIDTH in MODULE;
   returns: space name from libxml (in UTF-8) (for xmlFree ()) */
extern xmlChar *space_create_name (const struct cereal_module *module,
				   enum port_width width, size_t space);

/* Find port space with NAME (as got from libxml, i.e. in UTF-8) of WIDTH
   (NUM_WIDTHS = any) in MODULE, store its offset to *PSPACE; if not found, try
   to interpret NAME as width:offset or, if WIDTH != 0, as offset (with
   WIDTH implied);
   returns: space width if OK, NUM_WIDTHS on error */
extern enum port_width space_find_name (const struct cereal_module *module,
					const char *name,
					enum port_width width, size_t *pspace);

END_C_DECLS

#endif
