/* Insert Port Reference dialog
   Copyright (C) 2002, 2003, 2004, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */

#include <cassert>
#include <kcombobox.h>
#include <kpushbutton.h>
#include <kstdguiitem.h>
#include <libxml/tree.h>
#include <qlabel.h>
#include <qstringlist.h>

#include <me.h>
#include <module.h>
#include <port.h>

#include "kcinsertportdialog.h"

/** Create an "insert port reference" dialog for port type */
KCInsertPortDialog::KCInsertPortDialog (QWidget *parent, int type)
  : KCInsertPortDialogBase (parent), type_ (type)
{
  cancelButton->setGuiItem (KStdGuiItem::cancel ());
  OKButton->setGuiItem (KStdGuiItem::ok ());

  /* The order of building the data is not just "as the user sees it", it is
     also sorted so that variables are not accessed unitialized. */
  changing = true;
  QStringList modules;
  for (const struct me_entry *me = me_list; me != NULL; me = me->next)
    modules.append (me->name);
  modules.sort ();
  moduleName->insertStringList (modules);
  pi.me = me_find (moduleName->currentText ());

  rebuildSpace ();
  rebuildPort ();
  rebuildOK ();
  changing = false;
}

KCInsertPortDialog::~KCInsertPortDialog ()
{
}

/** @returns port name selected by user */
QString
KCInsertPortDialog::selectedPortName () const
{
  return QString ("%1/%2/%3").arg (pi.me->name).arg (spaceName->currentText ())
    .arg (portName->currentText ());
}

/** Module has changed */
void
KCInsertPortDialog::moduleChanged ()
{
  pi.me = me_find (moduleName->currentText ());
  assert (pi.me != NULL);
  if (changing == true)
    return;
  changing = true;
  rebuildSpace ();
  rebuildPort ();
  rebuildOK ();
  changing = false;
}

/** Space has changed */
void
KCInsertPortDialog::spaceChanged ()
{
  if (spaceName->count () == 0
      || (width = space_find_name (pi.me->module,
				   spaceName->currentText ().utf8 (),
				   NUM_WIDTHS, &pi.space)) == NUM_WIDTHS)
    {
      pi.space = 0;
      spaceDescr->setText ("");
    }
  else
    {
      xmlNode *node = space_get_xml (pi.me->module, width, pi.space);
      xmlChar *str;
      if (node != NULL
	  && (str = xmlGetProp (node, (const xmlChar *)"description"))
	  != NULL)
	{
	  spaceDescr->setText (QString::fromUtf8 ((const char *)str));
	  xmlFree (str);
	}
      else
	spaceDescr->setText ("");
    }
  if (changing == true)
    return;
  changing = true;
  rebuildPort ();
  rebuildOK ();
  changing = false;
}

/** Port has changed */
void
KCInsertPortDialog::portChanged ()
{
  QString val = portName->currentText ();
  if (val.isEmpty () || width == NUM_WIDTHS
      || port_find_name (width, &pi, val) != 0)
    {
      pi.port = 0;
      portDescr->setText ("");
    }
  else
    {
      xmlNode *node = space_get_xml (pi.me->module, width, pi.space);
      if (node != NULL)
	{
	  for (node = node->children; node != NULL; node = node->next)
	    {
	      if (node->type != XML_ELEMENT_NODE
		  || strcmp ((const char *)node->name, "port") != 0)
		continue;
	      xmlChar *str = xmlGetProp (node, (const xmlChar *)"name");
	      if (str == NULL)
		continue;
	      int res = val == QString::fromUtf8 ((const char *)str);
	      xmlFree (str);
	      if (res == 0)
		continue;
	      str = xmlGetProp (node, (const xmlChar *)"description");
	      if (str == NULL)
		continue;
	      portDescr->setText (QString::fromUtf8 ((const char *)str));
	      xmlFree (str);
	      goto found;
	    }
	}
      portDescr->setText ("");
    found:
      ;
    }
  if (changing == true)
    return;
  changing = true;
  rebuildOK ();
  changing = false;
}

/** Modify space list according to other selections, try to preserve currently
    selected item */
void
KCInsertPortDialog::rebuildSpace ()
{
  QString old_val = spaceName->currentText ();
  spaceName->clear ();
  const struct cereal_module *mod = pi.me->module;
  int index = -1;
  for (enum port_width w = WIDTH_FIRST; w < NUM_WIDTHS; w++)
    {
      const struct module_width *mw = mod->w + w;
      for (size_t s = 0; s < mw->num_spaces; s++)
	{
	  if (mw->spaces[s].get_fn[type_] != NULL)
	    {
	      xmlChar *str = space_create_name (mod, w, s);
	      QString val = QString::fromUtf8 ((const char *)str);
	      xmlFree (str);
	      spaceName->insertItem (val);
	      if (val == old_val)
		index = spaceName->count () - 1;
	    }
	}
    }
  if (index != -1)
    /* This retains the name, but may change the space index */
    spaceName->setCurrentItem (index);
  spaceChanged ();
}

/** Modify port list according to other selections, try to preserve currently
    selected item */
void
KCInsertPortDialog::rebuildPort ()
{
  QString old_val = portName->currentText ();
  portName->clear ();
  if (spaceName->count () == 0)
    return;
  xmlNodePtr node = space_get_xml (pi.me->module, width, pi.space);
  if (node != NULL)
    {
      for (node = node->children; node != NULL; node = node->next)
	{
	  xmlChar *str;
	  unsigned long port;
	  int res;
	  
	  if (node->type != XML_ELEMENT_NODE
	      || strcmp ((const char *)node->name, "port") != 0)
	    continue;
	  str = xmlGetProp (node, (const xmlChar *)"index");
	  if (str == NULL)
	    continue;
	  res = get_num ((const char *)str, 0, &port);
	  xmlFree (str);
	  if (res != 0)
	    continue;
	  str = xmlGetProp (node, (const xmlChar *)"name");
	  if (str == NULL)
	    continue;
	  switch (width)
	    {
#define DO_TYPE(TYPE, WIDTH)				\
	      do					\
		{					\
		  res = pf.fn.TYPE##_##WIDTH != NULL;	\
		  break;				\
		}					\
	      while (0)
#define DO_WIDTH(WIDTH)							\
	    case WIDTH_##WIDTH:						\
	      {								\
		struct port_fn pf;					\
		pi.me->module->w[width].spaces[pi.space].get_fn[type_]	\
		  (pi.me->mi, port, &pf);				\
		switch (type_)						\
		  {							\
		  case PORT_TYPE_READ:					\
		    DO_TYPE (read, WIDTH);				\
		    break;						\
		  case PORT_TYPE_WRITE:					\
		    DO_TYPE (write, WIDTH);				\
		    break;						\
		  case PORT_TYPE_DISPLAY:				\
		    DO_TYPE (display, WIDTH);				\
		    break;						\
		  case PORT_TYPE_MODIFY:				\
		    DO_TYPE (modify, WIDTH);				\
		    break;						\
		  default:						\
		    assert ("BUG" == NULL);				\
		  }							\
		break;							\
	      }
	      DO_WIDTH (1);
	      DO_WIDTH (8);
	      DO_WIDTH (16);

	    default:
	      assert ("BUG" == NULL);
#undef DO_WIDTH
#undef DO_TYPE
	    }
	  if (res != 0)
	    portName->insertItem (QString::fromUtf8 ((const char *)str));
	  xmlFree (str);
	}
    }
  if (old_val.isEmpty () || port_find_name (width, &pi, old_val) != 0)
    portName->setCurrentText (QString::number (0));
  else
    /* This retains the name, but may change the port index */
    portName->setCurrentText (old_val);
  portChanged ();
}

/** Update state of OKButton according to other selections */
void
KCInsertPortDialog::rebuildOK ()
{
  if (spaceName->count () == 0)
    goto invalid;
  if (port_find_name (width, &pi, portName->currentText ()) != 0)
    goto invalid;
  switch (width)
    {
#define DO_TYPE(TYPE, WIDTH)			\
      do					\
	{					\
	  if (pf.fn.TYPE##_##WIDTH == NULL)	\
	    goto invalid;			\
	}					\
      while (0)
#define DO_WIDTH(WIDTH)						\
    case WIDTH_##WIDTH:						\
      {								\
	struct port_fn pf;					\
	pi.me->module->w[width].spaces[pi.space].get_fn[type_]	\
	  (pi.me->mi, pi.port, &pf);				\
	switch (type_)						\
	  {							\
	  case PORT_TYPE_READ:					\
	    DO_TYPE (read, WIDTH);				\
	    break;						\
	  case PORT_TYPE_WRITE:					\
	    DO_TYPE (write, WIDTH);				\
	    break;						\
	  case PORT_TYPE_DISPLAY:				\
	    DO_TYPE (display, WIDTH);				\
	    break;						\
	  case PORT_TYPE_MODIFY:				\
	    DO_TYPE (modify, WIDTH);				\
	    break;						\
	  default:						\
	    assert ("BUG" == NULL);				\
	  }							\
	break;							\
      }
      DO_WIDTH (1);
      DO_WIDTH (8);
      DO_WIDTH (16);

    default:
      assert ("BUG" == NULL);
#undef DO_WIDTH
#undef DO_TYPE
    }
  OKButton->setEnabled (true);
  return;
  
 invalid:
  OKButton->setEnabled (false);
}

#include "kcinsertportdialog.moc"
