/* Main window of cereal_kde
   Copyright (C) 2002, 2003, 2004 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */

#include <cstring>
#include <kaction.h>
#include <kapplication.h>
#include <kfiledialog.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kparts/plugin.h>
#include <kpopupmenu.h>
#include <kshortcut.h>
#include <kstdaction.h>
#include <ktempfile.h>
#include <libxml/tree.h>
#include <qptrlist.h>
#include <qwidgetlist.h>
#include <qworkspace.h>

#include <cereal.h>
#include <breakpoint.h>
#include <me.h>
#include <setup.h>

#include "kcbreakwin.h"
#include "kcevalmodifydialog.h"
#include "kcmainwin.h"
#include "kcrundialog.h"
#include "kcselectmoduledialog.h"
#include "kcwatchwin.h"
#include "kcwindow.h"
#include "kcwindowprovider.h"

/* A window provided by a KCWindowProvider */
struct KCProvidedWindow
{
  KCProvidedWindow(KCWindowProvider *provider_, unsigned index_, QString name_)
    : provider (provider_), index (index_), name (name_), widget (0)
  { };
  KCWindowProvider *provider;
  unsigned index;
  QString name;
  QWidget *widget; /* May be 0 */
};

/** Create a new window, optionally loading file (setup or state) if it is
    non-NULL */
KCMainWin::KCMainWin (const char *file)
  : KParts::MainWindow (), noSetupLoaded_ (true), modified (false),
    filename (""), childViews (), windowProviders (), providedWindows (),
    stateHandlers ()
{
  windowProviders.setAutoDelete(true);

  KStdAction::open (this, SLOT (fileOpen ()), actionCollection ());
  KStdAction::save (this, SLOT (fileSave ()), actionCollection ());
  KStdAction::saveAs (this, SLOT (fileSaveAs ()), actionCollection ());
  KStdAction::quit (this, SLOT (fileQuit ()), actionCollection ());
  new KAction (i18n ("&Step"), KShortcut (Qt::Key_F8), this, SLOT (runStep ()),
	       actionCollection (), "runstep");
  new KAction (i18n ("&Run"), KShortcut (Qt::Key_F9), this, SLOT (runRun ()),
	       actionCollection (), "runrun");
  new KAction (i18n ("&Evaluate/Modify..."), KShortcut (Qt::Key_F4), this,
	       SLOT (evalModify ()), actionCollection (), "evalmodify");
  new KAction (i18n ("New &Watch Window"), 0, this, SLOT (newWatchWindow ()),
	       actionCollection (), "watchwindow");
  new KAction (i18n ("New &Breakpoint Window"), 0, this,
	       SLOT (newBreakpointWindow ()), actionCollection (),
	       "breakpointwindow");
  viewMenu = new KActionMenu (i18n ("&View"), actionCollection (), "viewmenu");
  windowCascadeAction = new KAction (i18n ("&Cascade"), 0, this,
				     SLOT (windowCascade ()),
				     actionCollection (), "cascade");
  windowTileAction = new KAction (i18n ("&Tile"), 0, this,
				  SLOT (windowTile ()), actionCollection (),
				  "tile");
  windowMenu = new KActionMenu (i18n ("&Window"), actionCollection (),
				"windowmenu");
  setXMLFile ("cerealui.rc");

  registerStateHandler (this); /* We "own" KCBreakWin and KCWatchWin */
  
  createGUI (0);
  setAutoSaveSettings ();

  workspace = new QWorkspace (this);
  setCentralWidget (workspace);
  QObject::connect (windowMenu->popupMenu (), SIGNAL (aboutToShow ()), this,
		    SLOT (rebuildWindowMenu ()));
  
  updateCaption ();
  if (file != 0)
    loadFile (file);
  else
    fileOpen ();
}

KCMainWin::~KCMainWin ()
{
}

/** Register an UI state handler */
void
KCMainWin::registerStateHandler (KCUIStateHandler *handler)
{
  stateHandlers.append (handler);
}

/** Update all views, because state has changed */
void
KCMainWin::updateViews ()
{
  for (KCWindow *w = childViews.first (); w != 0; w = childViews.next ())
    w->cerealUpdate ();
  modified = true;
  updateCaption ();
}

/** Register new view */
void
KCMainWin::registerView (KCWindow *view)
{
  childViews.append (view);
  if (view->provider)
    {
      for (KCProvidedWindow *w = providedWindows.first (); w != 0;
	   w = providedWindows.next ())
	{
	  if (w->provider == view->provider && w->name == view->name)
	    {
	      w->widget = view->widget;
	      break;
	    }
	}
    }
}

/** Unregister view */
void
KCMainWin::unregisterView (KCWindow *view)
{
  childViews.removeRef (view);
  for (KCProvidedWindow *w = providedWindows.first (); w != 0;
       w = providedWindows.next ())
    {
      if (w->widget == view->widget)
	{
	  w->widget = 0;
	  break;
	}
    }
  rebuildViewMenu ();
}

/** Register a view provider */
void
KCMainWin::registerProvider (KCWindowProvider *provider)
{
  windowProviders.append (provider);
  QStringList windows = provider->providedWindowList ();
  unsigned i = 0;
  for (QStringList::Iterator it = windows.begin (); it != windows.end ();
       ++it, i++)
    providedWindows.append (new KCProvidedWindow (provider, i, *it));
  rebuildViewMenu ();
}

/** viewMenu should be rebuilt based on new providedWindows */
void
KCMainWin::rebuildViewMenu ()
{
  KPopupMenu *menu = viewMenu->popupMenu ();
  menu->clear ();
  int i = 0;
  for (KCProvidedWindow *w = providedWindows.first (); w != 0;
       w = providedWindows.next (), i++)
    {
      int id = menu->insertItem (w->name, this,
				 SLOT (providedViewSelected (int)));
      menu->setItemParameter (id, i);
      if (w->widget)
	menu->setItemChecked (id, true);
    }
}

/** viewMenu entry has been selected */
void
KCMainWin::providedViewSelected (int id)
{
  KCProvidedWindow *w = providedWindows.at (id);
  if (w)
    {
      if (w->widget)
	w->widget->setFocus ();
      else
	w->provider->provideWindow (w->index);
    }
  rebuildViewMenu ();
}

/** Select a module instance of module, ask the user if ambiguous;
    @returns selected module or NULL if none */
struct me_entry *
KCMainWin::selectModule (const struct cereal_module *module)
{
  struct me_entry *res;

  res = NULL;
  for (struct me_entry *me = me_list; me != NULL; me = me->next)
    {
      if (me->module != module)
	continue;
      if (res == NULL)
	res = me;
      else
	{
	  KCSelectModuleDialog *dlg = new KCSelectModuleDialog (this, module);
	  if (dlg->exec () == QDialog::Accepted)
	    res = dlg->selectedModule ();
	  else
	    res = NULL;
	  delete dlg;
	  return res;
	}
    }
  return res;
}

/** Save current UI state under uiState */
void
KCMainWin::saveState (xmlNodePtr)
{ /* Nothing */
}

/** Load UI state from uiState of doc;
    returns: 0 if OK, -1 on error */
int
KCMainWin::loadState (xmlDocPtr doc, xmlNodePtr uiState)
{
  int res = 0;
  for (xmlNodePtr node = uiState->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE
	  || strcmp ((const char *)node->name, "ui_window") != 0)
	continue;
      xmlChar *str = xmlGetProp (node, (const xmlChar *)"class");
      if (str == NULL)
	continue;
      if (strcmp ((const char *)str, "KCBreakWin") == 0)
	{
	  KCBreakWin *w = new KCBreakWin (this, viewParent ());
	  res = w->loadState (doc, node);
	  if (res == 0)
	    w->show ();
	  else
	    delete w;
	}
      else if (strcmp ((const char *)str, "KCWatchWin") == 0)
	{
	  KCWatchWin *w = new KCWatchWin (this, viewParent ());
	  res = w->loadState (doc, node);
	  if (res == 0)
	    w->show ();
	  else
	    delete w;
	}
      xmlFree (str);
      if (res != 0)
	break;
    }
  return res;
}

/** @return true if the window can be closed */
bool
KCMainWin::queryClose ()
{
  return canLoseState ();
}

/** File/Open was selected */
void
KCMainWin::fileOpen ()
{
  if (!canLoseState ())
    return;
  /* Documentation states that this guarantees the file is present, which
     is apparently not true. */
  QString fn = KFileDialog::getOpenFileName (0, "*.xml", this);
  if (!fn.isEmpty ())
    loadFile (fn);
}

/** File/Save was selected */
void
KCMainWin::fileSave ()
{
  if (filename.isEmpty ())
    fileSaveAs ();
  else if (state_save (filename, saveStateCallback, (intptr_t)this) == 0
	   && modified)
    {
      modified = false;
      updateCaption ();
    }
}

/** Save UI state under node */
void
KCMainWin::saveStateCallback (xmlNodePtr node, intptr_t data)
{
  KCMainWin *p = (KCMainWin *)data;
  for (KCWindow *w = p->childViews.first (); w != 0; w = p->childViews.next ())
    w->saveState (node);
  for (KCUIStateHandler *h = p->stateHandlers.first (); h != 0;
       h = p->stateHandlers.next ())
    h->saveState (node);
}

/** File/Save as was selected */
void
KCMainWin::fileSaveAs ()
{
  QString fn = KFileDialog::getSaveFileName (0, "*.xml", this);
  if (!fn.isEmpty ()
      && state_save (fn, saveStateCallback, (intptr_t)this) == 0)
    {
      filename = fn;
      modified = false;
      updateCaption ();
    }
}

/** File/Quit was selected */
void
KCMainWin::fileQuit ()
{
  close ();
}

/** Run/Step was selected */
void
KCMainWin::runStep ()
{
  if (!schedule_pending ())
    {
      error (0, "nothing to emulate");
      return;
    }

  int res;
  do
    {
      res = schedule_run ();
      if (breakpoint_check () != NULL)
	break;
    }
  while ((res & (ER_HALT | ER_ERROR | ER_WARNING | ER_INSN)) == 0);
  updateViews ();
}

/** Run/Run was selected */
void
KCMainWin::runRun ()
{
  if (!schedule_pending ())
    {
      error (0, "nothing to emulate");
      return;
    }

  KCRunDialog *dlg = new KCRunDialog (this);
  dlg->show ();

  KApplication *app = KApplication::kApplication ();
  int res;
  int cnt = 0;
  do
    {
      /* Most of the time is spent in the processEvents call, so we may run
	 thousands of instructions before the user notices any slowdown.  In
	 fact, we have to, because the time spent actually emulating is zilch
	 when compared to the processEvents () call.  OTOH, calling it only
	 every 1E3, 1E4 or 1E5-th iteration makes no difference, so we try to
	 keep a reasonable balance (the % 1024 still gives us a noticeable
	 speedup compared to % 100). */
      if (cnt++ % 1024 == 0)
	app->processEvents ();
      if (dlg->isHidden ())
	break;
      res = schedule_run ();
    }
  while (breakpoint_check () == NULL
	 && (res & (ER_HALT | ER_ERROR | ER_WARNING)) == 0);
  /* FIXME: option not to stop on warnings */

  delete dlg;
  updateViews ();
}

/** Data/Evaluate/Modify was selected */
void
KCMainWin::evalModify ()
{
  /* Should this be made non-modal? */
  KCEvalModifyDialog *dlg = new KCEvalModifyDialog (this, this);
  dlg->exec ();
  delete dlg;
}

/** Data/New Watch Window was selected */
void
KCMainWin::newWatchWindow ()
{
  KCWatchWin *w = new KCWatchWin (this, viewParent ());
  w->show ();
}

/** Data/New Breapoint Window was selected */
void
KCMainWin::newBreakpointWindow ()
{
  KCBreakWin *w = new KCBreakWin (this, viewParent ());
  w->show ();
}

/** windowMenu is about to show */
void
KCMainWin::rebuildWindowMenu ()
{
  KPopupMenu *menu = windowMenu->popupMenu ();
  menu->clear ();
  windowMenu->insert (windowTileAction);
  windowMenu->insert (windowCascadeAction);

  QWidgetList list = workspace->windowList ();
  if (list.isEmpty ())
    {
      windowTileAction->setEnabled (false);
      windowCascadeAction->setEnabled (false);
    }
  else
    {
      windowTileAction->setEnabled (true);
      windowCascadeAction->setEnabled (true);
    }
  menu->insertSeparator ();

  QWidget *active = workspace->activeWindow ();
  int i = 0;
  for (QWidget *w = list.first (); w != 0; w = list.next (), i++)
    {
      int id = menu->insertItem (QString ("&%1 ").arg (i + 1) + w->caption (),
				 this, SLOT (windowSelected (int)));
      menu->setItemParameter (id, i);
      if (w == active)
	menu->setItemChecked (id, true);
    }
}

/** Window/Cascade was selected */
void
KCMainWin::windowCascade ()
{
  workspace->cascade ();
}

/** Window/Tile was selected */
void
KCMainWin::windowTile ()
{
  workspace->tile ();
}

/** windowMenu window entry has been selected */
void
KCMainWin::windowSelected (int id)
{
  QWidget *w = workspace->windowList ().at (id);
  if (w)
    w->setFocus ();
}

/** @return file name we display to the user */
QString
KCMainWin::prettyFileName ()
{
  if (filename.isEmpty ())
    return i18n ("Untitled");
  KURL url (filename);
  return url.fileName ();
}

/** Update window caption */
void
KCMainWin::updateCaption ()
{
  setCaption (prettyFileName ());
};

/** @return true if the user has confirmed he wants to lose current state */
bool
KCMainWin::canLoseState ()
{
  if (!modified)
    return true;
  switch (KMessageBox
	  ::warningYesNoCancel (0, i18n ("Emulation state has changed.\n"
					 "Do you want to save it?")))
    {
    case KMessageBox::Yes:
      fileSave ();
      return !modified;

    case KMessageBox::No:
      return true;

    default:
      return false;
    }
}

/** Setup has been loaded */
void
KCMainWin::loadStateSetupCallback (intptr_t data)
{
  KCMainWin *p = (KCMainWin *)data;
  for (KCWindowProvider *provider = p->windowProviders.first (); provider != 0;
       provider = p->windowProviders.next ())
    {
      QStringList windows = provider->providedWindowList ();
      unsigned i = 0;
      for (QStringList::Iterator it = windows.begin (); it != windows.end ();
	   ++it, i++)
	p->providedWindows.append (new KCProvidedWindow (provider, i, *it));
    }
}

/** Load UI state from node of doc;
    returns: 0 if OK, -1 on error */
int
KCMainWin::loadStateUICallback (xmlDocPtr doc, xmlNodePtr node, intptr_t data)
{
  KCMainWin *p = (KCMainWin *)data;
  for (KCUIStateHandler *h = p->stateHandlers.first (); h != 0;
       h = p->stateHandlers.next ())
    {
      if (h->loadState (doc, node) != 0)
	return -1;
    }
  return 0;
}

/** Load state or setup from file */
void
KCMainWin::loadFile (const QString &file)
{
  providedWindows.clear();
  for (KCWindow *w = childViews.first (); w != 0; w = childViews.next ())
    w->widget->close ();
  switch (state_load (file, loadStateSetupCallback, loadStateUICallback,
		      (intptr_t)this))
    {
    default:
      modified = false;
      setup_destroy (); /* Create a reasonable state */
      setup_init ();
      break;

    case 1: /* cereal_state */
      filename = file;
      /* Fall through */

    case 0: /* cereal_setup */
      noSetupLoaded_ = false;
      modified = false;
      updateCaption ();
      break;
    }
  rebuildViewMenu ();
}

#include "kcmainwin.moc"
