/* 8051 view window
   Copyright (C) 2002, 2004 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */

#include <cassert>
#include <climits>
#include <cstdio>
#include <klineedit.h>
#include <klistbox.h>
#include <klocale.h>
#include <libxml/tree.h>
#include <qgroupbox.h>
#include <qstringlist.h>
#include <qregexp.h>
#include <qwidget.h>

#include <cereal.h>
#include <disasm8051.h>
#include <expr.h>
#include "../kde/kcbitedit.h"
#include "../kde/kcportedit.h"
#include "../kde/kcuistatehandler.h"
#include "../kde/kde.h"
#include <lib.h>
#include <me.h>

#include "kc8051.h"
#include "kc8051viewwin.h"

/* Number of lines in diassembly listing */
enum { DISASM_LENGTH = 128 };

/* Names of SFR bits */
static const char *const PSWBits[8] =
  { "P", "PSW.2", "OV", "RS0", "RS1", "F0", "AC", "C"  };

static const char *const IEBits[8] =
  { "EX0", "ET0", "EX1", "ET1", "ES", "IE.5", "IE.6", "EA" };

static const char *const IPBits[8] =
  { "PX0", "PT0", "PX1", "PT1", "PS", "IP.5", "IP.6", "IP.7" };

static const char *const TMODBits[8] =
  { "0_M0", "0_M1", "0_CT", "0_GATE", "1_M0", "1_M1", "1_CT", "1_GATE" };

static const char *const TCONBits[8] =
  { "IT0", "IE0", "IT1", "IE1", "TR0", "TF0", "TR1", "TF1" };

static const char *const SCONBits[8] =
  { "RI", "TI", "RB8", "TB8", "REN", "SM2", "SM1", "SM0" };
    
static const char *const PCONBits[8] =
  { "IDL", "PD", "PCON.2", "PCON.3", "PCON.4", "PCON.5", "PCON.6", "SMOD" };


/* List of KCPortEdit children and their associated names and expressions */
struct PortChild
{
  /* Furrfu, I never thought I'll ever use this. */
  KCPortEdit *KC8051ViewWin::*edit;
  const char *label;
  const char *expr; /* '@' is replaced by me_->name */
  const char *const *names; /* Bit names or NULL */
};

static const struct PortChild portChildren[] =
  {
    { &KC8051ViewWin::ACCEdit, "&ACC", "[@/sfr/ACC]", NULL },
    /* These PSW-decoding expressions are "interesting".  Perfect excuse for
       adding indexed port adressing ;-) */
    { &KC8051ViewWin::R0Edit, "R&0", "([c:@/sfr/PSW] & 0x10) == 0 "
      "? (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/0] : [@/mem/8]) "
      ": (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/16] : [@/mem/24])", NULL },
    { &KC8051ViewWin::R1Edit, "R&1", "([c:@/sfr/PSW] & 0x10) == 0 "
      "? (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/1] : [@/mem/9]) "
      ": (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/17] : [@/mem/25])", NULL },
    { &KC8051ViewWin::R2Edit, "R&2", "([c:@/sfr/PSW] & 0x10) == 0 "
      "? (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/2] : [@/mem/10]) "
      ": (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/18] : [@/mem/26])", NULL },
    { &KC8051ViewWin::R3Edit, "R&3", "([c:@/sfr/PSW] & 0x10) == 0 "
      "? (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/3] : [@/mem/11]) "
      ": (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/19] : [@/mem/27])", NULL },
    { &KC8051ViewWin::R4Edit, "R&4", "([c:@/sfr/PSW] & 0x10) == 0 "
      "? (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/4] : [@/mem/12]) "
      ": (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/20] : [@/mem/28])", NULL },
    { &KC8051ViewWin::R5Edit, "R&5", "([c:@/sfr/PSW] & 0x10) == 0 "
      "? (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/5] : [@/mem/13]) "
      ": (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/21] : [@/mem/29])", NULL },
    { &KC8051ViewWin::R6Edit, "R&6", "([c:@/sfr/PSW] & 0x10) == 0 "
      "? (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/6] : [@/mem/14]) "
      ": (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/22] : [@/mem/30])", NULL },
    { &KC8051ViewWin::R7Edit, "R&7", "([c:@/sfr/PSW] & 0x10) == 0 "
      "? (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/7] : [@/mem/15]) "
      ": (([c:@/sfr/PSW] & 0x08) == 0 ? [@/mem/23] : [@/mem/31])", NULL },
    { &KC8051ViewWin::PSWEdit, "&PSW", "[@/sfr/PSW]", PSWBits },
    { &KC8051ViewWin::BEdit, "B", "[@/sfr/B]", NULL },
    { &KC8051ViewWin::SPEdit, "&SP", "[@/sfr/SP]", NULL },
    { &KC8051ViewWin::IEEdit, "IE", "[@/sfr/IE]", IEBits },
    { &KC8051ViewWin::IPEdit, "IP", "[@/sfr/IP]", IPBits },
    { &KC8051ViewWin::TMODEdit, "TMOD", "[@/sfr/TMOD]", TMODBits },
    { &KC8051ViewWin::TCONEdit, "TCON", "[@/sfr/TCON]", TCONBits },
    { &KC8051ViewWin::SCONEdit, "SCON", "[@/sfr/SCON]", SCONBits },
    { &KC8051ViewWin::PCONEdit, "PCON", "[@/sfr/PCON]", PCONBits },
    { &KC8051ViewWin::P0Edit, "P0", "[@/sfr/P0]", NULL },
    { &KC8051ViewWin::P1Edit, "P1", "[@/sfr/P1]", NULL },
    { &KC8051ViewWin::P2Edit, "P2", "[@/sfr/P2]", NULL },
    { &KC8051ViewWin::P3Edit, "P3", "[@/sfr/P3]", NULL },
    { &KC8051ViewWin::TH0Edit, "TH0", "[@/sfr/TH0]", NULL },
    { &KC8051ViewWin::TL0Edit, "TL0", "[@/sfr/TL0]", NULL },
    { &KC8051ViewWin::TH1Edit, "TH1", "[@/sfr/TH1]", NULL },
    { &KC8051ViewWin::TL1Edit, "Tl1", "[@/sfr/TL1]", NULL },
    { &KC8051ViewWin::DPHEdit, "DPH", "[@/sfr/DPH]", NULL },
    { &KC8051ViewWin::DPLEdit, "DPL", "[@/sfr/DPL]", NULL },
    { &KC8051ViewWin::PCEdit, "PC", "[@/misc16/PC]", NULL }
  };

/* List of KCBitEdit children and their associated names and expressions */
struct BitChild
{
  KCBitEdit *KC8051ViewWin::*edit;
};

/* List of KCBitEdit children */
static struct BitChild bitChildren[8] =
  {
    { &KC8051ViewWin::bit0Edit }, { &KC8051ViewWin::bit1Edit },
    { &KC8051ViewWin::bit2Edit }, { &KC8051ViewWin::bit3Edit },
    { &KC8051ViewWin::bit4Edit }, { &KC8051ViewWin::bit5Edit },
    { &KC8051ViewWin::bit6Edit }, { &KC8051ViewWin::bit7Edit }
  };


KC8051ViewWin::KC8051ViewWin (KCWindowProvider *provider, KCMainWin *mainWin,
			      struct me_entry *me, QWidget *parent,
			      const char *name)
  : KC8051ViewWinBase (parent, name),
    KCWindow (this, provider, me->name, mainWin), me_ (me),
    last_pc (-1) /* So that we redraw this time */
{
  setWFlags (getWFlags () | Qt::WDestructiveClose);

  setCaption (i18n ("8051: %1").arg (me->name));

  QString m (me->name);
  for (const struct PortChild *pc = portChildren;
       pc < portChildren + ARRAY_SIZE (portChildren); pc++)
    {
      (this->*pc->edit)->setMainWin (mainWin);
      (this->*pc->edit)->setExpression (pc->label, QString (pc->expr)
				  .replace (QRegExp ("@"), m));
    }
  for (const struct BitChild *pc = bitChildren;
       pc < bitChildren + ARRAY_SIZE (bitChildren); pc++)
    (this->*pc->edit)->setMainWin (mainWin);
  prev_pc = expr_compile (QString ("[%1/misc16/prev_pc]").arg (m), 0);
  if (prev_pc == NULL)
    {
      prev_pc = expr_compile ("0", 0);
      assert (prev_pc != NULL);
    }
  for (size_t i = 0; i < ARRAY_SIZE (prev_insn); i++)
    {
      prev_insn[i] = expr_compile (QString ("[%1/prev_insn/%2]").arg (m)
				   .arg (i), 0);
      if (prev_insn[i] == NULL)
	{
	  prev_insn[i] = expr_compile ("0", 0);
	  assert (prev_insn[i] != NULL);
	}
    }
  next_pc = expr_compile (QString ("[%1/misc16/PC]").arg (m), 0);
  if (prev_pc == NULL)
    {
      next_pc = expr_compile ("0", 0);
      assert (next_pc != NULL);
    }
  selectByte (PSWEdit);
  cerealUpdate ();
}

KC8051ViewWin::~KC8051ViewWin ()
{
  free (prev_pc);
  for (struct expr_code **p = prev_insn;
       p < prev_insn + ARRAY_SIZE (prev_insn); p++)
    free (*p);
  free (next_pc);
}

/* Display (well, not) error message */
static void
silent_error_handler (const QString &)
{
}

/** Update view because emulation state has changed */
void
KC8051ViewWin::cerealUpdate ()
{
  for (const struct PortChild *pc = portChildren;
       pc < portChildren + ARRAY_SIZE (portChildren); pc++)
    (this->*pc->edit)->cerealUpdate ();
  for (const struct BitChild *bc = bitChildren;
       bc < bitChildren + ARRAY_SIZE (bitChildren); bc++)
    (this->*bc->edit)->cerealUpdate ();

  intmax_t pc;
  uint8_t insn[3];
  const char *opcode;
  char buf[64];
  if ((emulation_result & ER_INSN) != 0 && expr_exec (prev_pc, &pc) == 0
      && last_pc != pc)
    {
      last_pc = pc;
      if (disasmList->count () == DISASM_LENGTH)
	disasmList->removeItem (0);
      assert (ARRAY_SIZE (insn) == ARRAY_SIZE (prev_insn));
      for (size_t i = 0; i < ARRAY_SIZE (insn); i++)
	{
	  intmax_t v;
	  
	  if (expr_exec (prev_insn[i], &v) != 0)
	    v = 0x00; /* NOP */
	  insn[i] = v;
	}
      i8051_disasm (pc, insn, &opcode, buf);
      disasmList->insertItem (QString ().sprintf ("0x%04X %-6s %s",
						  (unsigned)pc, opcode, buf));
      disasmList->setCurrentItem (disasmList->count () - 1);
    }
      
  void (*prev_error_handler) (const QString &msg) = error_handler;
  error_handler = silent_error_handler;
  if (expr_exec (next_pc, &pc) != 0)
    goto no_next;
  for (size_t i = 0; i < ARRAY_SIZE (insn); i++)
    {
      intmax_t v;

      if (expr_eval (QString ("[%1/next_insn/%2]").arg (me_->name).arg (i), &v)
	  != 0)
	{
	  if (i == 0)
	    goto no_next;
	  else
	    /* Allow the user to see the SJMP at the last 2 bytes of internal
	       memory ;-) */
	    v = 0x00;
	}
      insn[i] = v;
    }
  i8051_disasm (pc, insn, &opcode, buf);
  nextDisasmText->setText (QString ().sprintf ("0x%04X %-6s %s", (unsigned)pc,
					       opcode, buf));
  error_handler = prev_error_handler;
  return;

 no_next:
  nextDisasmText->setText (QString ().sprintf ("0x%04X ???", (unsigned)pc));
  error_handler = prev_error_handler;
}

/** Save current UI state under uiState */
void
KC8051ViewWin::saveState (xmlNodePtr uiState)
{
  xmlNodePtr node = KCUIStateHandler::saveWidget (this, "KC8051ViewWin",
						  uiState, NULL);
  xmlNsPtr ns = xmlNewNs (node, (const xmlChar *)KC8051XMLNs,
			  (const xmlChar *)"i8051");
  xmlSetNsProp (node, ns, (const xmlChar *)"me", (const xmlChar *)me_->name);
  char buf[32];
  assert (sizeof (last_pc) * CHAR_BIT <= 64);
  sprintf (buf, "%d", last_pc);
  xmlNewTextChild (node, ns, (const xmlChar *)"last_pc", (const xmlChar *)buf);
  QString data("");
  for (QListBoxItem *i = disasmList->firstItem (); i != 0; i = i->next ())
    data += i->text () + '\n';
  xmlNewTextChild (node, ns, (const xmlChar *)"disasm",
		   (const xmlChar *)(const char *)data.utf8 ()); /* Sheesh */
}

/** Load UI state from node of doc;
    returns: 0 if OK, -1 on error */
int
KC8051ViewWin::loadState (xmlDocPtr doc, xmlNodePtr node)
{
  if (KCUIStateHandler::loadWidget (this, doc, node) != 0)
    return -1;
  xmlNsPtr ns = xmlSearchNsByHref (doc, node, (const xmlChar *)KC8051XMLNs);
  if (ns == NULL)
    return 0;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE || node->ns != ns)
	continue;
      if (strcmp ((const char *)node->name, "last_pc") == 0)
	{
	  xmlChar *str = xmlNodeListGetString (doc, node->children, 1);
	  if (str == NULL)
	    goto err;
	  int val;
	  int res = sscanf ((const char *)str, "%d", &val) != 1;
	  xmlFree (str);
	  if (res != 0)
	    goto err;
	  last_pc = val;
	}
      else if (strcmp ((const char *)node->name, "disasm") == 0)
	{
	  disasmList->clear ();
	  xmlChar *str = xmlNodeListGetString (doc, node->children, 1);
	  if (str == NULL)
	    continue;
	  QStringList l
	    = QStringList::split ("\n", QString::fromUtf8 ((const char *)str));
	  xmlFree (str);
	  disasmList->insertStringList (l);
	  disasmList->setCurrentItem (disasmList->count () - 1);
	}
    }
  return 0;

 err:
  error (0, "Invalid 8051 ui_window state");
  return -1;
}

/** Byte edit widget was selected, use it in bits display */
void
KC8051ViewWin::selectByte (QWidget *widget)
{
  const struct PortChild *pc;
  for (pc = portChildren; pc < portChildren + ARRAY_SIZE (portChildren); pc++)
    {
      if (widget == this->*pc->edit)
	goto found;
    }
  return;

 found:
  QString byte (pc->label), m (me_->name);
  byte.replace (QRegExp ("&"), "");
  bitsBox->setTitle (i18n ("&Bits of %1").arg (byte));
  for (size_t i = 0; i < ARRAY_SIZE (bitChildren); i++)
    {
      KCBitEdit *e = this->*bitChildren[i].edit;
      QString name;
      if (pc->names != NULL)
	name = pc->names[i];
      else
	name = QString ("%1.%2").arg (byte).arg (i);
      e->setBit	(name, QString (pc->expr).replace (QRegExp ("@"), m), i);
    }
}

#include "kc8051viewwin.moc"
