/* Main window of khwconf
   Copyright (C) 2002, 2004 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */

#include <iostream>

#include <kaction.h>
#include <kapplication.h>
#include <kfiledialog.h>
#include <klineedit.h>
#include <klistview.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kpopupmenu.h>
#include <kstdaction.h>
#include <ktempfile.h>
#include <qobject.h>

#include <me.h>
#include <module.h>
#include <setup.h>

#include "khwinsertdialog.h"
#include "khwmainwin.h"
#include "khwmilist.h"
#include "khwpropertiesdialog.h"

/** Create a new window, optionally loading FILE if it is non-NULL */
KHwMainWin::KHwMainWin (const char *file)
  : KMainWindow (), modified (false), filename ("")
{
  KStdAction::openNew (this, SLOT (fileNew ()), actionCollection ());
  KStdAction::open (this, SLOT (fileOpen ()), actionCollection ());
  KStdAction::save (this, SLOT (fileSave ()), actionCollection ());
  KStdAction::saveAs (this, SLOT (fileSaveAs ()), actionCollection ());
  KStdAction::quit (this, SLOT (fileQuit ()), actionCollection ());
  new KAction (i18n ("&Insert..."), 0, this, SLOT (itemInsert ()),
	       actionCollection (), "iteminsert");
  renameAction = new KAction (i18n ("&Rename"), 0, this, SLOT (itemRename ()),
			     actionCollection (), "itemrename");
  renameAction->setEnabled (false);
  deleteAction = new KAction (i18n ("&Delete"), 0, this, SLOT (itemDelete ()),
			      actionCollection (), "itemdelete");
  deleteAction->setEnabled (false);
  propertiesAction = new KAction (i18n ("&Properties..."), 0, this,
				  SLOT (itemProperties ()),
				  actionCollection (), "itemproperties");
  propertiesAction->setEnabled (false);
  createGUI ();
  setAutoSaveSettings ();
  updateCaption ();

  miList = new KHwMiList (this);
  setCentralWidget (miList);
  miList->restoreLayout (KApplication::kApplication ()->config (), "MI List");
  QObject::connect (miList, SIGNAL (selectionChanged ()), this,
		    SLOT (listSelectionChanged ()));
  QObject::connect (miList, SIGNAL (deletePressed ()), this,
		    SLOT (itemDelete ()));
  QObject::connect (miList, SIGNAL (itemRenamed (QListViewItem *)), this,
		    SLOT (listItemRenamed (QListViewItem *)));
  QObject::connect (miList, SIGNAL (doubleClicked (QListViewItem *)), this,
		    SLOT (itemProperties ()));
  QObject::connect (miList, SIGNAL (returnPressed (QListViewItem *)), this,
		    SLOT (itemProperties ()));
  QObject::connect (miList, SIGNAL (contextMenu (KListView *, QListViewItem *,
						 const QPoint &)), this,
		    SLOT (listContextMenu (KListView *, QListViewItem *,
					   const QPoint &)));
  if (file != 0)
    loadFile (file);
}

KHwMainWin::~KHwMainWin ()
{
}

/** @return true if the window can be closed. */
bool
KHwMainWin::queryClose ()
{
  return canForgetData ();
}

/** Save settings before window is closed */
bool
KHwMainWin::queryExit ()
{
  miList->saveLayout (KApplication::kApplication ()->config (), "MI List");
  return KMainWindow::queryExit ();
}

/** File/New was selected */
void
KHwMainWin::fileNew ()
{
  if (!canForgetData ())
    return;
  setup_destroy ();
  setup_init ();
  modified = false;
  filename = "";
  updateCaption ();
  miList->reload ();
}

/** File/Open was selected */ 
void
KHwMainWin::fileOpen ()
{
  if (!canForgetData ())
    return;
  QString fn = KFileDialog::getOpenFileName (0, "*.xml", this);
  if (!fn.isEmpty ())
    loadFile (fn);
}

/** File/Save was selected */
void
KHwMainWin::fileSave ()
{
  if (filename.isEmpty ())
    fileSaveAs ();
  else if (setup_save (filename) == 0 && modified)
    {
      modified = false;
      updateCaption ();
    }
}

/** File/Save as was selected */
void
KHwMainWin::fileSaveAs ()
{
  QString fn = KFileDialog::getSaveFileName (0, "*.xml", this);
  if (!fn.isEmpty () && setup_save (fn) == 0)
    {
      filename = fn;
      modified = false;
      updateCaption ();
    }
}

/** File/Quit was selected */
void
KHwMainWin::fileQuit ()
{
  close ();
}

/** Item/Insert was selected */
void
KHwMainWin::itemInsert ()
{
  KHwInsertDialog *dlg = new KHwInsertDialog (this);
  if (dlg->exec () == QDialog::Accepted)
    {
      const struct cereal_module *module;

      module = cereal_module_find (dlg->getModuleType ());
      if (me_new (module, dlg->getModuleName ()) != NULL)
	{
	  miList->reload ();
	  modified = true;
	  updateCaption ();
	}
    }
  delete dlg;
}

/** Item/Rename was selected */
void
KHwMainWin::itemRename ()
{
  QListViewItem *curr = miList->selectedItem ();
  if (curr != 0)
    miList->rename (curr, 0);
}

/** Item/Delete was selected */
void
KHwMainWin::itemDelete ()
{
  QListViewItem *curr = miList->selectedItem ();
  if (curr == 0)
    return;
  QString name = curr->text (0);
  if (KMessageBox
      ::warningYesNo (0, i18n ("Do you really want to delete module `%1'?")
		      .arg (name)) == KMessageBox::Yes)
    {
      me_delete (name);
      miList->reload ();
      modified = true;
      updateCaption ();
    }
}

/** Item/Properties was selected */
void
KHwMainWin::itemProperties ()
{
  QListViewItem *curr = miList->selectedItem ();
  if (curr == 0)
    return;
  KTempFile tempFile;
  setup_save (tempFile.name ());
  KHwPropertiesDialog *dlg =
    new KHwPropertiesDialog (this, me_find (curr->text (0)));
  if (dlg->exec () != QDialog::Accepted)
    setup_load (tempFile.name ());
  else
    {
      modified = true;
      updateCaption ();
    }
  delete dlg;
  tempFile.unlink ();
}

/** Selection of miList has changed */
void
KHwMainWin::listSelectionChanged ()
{
  bool enable = miList->selectedItem () != 0;

  renameAction->setEnabled (enable);
  deleteAction->setEnabled (enable);
  propertiesAction->setEnabled (enable);
}

/** MI list item was renamed */
void
KHwMainWin::listItemRenamed (QListViewItem *item)
{
  if (me_rename (item->text (2), item->text (0)) == 0)
    item->setText (2, item->text (0));
  else
    item->setText (0, item->text (2));
}

/** MI list requests a context menu */
void
KHwMainWin::listContextMenu (KListView *, QListViewItem *, const QPoint &p)
{
  QPopupMenu *menu = static_cast <QPopupMenu *>
    (factory ()->container ("list_menu", this));
  menu->exec (p);
}

/** @returns true if current setup can be overwritten */
bool
KHwMainWin::canForgetData ()
{
  if (!modified)
    return true;
  switch (KMessageBox
	  ::warningYesNoCancel (0, i18n ("Current setup has been changed.\n"
					 "Do you want to save your changes?")))
    {
    case KMessageBox::Yes:
      fileSave ();
      return !modified;

    case KMessageBox::No:
      return true;
      
    default:
      return false;
    }
}

/** @returns file name we display to the user */
QString
KHwMainWin::prettyFilename ()
{
  if (filename.isEmpty ())
    return i18n ("Untitled");
  KURL url (filename);
  return url.fileName ();
}

/** Update window caption */
void
KHwMainWin::updateCaption (void)
{
  setCaption (prettyFilename (), modified);
}

/** Load setup from file */
void
KHwMainWin::loadFile (const QString &file)
{
  if (setup_load (file) != 0)
    {
      modified = false;
      fileNew (); /* Create a reasonable state */
      return;
    }
  filename = file;
  modified = false;
  updateCaption ();
  miList->reload ();
}

#include "khwmainwin.moc"
