/* Connection modify dialog
   Copyright (C) 2002, 2004, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */

#include <cassert>
#include <cstddef>
#include <libxml/tree.h>
#include <kcombobox.h>
#include <klocale.h>
#include <kpushbutton.h>
#include <kstdguiitem.h>
#include <qcheckbox.h>
#include <qstring.h>
#include <qstringlist.h>

#include <lib.h>
#include <me.h>
#include <module.h>
#include <port.h>

#include "khwconf.h"
#include "khwmodifyconnectiondialog.h"

/** Create a connection modify dialog for conn_ and possibly conn_peer_ (if
    non-NULL) */
KHwModifyConnectionDialog::KHwModifyConnectionDialog (QWidget *parent,
						      struct connect_entry
						      *conn_,
						      struct connect_entry
						      *conn_peer_)
  : KHwModifyConnectionDialogBase (parent), conn (conn_),
    conn_peer (conn_peer_), changing (false), peer_val (true)
{
  selectedConn.pi1.me = conn->pi1.me;
  init ();
}

/** Create a dialog to add a connection from module */
KHwModifyConnectionDialog::KHwModifyConnectionDialog (QWidget *parent,
						      struct me_entry *module)
  : KHwModifyConnectionDialogBase (parent), conn (NULL), conn_peer (NULL),
    changing (false), peer_val (true)
{
  setCaption (i18n ("Add Connection"));
  modifyPeerCheckbox->setText (i18n ("Add also &reverse connection"));
  selectedConn.pi1.me = module;
  init ();
}

/** The real "constructor", common to both of them */
void
KHwModifyConnectionDialog::init ()
{
  cancelButton->setGuiItem (KStdGuiItem::cancel ());
  OKButton->setGuiItem (KStdGuiItem::ok ());

  /* The order of building the data is not just "as the user sees it", it is
     also sorted so that elements of selectedConn variables are not accessed
     unitialized. */
  changing = true;
  for (int i = 0; i < NUM_PORT_TYPES; i++)
    connectionType->insertItem (port_type_name[i]);
  if (conn != NULL)
    connectionType->setCurrentItem (conn->type);
  selectedConn.type = connectionType->currentItem ();
  
  for (enum port_width w = WIDTH_FIRST; w < NUM_WIDTHS; w++)
    {
      connectionWidth->insertItem (QString::number (port_width_value[w]));
      if (conn != NULL && w == conn->width)
	connectionWidth->setCurrentItem (w - WIDTH_FIRST);
    }
  selectedConn.width = static_cast <enum port_width>
    (connectionWidth->currentItem () + WIDTH_FIRST);
  
  rebuildSourceSpace ();
  if (conn != NULL)
    {
      for (int i = 0; i < sourceSpace->count (); i++)
	{
	  size_t space;
	  if (space_find_name (selectedConn.pi1.me->module,
			       sourceSpace->text (i).utf8 (),
			       selectedConn.width, &space)
	      == selectedConn.width && space == conn->pi1.space)
	    {
	      sourceSpace->setCurrentItem (i);
	      break;
	    }
	}
    }
  sourceSpaceChanged ();

  rebuildSourcePort ();
  if (conn != NULL)
    {
      xmlChar *str = port_create_name (selectedConn.width, &conn->pi1);
      sourcePort->setCurrentText (QString::fromUtf8 ((const char *)str));
      xmlFree (str);
    }
  sourcePortChanged ();

  QStringList modules;
  for (const struct me_entry *me = me_list; me != NULL; me = me->next)
    modules.append (me->name);
  modules.sort ();
  destModule->insertStringList (modules);
  if (conn != NULL)
    {
      QString val = conn->pi2.me->name;
      for (int i = 0; i < destModule->count (); i++)
	{
	  if (destModule->text (i) == val)
	    {
	      destModule->setCurrentItem (i);
	      break;
	    }
	}
    }
  selectedConn.pi2.me = me_find (destModule->currentText ());

  rebuildDestSpace ();
  if (conn != NULL)
    {
      for (int i = 0; i < destSpace->count (); i++)
	{
	  size_t space;
	  if (space_find_name (selectedConn.pi2.me->module,
			       destSpace->text (i).utf8 (), selectedConn.width,
			       &space) == selectedConn.width
	      && space == conn->pi2.space)
	    {
	      destSpace->setCurrentItem (i);
	      break;
	    }
	}
    }
  destSpaceChanged ();

  rebuildDestPort ();
  if (conn != NULL)
    {
      xmlChar *str = port_create_name (selectedConn.width, &conn->pi2);
      destPort->setCurrentText (QString::fromUtf8 ((const char *)str));
      xmlFree (str);
    }
  destPortChanged ();

  rebuildPeerOK ();
  if (modifyPeerCheckbox->isEnabled ())
    {
      modifyPeerCheckbox->setChecked (true);
      peer_val = true;
    }
  changing = false;
}

KHwModifyConnectionDialog::~KHwModifyConnectionDialog ()
{
}

/** Is modifyPeerCheckbox checked? */
bool
KHwModifyConnectionDialog::modifyPeerChecked () const
{
  return modifyPeerCheckbox->isChecked ();
}

/** Port type has changed */
void
KHwModifyConnectionDialog::typeChanged ()
{
  selectedConn.type = connectionType->currentItem ();
  if (changing == true)
    return;
  changing = true;
  rebuildSourceSpace ();
  rebuildSourcePort ();
  rebuildDestSpace ();
  rebuildDestPort ();
  rebuildPeerOK ();
  changing = false;
}

/** Port width has changed */
void 
KHwModifyConnectionDialog::widthChanged ()
{
  selectedConn.width = static_cast <enum port_width>
    (connectionWidth->currentItem () + WIDTH_FIRST);
  if (changing == true)
    return;
  changing = true;
  rebuildSourceSpace ();
  rebuildSourcePort ();
  rebuildDestSpace ();
  rebuildDestPort ();
  rebuildPeerOK ();
  changing = false;
}

/** Source space has changed */
void 
KHwModifyConnectionDialog::sourceSpaceChanged ()
{
  if (sourceSpace->count () == 0
      || space_find_name (selectedConn.pi1.me->module,
			  sourceSpace->currentText ().utf8 (),
			  selectedConn.width, &selectedConn.pi1.space)
      != selectedConn.width)
    selectedConn.pi1.space = 0;
  if (changing == true)
    return;
  changing = true;
  rebuildSourcePort ();
  rebuildPeerOK ();
  changing = false;
}

/** Source port has changed */
void
KHwModifyConnectionDialog::sourcePortChanged ()
{
  QString val = sourcePort->currentText ();
  if (val.isEmpty ())
    selectedConn.pi1.port = 0;
  else
    port_find_name (selectedConn.width, &selectedConn.pi1, val);
  if (changing == true)
    return;
  changing = true;
  rebuildPeerOK ();
  changing = false;
}

/** Destination port has changed */
void 
KHwModifyConnectionDialog::destModuleChanged ()
{
  selectedConn.pi2.me = me_find (destModule->currentText ());
  assert (selectedConn.pi2.me != NULL);
  if (changing == true)
    return;
  changing = true;
  rebuildDestSpace ();
  rebuildDestPort ();
  rebuildPeerOK ();
  changing = false;
}

/** Destination space has changed */
void 
KHwModifyConnectionDialog::destSpaceChanged ()
{
  if (destSpace->count () == 0
      || space_find_name (selectedConn.pi2.me->module,
			  destSpace->currentText ().utf8 (),
			  selectedConn.width, &selectedConn.pi2.space)
      != selectedConn.width)
    selectedConn.pi2.space = 0;
  if (changing == true)
    return;
  changing = true;
  rebuildDestPort ();
  rebuildPeerOK ();
  changing = false;
}

/** Destination port has changed */
void 
KHwModifyConnectionDialog::destPortChanged ()
{
  QString val = destPort->currentText ();
  if (val.isEmpty ())
    selectedConn.pi2.port = 0;
  else
    port_find_name (selectedConn.width, &selectedConn.pi2, val);
  if (changing == true)
    return;
  changing = true;
  rebuildPeerOK ();
  changing = false;
}

/** Make dest a peer connection to src (which is PORT_TYPE_{READ,WRITE} */
static void
makePeer (struct connect_entry *dest, const struct connect_entry *src)
{
  dest->type = src->type == PORT_TYPE_READ ? PORT_TYPE_WRITE : PORT_TYPE_READ;
  dest->width = src->width;
  dest->pi1 = src->pi2;
  dest->pi2 = src->pi1;
}

/** Check if data is OK, and accept () if so */
void
KHwModifyConnectionDialog::tryAccept ()
{
  if (conn != NULL)
    {
      if (conn->type == selectedConn.type
	  && conn->width == selectedConn.width
	  && conn->pi1.me == selectedConn.pi1.me
	  && conn->pi2.me == selectedConn.pi2.me
	  && conn->pi1.space == selectedConn.pi1.space
	  && conn->pi2.space == selectedConn.pi2.space
	  && conn->pi1.port == selectedConn.pi1.port
	  && conn->pi2.port == selectedConn.pi2.port)
	reject (); /* No change */
      port_do_disconnect (conn);
      if (modifyPeerCheckbox->isChecked ())
	port_do_disconnect (conn_peer);
      conn->type = selectedConn.type;
      conn->width = selectedConn.width;
      conn->pi1 = selectedConn.pi1;
      conn->pi2 = selectedConn.pi2;
      port_do_connect (conn);
      if (modifyPeerCheckbox->isChecked ())
	{
	  makePeer (conn_peer, &selectedConn);
	  port_do_connect (conn_peer);
	}
    }
  else
    {
      port_connect (&selectedConn);
      if (modifyPeerCheckbox->isChecked ())
	{
	  struct connect_entry peer_skel;
	  makePeer (&peer_skel, &selectedConn);
	  port_connect (&peer_skel);
	}
    }
  accept ();
}

/** Modify source space list according to other selections, try to preserve
    currently selected item */
void
KHwModifyConnectionDialog::rebuildSourceSpace ()
{
  QString old_val = sourceSpace->currentText ();
  sourceSpace->clear ();
  const struct cereal_module *mod = selectedConn.pi1.me->module;
  const struct module_width *mw = mod->w + selectedConn.width;
  int index = -1;
  for (size_t s = 0; s < mw->num_spaces; s++)
    {
      if (mw->spaces[s].set_fn[selectedConn.type] != NULL)
	{
	  xmlChar *str = space_create_name (mod, selectedConn.width, s);
	  QString val = QString::fromUtf8 ((const char *)str);
	  xmlFree (str);
	  sourceSpace->insertItem (val);
	  if (val == old_val)
	    index = sourceSpace->count () - 1;
	}
    }
  if (index != -1)
    /* This retains the name, but may change the space index */
    sourceSpace->setCurrentItem (index);
  sourceSpaceChanged ();
}

/** Modify source port list according to other selections, try to preserve
    currently selected item */
void
KHwModifyConnectionDialog::rebuildSourcePort ()
{
  QString old_val = sourcePort->currentText ();
  sourcePort->clear ();
  if (sourceSpace->count () == 0)
    return;
  struct port_id *const pi1 = &selectedConn.pi1;
  xmlNodePtr node = space_get_xml (pi1->me->module, selectedConn.width,
				   pi1->space);
  /* Get current connections out of the way when we're looking for possible
     ports. */
  if (conn != NULL)
    {
      port_do_disconnect (conn);
      if (conn_peer != NULL)
	port_do_disconnect (conn_peer);
    }
  if (node != NULL)
    {
      for (node = node->children; node != NULL; node = node->next)
	{
	  xmlChar *str;
	  unsigned long port;
	  int res;
	  
	  if (node->type != XML_ELEMENT_NODE
	      || strcmp ((const char *)node->name, "port") != 0)
	    continue;
	  str = xmlGetProp (node, (const xmlChar *)"index");
	  if (str == NULL)
	    continue;
	  res = get_num ((const char *)str, 0, &port);
	  xmlFree (str);
	  if (res != 0)
	    continue;
	  str = xmlGetProp (node, (const xmlChar *)"name");
	  if (str == NULL)
	    continue;
	  const struct space *sp = pi1->me->module->w[selectedConn.width]
	    .spaces + pi1->space;
	  switch (selectedConn.width)
	    {
#define DO_TYPE(TYPE, WIDTH)						      \
	      {								      \
		struct port_fn pf;					      \
									      \
		/* We need to get a non-NULL pointer of the given type.	      \
		   We know this conversion is allowed and we can convert      \
		   back, so we also know that the converted value is	      \
		   non-NULL, assuming there is only one NULL value.  This     \
		   is not  portable, but still better than (xxx_t)1 */	      \
		pf.fn.TYPE##_##WIDTH = (port_##WIDTH##_##TYPE##_t)get_num;    \
		assert (pf.fn.TYPE##_##WIDTH != NULL);			      \
		res = sp->set_fn[selectedConn.type] (pi1->me->mi, port, &pf); \
		if (res == 0)						      \
		  {							      \
		    pf.fn.TYPE##_##WIDTH = NULL;			      \
		    sp->set_fn[selectedConn.type] (pi1->me->mi, port, &pf);   \
		  }							      \
		break;							      \
	      }
#define DO_WIDTH(WIDTH)				\
	    case WIDTH_##WIDTH:			\
	      {					\
		switch (selectedConn.type)	\
		  {				\
		  case PORT_TYPE_READ:		\
		    DO_TYPE (read, WIDTH);	\
		    break;			\
		  case PORT_TYPE_WRITE:		\
		    DO_TYPE (write, WIDTH);	\
		    break;			\
		  case PORT_TYPE_DISPLAY:	\
		    DO_TYPE (display, WIDTH);	\
		    break;			\
		  case PORT_TYPE_MODIFY:	\
		    DO_TYPE (modify, WIDTH);	\
		    break;			\
		  default:			\
		    assert ("BUG" == NULL);	\
		  }				\
		break;				\
	      }
	      DO_WIDTH (1);
	      DO_WIDTH (8);
	      DO_WIDTH (16);

	    default:
	      assert ("BUG" == NULL);
#undef DO_WIDTH
#undef DO_TYPE
	    }
	  if (res == 0)
	    sourcePort->insertItem (QString::fromUtf8 ((const char *)str));
	  xmlFree (str);
	}
    }
  if (conn != NULL)
    {
      port_do_connect (conn);
      if (conn_peer != NULL)
	port_do_connect (conn_peer);
    }
  if (old_val.isEmpty ()
      || port_find_name (selectedConn.width, pi1, old_val) != 0)
    sourcePort->setCurrentText (QString::number (0));
  else
    /* This retains the name, but may change the port index */
    sourcePort->setCurrentText (old_val);
  sourcePortChanged ();
}

/** Modify destination space list according to other selections, try to
    preserve currently selected item */
void
KHwModifyConnectionDialog::rebuildDestSpace ()
{
  QString old_val = destSpace->currentText ();
  destSpace->clear ();
  const struct cereal_module *mod = selectedConn.pi2.me->module;
  const struct module_width *mw = mod->w + selectedConn.width;
  int index = -1;
  for (size_t s = 0; s < mw->num_spaces; s++)
    {
      if (mw->spaces[s].get_fn[selectedConn.type] != NULL)
	{
	  xmlChar *str = space_create_name (mod, selectedConn.width, s);
	  QString val = QString::fromUtf8 ((const char *)str);
	  xmlFree (str);
	  destSpace->insertItem (val);
	  if (val == old_val)
	    index = destSpace->count () - 1;
	}
    }
  if (index != -1)
    /* This retains the name, but may change the space index */
    destSpace->setCurrentItem (index);
  destSpaceChanged ();
}

/** Modify destination port list according to other selections, try to preserve
    currently selected item */
void
KHwModifyConnectionDialog::rebuildDestPort ()
{
  QString old_val = destPort->currentText ();
  destPort->clear ();
  if (destSpace->count () == 0)
    return;
  struct port_id *const pi2 = &selectedConn.pi2;
  xmlNodePtr node = space_get_xml (pi2->me->module, selectedConn.width,
				   pi2->space);
  if (node != NULL)
    {
      for (node = node->children; node != NULL; node = node->next)
	{
	  xmlChar *str;
	  unsigned long port;
	  int res;
	  
	  if (node->type != XML_ELEMENT_NODE
	      || strcmp ((const char *)node->name, "port") != 0)
	    continue;
	  str = xmlGetProp (node, (const xmlChar *)"index");
	  if (str == NULL)
	    continue;
	  res = get_num ((const char *)str, 0, &port);
	  xmlFree (str);
	  if (res != 0)
	    continue;
	  str = xmlGetProp (node, (const xmlChar *)"name");
	  if (str == NULL)
	    continue;
	  switch (selectedConn.width)
	    {
#define DO_TYPE(TYPE, WIDTH)				\
	      {						\
		res = pf.fn.TYPE##_##WIDTH != NULL;	\
		break;					\
	      }
#define DO_WIDTH(WIDTH)							  \
	    case WIDTH_##WIDTH:						  \
	      {								  \
		struct port_fn pf;					  \
		pi2->me->module->w[selectedConn.width].spaces[pi2->space] \
		  .get_fn[selectedConn.type] (pi2->me->mi, port, &pf);	  \
		switch (selectedConn.type)				  \
		  {							  \
		  case PORT_TYPE_READ:					  \
		    DO_TYPE (read, WIDTH);				  \
		    break;						  \
		  case PORT_TYPE_WRITE:					  \
		    DO_TYPE (write, WIDTH);				  \
		    break;						  \
		  case PORT_TYPE_DISPLAY:				  \
		    DO_TYPE (display, WIDTH);				  \
		    break;						  \
		  case PORT_TYPE_MODIFY:				  \
		    DO_TYPE (modify, WIDTH);				  \
		    break;						  \
		  default:						  \
		    assert ("BUG" == NULL);				  \
		  }							  \
		break;							  \
	      }
	      DO_WIDTH (1);
	      DO_WIDTH (8);
	      DO_WIDTH (16);

	    default:
	      assert ("BUG" == NULL);
#undef DO_WIDTH
#undef DO_TYPE
	    }
	  if (res != 0)
	    destPort->insertItem (QString::fromUtf8 ((const char *)str));
	  xmlFree (str);
	}
    }
  if (old_val.isEmpty ()
      || port_find_name (selectedConn.width, pi2, old_val) != 0)
    destPort->setCurrentText (QString::number (0));
  else
    /* This retains the name, but may change the port index */
    destPort->setCurrentText (old_val);
  destPortChanged ();
}

/* Display (well, not) error message */
static void
silent_error_handler (const QString &)
{
}

/** Update state of modifyPeerCheckbox and OKButton according to other
    selections, try to preserve currently selected item */
void
KHwModifyConnectionDialog::rebuildPeerOK ()
{
  void (*prev_error_handler) (const QString &msg) = error_handler;
  error_handler = silent_error_handler;
  
  struct connect_entry *t1, *t2;
  if (sourceSpace->count () == 0 || destSpace->count () == 0)
    goto invalid;
  if (port_find_name (selectedConn.width, &selectedConn.pi1,
		      sourcePort->currentText ()) != 0)
    goto invalid;
  if (port_find_name (selectedConn.width, &selectedConn.pi2,
		      destPort->currentText ()) != 0)
    goto invalid;
  /* We play games with connecting/disconnecting, but we must keep conn and
     conn_peer at their original addresses, because KHwPropertiesDialog may be
     repainted any time, which requires access to the connect_entries (the list
     caches the pointers).  This also spares us some malloc () and free ()
     calls, but we must be careful. */
  if (conn != NULL)
    {
      port_do_disconnect (conn);
      if (conn_peer != NULL)
	port_do_disconnect (conn_peer);
    }
  t1 = port_connect (&selectedConn);
  if (conn == NULL || conn_peer != NULL)
    {
      switch (selectedConn.type)
	{
	case PORT_TYPE_READ: case PORT_TYPE_WRITE:
	  {
	    struct connect_entry peer_skel;
	    makePeer (&peer_skel, &selectedConn);
	    t2 = port_connect (&peer_skel);
	    break;
	  }
	  
	case PORT_TYPE_DISPLAY: case PORT_TYPE_MODIFY:
	  t2 = NULL;
	  break;
	  
	default:
	  assert ("BUG" == NULL);
	  goto invalid;
	}
    }
  else
    t2 = NULL;
  if (t1 != NULL)
    port_disconnect (t1);
  if (t2 != NULL)
    port_disconnect (t2);
  if (conn != NULL)
    {
      port_do_connect (conn);
      if (conn_peer != NULL)
	port_do_connect (conn_peer);
    }
  if (t1 == NULL)
    goto invalid;
  enablePeer (t2 != NULL);
  OKButton->setEnabled (true);
  error_handler = prev_error_handler;
  return;
  
 invalid:
  enablePeer (false);
  OKButton->setEnabled (false);
  error_handler = prev_error_handler;
}

/** Enable/disable modifyPeerCheckbox, preserving user's last selection
    during the time the checkbox is disabled */
void
KHwModifyConnectionDialog::enablePeer (bool enable)
{
  if (enable == false)
    {
      if (modifyPeerCheckbox->isEnabled ())
	{
	  peer_val = modifyPeerCheckbox->isChecked ();
	  if (peer_val == true)
	    modifyPeerCheckbox->setChecked (false);
	}
    }
  else if (!modifyPeerCheckbox->isEnabled () && peer_val == true)
    modifyPeerCheckbox->setChecked (true);
  modifyPeerCheckbox->setEnabled (enable);
}

#include "khwmodifyconnectiondialog.moc"
