/* Item/Properties dialog
   Copyright (C) 2002, 2004, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */

#include <cstdlib>
#include <klineedit.h>
#include <klistview.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kpushbutton.h>
#include <kstdguiitem.h>
#include <libxml/tree.h>
#include <qlabel.h>
#include <qstring.h>
#include <qtabwidget.h>

#include <me.h>
#include <module.h>
#include <port.h>

#include "khwdeleteconnectiondialog.h"
#include "khwmodifyconnectiondialog.h"
#include "khwpropertiesdialog.h"

class KHwConnListItem: public KListViewItem
{
public:
  KHwConnListItem (QListView *parent, struct connect_entry *conn);
  ~KHwConnListItem ();
  QString text (int column) const;
  int compare (QListViewItem *i, int col, bool ascending) const;
  struct connect_entry *conn () const { return conn_; };

private:
  struct connect_entry *conn_;
};

KHwConnListItem::KHwConnListItem (QListView *parent,
				  struct connect_entry *conn)
  : KListViewItem (parent), conn_ (conn)
{
}

KHwConnListItem::~KHwConnListItem ()
{
}

int
KHwConnListItem::compare (QListViewItem *i_, int col, bool) const
{
  const KHwConnListItem *i = static_cast <KHwConnListItem *> (i_);
  switch (col)
    {
    case 0:
      {
	int res = QString::localeAwareCompare (text (0), i->text (0));
	if (res < 0)
	  return -1;
	if (res > 0)
	  return 1;
	return 0;
      }

    case 1:
      {
	unsigned w1 = port_width_value[conn ()->width];
	unsigned w2 = port_width_value[i->conn ()->width];

	if (w1 < w2)
	  return -1;
	if (w1 > w2)
	  return 1;
	return 0;
      }

    case 2:
      {
	const struct connect_entry *ce1 = conn (), *ce2 = i->conn ();

	if (ce1->pi1.space < ce2->pi1.space)
	  return -1;
	if (ce1->pi1.space > ce2->pi1.space)
	  return 1;
	if (ce1->pi1.port < ce2->pi1.port)
	  return -1;
	if (ce1->pi1.port > ce2->pi1.port)
	  return 1;
	return 0;
      }

    case 3:
      {
	const struct connect_entry *ce1 = conn (), *ce2 = i->conn ();
	int res = QString::localeAwareCompare (ce1->pi2.me->name,
					       ce2->pi2.me->name);
	if (res < 0)
	  return -1;
	if (res> 0)
	  return 1;
	
	if (ce1->pi2.space < ce2->pi2.space)
	  return -1;
	if (ce1->pi2.space > ce2->pi2.space)
	  return 1;
	
	if (ce1->pi2.port < ce2->pi2.port)
	  return -1;
	if (ce1->pi2.port > ce2->pi2.port)
	  return 1;
	return 0;
      }
      
    default:
      return 0;
    }
}

/** Create name for pi of width */
static QString
name_space_port (enum port_width width, const struct port_id *pi)
{
  xmlChar *str = space_create_name (pi->me->module, width, pi->space);
  QString res = QString::fromUtf8 ((const char *)str);
  xmlFree (str);
  str = port_create_name (width, pi);
  res.append ('/').append (QString::fromUtf8 ((const char *)str));
  xmlFree (str);
  return res;
}

QString
KHwConnListItem::text (int column) const
{
  const struct connect_entry *ce = conn ();;
  switch (column)
    {
    case 0:
      return i18n (port_type_name[ce->type]);

    case 1:
      return QString::number (port_width_value[ce->width]);

    case 2:
      return name_space_port (ce->width, &ce->pi1);

    case 3:
      {
	QString res;
	res.sprintf ("%s/", ce->pi2.me->name);
	return res.append (name_space_port (ce->width, &ce->pi2));
      }

    default:
      return "";
    }
}

KHwPropertiesDialog::KHwPropertiesDialog (QWidget *parent,
					  struct me_entry *me_)
  : KHwPropertiesDialogBase (parent), me (me_), prev_option (0),
    descriptions (), modified (false), last_err_option (0)
{
  setCaption (i18n ("Properties of %1").arg (me->name));
  cancelButton->setGuiItem (KStdGuiItem::cancel ());
  OKButton->setGuiItem (KStdGuiItem::ok ());
  optionValue->setEnabled (false);
  deleteConnectionButton->setEnabled (false);
  modifyConnectionButton->setEnabled (false);
  if (me->module->xml != 0)
    {
      xmlNodePtr node = xmlDocGetRootElement (me->module->xml);
      for (node = node->children; node != 0; node = node->next)
	{
	  if (node->type != XML_ELEMENT_NODE)
	    continue;
	  if (strcmp ((const char *)node->name, "options") == 0)
	    {
	      for (xmlNodePtr option = node->children; option != 0;
		   option = option->next)
		{
		  xmlChar *name, *descr;

		  name = xmlGetProp (option, (const xmlChar *)"name");
		  if (name == NULL)
		    continue;
		  QString qname(QString::fromUtf8 ((const char *)name));
		  new KListViewItem (optionName, qname);
		  descr = xmlNodeListGetString (me->module->xml,
						option->children, 1);
		  descriptions[qname] =
		    QString::fromUtf8 ((const char *)descr);
		  xmlFree (descr);
		  xmlFree (name);
		}
	    }
	}
      optionName->sort ();
      QListViewItem *first = optionName->firstChild ();
      if (first != 0)
	{
	  optionName->setCurrentItem (first);
	  optionName->setSelected (first, false);
	}
    }
  rebuildConnectionList ();
}

KHwPropertiesDialog::~KHwPropertiesDialog ()
{
}

/** connection selection has changed */
void
KHwPropertiesDialog::connectionSelectionChanged ()
{
  bool enable = connectionList->selectedItem () != 0;
  deleteConnectionButton->setEnabled (enable);
  modifyConnectionButton->setEnabled (enable);
}

/** Delete selected connection */
void
KHwPropertiesDialog::deleteConnection ()
{
  QListViewItem *curr = connectionList->selectedItem ();
  if (curr == 0)
    return;
  KHwConnListItem *c = static_cast <KHwConnListItem *> (curr);
  struct connect_entry *ce = c->conn ();
  struct connect_entry *ce2 = port_peer_connection (ce);
  KHwDeleteConnectionDialog *dlg = new KHwDeleteConnectionDialog (this);
  dlg->setDeletePeerChecked (ce2 != NULL);
  if (ce2 == NULL)
    dlg->setDeletePeerEnabled (false);
  if (dlg->exec () == QDialog::Accepted)
    {
      if (dlg->deletePeerChecked ())
	{
	  KHwConnListItem *c2 = findConn (ce2);
	  if (c2 != 0)
	    delete c2;
	  port_disconnect (ce2);
	}
      delete c;
      port_disconnect (ce);
      modified = true;
    }
  delete dlg;
}

/** Modify selected connection */
void
KHwPropertiesDialog::modifyConnection()
{
  QListViewItem *curr = connectionList->selectedItem ();
  if (curr == 0)
    return;
  KHwConnListItem *c = static_cast <KHwConnListItem *> (curr);
  struct connect_entry *ce = c->conn ();
  struct connect_entry *ce2 = port_peer_connection (ce);
  KHwModifyConnectionDialog *dlg = new KHwModifyConnectionDialog (this, ce,
								  ce2);
  if (dlg->exec () == QDialog::Accepted)
    {
      c->repaint ();
      if (dlg->modifyPeerChecked ())
	findConn (ce2)->repaint ();
      modified = true;
    }
  delete dlg;
}

/** Create a new connection */
void
KHwPropertiesDialog::newConnection()
{
  KHwModifyConnectionDialog *dlg = new KHwModifyConnectionDialog (this, me);
  if (dlg->exec () == QDialog::Accepted)
    {
      rebuildConnectionList ();
      modified = true;
    }
  delete dlg;
}

/** Option selection has changed */
void
KHwPropertiesDialog::optionSelectionChanged ()
{
  QListViewItem *new_option = optionName->selectedItem ();
  if (new_option == prev_option)
    return;
  if (!setOption ())
    {
      optionName->setSelected (prev_option, true);
      return;
    }
  optionValue->setEnabled (new_option != 0);
  if (new_option != 0)
    {
      const char *val = me->module->get_option (me->mi, new_option->text (0));
      if (val != NULL)
	{
	  optionValue->setText (val);
	  free ((void *)val);
	}
      optionDescr->setText (descriptions[new_option->text (0)]);
    }      
  prev_option = new_option;
}

/** Tab was switched */
void
KHwPropertiesDialog::tabSwitched (QWidget *tab)
{
  if (tabWidget->indexOf (tab) == 0)
    return;
  if (!setOption ())
    tabWidget->setCurrentPage (0);
}

/** Check if optionValue is OK, and accept () if so */
void
KHwPropertiesDialog::tryAccept ()
{
  if (setOption ())
    {
      if (modified)
	accept ();
      else
	/* We don't want the whole setup to be marked modified */
	reject ();
    }
}

/** Set prev_option to optionValue.  @returns true on success */
bool
KHwPropertiesDialog::setOption ()
{
  if (prev_option == 0)
    return true;
  /* Report the error only every second attempt until the input changes,
     because we get optionSelected () *twice* if we try to return it back
     and it was generated using a mouse (not keyboard), and there seems to be
     no better way to solve it than this kludge. */
  if (prev_option == last_err_option && optionValue->text () == last_err_val)
    {
      last_err_option = 0;
      return false;
    }
  if (me->module->set_option (me->mi, prev_option->text (0),
			      optionValue->text ()) == 0)
    {
      modified = true;
      last_err_option = 0;
      return true;
    }
  last_err_option = prev_option;
  last_err_val = optionValue->text ();
  return false;
}

/** Find KHwConnListItem for peer of conn */
KHwConnListItem *
KHwPropertiesDialog::findConn (const struct connect_entry *conn)
{
  for (QListViewItem *i = connectionList->firstChild (); i != 0;
       i = i->nextSibling ())
    {
      KHwConnListItem *c = static_cast <KHwConnListItem *> (i);
      if (c->conn () == conn)
	return c;
    }
  return 0;
}

/** Rebuild connectionList */
void
KHwPropertiesDialog::rebuildConnectionList ()
{
  connectionList->clear ();
  for (struct connect_entry *ce = connect_list; ce != 0; ce = ce->next)
    {
      if (ce->pi1.me == me)
	new KHwConnListItem (connectionList, ce);
    }
  connectionList->sort ();
  QListViewItem *first = connectionList->firstChild ();
  if (first != 0)
    {
      connectionList->setCurrentItem (first);
      connectionList->setSelected (first, false);
    }
}

#include "khwpropertiesdialog.moc"
