/* 8051 disassembler
   Copyright (C) 2002, 2004, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <stdio.h>
#include <string.h>

#include <disasm8051.h>
#include <lib.h>

/* If we wanted to be really elegant, we would get bit and SFR names from the
   XML files, but this has some layering and speed issues.  It can wait at
   least until users are able to define their own names / aliases. */

/* Name BIT, store to BUF;
   returns: next position in buffer (contains trailing '\0') */
static char *
get_bit_name (char *buf, uint8_t bit)
{
  const char *name;
  size_t len;
  
  switch (bit >> 4)
    {
    case 0x8: case 0x9: case 0xA: case 0xB:
      if ((bit & 0x08) == 0)
	{
	  sprintf (buf, "P%c.%c", (bit >> 4) + '0' - 0x8, (bit & 0x07) + '0');
	  return buf + 4;
	}
      else
	{
	  static const char *const names[0xC - 0x8][0x10 - 0x8] =
	    {
	      [0x8 - 0x8] =
	      {
		[0x8 - 0x8] = "IT0", [0x9 - 0x8] = "IE0", [0xA - 0x8] = "IT1",
		[0xB - 0x8] = "IE1", [0xC - 0x8] = "TR0", [0xD - 0x8] = "TF0",
		[0xE - 0x8] = "TR1", [0xF - 0x8] = "TF1"
	      },
	      [0x9 - 0x8] =
	      {
		[0x8 - 0x8] = "RI", [0x9 - 0x8] = "TI", [0xA - 0x8] = "RB8",
		[0xB - 0x8] = "TB8", [0xC - 0x8] = "REN",
		[0xD - 0x8] = "SM2", [0xE - 0x8] = "SM1",
		[0xF - 0x8] = "SM0",
	      },
	      [0xA - 0x8] =
	      {
		[0x8 - 0x8] = "EX0", [0x9 - 0x8] = "ET0", [0xA - 0x8] = "EX1",
		[0xB - 0x8] = "ET1", [0xC - 0x8] = "ES", [0xD - 0x8] = "IE.5",
		[0xE - 0x8] = "IE.6", [0xF - 0x8] = "EA",
	      },
	      [0xB - 0x8] =
	      {
		[0x8 - 0x8] = "PX0", [0x9 - 0x8] = "PT0", [0xA - 0x8] = "PX1",
		[0xB - 0x8] = "PT1", [0xC - 0x8] = "PS", [0xD - 0x8] = "IP.5",
		[0xE - 0x8] = "IP.6", [0xF - 0x8] = "IP.7"
	      }
	    };

	  name = names[(bit >> 4) - 0x8][bit & 0x07];
	  goto got_name;
	}
      break;

    case 0xD:
      if ((bit & 0x08) == 0)
	{
	  static const char *const names[8] =
	    {
	      [0] = "P", [1] = "PSW.1", [2] = "OV", [3] = "RS0", [4] = "RS1",
	      [5] = "F0", [6] = "AC", [7] = "C"
	    };
	  
	  name = names[bit & 0x07];
	  goto got_name;
	}
      break;
      
    case 0x0E:
      if ((bit & 0x08) == 0)
	{
	  sprintf (buf, "ACC.%c", (bit & 0x07) + '0');
	  return buf + 5;
	}
      break;

    case 0x0F:
      if ((bit & 0x08) == 0)
	{
	  sprintf (buf, "B.%c", (bit & 0x07) + '0');
	  return buf + 3;
	}
      break;
    }
  sprintf (buf, "0x%02X", bit);
  return buf + 4;
  
 got_name:
  len = strlen (name);
  memcpy (buf, name, len + 1);
  return buf + len;
}

/* Name direct-addressed BYTE, store to BUF;
   returns: next position in buffer (contains trailing '\0') */
static char *
get_direct_name (char *buf, uint8_t byte)
{
  const char *name;
  size_t len;
  
  /* Ugly, but better than array of 256 values */
  if (byte >= 0x80 && byte < 0x90)
    {
      static const char *const names[0x10] =
	{
	  [0x0] = "P0", [0x1] = "SP", [0x2] = "DPL", [0x3] = "DPH",
	  [0x4] = NULL, [0x5] = NULL, [0x6] = NULL, [0x7] = "PCON",
	  [0x8] = "TCON", [0x9] = "TMOD", [0xA] = "TL0", [0xB] = "TL1",
	  [0xC] = "TH0", [0xD] = "TH1", [0xE] = NULL, [0xF] = NULL
	};
      
      name = names[byte & 0x0F];
      if (name != NULL)
	goto got_name;
    }
  else if ((byte & 0xC7) == 0x80) /* 0x90, 0x98, 0xA0, 0xA8, 0xB0, 0xB8 */
    {
      static const char *const names[(0xC0 - 0x90) >> 3] =
	{
	  [(0x90 - 0x90) >> 3] = "P1", [(0x98 - 0x90) >> 3] = "SCON",
	  [(0xA0 - 0x90) >> 3] = "P2", [(0xA8 - 0x90) >> 3] = "IE",
	  [(0xB0 - 0x90) >> 3] = "P3", [(0xB8 - 0x90) >> 3] = "IP"
	};

      name = names[(byte >> 3) - (0x90 >> 3)];
      goto got_name;
    }
  else if (byte == 0x99) /* Special case :-( */
    {
      name = "SBUF";
      goto got_name;
    }
  else if ((byte & 0xCF) == 0xC0) /* 0xC0, 0xD0, 0xE0, 0xF0 */
    {
      static const char *const names[0x10 - 0x0C] =
	{
	  [0xC - 0xC] = NULL, [0xD - 0xC] = "PSW", [0xE - 0xC] = "ACC",
	  [0xF - 0xC] = "B"
	};

      name = names[(byte >> 4) - 0x0C];
      if (name != NULL)
	goto got_name;
    }
  sprintf (buf, "0x%02X", byte);
  return buf + 4;
  
 got_name:
  len = strlen (name);
  memcpy (buf, name, len + 1);
  return buf + len;
}

/* Disassemble source operand of INSN, store to BUF;
   returns: next position in buffer (contains trailing '\0') */
static char *
get_insn_source (char *buf, const uint8_t *insn)
{
  switch (*insn & 0x0F)
    {
    case 0x04:
      sprintf (buf, "#0x%02X", insn[1]);
      return buf + 5;

    case 0x05:
      return get_direct_name (buf, insn[1]);

    case 0x06: case 0x07:
      sprintf (buf, "@R%c", '0' + (*insn & 0x01));
      return buf + 3;

    case 0x08: case 0x09: case 0x0A: case 0x0B: case 0x0C: case 0x0D:
    case 0x0E: case 0x0F:
      sprintf (buf, "R%c", '0' + (*insn & 0x07));
      return buf + 2;

    default:
      assert ("BUG" == NULL);
      return NULL;
    }
}

/* Length of data needed for source operand of INSN */
inline static size_t
insn_source_len (const uint8_t *insn)
{
  return (*insn & 0x0F) <= 0x05; /* 0x04, 0x05 */
}

/* Disassemble destination operand of INSN, store to BUF;
   returns: next position in buffer (contains trailing '\0') */
static char *
get_insn_dest (char *buf, const uint8_t *insn)
{
  switch (*insn & 0x0F)
    {
    case 0x04:
      memcpy (buf, "A", 2);
      return buf + 1;

    case 0x05:
      return get_direct_name (buf, insn[1]);

    case 0x06: case 0x07:
      sprintf (buf, "@R%c", '0' + (*insn & 0x01));
      return buf + 3;

    case 0x08: case 0x09: case 0x0A: case 0x0B: case 0x0C: case 0x0D:
    case 0x0E: case 0x0F:
      sprintf (buf, "R%c", '0' + (*insn & 0x07));
      return buf + 2;

    default:
      assert ("BUG" == NULL);
      return NULL;
    }
}

/* Length of data needed for destination operand of INSN */
inline static size_t
insn_dest_len (const uint8_t *insn)
{
  return (*insn & 0x0F) == 0x05;
}

/* Get target of jumb by DISP relative to PC (past current insn), store to
   BUF */
static void
get_rel_target (char *buf, uint16_t pc, int8_t disp)
{
  sprintf (buf, "0x%04X", (pc + disp) & 0xFFFF);
}

/* Get target of jumb by DISP relative to PC (past current insn), store to
   BUF (with leading ", ") */
static void
append_rel_target (char *buf, uint16_t pc, int8_t disp)
{
  sprintf (buf, ", 0x%04X", (pc + disp) & 0xFFFF);
}

/* Disassemble INSN at PC, store opcode to *OPCODE (max. 5 chars,
   not to be freed), operands to BUF (64 chars);
   returns: number of bytes in instruction */
size_t
i8051_disasm (uint16_t pc, const uint8_t *insn, const char **opcode,
	      char *buf)
{
  char *p;

  switch (*insn)
    {
    case 0x00: case 0x22: case 0x32: case 0xA5:
      {
	static const char *const ops[0x10] =
	  {
	    [0x0] = "NOP", [0x1] = NULL, [0x2] = "RET", [0x3] = "RETI",
	    [0x4] = NULL, [0x5] = NULL, [0x6] = NULL, [0x7] = NULL,
	    [0x8] = NULL, [0x9] = NULL, [0xA] = "!0xA5!", [0xB] = NULL,
	    [0xC] = NULL, [0xD] = NULL, [0xE] = NULL, [0xF] = NULL
	  };
	
	*opcode = ops[*insn >> 4];
	assert (*opcode != NULL);
	*buf = 0;
	return 1;
      }
      
      /* (*insn & 0x0F) == 0x01 */
    case 0x01: case 0x11: case 0x21: case 0x31: case 0x41: case 0x51:
    case 0x61: case 0x71: case 0x81: case 0x91: case 0xA1: case 0xB1:
    case 0xC1: case 0xD1: case 0xE1: case 0xF1:
      *opcode = (*insn & 0x10) == 0 ? "AJMP" : "ACALL";
      sprintf (buf, "0x%04X",
	       ((pc + 2) & 0xF800) | (*insn & 0xE0) << 3 | insn[1]);
      return 2;

    case 0x02: case 0x12:
      *opcode = *insn == 0x02 ? "LJMP" : "LCALL";
      sprintf (buf, "0x%04X", insn[1] << 8 | insn[2]);
      return 3;
      
    case 0x03: case 0x13: case 0x23: case 0x33: case 0xC4:
    case 0xD4: case 0xE4:
    case 0xF4:
      {
	static const char *const ops[0x10] =
	  {
	    [0x0] = "RR", [0x1] = "RRC", [0x2] = "RL", [0x3] = "RLC",
	    [0x4] = NULL, [0x5] = NULL, [0x6] = NULL, [0x7] = NULL,
	    [0x8] = NULL, [0x9] = NULL, [0xA] = NULL, [0xB] = NULL,
	    [0xC] = "SWAP", [0xD] = "DA", [0xE] = "CLR", [0xF] = "CPL"
	  };
	
	*opcode = ops[*insn >> 4];
	assert (*opcode != NULL);
	memcpy (buf, "A", 2);
	return 1;
      }
      
    case 0x04: case 0x05: case 0x06: case 0x07: case 0x08: case 0x09:
    case 0x0A: case 0x0B: case 0x0C: case 0x0D: case 0x0E: case 0x0F:
    case 0x14: case 0x15: case 0x16: case 0x17: case 0x18: case 0x19:
    case 0x1A: case 0x1B: case 0x1C: case 0x1D: case 0x1E: case 0x1F:
      *opcode = (*insn & 0xF0) == 0 ? "INC" : "DEC";
      get_insn_dest (buf, insn);
      return 1 + insn_dest_len (insn);
      
    case 0x24: case 0x25: case 0x26: case 0x27: case 0x28: case 0x29:
    case 0x2A: case 0x2B: case 0x2C: case 0x2D: case 0x2E: case 0x2F:
    case 0x34: case 0x35: case 0x36: case 0x37: case 0x38: case 0x39:
    case 0x3A: case 0x3B: case 0x3C: case 0x3D: case 0x3E: case 0x3F:
    case 0x44: case 0x45: case 0x46: case 0x47: case 0x48: case 0x49:
    case 0x4A: case 0x4B: case 0x4C: case 0x4D: case 0x4E: case 0x4F:
    case 0x54: case 0x55: case 0x56: case 0x57: case 0x58: case 0x59:
    case 0x5A: case 0x5B: case 0x5C: case 0x5D: case 0x5E: case 0x5F:
    case 0x64: case 0x65: case 0x66: case 0x67: case 0x68: case 0x69:
    case 0x6A: case 0x6B: case 0x6C: case 0x6D: case 0x6E: case 0x6F:
    case 0x94: case 0x95: case 0x96: case 0x97: case 0x98: case 0x99:
    case 0x9A: case 0x9B: case 0x9C: case 0x9D: case 0x9E: case 0x9F:
    case 0xC5: case 0xC6: case 0xC7: case 0xC8: case 0xC9: case 0xCA:
    case 0xCB: case 0xCC: case 0xCD: case 0xCE: case 0xCF:
    case 0xD6: case 0xD7:
      {
	static const char *const ops[0x10] =
	  {
	    [0x0] = NULL, [0x1] = NULL, [0x2] = "ADD", [0x3] = "ADDC",
	    [0x4] = "ORL", [0x5] = "ANL", [0x6] = "XRL", [0x7] = NULL,
	    [0x8] = NULL, [0x9] = "SUBB", [0xA] = NULL, [0xB] = NULL,
	    [0xC] = "XCH", [0xD] = "XCHD", [0xE] = NULL, [0xF] = NULL,
	  };

	*opcode = ops[*insn >> 4];
	assert (*opcode != NULL);
	memcpy (buf, "A, ", 3);
	get_insn_source (buf + 3, insn);
	return 1 + insn_source_len (insn);
      }
      
    case 0x10: case 0x20: case 0x30:
      {
	static const char *const ops[0x4 - 0x1] =
	  { [0x1 - 0x1] = "JBC", [0x2 - 0x1] = "JB", [0x3 - 0x1] = "JNB" };

	*opcode = ops[(*insn >> 4) - 0x1];
	p = get_bit_name (buf, insn[1]);
	append_rel_target (p, pc + 3, insn[2]);
	return 3;
      }

    case 0x40: case 0x50: case 0x60: case 0x70: case 0x80:
      {
	static const char *const ops[0x9 - 0x4] =
	  {
	    [0x4 - 0x4] = "JC", [0x5 - 0x4] = "JNC", [0x6 - 0x4] = "JZ",
	    [0x7 - 0x4] = "JNZ", [0x8 - 0x4] = "SJMP"
	  };

	*opcode = ops[(*insn >> 4) - 0x4];
	get_rel_target (buf, pc + 2, insn[1]);
	return 2;
      }
      
    case 0x42: case 0x43: case 0x52: case 0x53: case 0x62: case 0x63:
      {
	static const char *const ops[0x7 - 0x4] =
	  { [0x4 - 0x4] = "ORL", [0x5 - 0x4] = "ANL", [0x6 - 0x4] = "XRL" };

	*opcode = ops[(*insn >> 4) - 0x4];
	p = get_direct_name (buf, insn[1]);
	if ((*insn & 0x01) == 0)
	  {
	    memcpy (p, ", A", 4);
	    return 2;
	  }
	else
	  {
	    sprintf (p, ", #0x%02X", insn[2]);
	    return 3;
	  }
      }	

    case 0x73:
      *opcode = "JMP";
      memcpy (buf, "@A + DPTR", 10);
      return 1;
      
    case 0x74: case 0x76: case 0x77: case 0x78: case 0x79: case 0x7A:
    case 0x7B: case 0x7C: case 0x7D: case 0x7E: case 0x7F:
      *opcode = "MOV";
      p = get_insn_dest (buf, insn);
      sprintf (p, ", #0x%02X", insn[1]);
      return 2;

    case 0x75:
      *opcode = "MOV";
      p = get_direct_name (buf, insn[1]);
      sprintf (p, ", #0x%02X", insn[2]);
      return 3;
      
    case 0x72: case 0x82: case 0xA2:
      {
	static const char *const ops[0xB - 0x7] =
	  {
	    [0x7 - 0x7] = "ORL", [0x8 - 0x7] = "ANL", [0x9 - 0x7] = NULL,
	    [0xA - 0x7] = "MOV"
	  };

	*opcode = ops[(*insn >> 4) - 0x7];
	assert (*opcode != NULL);
	memcpy (buf, "C, ", 3);
	get_bit_name (buf + 3, insn[1]);
	return 2;
      }

    case 0x83:
      *opcode = "MOVC";
      memcpy (buf, "A, @A + PC", 11);
      return 1;
      
    case 0x84: case 0xA4:
      *opcode = *insn == 0x84 ? "DIV" : "MUL";
      memcpy (buf, "AB", 3);
      return 1;

    case 0x85:
      *opcode = "MOV";
      p = get_direct_name (buf, insn[2]);
      memcpy (p, ", ", 2);
      get_direct_name (p + 2, insn[1]);
      return 3;
      
    case 0x86: case 0x87: case 0x88: case 0x89: case 0x8A: case 0x8B:
    case 0x8C: case 0x8D: case 0x8E: case 0x8F:
      *opcode = "MOV";
      p = get_direct_name (buf, insn[1]);
      memcpy (p, ", ", 2);
      get_insn_dest (p + 2, insn);
      return 2;

    case 0x90:
      *opcode = "MOV";
      sprintf (buf, "DPTR, 0x%04X", insn[1] << 8 | insn[2]);
      return 3;
      
    case 0x92:
      *opcode = "MOV";
      p = get_bit_name (buf, insn[1]);
      memcpy (p, ", C", 4);
      return 2;
      
    case 0x93:
      *opcode = "MOVC";
      memcpy (buf, "A, @A + DPTR", 13);
      return 1;
      
    case 0xA0: case 0xB0:
      *opcode = *insn == 0xA0 ? "ORL" : "ANL";
      memcpy (buf, "C, /", 4);
      get_bit_name (buf + 4, insn[1]);
      return 2;

    case 0xA3:
      *opcode = "INC";
      memcpy (buf, "DPTR", 5);
      return 1;

    case 0xA6: case 0xA7: case 0xA8: case 0xA9: case 0xAA: case 0xAB:
    case 0xAC: case 0xAD: case 0xAE: case 0xAF:
      *opcode = "MOV";
      p = get_insn_dest (buf, insn);
      memcpy (p, ", ", 2);
      get_direct_name (p + 2, insn[1]);
      return 2;
      
    case 0xB2: case 0xB3: case 0xC2: case 0xC3: case 0xD2: case 0xD3:
      {
	static const char *const ops[0xE - 0xB] =
	  { [0xB - 0xB] = "CPL", [0xC - 0xB] = "CLR", [0xD - 0xB] = "SETB" };

	*opcode = ops[(*insn >> 4) - 0xB];
	if ((*insn & 0x01) == 0)
	  {
	    get_bit_name (buf, insn[1]);
	    return 2;
	  }
	else
	  {
	    memcpy (buf, "C", 2);
	    return 1;
	  }
      }
      
    case 0xB4: case 0xB5:
      *opcode = "CJNE";
      memcpy (buf, "A, ", 3);
      p = get_insn_source (buf + 3, insn);
      append_rel_target (p, pc + 3, insn[2]);
      return 3;

    case 0xB6: case 0xB7: case 0xB8: case 0xB9: case 0xBA: case 0xBB:
    case 0xBC: case 0xBD: case 0xBE: case 0xBF:
      *opcode = "CJNE";
      p = get_insn_source (buf, insn);
      sprintf (p, ", #0x%02X", insn[1]);
      append_rel_target (p + 7, pc + 3, insn[2]);
      return 3;

    case 0xC0: case 0xD0:
      *opcode = *insn == 0xC0 ? "PUSH" : "POP";
      get_direct_name (buf, insn[1]);
      return 2;
      
    case 0xD5:
      *opcode = "DJNZ";
      p = get_direct_name (buf, insn[1]);
      append_rel_target (p, pc + 3, insn[2]);
      return 3;
      
    case 0xD8: case 0xD9: case 0xDA: case 0xDB: case 0xDC: case 0xDD:
    case 0xDE: case 0xDF:
      *opcode = "DJNZ";
      p = get_insn_dest (buf, insn);
      append_rel_target (p, pc + 2, insn[1]);
      return 2;
      
    case 0xE0:
      *opcode = "MOVX";
      memcpy (buf, "A, @DPTR", 9);
      return 1;
      
    case 0xE2: case 0xE3:
      *opcode = "MOVX";
      sprintf (buf, "A, @R%c", (*insn & 0x01) + '0');
      return 1;
      
    case 0xE5: case 0xE6: case 0xE7: case 0xE8: case 0xE9: case 0xEA:
    case 0xEB: case 0xEC: case 0xED: case 0xEE: case 0xEF:
      *opcode = "MOV";
      memcpy (buf, "A, ", 3);
      get_insn_source (buf + 3, insn);
      return 1 + insn_source_len (insn);

    case 0xF0:
      *opcode = "MOVX";
      memcpy (buf, "@DPTR, A", 9);
      return 1;
      
    case 0xF2: case 0xF3:
      *opcode = "MOVX";
      sprintf (buf, "@R%c, A", (*insn & 0x01) + '0');
      return 1;
      
    case 0xF5: case 0xF6: case 0xF7: case 0xF8: case 0xF9: case 0xFA:
    case 0xFB: case 0xFC: case 0xFD: case 0xFE: case 0xFF:
      *opcode = "MOV";
      p = get_insn_dest (buf, insn);
      memcpy (p, ", A", 4);
      return 1 + insn_dest_len (insn);

    default:
      assert ("BUG" == NULL);
      return 0; /* Infinite loop is hart not to notice even with NDEBUG ;-) */
    }
}
