/* Intel hex format handling
   Copyright (C) 2002 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <lib.h>

#include "hex.h"

/* Read one hexadecimal digit from POS;
   returns: digit value if OK, -1 on error */
static int
hex_read_digit (struct file_pos *pos)
{
  int c;

  c = pos->ops->next_char (pos);
  switch (c)
    {
    case '0': case '1': case '2': case '3': case '4': case '5': case '6':
    case '7': case '8': case '9':
      return c - '0';
      /* Letters are not required to be consecutive by ISO C */
    case 'A':
      return 10;
    case 'B':
      return 11;
    case 'C':
      return 12;
    case 'D':
      return 13;
    case 'E':
      return 14;
    case 'F':
      return 15;
    case EOF:
      if (pos->ops->was_error (pos))
	error (errno, "cannot read `%s'", pos->filename);
      else
	error (0, "%s:%u: unexpected end of file", pos->filename, pos->line);
      return -1;
    default:
      error (0, "%s:%u: unexpected character `%c'", pos->filename, pos->line,
	     c);
      return -1;
    }
}

/* Read one byte from POS;
   returns: read value if OK, -1 on error */
static int
hex_read_byte (struct file_pos *pos)
{
  int val1, val2;

  val1 = hex_read_digit (pos);
  if (val1 < 0)
    return -1;
  val2 = hex_read_digit (pos);
  if (val2 < 0)
    return -1;
  return val1 * 16 + val2;
}

/* Read line from POS *BUF of *BUF_SIZE;
   returns: 1 if OK, 0 at EOF, -1 on error */
inline static int
hex_read_line (struct file_pos *pos, uint8_t **buf, size_t *buf_size)
{
  size_t line_size, i;
  uint8_t *dest, checksum;
  int val;

  val = pos->ops->next_char (pos);
  if (val == EOF)
    {
      if (pos->ops->was_error (pos))
	goto read_error;
      return 0;
    }
  else if (val != ':')
    {
      error (0, "%s:%u: line does not start with `:'", pos->filename,
	     pos->line);
      return -1;
    }
  val = hex_read_byte (pos);
  if (val < 0)
    return -1;
  /* Because val is only one byte, the value is always "reasonable". */
  line_size = IH_DATA + val;
  if (*buf_size < line_size)
    {
      *buf = xrealloc (*buf, line_size);
      *buf_size = line_size;
    }
  dest = *buf;
  *dest++ = val;
  checksum = val;
  for (i = 1; i < line_size; i++)
    {
      val = hex_read_byte (pos);
      if (val < 0)
	return -1;
      *dest++ = val;
      checksum += val;
    }
  val = hex_read_byte (pos);
  if (val < 0)
    return -1;
  checksum += val;
  if (checksum != 0)
    {
      error (0, "%s:%u: checksum error", pos->filename, pos->line);
      return -1;
    }
  val = pos->ops->next_char (pos);
  if (val == EOF)
    {
      if (pos->ops->was_error (pos))
	goto read_error;
    }
  else
    {
      /* Accept files that don't have CR, because we're probably running
	 on a POSIX system and because XML mandates normalization of line
	 breaks to '\n' */
      if (val == '\r')
	{
	  val = pos->ops->next_char (pos);
	  if (val == EOF)
	    {
	      if (pos->ops->was_error (pos))
		goto read_error;
	    }
	}
      if (val != '\n')
	{
	  error (0, "%s:%u: end of line (LF) expected", pos->filename,
		 pos->line);
	  return -1;
	}
    }
  return 1;

 read_error:
  error (errno, "cannot read `%s'", pos->filename);
  return -1;
}

/* Load Intel hex from POS to DATA of LEN;
   returns: 0 if OK, -1 on error */
int
hex_read_file (struct file_pos *pos, uint8_t *data, size_t len)
{
  uint8_t *copy, *buf;
  size_t buf_size;

  copy = xmalloc (len);
  /* If we load more than one file, merge them (the later loaded override the
     former ones). */
  memcpy (copy, data, len);
  pos->line = 1;
  buf = NULL;
  buf_size = 0;
  for (;;)
    {
      int res;

      res = hex_read_line (pos, &buf, &buf_size);
      if (res < 0)
	goto err_buf;
      if (res == 0)
	break;
      switch (buf[IH_TYPE])
	{
	case IHT_DATA:
	  {
	    size_t addr;
	    
	    addr = buf[IH_ADDR_HI] * 256 + buf[IH_ADDR_LO];
	    if (addr + buf[IH_DATA_LEN] > len)
	      {
		error (0, "%s:%u: address out of bounds", pos->filename,
		       pos->line);
		goto err_buf;
	      }
	    memcpy (copy + addr, buf + IH_DATA, buf[IH_DATA_LEN]);
	    break;
	  }
	default:
	  /* Print a warning (assuming we ignore at least type 1)? */
	  break;
	}
      pos->line++;
    }
  free (buf);
  /* Everything went OK, so "commit". */
  memcpy (data, copy, len);
  free (copy);
  return 0;

 err_buf:
  free (buf);
  free (copy);
  return -1;
}
