/* Expression evaluation
   Copyright (C) 2002, 2004 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdlib.h>

#include <expr.h>
#include <lib.h>
#include <port.h>

#include "expr_code.h"

/* Value stack */
struct stack
{
  intmax_t *bottom; /* Deepest entry */
  intmax_t *top; /* *AFTER* TOS */
  intmax_t *end; /* After allocated area */
};

/* Initialize STACK */
inline static void
stack_init (struct stack *stack)
{
  stack->bottom = NULL;
  stack->top = NULL;
  stack->end = NULL;
};

/* Destroy STACK */
inline static void
stack_destroy (struct stack *stack)
{
  free (stack->bottom);
}

/* Check whether STACK is empty */
inline static int
stack_empty (const struct stack *stack)
{
  return stack->top == stack->bottom;
}

/* Push VAL to STACK */
static void
stack_push (struct stack *stack, intmax_t val)
{
  if (stack->top == stack->end)
    {
      intmax_t *copy;
      size_t size;

      if (stack->bottom == NULL)
	size = 8;
      else
	size = 2 * (stack->end - stack->bottom);
      copy = xrealloc (stack->bottom, size * sizeof (*stack->bottom));
      stack->top = copy + (stack->top - stack->bottom);
      stack->end = copy + size;
      stack->bottom = copy;
    }
  *stack->top++ = val;
}

/* Return top of STACK, pop it */
static intmax_t
stack_pop (struct stack *stack)
{
  assert (stack_empty (stack) == 0);
  return *--stack->top;
}

/* Evaluate compiled EXPR, store result to DEST;
   returns: 0 if OK, -1 on error */
int
expr_exec (const struct expr_code *expr, intmax_t *dest)
{
  struct stack stack;

  assert (expr->code != E_END);
  stack_init (&stack);
  for (;;)
    {
      intmax_t val;
      
      switch (expr->code)
	{
	case E_END:
	  *dest = stack_pop (&stack);
	  assert (stack_empty (&stack));
	  stack_destroy (&stack);
	  return 0;
	  
	case E_VAL:
	  val = expr->v.val;
	  break;

	case E_PORT: case E_CPORT:
	  switch (expr->v.port.width)
	    {
#define DO_WIDTH(WIDTH, TYPE)						\
	    case WIDTH_##WIDTH:						\
	      {								\
		TYPE v;							\
									\
		if (port_disp_##WIDTH (&expr->v.port.pi, &v) != 0)	\
		  goto err;						\
		val = v;						\
		break;							\
	      } 
	      DO_WIDTH (1, int);
	      DO_WIDTH (8, uint8_t);
	      DO_WIDTH (16, uint16_t);
	      
	    default:
	      assert ("BUG" == NULL);
	      goto err;
#undef DO_WIDTH
	    }
	  break;

	case E_NEG: case E_NOT:
	  {
	    intmax_t a;

	    a = stack_pop (&stack);
	    val = expr->code == E_NEG ? -a : !a;
	    break;
	  }

	case E_MUL: case E_DIV: case E_MOD: case E_ADD: case E_SUB: case E_LSH:
	case E_RSH: case E_LT: case E_GT: case E_LE: case E_GE: case E_EQ:
	case E_NE: case E_BAND: case E_BXOR: case E_BOR: case E_LAND:
	case E_LOR:
	  {
	    intmax_t a, b;

	    b = stack_pop (&stack);
	    a = stack_pop (&stack);
	    switch (expr->code)
	      {
	      case E_MUL:
		val = a * b;
		break;

	      case E_DIV:
		if (b == 0)
		  goto err_div_by_zero;
		val = a / b;
		break;
		
	      case E_MOD:
		if (b == 0)
		  goto err_div_by_zero;
		val = a % b;
		break;

	      case E_ADD:
		val = a + b;
		break;
		
	      case E_SUB:
		val = a - b;
		break;

	      case E_LSH:
		val = a << b;
		break;

	      case E_RSH:
		val = a >> b;
		break;

	      case E_LT:
		val = a < b;
		break;

	      case E_GT:
		val = a > b;
		break;
		
	      case E_LE:
		val = a <= b;
		break;

	      case E_GE:
		val = a >= b;
		break;

	      case E_EQ:
		val = a == b;
		break;

	      case E_NE:
		val = a != b;
		break;

	      case E_BAND:
		val = a & b;
		break;

	      case E_BXOR:
		val = a ^ b;
		break;

	      case E_BOR:
		val = a | b;
		break;

	      case E_LAND:
		val = a && b;
		break;

	      case E_LOR:
		val = a || b;
		break;

	      default:
		assert ("BUG" == NULL);
		goto err;
	      }
	    break;
	  }
	  
	case E_COND:
	  {
	    intmax_t a, b, c;

	    c = stack_pop (&stack);
	    b = stack_pop (&stack);
	    a = stack_pop (&stack);
	    val = a ? b : c;
	    break;
	  }

	default:
	  assert ("BUG" == NULL);
	  goto err;
	}
      stack_push (&stack, val);
      expr++;
    }

 err_div_by_zero:
  error (0, "division by zero");
 err:
  stack_destroy (&stack);
  return 1;
}

/* Make a copy of SRC with E_CPORT codes converted to E_VAL with current values
   and constant subexpressions evaluated;
   returns: modified expression (for free ()) if OK, NULL on error */
inline static struct expr_code *
expr_cport_eval (const struct expr_code *src)
{
  struct expr_code *dest, *p;
  const struct expr_code *sp;

  for (sp = src; sp->code != E_END; sp++)
    ;
  dest = xmalloc ((sp - src + 1) * sizeof (*dest));
  p = dest;
  for (;;)
    {
      switch (src->code)
	{
	case E_END:
	  p->code = E_END;
	  return dest;

	case E_VAL: case E_PORT:
	  *p++ = *src;
	  break;

	case E_CPORT:
	  p->code = E_VAL;
	  switch (src->v.port.width)
	    {
#define DO_WIDTH(WIDTH, TYPE)						\
	    case WIDTH_##WIDTH:						\
	      {								\
		TYPE val;						\
									\
		if (port_disp_##WIDTH (&src->v.port.pi, &val) != 0)	\
		  goto err;						\
		p->v.val = val;						\
		break;							\
	      }
	      DO_WIDTH (1, int);
	      DO_WIDTH (8, uint8_t);
	      DO_WIDTH (16, uint16_t);

	    default:
	      assert ("BUG" == NULL);
#undef DO_WIDTH
	    }
	  p++;
	  break;

#define U(CODE, OP)				\
	case CODE:				\
	  if (p[-1].code != E_VAL)		\
	    goto copy_op;			\
	  p[-1].v.val = OP p[-1].v.val;		\
	  break;

	  U (E_NEG, -);
	  U (E_NOT, !);
#undef U
	  
	  /* BA stands for binary, assignable (OP= is valid) */
#define BA(CODE, OP)						\
	case CODE:						\
	  if (p[-2].code != E_VAL || p[-1].code != E_VAL)	\
	    goto copy_op;					\
	  p[-2].v.val OP##= p[-1].v.val;			\
	  p--;							\
	  break;

	  /* and BN for binary, nonassignable */
#define BN(CODE, OP)						\
	case CODE:						\
	  if (p[-2].code != E_VAL || p[-1].code != E_VAL)	\
	    goto copy_op;					\
	  p[-2].v.val = p[-2].v.val OP p[-1].v.val;		\
	  p--;							\
	  break;

	  BA (E_MUL, *);
	  
	case E_DIV:
	  if (p[-2].code != E_VAL || p[-1].code != E_VAL)
	    goto copy_op;
	  if (p[-2].v.val == 0)
	    goto err_div_by_zero;
	  p[-2].v.val /= p[-1].v.val;
	  p--;
	  break;

	case E_MOD:
	  if (p[-2].code != E_VAL || p[-1].code != E_VAL)
	    goto copy_op;
	  if (p[-2].v.val == 0)
	    goto err_div_by_zero;
	  p[-2].v.val %= p[-1].v.val;
	  p--;
	  break;

	  BA (E_ADD, +);
	  BA (E_SUB, -);
	  BA (E_LSH, <<);
	  BA (E_RSH, >>);
	  BN (E_LT, <);
	  BN (E_GT, >);
	  BN (E_LE, <=);
	  BN (E_GE, >=);
	  BN (E_EQ, ==);
	  BN (E_NE, !=);
	  BA (E_BAND, &);
	  BA (E_BXOR, ^);
	  BA (E_BOR, |);
	  BN (E_LAND, &&);
	  BN (E_LOR, ||);
#undef BA
#undef BN
	  
	case E_COND:
	  if (p[-3].code != E_VAL || p[-2].code != E_VAL
	      || p[-1].code != E_VAL)
	    goto copy_op;
	  p[-3].v.val = p[-3].v.val ? p[-2].v.val : p[-1].v.val;
	  p -= 2;
	  break;
	  
	copy_op:
	  p->code = src->code;
	  p++;
	  break;

	default:
	  assert ("BUG" == NULL);
	}
      src++;
    }

 err_div_by_zero:
  error (0, "division by zero");
 err:
  free (dest);
  return NULL;
}

/* Fill expr_code.v.op pointers in CODE;
   returns: pointer to root of the expression tree */
inline static struct expr_code *
expr_make_tree (struct expr_code *code)
{
  struct expr_code *p;
  struct stack stack;
  
  assert (code->code != E_END);
  stack_init (&stack);
  for (p = code; p->code != E_END; p++)
    {
      switch (p->code)
	{
	case E_VAL: case E_PORT:
	  break;

	case E_NEG: case E_NOT:
	  p->v.op[0] = code + stack_pop (&stack);
	  break;

	case E_MUL: case E_DIV: case E_MOD: case E_ADD: case E_SUB: case E_LSH:
	case E_RSH: case E_LT: case E_GT: case E_LE: case E_GE: case E_EQ:
	case E_NE: case E_BAND: case E_BXOR: case E_BOR: case E_LAND:
	case E_LOR:
	  p->v.op[1] = code + stack_pop (&stack);
	  p->v.op[0] = code + stack_pop (&stack);
	  break;

	case E_COND:
	  p->v.op[2] = code + stack_pop (&stack);
	  p->v.op[1] = code + stack_pop (&stack);
	  p->v.op[0] = code + stack_pop (&stack);
	  break;
	  
	case E_CPORT: default:
	  assert ("BUG" == NULL);
	}
      stack_push (&stack, p - code);
    }
  p = code + stack_pop (&stack);
  assert (stack_empty (&stack));
  stack_destroy (&stack);
  return p;
}

/* Modify port referenced in DEST so that DEST == SRC;
   returns: 0 if OK, -1 on error */
int
expr_modify (const char *dest, const char *src)
{
  intmax_t value;
  struct expr_code *compiled, *code, *ptr;

  if (expr_eval (src, &value) != 0)
    goto err;
  compiled = expr_compile (dest, EC_MODIFY);
  if (compiled == NULL)
    goto err;
  code = expr_cport_eval (compiled);
  free (compiled);
  if (code == NULL)
    goto err;
  ptr = expr_make_tree (code);
  for (;;)
    {
      switch (ptr->code)
	{
	case E_VAL:
	  if (ptr->v.val != value)
	    goto impossible;
	  goto done; /* Tell the user he is stupid? */

	case E_PORT:
	  if ((value
	       & ~((INTMAX_C (1) << port_width_value[ptr->v.port.width]) - 1))
	      != 0)
	    goto impossible;
	  switch (ptr->v.port.width)
	    {
#define DO_WIDTH(WIDTH)							\
	    case WIDTH_##WIDTH:						\
	      if (port_set_##WIDTH (&ptr->v.port.pi, value) != 0)	\
		goto err_code;						\
	      break;
	      DO_WIDTH (1);
	      DO_WIDTH (8);
	      DO_WIDTH (16);

	    default:
	      assert ("BUG" == NULL);
#undef DO_WIDTH
	    }
	  goto done;

	case E_NEG:
	  value = -value;
	  ptr = ptr->v.op[0];
	  break;

	case E_NOT:
	  if (value == 0)
	    goto unknown;
	  if (value != 1)
	    goto impossible;
	  value = 0;
	  ptr = ptr->v.op[0];
	  break;

	  /* Associative operators */
	case E_MUL: case E_ADD: case E_EQ: case E_NE: case E_BAND: case E_BXOR:
	case E_BOR: case E_LAND: case E_LOR:
	  {
	    struct expr_code *sub, *val;

	    if (ptr->v.op[0]->code == E_VAL)
	      {
		val = ptr->v.op[0];
		sub = ptr->v.op[1];
		assert (sub->code != E_VAL);
	      }
	    else if (ptr->v.op[1]->code == E_VAL)
	      {
		sub = ptr->v.op[0];
		val = ptr->v.op[1];
	      }
	    else
	      goto unknown;
	    switch (ptr->code)
	      {
	      case E_MUL:
		if (value % val->v.val != 0)
		  goto impossible;
		value /= val->v.val;
		break;

	      case E_ADD:
		value -= val->v.val;
		break;

	      case E_EQ:
		if (value == 0)
		  goto unknown;
		else if (value != 1)
		  goto impossible;
		value = val->v.val;
		break;
		
	      case E_NE:
		if (value == 1)
		  goto unknown;
		else if (value != 0)
		  goto impossible;
		value = val->v.val;
		break;

	      case E_BAND:
		if ((value & ~val->v.val) != 0)
		  goto impossible;
		goto unknown;

	      case E_BXOR:
		value ^= val->v.val;
		break;
		
	      case E_BOR:
		if ((value & val->v.val) != val->v.val)
		  goto impossible;
		if (val->v.val != 0)
		  goto unknown;
		break;
		
	      case E_LAND:
		if (value == 0)
		  {
		    if (val->v.val == 0)
		      goto unknown;
		  }
		else if (value == 1)
		  {
		    if (val->v.val == 0)
		      goto impossible;
		    goto unknown;
		  }
		else
		  goto impossible;
		break;

	      case E_LOR:
		if (value == 0)
		  {
		    if (val->v.val != 0)
		      goto impossible;
		  }
		else if (value == 1)
		  goto unknown;
		else
		  goto impossible;
		break;

	      default:
		assert ("BUG" == NULL);
	      }
	    ptr = sub;
	    break;
	  }

	  /* Non-associative operators */
	case E_DIV: case E_MOD: case E_SUB: case E_LSH: case E_RSH: case E_LT:
	case E_GT: case E_LE: case E_GE:
	  {
	    struct expr_code *left, *right;

	    left = ptr->v.op[0];
	    right = ptr->v.op[1];
	    if (left->code == E_VAL)
	      {
		assert (right->code != E_VAL);
		switch (ptr->code)
		  {
		  case E_DIV:
		    {
		      intmax_t v;
		      
		      if (value == 0)
			{
			  if ((uintmax_t)imaxabs (left->v.val) ==
			      ((uintmax_t)INTMAX_MAX >= -(uintmax_t)INTMAX_MIN
			       ? (uintmax_t)INTMAX_MAX
			       : -(uintmax_t)INTMAX_MIN))
			    goto impossible;
			  goto unknown;
			}
		      if (imaxabs (left->v.val) < imaxabs (value))
			goto impossible;
		      v = left->v.val / value;
		      assert (v != 0);
		      if (left->v.val / v != value)
			goto impossible;
		      value = v;
		      break;
		    }

		  case E_MOD:
		    if (imaxabs (left->v.val) <= imaxabs (value))
		      goto impossible;
		    /* There are some cases when we know the solution:
		       given a % x == b, we know |b| < |x| and
		       (a - b) % x == 0, or (a - b) == k * x for some
		       integer k.  So we would need to factorize (a - b) and
		       figure out how many possibilities are there.
		       Factorization of imax_t (which is at least 64b) takes
		       too much time in the worst case to be usable. */
		    goto unknown;

		  case E_SUB:
		    value = left->v.val - value;
		    break;
		    
		  case E_LSH:
		    {
		      int v, i;

		      v = -1;
		      /* Note that shifting by >= INTMAX_BITS (if the constant
			 existed) is undefined by C standard.  OTOH we are
			 guarranteed intmax_t and uintmax_t have the same
			 width. */
		      for (i = 0; i == 0 || UINTMAX_C (2) << (i - 1) != 0; i++)
			{
			  if (left->v.val << i == value)
			    {
			      if (v != -1)
				goto unknown;
			      v = i;
			    }
			}
		      if (v == -1)
			goto impossible;
		      value = v;
		      break;
		    }
		    
		  case E_RSH:
		    {
		      int v, i;

		      v = -1;
		      /* Same as above. */
		      for (i = 0; i == 0 || UINTMAX_C (2) << (i - 1) != 0; i++)
			{
			  if (left->v.val >> i == value)
			    {
			      if (v != -1)
				goto unknown;
			      v = i;
			    }
			}
		      if (v == -1)
			goto impossible;
		      value = v;
		      break;
		    }

		  case E_LT:
		    if (value == 0)
		      {
			if (left->v.val != INTMAX_MIN)
			  goto unknown;
			value = INTMAX_MIN;
		      }
		    else if (value == 1)
		      {
			if (left->v.val == INTMAX_MAX)
			  goto impossible;
			if (left->v.val != INTMAX_MAX - 1)
			  goto unknown;
			value = INTMAX_MAX;
		      }
		    else
		      goto impossible;
		    break;
		    
		  case E_GT:
		    if (value == 0)
		      {
			if (left->v.val != INTMAX_MAX)
			  goto unknown;
			value = INTMAX_MAX;
		      }
		    else if (value == 1)
		      {
			if (left->v.val == INTMAX_MIN)
			  goto impossible;
			if (left->v.val != INTMAX_MIN + 1)
			  goto unknown;
			value = INTMAX_MIN;
		      }
		    else
		      goto impossible;
		    break;

		  case E_LE:
		    if (value == 0)
		      {
			if (left->v.val == INTMAX_MIN)
			  goto impossible;
			if (left->v.val != INTMAX_MIN + 1)
			  goto unknown;
			value = INTMAX_MIN;
		      }
		    else if (value == 1)
		      {
			if (left->v.val != INTMAX_MAX)
			  goto unknown;
			value = INTMAX_MAX;
		      }
		    else
		      goto impossible;
		    break;
		    
		  case E_GE:
		    if (value == 0)
		      {
			if (left->v.val == INTMAX_MAX)
			  goto impossible;
			if (left->v.val != INTMAX_MAX - 1)
			  goto unknown;
			value = INTMAX_MAX;
		      }
		    else if (value == 1)
		      {
			if (left->v.val != INTMAX_MIN)
			  goto unknown;
			value = INTMAX_MIN;
		      }
		    else
		      goto impossible;
		    break;

		  default:
		    assert ("BUG" == NULL);
		  }
		ptr = right;
	      }
	    else if (right->code == E_VAL)
	      {
		switch (ptr->code)
		  {
		  case E_DIV:
		    if (right->v.val == 0)
		      {
			error (0, "division by zero");
			goto err_code;
		      }
		    if (imaxabs (right->v.val) != 1)
		      goto unknown;
		    value *= right->v.val;
		    break;
		    
		  case E_MOD:
		    if (right->v.val == 0)
		      {
			error (0, "division by zero");
			goto err_code;
		      }
		    if (value == 0)
		      goto unknown; /* At least 0 and right->v.val */
		    if (((value > 0) ^ (right->v.val > 0)) == 0)
		      { /* left->v.val > 0 */
			if ((uintmax_t)imaxabs (value)
			    + (uintmax_t)imaxabs (right->v.val)
			    <= (uintmax_t)INTMAX_MAX)
			  goto unknown;
		      }
		    else /* left->v.val < 0 */
		      {
			if ((uintmax_t)imaxabs (value)
			    + (uintmax_t)imaxabs (right->v.val)
			    <= -(uintmax_t)INTMAX_MIN)
			  goto unknown;
			value = -value;
		      }
		    break;

		  case E_SUB:
		    value += right->v.val;
		    break;
		    
		  case E_LSH:
		    if ((value & ((UINTMAX_C (1) << right->v.val) - 1)) != 0)
		      goto impossible;
		    value >>= right->v.val;
		    break;

		  case E_RSH:
		    {
		      intmax_t v;

		      v = value << right->v.val;
		      /* Shifting negative values right is
			 implementation-defined, and in the usual case (sign
			 extension) we still may not get what we want
			 (we get (01xyz << 1) >> 1 == 11xyz), so check the
			 result. */
		      if (v >> right->v.val != value)
			goto impossible;
		      value = v;
		      break;
		    }
		    
		  case E_LT:
		    if (value == 0)
		      {
			if (right->v.val != INTMAX_MAX)
			  goto unknown;
			value = INTMAX_MAX;
		      }
		    else if (value == 1)
		      {
			if (right->v.val == INTMAX_MIN)
			  goto impossible;
			if (right->v.val != INTMAX_MIN + 1)
			  goto unknown;
			value = INTMAX_MIN;
		      }
		    else
		      goto impossible;
		    break;
		    
		  case E_GT:
		    if (value == 0)
		      {
			if (right->v.val != INTMAX_MIN)
			  goto unknown;
			value = INTMAX_MIN;
		      }
		    else if (value == 1)
		      {
			if (right->v.val == INTMAX_MAX)
			  goto impossible;
			if (right->v.val != INTMAX_MAX - 1)
			  goto unknown;
			value = INTMAX_MAX;
		      }
		    else
		      goto impossible;
		    break;

		  case E_LE:
		    if (value == 0)
		      {
			if (right->v.val == INTMAX_MAX)
			  goto impossible;
			if (right->v.val != INTMAX_MAX - 1)
			  goto unknown;
			value = INTMAX_MAX;
		      }
		    else if (value == 1)
		      {
			if (right->v.val != INTMAX_MIN)
			  goto unknown;
			value = INTMAX_MIN;
		      }
		    else
		      goto impossible;
		    break;
		    
		  case E_GE:
		    if (value == 0)
		      {
			if (right->v.val == INTMAX_MIN)
			  goto impossible;
			if (right->v.val != INTMAX_MIN + 1)
			  goto unknown;
			value = INTMAX_MIN;
		      }
		    else if (value == 1)
		      {
			if (right->v.val != INTMAX_MAX)
			  goto unknown;
			value = INTMAX_MAX;
		      }
		    else
		      goto impossible;
		    break;
		    
		  default:
		    assert ("BUG" == NULL);
		  }
		ptr = left;
	      }
	    else
	      goto unknown;
	    break;
	  }

	case E_COND:
	  {
	    struct expr_code *cond;

	    cond = ptr->v.op[0];
	    if (cond->code == E_VAL)
	      ptr = ptr->v.op[cond->v.val ? 1 : 2];
	    else
	      {
		struct expr_code *left, *right;

		left = ptr->v.op[1];
		right = ptr->v.op[2];
		if (left->code != E_VAL || right->code != E_VAL)
		  goto unknown;
		if (value == left->v.val)
		  goto unknown;
		else if (value != right->v.val)
		  goto impossible;
		value = 0;
		ptr = cond;
	      }
	    break;
	  }
	  
	case E_END: case E_CPORT: default:
	  assert ("BUG" == NULL);
	}
    }
 done:
  free (code);
  return 0;

 impossible:
  error (0, "expression cannot have the desired value");
  goto err_code;

 unknown:
  error (0, "cannot compute value to set");
 err_code:
  free (code);
 err:
  return -1;
}
