/* Expression compilation and evaluation
   Copyright (C) 2002, 2004 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
%{
#include <config.h>

#include <assert.h>
#include <errno.h>
#include <inttypes.h>  
#include <stdlib.h>
#include <string.h>

#include <expr.h>
#include <lib.h>
#include <port.h>

#include "expr_code.h"
  
/* Parser information */
struct pi
{
  const char *expr; /* Current position in parsed expression */
  struct expr_code *start, *code, *end;
  unsigned flags; /* See EC_* in <expr.h> */
};

/* Semantic data type */
struct semantic
{
  /* In NUM, there was an error evaluating the port referece, so call YYABORT
     when possible */
  int err; 
};

#define YYSTYPE struct semantic

#define YYPARSE_PARAM pi_
#define YYLEX_PARAM pi_
#define YYERROR_VERBOSE
 
static void expre_error (const char *s);
static int expre_lex (YYSTYPE *lvalp, void *pi_);

/* Add OP to expression */
static void expre_add_op (struct pi *pi, int op);
 
%}

/* I like to see things bottom-up, not top-down.  Everybody must think I'm
   crazy ;-). */
%start expression

%pure_parser
%debug

%token NUM
%token LSH "<<"
%token RSH ">>"
%token LEQ "<="
%token GEQ ">="
%token EQ "=="
%token NEQ "!="
%token LAND "&&"
%token LOR "||"

%right '?' ':'
%left "||"
%left "&&"
%left '|'
%left '^'
%left '&'
%left "==" "!="
%left '<' '>' "<=" ">="
%left "<<" ">>"
%left '+' '-'
%left '*' '/' '%'
%right UPLUS UNEG '!'

%%

primary:	NUM
			{ 
			  if ($$.err != 0)
			    YYABORT;
			}
		| '(' cond ')'
		;

unary:	primary
	| '+' unary %prec UPLUS
	| '-' unary %prec UNEG
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c->code == E_VAL)
		    c->v.val = -c->v.val;
		  else
		    expre_add_op (pi_, E_NEG);
		}
 	| '!' unary
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c->code == E_VAL)
		    c->v.val = !c->v.val;
		  else
		    expre_add_op (pi_, E_NOT);
		}
	/* '~' is left out on purpose.  We can't guess what type the user
	   thinks this expression is evaluated in.  Instead of ~x something
	   like (x ^ 0xFF) may be used, and the result is obvious. */
	;

mult:	unary
	| mult '*' unary
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val *= c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_MUL);
		}
	| mult '/' unary
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      if (c->v.val == 0)
			{
			  expre_error ("division by zero");
			  YYABORT;
			}
		      c[-1].v.val /= c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_DIV);
		}
	| mult '%' unary
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      if (c->v.val == 0)
			{
			  expre_error ("division by zero");
			  YYABORT;
			}
		      c[-1].v.val %= c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_MOD);
		}
	;

add:	mult
	| add '+' mult
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val += c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_ADD);
		}
	| add '-' mult
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val -= c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_SUB);
		}
	;

shift:	add
	| shift "<<" add
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val <<= c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_LSH);
		}
	| shift ">>" add
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val >>= c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_RSH);
		}
	;

relation:	shift
		| relation '<' shift
			{
			  struct expr_code *c;
			  
			  c = ((struct pi *)pi_)->code - 1;
			  if (c[-1].code == E_VAL && c->code == E_VAL)
			    {
			      c[-1].v.val = c[-1].v.val < c->v.val;
			      ((struct pi *)pi_)->code = c;
			    }
			  else
			    expre_add_op (pi_, E_LT);
			}
		| relation '>' shift
			{
			  struct expr_code *c;
			  
			  c = ((struct pi *)pi_)->code - 1;
			  if (c[-1].code == E_VAL && c->code == E_VAL)
			    {
			      c[-1].v.val = c[-1].v.val > c->v.val;
			      ((struct pi *)pi_)->code = c;
			    }
			  else
			    expre_add_op (pi_, E_GT);
			}
		| relation "<=" shift
			{
			  struct expr_code *c;
			  
			  c = ((struct pi *)pi_)->code - 1;
			  if (c[-1].code == E_VAL && c->code == E_VAL)
			    {
			      c[-1].v.val = c[-1].v.val <= c->v.val;
			      ((struct pi *)pi_)->code = c;
			    }
			  else
			    expre_add_op (pi_, E_LE);
			}
		| relation ">=" shift
			{
			  struct expr_code *c;
			  
			  c = ((struct pi *)pi_)->code - 1;
			  if (c[-1].code == E_VAL && c->code == E_VAL)
			    {
			      c[-1].v.val = c[-1].v.val >= c->v.val;
			      ((struct pi *)pi_)->code = c;
			    }
			  else
			    expre_add_op (pi_, E_GE);
			}
		;

equal:	relation
	| equal "==" relation
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val = c[-1].v.val == c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		  expre_add_op (pi_, E_EQ);
		}
	| equal "!=" relation
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val = c[-1].v.val != c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_NE);
		}
	;

band:	equal
	| band '&' equal
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val &= c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_BAND);
		}
	;

bxor:	band
	| bxor '^' band
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val ^= c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_BXOR);
		}
	;

bor:	bxor
	| bor '|' bxor
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val |= c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_BOR);
		}
	;

land:	bor
	| land "&&" bor
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val = c[-1].v.val && c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_LAND);
		}
	;

lor:	land
	| lor "||" land
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 1;
		  if (c[-1].code == E_VAL && c->code == E_VAL)
		    {
		      c[-1].v.val = c[-1].v.val || c->v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_LOR);
		}
	;

cond:	lor
	| lor '?' cond ':' cond
		{
		  struct expr_code *c;

		  c = ((struct pi *)pi_)->code - 2;
		  if (c[-1].code == E_VAL && c->code == E_VAL
		      && c[1].code == E_VAL)
		    {
		      c[-1].v.val = c[-1].v.val ? c->v.val : c[1].v.val;
		      ((struct pi *)pi_)->code = c;
		    }
		  else
		    expre_add_op (pi_, E_COND);
		}
	;

expression:	cond
		{ expre_add_op (pi_, E_END); }
	;

%%

/* Add *CODE to expression */
static void
expre_add (struct pi *pi, const struct expr_code *code)
{
  if (pi->code == pi->end)
    {
      struct expr_code *copy;
      size_t size;

      size = 2 * (pi->end - pi->start);
      copy = xrealloc (pi->start, size * sizeof (*pi->start));
      pi->code = copy + (pi->code - pi->start);
      pi->end = copy + size;
      pi->start = copy;
    }
  *pi->code++ = *code;
}

/* Add OP to expression */
static void
expre_add_op (struct pi *pi, int op)
{
  struct expr_code code;

  code.code = op;
  expre_add (pi, &code);
}

static void
expre_error (const char *s)
{
  error (0, "%s", s);
}

/* Is C a hexadecimal digit? */
static int
is_hex_digit (char c)
{
  static const char valid[22] = "0123456789ABCDEFabcdef";

  return memchr (valid, c, sizeof (valid)) != NULL;
}

/* If we set lvalp->err = 1 in one of the parsing functions, we can leave
   pi->expr pointing anywhere, because as soon as we return to the parser,
   the parser aborts and pi->expr is not used again. */

/* Parse a character constant from PI to LVALP */
inline static void
parse_char_constant (YYSTYPE *lvalp, struct pi *pi)
{
  struct expr_code code;
  intmax_t val;
  char c;

  c = *pi->expr++;
  switch (c)
    {
    case '\'':
      expre_error ("empty character constant");
      lvalp->err = 1;
      return;

    case 0:
      goto unterminated;
      
    default:
      val = (unsigned char)c;
      break;

    case '\\':
      switch (*pi->expr++)
	{
	case '\'': case '\"': case '\?': case '\\':
	  val = (unsigned char)pi->expr[-1];
	  break;

	case 'a':
	  val = '\a';
	  break;

	case 'b':
	  val = '\b';
	  break;

	case 'f':
	  val = '\f';
	  break;

	case 'n':
	  val = '\n';
	  break;

	case 't':
	  val = '\t';
	  break;

	case 'r':
	  val = '\r';
	  break;

	case 'v':
	  val = '\v';
	  break;

	case '0': case '1': case '2': case '3': case '4': case '5': case '6':
	case '7':
	  val = pi->expr[-1] - '0';
	  if (*pi->expr >= '0' && *pi->expr < '8')
	    {
	      val = val * 8 + *pi->expr++ - '0';
	      if (*pi->expr >= '0' && *pi->expr < '8')
		val = val * 8 + *pi->expr++ - '0';
	    }
	  break;

	case 'x':
	  if (!is_hex_digit (*pi->expr))
	    goto bad_hex;
	  val = 0;
	  do
	    {
	      unsigned digit;
	      
	      switch (*pi->expr++)
		{
		case '0': case '1': case '2': case '3': case '4': case '5':
		case '6': case '7': case '8': case '9':
		  digit = pi->expr[-1] - '0';
		  break;

		case 'A': case 'a':
		  digit = 10;
		  break;
		case 'B': case 'b':
		  digit = 11;
		  break;
		case 'C': case 'c':
		  digit = 12;
		  break;
		case 'D': case 'd':
		  digit = 13;
		  break;
		case 'E': case 'e':
		  digit = 14;
		  break;
		case 'F': case 'f':
		  digit = 15;
		  break;

		default:
		  assert ("BUG" == NULL);
		  goto bad_hex;
		}
	      if (val > INTMAX_MAX / 16
		  || (val == INTMAX_MAX / 16 && digit > INTMAX_MAX % 16))
		goto bad_hex;
	      val = val * 16 + digit;
	    }
	  while (is_hex_digit (*pi->expr));
	  break;

	default:
	  expre_error ("invalid excape sequence");
	  lvalp->err = 1;
	  return;
	}
      break;
    }
  if (*pi->expr++ != '\'')
    goto unterminated;
  lvalp->err = 0;
  code.code = E_VAL;
  code.v.val = val;
  expre_add (pi, &code);
  return;

 bad_hex:
  expre_error ("invalid hexadecimal escape sequence");
  lvalp->err = 1;
  return;

 unterminated:
  expre_error ("unterminated character constant");
  lvalp->err = 1;
}

/* Parse a port value reference from PI to LVALP.  In case this never gets
   documented ;-), the expected format is [PORT] or [c:PORT] if you want to
   reference the port in evaluate/modify, but don't want to assign to it. */
static void
parse_port (YYSTYPE *lvalp, struct pi *pi)
{
  struct expr_code code;
  char *end;
  size_t len;

  if (*pi->expr == 'c' && pi->expr[1] == ':')
    {
      pi->expr += 2;
      code.code = E_CPORT;
    }
  else
    code.code = E_PORT;
  /* FIXME: provide a way to define aliases */
  errno = 0;
  end = strchr (pi->expr, ']');
  if (end == NULL)
    {
      expre_error ("unterminated port reference");
      lvalp->err = 1;
      return;
    }
  len = end - pi->expr;
  {
    char copy[len + 1];
    struct port_id id;

    memcpy (copy, pi->expr, len);
    copy[len] = 0;
    code.v.port.width = get_port_width (port_find (NUM_WIDTHS, copy, &id));
    if (code.v.port.width == NUM_WIDTHS)
      {
	lvalp->err = 1;
	return;
      }
    switch (code.v.port.width)
      {
#define DO_TYPE(WIDTH, TYPE)						\
       case WIDTH_##WIDTH:						\
	{								\
	  TYPE val;							\
									\
	  if (((pi->flags & EC_MODIFY) == 0 || code.code == E_CPORT)	\
	      && port_disp_##WIDTH (&id, &val) != 0)			\
	    {								\
	      lvalp->err = 1;						\
	      return;							\
	    }								\
	  if ((pi->flags & EC_VALUE) == 0)				\
	    code.v.port.pi = id;					\
	  else								\
	    {								\
	      code.code = E_VAL;					\
	      code.v.val = val;						\
	    }								\
	  break;							\
	}
	DO_TYPE (1, int);
	DO_TYPE (8, uint8_t);
	DO_TYPE (16, uint16_t);

      default:
	assert ("BUG" == NULL);
#undef DO_TYPE
      }
  }
  pi->expr = end + 1;
  lvalp->err = 0;
  expre_add (pi, &code);
}

static int
expre_lex (YYSTYPE *lvalp, void *pi_)
{
  struct pi *pi;
  char c;

  pi = pi_;
  while ((c = *pi->expr++) != 0)
    {
      switch (c)
	{
	case ' ': case '\t': case '\n': case '\v': case '\f':
	  break; /* Ignore */

	case '0': case '1': case '2': case '3': case '4': case '5': case '6':
	case '7': case '8': case '9':
	  {
	    char *end;
	    struct expr_code code;

	    pi->expr--; /* Back to the first digit */
	    errno = 0;
	    code.v.val = strtoimax (pi->expr, &end, 0);
	    if (end == pi->expr || errno != 0)
	      {
		expre_error ("invalid integer constant");
		lvalp->err = 1;
		return NUM;
	      }
	    pi->expr = end;
	    lvalp->err = 0;
	    code.code = E_VAL;
	    expre_add (pi, &code);
	    return NUM;
	  }

	case '\'': /* L' ' not supported */
	  parse_char_constant (lvalp, pi);
	  return NUM;

	case '[':
	  parse_port (lvalp, pi);
	  return NUM;

#define CHECK(CHAR, SYM)			\
	  if (*pi->expr == CHAR)		\
	    {					\
	      pi->expr++;			\
	      return SYM;			\
	    }
	case '!':
	  CHECK ('=', NEQ);
	  return c;
	  
	case '&':
	  CHECK ('&', LAND);
	  return c;

	case '<':
	  CHECK ('<', LSH);
	  CHECK ('=', LEQ);
	  return c;

	case '=':
	  CHECK ('=', EQ);
	  return c;
	  
	case '>':
	  CHECK ('>', RSH);
	  CHECK ('=', GEQ);
	  return c;

	case '|':
	  CHECK ('|', LOR);
	  return c;
#undef CHECK
	  
	case 0: case '%': case '(': case ')': case '*': case '+': case '-':
	case '/': case ':': case '?': case '^': default:
	  return c;
	}
      
    }
  return 0;
}

/* Evaluate EXPR, store result to DEST;
   returns: 0 if OK, -1 on error */
int
expr_eval (const char *expr, intmax_t *dest)
{
  struct pi pi;
  int res;

  pi.expr = expr;
  pi.start = xmalloc (8 * sizeof (*pi.start));
  pi.code = pi.start;
  pi.end = pi.start + 8;
  pi.flags = EC_VALUE;
  if (expre_parse (&pi) != 0)
    {
      res = -1;
      goto err;
    }
  assert (pi.start->code == E_VAL && pi.start[1].code == E_END);
  *dest = pi.start->v.val;
  res = 0;
 err:
  free (pi.start);
  return res;
}

/* Compile EXPR according to FLAGS (see EC_* below);
   returns: expression (for free ()) if OK, NULL on error */
struct expr_code *
expr_compile (const char *expr, unsigned flags)
{
  struct pi pi;

  pi.expr = expr;
  pi.start = xmalloc (8 * sizeof (*pi.start));
  pi.code = pi.start;
  pi.end = pi.start + 8;
  pi.flags = flags;
  if (expre_parse (&pi) == 0)
    {
      pi.start = xrealloc (pi.start,
			   (pi.code - pi.start) * sizeof (*pi.start));
      return pi.start;
    }
  free (pi.start);
  return NULL;
}
