/* Module instance management
   Copyright (C) 2002, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>

#include <lib.h>
#include <me.h>
#include <module.h>
#include <port.h>

/* me name hash table size */
enum { HASH_ENTRIES = 31 };

/* List of all me's */
struct me_entry *me_list; /* = NULL; */

/* Type used for hash values */
typedef uint32_t hash_t;

/* me hash table chain entry */
struct hash_entry
{
  struct me_entry me;
  struct hash_entry *next;
  hash_t hash;
  struct me_entry **list_pprev; /* in me_list */
};

/* me name hash table */
static struct hash_entry *hash_table[HASH_ENTRIES]; /* = { 0, }; */

/* Compute hash value for NAME */
static hash_t
hash_fn (const char *name)
{
  hash_t hash;

  hash = 0;
  while (*name != 0)
    hash = (hash << 8) + (hash >> 24) + (unsigned char)*name++;
  return hash;
}

/* Find hash_entry for NAME with HASH;
   returns: pointer to me_entry or NULL if not fount */
static struct hash_entry **
hash_find_ptr (const char *name, hash_t hash)
{
  struct hash_entry **l;

  for (l = hash_table + hash % HASH_ENTRIES; *l != NULL; l = &(*l)->next)
    {
      if ((*l)->hash == hash && strcmp (name, (*l)->me.name) == 0)
	return l;
    }
  return NULL;
}

/* Find hash_entry for NAME with HASH;
   returns: me_entry or NULL if not found */
static struct me_entry *
hash_find_known (const char *name, hash_t hash)
{
  struct hash_entry **l;

  l = hash_find_ptr (name, hash);
  return l != NULL ? &(*l)->me : NULL;
}

/* Insert ENTRY with HASH to hash table */
static void
hash_insert (struct hash_entry *entry, hash_t hash)
{
  struct hash_entry **list;
  
  entry->hash = hash;
  list = hash_table + hash % HASH_ENTRIES;
  entry->next = *list;
  *list = entry;
}

/* Find me NAME;
   returns: me if found, NULL otherwise */
struct me_entry *
me_find (const char *name)
{
  return hash_find_known (name, hash_fn (name));
}

/* Find hash_entry for NAME with HASH, report error if not found;
   returns: pointer to me_entry or NULL if not fount */
static struct hash_entry **
hash_get_ptr (const char *name, hash_t hash)
{
  struct hash_entry **entry;

  entry = hash_find_ptr (name, hash);
  if (entry == NULL)
    error (0, "unknown module `%s'", name);
  return entry;
}

/* Find me NAME, report error if not found;
   returns: me if found, NULL otherwise */
struct me_entry *
me_get (const char *name)
{
  struct me_entry *me;

  me = me_find (name);
  if (me == NULL)
    error (0, "unknown module `%s'", name);
  return me;
}

/* Check if NAME is OK
   returns: 1 if it is, 0 otherwise */
int
me_name_ok (const char *name)
{
  const char *p;

  if (*name == 0)
    {
      error (0, "module name cannot be empty");
      return 0;
    }
  /* FIXME: allow other Unicode letters? */
  if (strchr ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_", *name)
      == NULL)
    {
      /* How to tell the user not to expect \aleph to work? */
      error (0, "module name must start with a letter or underscore");
      return 0;
    }
  for (p = name + 1; *p != 0; p++)
    {
      if (strchr ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_"
		  "0123456789", *p) == NULL)
	{
	  error (0, "character `%c' (%hhu) is not allowed in module name", *p,
		 (unsigned char)*p);
	  return 0;
	}
    }
  return 1;
}

/* Create new me for MODULE called NAME;
   returns: me if OK, NULL on error */
struct me_entry *
me_new (const struct cereal_module *module, const char *name)
{
  struct hash_entry *entry;
  void *mi;
  hash_t hash;

  if (me_name_ok (name) == 0)
    return NULL;
  hash = hash_fn (name);
  if (hash_find_known (name, hash) != NULL)
    {
      error (0, "module named `%s' already exists", name);
      return NULL;
    }
  mi = module->mi_new ();
  if (mi == NULL)
    return NULL;
  entry = xmalloc (sizeof (*entry));
  entry->me.name = xstrdup (name);
  entry->me.module = module;
  entry->me.mi = mi;
  hash_insert (entry, hash);
  /* Guarranteed by the C standard */
  assert (offsetof (struct hash_entry, me) == 0);
  entry->list_pprev = &me_list;
  entry->me.next = me_list;
  if (entry->me.next != NULL)
    ((struct hash_entry *)entry->me.next)->list_pprev = &entry->me.next;
  me_list = &entry->me;
  return &entry->me;
}

/* Rename me MODULE to NEW_NAME;
   returns: 0 if OK, -1 on error */
int
me_rename (const char *old, const char *new_name)
{
  hash_t hash;
  struct hash_entry **p, *entry;

  if (me_name_ok (new_name) == 0)
    return -1;
  p = hash_get_ptr (old, hash_fn (old));
  if (p == NULL)
    return -1;
  if (strcmp (old, new_name) == 0)
    return 0;
  hash = hash_fn (new_name);
  if (hash_find_known (new_name, hash) != NULL)
    {
      error (0, "module named `%s' already exists", new_name);
      return -1;
    }
  entry = *p;
  free (entry->me.name);
  entry->me.name = xstrdup (new_name);
  *p = entry->next;
  hash_insert (entry, hash);
  return 0;
}

/* Delete ENTRYa */
static void
me_do_delete (struct hash_entry *entry)
{
  port_disconnect_me (&entry->me);
  entry->me.module->mi_delete (entry->me.mi);
  free (entry->me.name);
  /* Guarranteed by the C standard */
  assert (offsetof (struct hash_entry, me) == 0);
  *entry->list_pprev = entry->me.next;
  if (entry->me.next != NULL)
    ((struct hash_entry *)entry->me.next)->list_pprev = entry->list_pprev;
}

/* Delete me NAME */
void
me_delete (const char *name)
{
  struct hash_entry **p, *entry;
  
  p = hash_get_ptr (name, hash_fn (name));
  if (p == NULL)
    return;
  entry = *p;
  me_do_delete (entry);
  *p = entry->next;
  free (entry);
}

/* Delete all known me's */
void
me_delete_all (void)
{
  struct hash_entry **list;

  me_list = NULL;
  for (list = hash_table; list < hash_table + HASH_ENTRIES; list++)
    {
      struct hash_entry *e;
      
      while ((e = *list) != NULL)
	{
	  me_do_delete (e);
	  *list = e->next;
	  free (e);
	}
    }
}

/* Save setup (configuration options) of all me's as children of NODE */
void
me_save_setup (xmlNodePtr node)
{
  struct me_entry *me;

  for (me = me_list; me != NULL; me = me->next)
    {
      xmlNodePtr child;
      
      child = xmlNewChild (node, NULL, (const xmlChar *)"instance", NULL);
      xmlSetProp (child, (const xmlChar *)"module",
		  (const xmlChar *)me->module->name);
      xmlSetProp (child, (const xmlChar *)"name", (const xmlChar *)me->name);
      me->module->save_setup (me->mi, child);
    }
}

/* Load setup (configuration options of me's according to NODE of DOC;
   returns: 0 if OK, -1 on error */
int
me_load_setup (xmlDocPtr doc, xmlNodePtr node)
{
  int res;

  res = 0;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE)
	continue;
      if (strcmp ((const char *)node->name, "instance") == 0)
	{
	  const struct cereal_module *mod;
	  struct me_entry *me;
	  xmlChar *module, *name;

	  module = xmlGetProp (node, (const xmlChar *)"module");
	  if (module == NULL)
	    {
	      error (0, "missing `module' attribute");
	      res = -1;
	      goto err;
	    }
	  name = xmlGetProp (node, (const xmlChar *)"name");
	  if (name == NULL)
	    {
	      error (0, "missing `name' attribute");
	      res = -1;
	      goto err_module;
	    }
	  mod = cereal_module_find ((const char *)module);
	  if (mod == NULL)
	    {
	      error (0, "unknown module type `%s'", module);
	      res = -1;
	      goto err_name;
	    }
	  me = me_new (mod, (const char *)name);
	  if (me == NULL)
	    {
	      res = -1;
	      goto err_name;
	    }
	  if (me->module->load_setup (me->mi, doc, node) != 0)
	    res = -1;
	err_name:
	  xmlFree (name);
	err_module:
	  xmlFree (module);
	err:
	  ;
	}
    }
  return res;
}
     

/* Save state of all me's as children of NODE */
void
me_save_state (xmlNodePtr node)
{
  struct me_entry *me;

  for (me = me_list; me != NULL; me = me->next)
    {
      xmlNodePtr child;

      child = xmlNewChild (node, NULL, (const xmlChar *)"instance_state",
			   NULL);
      xmlSetProp (child, (const xmlChar *)"instance",
		  (const xmlChar *)me->name);
      me->module->save_state (me->mi, child);
    }
}

/* Load state of me's according to NODE of DOC;
   returns: 0 if OK, -1 on error */
int
me_load_state (xmlDocPtr doc, xmlNodePtr node)
{
  int res;

  res = 0;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE)
	continue;
      if (strcmp ((const char *)node->name, "instance_state") == 0)
	{
	  struct me_entry *me;
	  xmlChar *name;

	  name = xmlGetProp (node, (const xmlChar *)"instance");
	  if (name == NULL)
	    {
	      error (0, "missing `instance' attribute");
	      res = -1;
	      goto err;
	    }
	  me = me_find ((const char *)name);
	  if (me == NULL)
	    {
	      error (0, "unknown module `%s'", name);
	      res = -1;
	      goto err_name;
	    }
	  if (me->module->load_state (me->mi, doc, node) != 0)
	    res = -1;
	err_name:
	  xmlFree (name);
	err:
	  ;
	}
    }
  return res;
}
