/* Module registration and lookup
   Copyright (C) 2002, 2004, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <dlfcn.h>
#include <errno.h>
#include <limits.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>

#include <lib.h>
#include <module.h>

/* Port width values */
const unsigned port_width_value[NUM_WIDTHS] =
  { [WIDTH_1] = 1, [WIDTH_8] = 8, [WIDTH_16] = 16 };

/* Get port width from port width value;
   returns: port width or NUM_WIDTHS if not found */
enum port_width
get_port_width (unsigned value)
{
  const unsigned *p;

  for (p = port_width_value; p < port_width_value + NUM_WIDTHS; p++)
    {
      if (value == *p)
	break;
    }
  return p - port_width_value;
}

/* Module name hash table size */
enum { HASH_ENTRIES = 31 };

/* Type used for hash values */
typedef uint32_t hash_t;

/* Module name hash table chain entry */
struct hash_entry
{
  struct hash_entry *next, **pprev;
  hash_t hash;
  const char *name; /* Points to module->name or in module->compat->names */
  struct cereal_module *module;
};

/* Module name hash table */
static struct hash_entry *hash_table[HASH_ENTRIES]; /* = { 0, }; */

/* List of currently registered modules */
struct cereal_module *cereal_module_list; /* = NULL; */

/* Compute hash value for NAME */
static hash_t
hash_fn (const char *name)
{
  hash_t hash;

  hash = 0;
  while (*name != 0)
    hash = (hash << 8) + (hash >> 24) + (unsigned char)*name++;
  return hash;
}

/* Find hash_entry for NAME with HASH;
   returns: hash_entry or NULL if not found */
static struct hash_entry *
hash_find_known (const char *name, hash_t hash)
{
  struct hash_entry *e;

  for (e = hash_table[hash % HASH_ENTRIES]; e != NULL; e = e->next)
    {
      if (e->hash == hash && strcmp (name, e->name) == 0)
	break;
    }
  return e;
}

/* Find hash_entry for NAME;
   returns: hash_entry or NULL if not found */
static struct hash_entry *
hash_find (const char *name)
{
  return hash_find_known (name, hash_fn (name));
}

/* Add MODULE with NAME to hash table */
static void
hash_add (struct cereal_module *module, const char *name)
{
  hash_t hash;
  struct hash_entry *entry, **list;

  hash = hash_fn (name);
  entry = xmalloc (sizeof (*entry));
  entry->hash = hash;
  entry->name = name;
  entry->module = module;
  list = hash_table + hash % HASH_ENTRIES;
  entry->pprev = list;
  entry->next = *list;
  if (entry->next != NULL)
    entry->next->pprev = &entry->next;
  *list = entry;
}

/* Register MODULE to make it accessible by its name.
   returns: 0 if OK, -1 on error */
int
cereal_module_register (struct cereal_module *module)
{
  static const char xml_dir[] = CEREAL_XML_DIR;
  static const char xml_suffix[] = ".xml";
  size_t name_len;
  const char *env_dir;
  struct module_width *mw;

  if (hash_find (module->name) != NULL)
    {
      error (0, "module `%s' already exists", module->name);
      return -1;
    }
  if (module->compat_names != NULL)
    {
      const char *const *compat;

      for (compat = module->compat_names; *compat != NULL; compat++)
	{
	  if (hash_find (*compat) != NULL)
	    {
	      error (0, "obsolete module `%s' already exists", *compat);
	      return -1;
	    }
	}
      for (compat = module->compat_names; *compat != NULL; compat++)
	hash_add (module, *compat);
    }
  hash_add (module, module->name);
  module->pprev = &cereal_module_list;
  module->next = cereal_module_list;
  if (module->next != NULL)
    module->next->pprev = &module->next;
  cereal_module_list = module;
  module->xml = NULL;
  name_len = strlen (module->name);
  env_dir = getenv ("CEREAL_XML_DIR");
  if (env_dir != NULL)
    {
      size_t dir_len;

      dir_len = strlen (env_dir);
      {
	char buf[dir_len + 1 + name_len + sizeof (xml_suffix)];

	memcpy (buf, env_dir, dir_len);
	buf[dir_len] = '/';
	memcpy (buf + dir_len + 1, module->name, name_len);
	/* Including terminating '\0' */
	memcpy (buf + dir_len + 1 + name_len, xml_suffix, sizeof (xml_suffix));
	module->xml = xmlParseFile (buf);
      }
    }
  if (module->xml == NULL)
    {
      char buf[sizeof (xml_dir) + name_len + sizeof (xml_suffix)];
      
      memcpy (buf, xml_dir, sizeof (xml_dir) - 1);
      buf[sizeof (xml_dir) - 1] = '/';
      memcpy (buf + sizeof (xml_dir), module->name, name_len);
      /* Including terminating '\0' */
      memcpy (buf + sizeof (xml_dir) + name_len, xml_suffix,
	      sizeof (xml_suffix));
      module->xml = xmlParseFile (buf);
    }
  for (mw = module->w; mw < module->w + NUM_WIDTHS; mw++)
    {
      size_t len;

      len = mw->num_spaces * sizeof (*mw->xml_spaces);
      mw->xml_spaces = xmalloc (len);
      memset (mw->xml_spaces, 0, len);
    }
  if (module->xml != NULL)
    {
      xmlNodePtr node;
      int ok;

      node = xmlDocGetRootElement (module->xml);
      ok = 0;
      if (node->type == XML_ELEMENT_NODE
	  && strcmp ((const char *)node->name, "cereal_module") == 0)
	{
	  xmlChar *name;
	  
	  name = xmlGetProp (node, (const xmlChar *)"name");
	  if (name != NULL)
	    {
	      ok = strcmp ((const char *)name, module->name) == 0;
	      xmlFree (name);
	    }
	}
      if (ok == 0)
	{
	  error (0, "invalid XML description file found for module `%s' will "
		 "be ignored", module->name);
	  xmlFreeDoc (module->xml);
	  module->xml = NULL;
	}
      for (node = node->children; node != NULL; node = node->next)
	{
	  xmlNodePtr s;
	  
	  if (node->type != XML_ELEMENT_NODE
	      || strcmp ((const char *)node->name, "spaces") != 0)
	    continue;
	  for (s = node->children; s != NULL; s = s->next)
	    {
	      xmlChar *str;
	      unsigned long val;
	      enum port_width width;
	      size_t idx;
	      int res;
	      
	      if (s->type != XML_ELEMENT_NODE
		  || strcmp ((const char *)s->name, "space") != 0)
		continue;
	      /* Check that required fields are present */
	      str = xmlGetProp (s, (const xmlChar *)"name");
	      if (str == NULL)
		continue;
	      xmlFree (str);
	      str = xmlGetProp (s, (const xmlChar *)"description");
	      if (str == NULL)
		continue;
	      xmlFree (str);
	      str = xmlGetProp (s, (const xmlChar *)"width");
	      if (str == NULL)
		continue;
	      res = get_num ((const char *)str, 0, &val);
	      xmlFree (str);
	      if (res != 0)
		continue;
	      width = get_port_width (val);
	      if (width == NUM_WIDTHS)
		continue;
	      str = xmlGetProp (s, (const xmlChar *)"index");
	      if (str == NULL)
		continue;
	      res = get_num ((const char *)str, 0, &val);
	      xmlFree (str);
	      if (res != 0)
		continue;
	      idx = val;
	      mw = module->w + width;
	      if (idx >= mw->num_spaces)
		continue;
	      mw->xml_spaces[idx] = s;
	    }
	}
    }
  return 0;
}

/* Unregister module ENTRY (already removed from hash table) */
static void
module_do_unregister (struct hash_entry *entry)
{
  struct cereal_module *m;
  struct module_width *mw;

  m = entry->module;
  if (m->compat_names != NULL)
    {
      const char *const *compat;
      
      for (compat = m->compat_names; *compat != NULL; compat++)
	{
	  struct hash_entry *e;

	  e = hash_find (*compat);
	  assert (e != NULL);
	  *e->pprev = e->next;
	  if (e->next != NULL)
	    e->next->pprev = e->pprev;
	  free (e);
	}
    }
  if (m->xml != NULL)
    xmlFreeDoc (m->xml);
  for (mw = m->w; mw < m->w + NUM_WIDTHS; mw++)
    free (mw->xml_spaces);
  *m->pprev = m->next;
  if (m->next != NULL)
    m->next->pprev = m->pprev;
  free (entry);
  if (m->handle != NULL)
    dlclose (m->handle);
}

/* Unregister registered MODULE */
void
cereal_module_unregister (const struct cereal_module *module)
{
  struct hash_entry *entry;
  
  entry = hash_find (module->name);
  assert (entry != NULL);
  *entry->pprev = entry->next;
  if (entry->next != NULL)
    entry->next->pprev = entry->pprev;
  module_do_unregister (entry);
}

/* Find module NAME;
   returns: module or NULL if not found */
const struct cereal_module *
cereal_module_find (const char *name)
{
  struct hash_entry *entry;

  entry = hash_find (name);
  if (entry == NULL)
    return NULL;
  return entry->module;
}

/* Unregister all known modules */
void
cereal_modules_unregister_all (void)
{
  struct hash_entry **list;

  cereal_module_list = NULL;
  for (list = hash_table; list < hash_table + HASH_ENTRIES; list++)
    {
      struct hash_entry **l, *e;

      l = list;
      while ((e = *l) != NULL)
	{
	  if (e->name == e->module->name)
	    {
	      *l = e->next;
	      module_do_unregister (e);
	    }
	  else
	    l = &e->next;
	}
    }
}

/* Save list of all known modules as children of NODE */
void
cereal_modules_save (xmlNodePtr node)
{
  struct hash_entry **list;

  for (list = hash_table; list < hash_table + HASH_ENTRIES; list++)
    {
      struct hash_entry *e;

      for (e = *list; e != NULL; e = e->next)
	{
	  if (e->name == e->module->name)
	    xmlNewTextChild (node, NULL, (const xmlChar *)"module",
			     (const xmlChar *)e->name);
	}
    }
}

/* Load modules according to NODE of DOC;
   returns: 0 if OK, -1 on error */
int
cereal_modules_load (xmlDocPtr doc, xmlNodePtr node)
{
  int res;
  
  res = 0;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE)
	continue;
      if (strcmp ((const char *)node->name, "module") == 0)
	{
	  xmlChar *name;

	  name = xmlNodeListGetString (doc, node->children, 1);
	  if (cereal_module_find ((const char *)name) == NULL)
	    {
	      error (0, "cannot load module `%s'", name);
	      res = -1;
	    }
	  xmlFree (name);
	}
    }
  return res;
}

/* Set option NAME of MI_ to VALUE;
   returns: 0 if OK, -1 on error;
   Just say that the option is unknown */
int
generic_module_set_option (void *mi_, const char *name, const char *value)
{
  (void)mi_;
  (void)value;
  error (0, "unknown option `%s'", name);
  return -1;
}

/* Get value of option NAME of MI_;
   returns: string for free () or NULL on error;
   Just say that the option is unknown */
char *
generic_module_get_option (void *mi_, const char *name)
{
  (void)mi_;
  error (0, "unknown option `%s'", name);
  return NULL;
}

/* Save setup (configuration options) of MI_ to NODE;
   Do nothing */
void
generic_module_save_setup (void *mi_, xmlNodePtr node)
{
  (void)mi_;
  (void)node;
}

/* Load setup (configuration options) of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error;
   Just return 0 */
int
generic_module_load_setup (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  (void)mi_;
  (void)doc;
  (void)node;
  return 0;
}

/* Load setup (configuration options) of MI_ from NODE of DOC by setting
   options of MODULE according to elements of namespace NS;
   returns: 0 if OK, -1 on error */
int
generic_module_load_setup_options (void *mi_, xmlDocPtr doc, xmlNodePtr node,
				   const struct cereal_module *module,
				   const xmlChar *ns)
{
  xmlNsPtr xns;
  int res;

  xns = xmlSearchNsByHref (doc, node, ns);
  res = 0;
  if (xns == NULL)
    goto err;
  for (node = node->children; node != NULL; node = node->next)
    {
      xmlChar *val;
      
      if (node->type != XML_ELEMENT_NODE || node->ns != xns)
	continue;
      val = xmlNodeListGetString (doc, node->children, 1);
      if (val == NULL)
	continue;
      if (module->set_option (mi_, (const char *)node->name, (const char *)val)
	  != 0)
	res = -1;
      xmlFree (val);
    }
 err:
  return res;
}

/* Get XML node for SPACE of WIDTH in MODULE;
   returns: XML node pointer or NULL */
xmlNodePtr
space_get_xml (const struct cereal_module *module, enum port_width width,
	       size_t space)
{
  return module->w[width].xml_spaces[space];
}

/* Create name for SPACE of WIDTH in MODULE;
   returns: space name from libxml (in UTF-8) (for xmlFree ()) */
xmlChar *
space_create_name (const struct cereal_module *module, enum port_width width,
		   size_t space)
{
  xmlNodePtr node;
  xmlChar *name;

  node = space_get_xml (module, width, space);
  if (node != NULL)
    {
      name = xmlGetProp (node, (const xmlChar *)"name");
      if (name != NULL)
	return name;
    }
  assert (sizeof (space) * CHAR_BIT <= 64);
  name = xxmlMalloc (32);
  sprintf ((char *)name, "%zu", space);
  return name;
}

/* Find port space with NAME (as got from libxml, i.e. in UTF-8) of WIDTH
   (NUM_WIDTHS = any) in MODULE, store its offset to *PSPACE; if not found, try
   to interpret NAME as width:offset or, if WIDTH != 0, as offset (with
   WIDTH implied);
   returns: space width if OK, NUM_WIDTHS on error */
enum port_width
space_find_name (const struct cereal_module *module, const char *name,
		 enum port_width width, size_t *pspace)
{
  char *end;
  unsigned long val;
  enum port_width w;

  for (w = WIDTH_FIRST; w < NUM_WIDTHS; w++)
    {
      xmlNodePtr *node;
      const struct module_width *mw;

      if (width != NUM_WIDTHS && width != w)
	continue;
      mw = module->w + w;
      for (node = mw->xml_spaces; node < mw->xml_spaces + mw->num_spaces;
	   node++)
	{
	  xmlChar *str;
	  int res;
	  
	  str = xmlGetProp (*node, (const xmlChar *)"name");
	  if (str == NULL)
	    continue;
	  res = strcmp ((const char *)str, name);
	  xmlFree (str);
	  if (res == 0)
	    {
	      *pspace = node - mw->xml_spaces;
	      return w;
	    }
	}
    }
  errno = 0;
  val = strtoul (name, &end, 0);
  if (end == name || errno != 0)
    goto err;
  if (*end == ':')
    {
      char *e;
      
      if (width != NUM_WIDTHS && val != port_width_value[width])
	goto err;
      width = get_port_width (val);
      if (width == NUM_WIDTHS)
	goto err;
      end++;
      val = strtoul (end, &e, 0);
      if (e == end || errno != 0 || *e != 0)
	goto err;
    }
  else if (width == NUM_WIDTHS || *end != 0)
    goto err;
  if (val >= module->w[width].num_spaces)
    goto err;
  *pspace = val;
  return width;
  
 err:
  error (0, "unknown space `%s'", name);
  return NUM_WIDTHS;
}
