/* Port access
   Copyright (C) 2002, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <libxml/tree.h>

#include <lib.h>
#include <me.h>
#include <module.h>
#include <port.h>

/* List of current port connections */
struct connect_entry *connect_list; /* = NULL; */

/* Human-readable port type names */
const char *const port_type_name[NUM_PORT_TYPES] =
{
  [PORT_TYPE_READ] = "read", [PORT_TYPE_WRITE] = "write",
  [PORT_TYPE_DISPLAY] = "display", [PORT_TYPE_MODIFY] = "modify"
};

/* Read WIDTH (NUM_WIDTHS = any) port specification from SPEC, store to *PI;
   returns: 0 if OK, -1 on error */
unsigned
port_find (enum port_width width, const char *spec, struct port_id *pi)
{
  const struct me_entry *me;
  size_t space, len;

  /* We currently handle only MODULE/SPACE/PORT, the idea is to handle
     configurable aliases etc. */
  len = strlen (spec) + 1;
  {
    char buf[len];
    char *s2, *s3;
    enum port_width w;

    memcpy (buf, spec, len);
    for (s2 = buf; *s2 != '/'; s2++)
      {
	if (*s2 == 0)
	  goto syntax_err;
      }
    *s2++ = 0;
    for (s3 = s2; *s3 != '/'; s3++)
      {
	if (*s3 == 0)
	  goto syntax_err;
      }
    *s3++ = 0;
    if (strchr (s3, '/') != NULL)
      goto syntax_err;
    me = me_get (buf);
    if (me == NULL)
      return 0;
    w = space_find_name (me->module, s2, width, &space);
    if (width == NUM_WIDTHS)
      width = w;
    if (w == NUM_WIDTHS || w != width)
      return 0;
    pi->me = me;
    pi->space = space;
    if (port_find_name (width, pi, s3) != 0)
      {
	error (0, "unknown port `%s'", s3);
	return 0;
      }
  }
  return port_width_value[width];

 syntax_err:
  error (0, "port specification is not in MODULE/SPACE/PORT syntax");
  return 0;
}

#define DO_TYPE(WIDTH, TYPE, NAME, FMT)					  \
/* Get value of port *PI, store it to DEST;				  \
   returns: 0 if OK, -1 on error */					  \
int									  \
port_disp_##WIDTH (const struct port_id *pi, TYPE *dest)		  \
{									  \
  struct port_fn pf;							  \
  const struct space *sp;						  \
									  \
  sp = pi->me->module->w[WIDTH_##WIDTH].spaces + pi->space;		  \
  if (sp->get_fn[PORT_TYPE_DISPLAY] == NULL)				  \
    {									  \
      error (0, "address space `%td' doesn't support displaying value",	  \
	     pi->space);						  \
      return -1;							  \
    }									  \
  sp->get_fn[PORT_TYPE_DISPLAY] (pi->me->mi, pi->port, &pf);		  \
  if (pf.fn.display_##WIDTH == NULL)					  \
    {									  \
      error (0, NAME " %zu in address space `%td' doesn't support "	  \
	     "displaying value", pi->port, pi->space);			  \
      return -1;							  \
    }									  \
  *dest = pf.fn.display_##WIDTH (pf.mi, pf.data);			  \
  return 0;								  \
}									  \
									  \
/* Set port *PI to VAL;							  \
   returns: 0 if OK, -1 on error */					  \
int									  \
port_set_##WIDTH (const struct port_id *pi, unsigned long val)		  \
{									  \
  struct port_fn pf;							  \
  const struct space *sp;						  \
									  \
  if (val >= (1ul << WIDTH))						  \
    {									  \
      error (0, "value %" FMT "lX is too large", val);			  \
      return -1;							  \
    }									  \
  sp = pi->me->module->w[WIDTH_##WIDTH].spaces + pi->space;		  \
  if (sp->get_fn[PORT_TYPE_MODIFY] == NULL)				  \
    {									  \
      error (0, "address space `%td' doesn't support value modification", \
	     pi->space);						  \
      return -1;							  \
    }									  \
  sp->get_fn[PORT_TYPE_MODIFY] (pi->me->mi, pi->port, &pf);		  \
  if (pf.fn.modify_##WIDTH == NULL)					  \
    {									  \
      error (0, NAME " %zu in address space `%td' doesn't support value " \
	     "modification", pi->port, pi->space);			  \
      return -1;							  \
    }									  \
  pf.fn.modify_##WIDTH (pf.mi, pf.data, val);				  \
  return 0;								  \
}

DO_TYPE(1, int, "bit", "");
DO_TYPE(8, uint8_t, "byte", "02");
DO_TYPE(16, uint16_t, "16b", "04");

/* Create port name for PI of WIDTH;
   returns: port name from libxml (in UTF-8) (for xmlFree ()) */
xmlChar *
port_create_name (enum port_width width, const struct port_id *pi)
{
  xmlNodePtr node;
  xmlChar *str;

  node = space_get_xml (pi->me->module, width, pi->space);
  if (node != NULL)
    {
      for (node = node->children; node != NULL; node = node->next)
	{
	  unsigned long val;
	  int ret;
	  
	  if (node->type != XML_ELEMENT_NODE
	      || strcmp ((const char *)node->name, "port") != 0)
	    continue;
	  str = xmlGetProp (node, (const xmlChar *)"index");
	  if (str == NULL)
	    continue;
	  ret = get_num ((const char *)str, 0, &val);
	  xmlFree (str);
	  if (ret != 0 || val != pi->port)
	    continue;
	  str = xmlGetProp (node, (const xmlChar *)"name");
	  if (str == NULL)
	    continue;
	  return str;
	}
    }
  assert (sizeof (pi->port) * CHAR_BIT <= 64);
  str = xxmlMalloc (32);
  sprintf ((char *)str, "%zu", pi->port);
  return str;
}

/* Read port number from NAME (in UTF-8), store to PI of WIDTH;
   returns: 0 if OK, -1 on error (PI->port not modified) */
int
port_find_name (enum port_width width, struct port_id *pi, const char *name)
{
  unsigned long val;
  size_t max;
  int res;
  xmlNodePtr node;
  char *tail;
  const struct cereal_module *module;

  module = pi->me->module;
  node = space_get_xml (module, width, pi->space);
  max = module->w[width].spaces[pi->space].get_size (pi->me->mi);
  if (node != NULL)
    {
      for (node = node->children; node != NULL; node = node->next)
	{
	  xmlChar *str;
	  
	  if (node->type != XML_ELEMENT_NODE
	      || strcmp ((const char *)node->name, "port") != 0)
	    continue;
	  str = xmlGetProp (node, (const xmlChar *)"name");
	  if (str == NULL)
	    continue;
	  res = strcmp ((const char *)str, name);
	  xmlFree (str);
	  if (res != 0)
	    continue;
	  str = xmlGetProp (node, (const xmlChar *)"index");
	  if (str == NULL)
	    continue;
	  res = get_num ((const char *)str, 0, &val);
	  xmlFree (str);
	  if (res != 0 || val >= max)
	    continue;
	  pi->port = val;
	  return 0;
	}
    }
  val = strtoul (name, &tail, 0);
  if (tail == name || *tail != 0 || errno != 0 || val >= max)
    res = -1;
  else
    {
      res = 0;
      pi->port = val;
    }
  return res;
}

/* Resolve port type from NAME;
   returns: port type or -1 if not found */
int
port_type_find (const char *name)
{
  const char *const *p;

  for (p = port_type_name; p < port_type_name + ARRAY_SIZE (port_type_name);
       p++)
    {
      if (strcmp (*p, name) == 0)
	return p - port_type_name;
    }
  return -1;
}

/* Find connection going "the other way" (read / write) than CE;
   returns: peer connection or NULL if not found */
struct connect_entry *
port_peer_connection (const struct connect_entry *ce)
{
  struct connect_entry *res;
  int type;

  switch (ce->type)
    {
    case PORT_TYPE_READ:
      type = PORT_TYPE_WRITE;
      break;

    case PORT_TYPE_WRITE:
      type = PORT_TYPE_READ;
      break;

    case PORT_TYPE_DISPLAY: case PORT_TYPE_MODIFY:
      return NULL;

    default:
      assert ("BUG" == NULL);
      return NULL;
    }
  for (res = connect_list; res != NULL; res = res->next)
    {
      if (res->type == type && res->width == ce->width
	  && res->pi1.me == ce->pi2.me && res->pi1.space == ce->pi2.space
	  && res->pi1.port == ce->pi2.port && res->pi2.me == ce->pi1.me
	  && res->pi2.space == ce->pi1.space && res->pi2.port == ce->pi1.port)
	break;
    }
  return res;
}

/* Create a new connection using CE;
   returns: CE if OK, NULL on error */
struct connect_entry *
port_do_connect (struct connect_entry *ce)
{
  const char *size_name;
  int err;
  const struct space *space1, *space2;
  struct port_fn pf;

  space1 = ce->pi1.me->module->w[ce->width].spaces + ce->pi1.space;
  space2 = ce->pi2.me->module->w[ce->width].spaces + ce->pi2.space;
  if (space2->get_fn[ce->type] == NULL)
    goto dest_space_unsup;
  if (space1->set_fn[ce->type] == NULL)
    goto src_space_unsup;
  space2->get_fn[ce->type] (ce->pi2.me->mi, ce->pi2.port, &pf);
  switch (ce->width)
    {
#define CONNECT_TYPE(TN, WIDTH)			\
      do					\
	{					\
	  if (pf.fn.TN##_##WIDTH == NULL)	\
	    goto dest_port_unsup;		\
	}					\
      while (0)
#define DO_WIDTH(WIDTH, NAME)			\
    case WIDTH_##WIDTH:				\
      {						\
	size_name = NAME;			\
	switch (ce->type)			\
	  {					\
	  case PORT_TYPE_READ:			\
	    CONNECT_TYPE (read, WIDTH);		\
	    break;				\
						\
	  case PORT_TYPE_WRITE:			\
	    CONNECT_TYPE (write, WIDTH);	\
	    break;				\
						\
	  case PORT_TYPE_DISPLAY:		\
	    CONNECT_TYPE (display, WIDTH);	\
	    break;				\
						\
	  case PORT_TYPE_MODIFY:		\
	    CONNECT_TYPE (modify, WIDTH);	\
	    break;				\
						\
	  default:				\
	    assert ("BUG" == NULL);		\
	    return NULL;			\
	  }					\
	break;					\
      }
      DO_WIDTH (1, "bit");
      DO_WIDTH (8, "byte");
      DO_WIDTH (16, "16b");
      
    default:
      assert ("BUG" == NULL);
      return NULL;
#undef DO_WIDTH
    }
  err = space1->set_fn[ce->type] (ce->pi1.me->mi, ce->pi1.port, &pf);
  switch (err)
    {
    case 0:
      break;

    case ENOTSUP:
      goto src_port_unsup;

    case EBUSY:
      goto src_port_busy;

    default:
      goto unknown_err;
    }
  ce->pprev = &connect_list;
  ce->next = connect_list;
  if (ce->next != NULL)
    ce->next->pprev = &ce->next;
  connect_list = ce;
  return ce;

 dest_space_unsup:
  error (0, "destination address space doesn't support %s access",
	 port_type_name[ce->type]);
  return NULL;

 src_space_unsup:
  error (0, "source address space doesn't support %s access",
	 port_type_name[ce->type]);
  return NULL;

 dest_port_unsup:
  error (0, "destination %s doesn't support %s access", size_name,
	 port_type_name[ce->type]);
  return NULL;

 src_port_unsup:
  error (0, "source %s doesn't support %s access", size_name,
	 port_type_name[ce->type]);
  return NULL;

 src_port_busy:
  error (0, "source %s is already connected for %s", size_name,
	 port_type_name[ce->type]);
  return NULL;

 unknown_err:
  error (err, "unknown error");
  return NULL;
}

/* Create a new connection according to data in *SKEL;
   returns: connection entry if OK, NULL on error */
struct connect_entry *
port_connect (const struct connect_entry *skel)
{
  struct connect_entry *ce;

  ce = xmalloc (sizeof (*ce));
  *ce = *skel;
  if (port_do_connect (ce) != NULL)
    return ce;
  free (ce);
  return NULL;
}

/* Disconnect CE without deallocating it */
void
port_do_disconnect (struct connect_entry *ce)
{
  const struct space *sp;
  struct port_fn pf;

  sp = ce->pi1.me->module->w[ce->width].spaces + ce->pi1.space;
  switch (ce->width)
    {
#define DISCONNECT_TYPE(TN, WIDTH)		\
      do					\
	{					\
	  pf.fn.TN##_##WIDTH = NULL;		\
	}					\
      while (0)
#define DO_WIDTH(WIDTH)				\
    case WIDTH_##WIDTH:				\
      {						\
	switch (ce->type)			\
	  {					\
	  case PORT_TYPE_READ:			\
	    DISCONNECT_TYPE (read, WIDTH);	\
	    break;				\
						\
	  case PORT_TYPE_WRITE:			\
	    DISCONNECT_TYPE (write, WIDTH);	\
	    break;				\
						\
	  case PORT_TYPE_DISPLAY:		\
	    DISCONNECT_TYPE (display, WIDTH);	\
	    break;				\
						\
	  case PORT_TYPE_MODIFY:		\
	    DISCONNECT_TYPE (modify, WIDTH);	\
	    break;				\
						\
	  default:				\
	    assert ("BUG" == NULL);		\
	  }					\
	break;					\
      }
      DO_WIDTH (1);
      DO_WIDTH (8);
      DO_WIDTH (16);

    default:
      assert ("BUG" == NULL);
#undef DO_WIDTH
    }
  sp->set_fn[ce->type] (ce->pi1.me->mi, ce->pi1.port, &pf);
  *ce->pprev = ce->next;
  if (ce->next != NULL)
    ce->next->pprev = ce->pprev;
}

/* Disconnect CE */
void
port_disconnect (struct connect_entry *ce)
{
  port_do_disconnect (ce);
  free (ce);
}

/* Delete all connections involving ME */
void
port_disconnect_me (const struct me_entry *me)
{
  struct connect_entry **p;

  p = &connect_list;
  while (*p != NULL)
    {
      struct connect_entry *ce;

      ce = *p;
      if (me != ce->pi1.me && me != ce->pi2.me)
	p = &(*p)->next;
      else
	port_disconnect (ce);
    }
}

/* Forget about all known port connections */
void
port_disconnect_all (void)
{
  struct connect_entry *ce;

  while ((ce = connect_list) != NULL)
    {
      connect_list = ce->next;
      free (ce);
    }
}

/* Save list of all known port connectios as children of NODE */
void
port_connections_save (xmlNodePtr node)
{
  const struct connect_entry *ce;

  for (ce = connect_list; ce != NULL; ce = ce->next)
    {
      xmlNodePtr child;
      char buf[32];
      size_t len;

      child = xmlNewChild (node, NULL, (const xmlChar *)"connection", NULL);
      xmlSetProp (child, (const xmlChar *)"type",
		  (const xmlChar *)port_type_name[ce->type]);
      assert (sizeof (ce->type) * CHAR_BIT <= 64);
      sprintf (buf, "%u", port_width_value[ce->width]);
      xmlSetProp (child, (const xmlChar *)"width", (xmlChar *)buf);
      len = strlen (ce->pi1.me->name);
      assert (sizeof (ce->pi1.space) * CHAR_BIT <= 64);
      assert (sizeof (ce->pi1.port) * CHAR_BIT <= 64);
      {
	char b2[len + 64];

	sprintf (b2, "%s/%zu/%zu", ce->pi1.me->name, ce->pi1.space,
		 ce->pi1.port);
	xmlSetProp (child, (const xmlChar *)"port1", (xmlChar *)b2);
      }
      len = strlen (ce->pi2.me->name);
      assert (sizeof (ce->pi2.space) * CHAR_BIT <= 64);
      assert (sizeof (ce->pi2.port) * CHAR_BIT <= 64);
      {
	char b2[len + 64];

	sprintf (b2, "%s/%zu/%zu", ce->pi2.me->name, ce->pi2.space,
		 ce->pi2.port);
	xmlSetProp (child, (const xmlChar *)"port2", (xmlChar *)b2);
      }
    }
}

/* Load port connectios according to NODE of DOC;
   returns: 0 if OK, -1 on error */
int
port_connections_load (xmlDocPtr doc, xmlNodePtr node)
{
  int res;

  (void)doc;
  res = 0;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE)
	continue;
      if (strcmp ((const char *)node->name, "connection") == 0)
	{
	  xmlChar *p1, *p2, *buf;
	  int type;
	  unsigned long width;
	  enum port_width w;
	  struct connect_entry skel;

	  buf = xmlGetProp (node, (const xmlChar *)"type");
	  if (buf == NULL)
	    goto err;
	  type = port_type_find ((const char *)buf);
	  if (type < 0)
	    {
	      error (0, "unknown port type `%s'", buf);
	      xmlFree (buf);
	      res = -1;
	      goto err;
	    }
	  xmlFree (buf);
	  buf = xmlGetProp (node, (const xmlChar *)"width");
	  if (buf == NULL)
	    goto err;
	  if (get_num ((const char *)buf, 0, &width) != 0)
	    {
	      xmlFree (buf);
	      res = -1;
	      goto err;
	    }
	  xmlFree (buf);
	  w = get_port_width (width);
	  if (w == NUM_WIDTHS)
	    {
	      error (0, "unsupported port width `%lu'", width);
	      res = -1;
	      goto err;
	    }
	  p1 = xmlGetProp (node, (const xmlChar *)"port1");
	  if (p1 == NULL)
	    goto err;
	  p2 = xmlGetProp (node, (const xmlChar *)"port2");
	  if (p2 == NULL)
	    goto err_p1;
	  skel.type = type;
	  skel.width = w;
	  if (port_find (w, (const char *)p1, &skel.pi1) != width
	      || port_find (w, (const char *)p2, &skel.pi2) != width
	      || port_connect (&skel) == NULL)
	    {
	      res = -1;
	      goto err_p2;
	    }
	err_p2:
	  xmlFree (p2);
	err_p1:
	  xmlFree (p1);
	err:
	  ;
	}
    }
  return res;
}
