/* Emulation event scheduler
   Copyright (C) 2002, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <float.h>
#include <limits.h>
#include <locale.h>
#include <math.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>

#include <cereal.h>
#include <lib.h>

/* Detected conditions in this emulation step (see ER_* below), modified
   by emulation functions */
int emulation_result;

/* Schedule event representation. */
struct schedule_event
{
  size_t heap_index; /* 0 if not scheduled */
  struct schedule_time time;
  schedule_fn_t fn;
  void *mi_;
};

/* Current time */
struct schedule_time cereal_time = { .second = 0, .subsecond = 0.0 };

/* Scheduled events heap (has unused entry 0) */
static struct schedule_event **event_heap; /* = NULL; */
/* Number of valid heap entries, starting from entry 1 */
static size_t event_heap_size; /* = 0; */
/* Number of allocated heap entries, starting from entry 1 */
static size_t event_heap_alloc; /* = 0; */

/* Number of allocated events. */
static size_t total_events; /* = 0; */

/* Is A before B? */
inline static int
event_before (const struct schedule_event *a, const struct schedule_event *b)
{
  return a->time.second < b->time.second
    || (a->time.second == b->time.second
	&& a->time.subsecond < b->time.subsecond);
}

/* Add EVENT to event_heap */
inline static void
heap_add (struct schedule_event *event)
{
  size_t x;

  x = ++event_heap_size;
  if (event_heap_size > event_heap_alloc)
    {
      /* Do the realloc once for all conceivable events. */
      event_heap = xrealloc (event_heap,
			     (total_events + 1) * sizeof (*event_heap));
      event_heap_alloc = total_events;
    }
  while (x != 1)
    {
      size_t p;
      struct schedule_event *parent;

      p = x / 2;
      parent = event_heap[p];
      if (event_before (parent, event))
	break;
      event_heap[x] = parent;
      parent->heap_index = x;
      x = p;
    }
  event_heap[x] = event;
  event->heap_index = x;
}

/* Del EVENT from event_heap */
inline static void
heap_del (struct schedule_event *event)
{
  struct schedule_event *ev;
  size_t x;

  assert (event_heap_size != 0);
  ev = event_heap[event_heap_size];
  event_heap_size--;
  if (ev == event)
    return;
  x = event->heap_index;
  for (;;)
    {
      size_t j;
      struct schedule_event *child;

      j = x * 2;
      if (j > event_heap_size)
	break;
      if (j < event_heap_size
	  && event_before (event_heap[j + 1], event_heap[j]))
	j++;
      child = event_heap[j];
      if (event_before (ev, child))
	break;
      event_heap[x] = child;
      child->heap_index = x;
      x = j;
    }
  while (x != 1)
    {
      size_t p;
      struct schedule_event *parent;

      p = x / 2;
      parent = event_heap[p];
      if (event_before (parent, ev))
	break;
      event_heap[x] = parent;
      parent->heap_index = x;
      x = p;
    }
  event_heap[x] = ev;
  ev->heap_index = x;
}

/* Schedule emulation FN to be called with EVENT, TIME seconds later */
void
schedule_add (struct schedule_event *event, double time, schedule_fn_t fn)
{
  assert (event->heap_index == 0);
  event->time.second = cereal_time.second;
  time += cereal_time.subsecond;
  if (time >= 1.0L)
    {
      double tmp;

      tmp = floor (time);
      event->time.second += tmp;
      time -= tmp;
    }
  event->time.subsecond = time;
  event->fn = fn;
  heap_add (event);
}

/* Cancel a scheduled, not yet activated EVENT */
void
schedule_cancel (struct schedule_event *event)
{
  assert (event->heap_index != 0);
  heap_del (event);
  event->heap_index = 0;
}

/* Check whether there is an event scheduled */
int
schedule_pending (void)
{
  return event_heap_size > 0;
}

/* Run next scheduled event;
   returns: emulation result */
int
schedule_run (void)
{
  struct schedule_event *event;

  if (event_heap_size < 1)
    return ER_HALT;
  event = event_heap[1];
  schedule_cancel (event); /* Remove event from heap */
  /* FIXME: wait if we are too fast */
  cereal_time = event->time;
  emulation_result = 0;
  event->fn (event->mi_);
  if (event_heap_size < 1)
    emulation_result |= ER_HALT;
  return emulation_result;
}

/* Create a new schedulable event for use by MI_;
   returns: schedulable event */
struct schedule_event *
schedule_new (void *mi_)
{
  struct schedule_event *event;

  total_events++;
  event = xmalloc (sizeof (*event));
  event->heap_index = 0;
  event->mi_ = mi_;
  return event;
}

/* Delete EVENT (canceling it if it is already scheduled) */
void
schedule_delete (struct schedule_event *event)
{
  if (event->heap_index != 0)
    schedule_cancel (event);
  free (event);
  total_events--;
}

/* Return difference between A and B, in seconds */
double
schedule_difftime (const struct schedule_time *a,
		   const struct schedule_time *b)
{
  return (a->subsecond - b->subsecond)
    + ((double)a->second - (double)b->second);
}

/* Store TIME in human-readable format to BUF
   (of size at least 33 + DBL_DIG) */
static void
time_save (char *buf, const struct schedule_time *time)
{
  char *lc;

  assert (sizeof (time->second) * CHAR_BIT <= 64);
  lc = setlocale (LC_NUMERIC, NULL);
  if (lc != NULL)
    lc = xstrdup (lc);
  setlocale (LC_NUMERIC, "C");
  sprintf (buf, "%u,%.*f", time->second, DBL_DIG, time->subsecond);
  if (lc != NULL)
    {
      setlocale (LC_NUMERIC, lc);
      free (lc);
    }
}

/* Load TIME from BUF generated by time_save ();
   returns: 0 if OK, -1 on error */
static int
time_load (struct schedule_time *time, const xmlChar *buf)
{
  char *lc;
  int res;
  
  lc = setlocale (LC_NUMERIC, NULL);
  if (lc != NULL)
    lc = xstrdup (lc);
  setlocale (LC_NUMERIC, "C");
  res = sscanf ((const char *)buf, "%u,%lf", &time->second,  &time->subsecond);
  if (lc != NULL)
    {
      setlocale (LC_NUMERIC, lc);
      free (lc);
    }
  return res == 2 ? 0 : -1;
}

/* Save EVENT state ax XML element NS:NAME under NODE using (NULL-terminated)
   LIST of possible functions */
void
schedule_save_event (struct schedule_event *event, xmlNodePtr node,
		     xmlNsPtr ns, const char *name,
		     const struct schedule_fn_desc *list)
{
  node = xmlNewChild (node, ns, (const xmlChar *)name, NULL);
  if (event->heap_index == 0)
    xmlSetProp (node, (const xmlChar *)"time", (const xmlChar *)"off");
  else
    {
      char buf[33 + DBL_DIG];

      time_save (buf, &event->time);
      xmlSetProp (node, (const xmlChar *)"time", (xmlChar *)buf);
      while (list->fn != event->fn)
	{
	  list++;
	  assert (list->fn != NULL);
	}
      xmlSetProp (node, (const xmlChar *)"function",
		  (const xmlChar *)list->name);
    }
}

/* Restore EVENT state from NODE of DOC using (NULL-terminated) LIST of
   possible functions;
   returns: 0 if OK, -1 on error */
int
schedule_load_event (struct schedule_event *event, xmlNodePtr node,
		     const struct schedule_fn_desc *list)
{
  xmlChar *str;

  str = xmlGetProp (node, (const xmlChar *)"time");
  if (str == NULL)
    goto err;
  if (event->heap_index != 0)
    schedule_cancel (event);
  if (strcmp ((const char *)str, "off") != 0)
    {
      xmlChar *fn;

      if (time_load (&event->time, str) != 0)
	goto err;
      fn = xmlGetProp (node, (const xmlChar *)"function");
      if (fn == NULL)
	goto err;
      while (list->fn != NULL)
	{
	  if (strcmp (list->name, (const char *)fn) == 0)
	    goto found;
	  list++;
	}
      xmlFree (fn);
      goto err;
      
    found:
      xmlFree (fn);
      event->fn = list->fn;
      heap_add (event);
    }
  xmlFree (str);
  return 0;

 err:
  error (0, "invalid schedule event element");
  return -1;
}

/* Save global scheduling state to NODE */
void
schedule_save_state (xmlNodePtr node)
{
  char buf[33 + DBL_DIG];

  time_save (buf, &cereal_time);
  xmlNewTextChild (node, NULL, (const xmlChar *)"schedule_time",
		   (xmlChar *)buf);
}

/* Restore global scheduling state from NODE of DOC;
   returns: 0 if OK, -1 on error */
int
schedule_load_state (xmlDocPtr doc, xmlNodePtr node)
{
  int res;

  res = 0;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE)
	continue;
      if (strcmp ((const char *)node->name, "schedule_time") == 0)
	{
	  xmlChar *str;
	  
	  str = xmlNodeListGetString (doc, node->children, 1);
	  if (str == NULL)
	    {
	      res = -1;
	      goto err;
	    }
	  res = time_load (&cereal_time, str);
	  xmlFree (str);
	}
      if (res != 0)
	goto err;
    }
 err:
  return res;
}
