/* Setup handling
   Copyright (C) 2002, 2003, 2004, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <dirent.h>
#include <dlfcn.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include <breakpoint.h>
#include <cereal.h>
#include <lib.h>
#include <me.h>
#include <module.h>
#include <port.h>
#include <setup.h>

/* Create an "inital" setup */
void
setup_init (void)
{
  char *module_dir;
#define MODULE(X)				\
  if (cereal_module_register (&(X)) != 0)	\
    exit (EXIT_FAILURE);
#include <module_list.h> /* Register built-in modules */
#undef MODULE

  module_dir = getenv ("CEREAL_MODULE_DIR");
  if (module_dir != NULL)
    {
      DIR *dir;
      struct dirent *de;
      union
      {
	struct dirent d;
	char b[offsetof (struct dirent, d_name) + NAME_MAX + 1];
      } u;
      size_t dir_len;

      dir_len = strlen (module_dir);
      dir = opendir (module_dir);
      if (dir == NULL)
	goto err;
      while (readdir_r (dir, &u.d, &de) == 0 && de != NULL)
	{
	  size_t len;

	  len = strlen (de->d_name);
	  if (memcmp (de->d_name, "libcereal", 9) == 0
	      && memcmp (de->d_name + len - 3, ".so", 3) == 0)
	    {
	      char name[dir_len + len + 2];
	      void *handle;
	      struct cereal_module *(*fn) (void), *m;

	      memcpy (name, module_dir, dir_len);
	      name[dir_len] = '/';
	      memcpy (name + dir_len + 1, de->d_name, len + 1);
	      /* FIXME: use libltdl? */
	      handle = dlopen (name, RTLD_LAZY | RTLD_LOCAL);
	      if (handle == NULL)
		continue;
	      fn = (struct cereal_module *(*) (void))dlsym (handle,
							    "register_self");
	      if (fn != NULL && (m = fn ()) != NULL)
		m->handle = handle;
	      else
		dlclose (handle);
	    }
	}
      closedir (dir);
    err:
      ;
    }
}

/* Delete current setup */
void
setup_destroy (void)
{
  breakpoint_delete_all ();
  port_disconnect_all ();
  me_delete_all ();
  cereal_modules_unregister_all ();
}

/* Save current setup to NODE */
static void
setup_do_save (xmlNodePtr node)
{
  cereal_modules_save (xmlNewChild (node, NULL, (const xmlChar *)"modules",
				    NULL));
  me_save_setup (xmlNewChild (node, NULL, (const xmlChar *)"instances", NULL));
  port_connections_save (xmlNewChild (node, NULL,
				      (const xmlChar *)"connections", NULL));
}

/* Save current setup to FILE;
   returns: 0 if OK, -1 on error */
int
setup_save (const char *file)
{
  int res;
  xmlDocPtr doc;

  doc = xmlNewDoc ((const xmlChar *)XML_DEFAULT_VERSION);
  doc->children = xmlNewDocNode (doc, NULL, (const xmlChar *)"cereal_setup",
				 NULL);
  setup_do_save (doc->children);
  res = xmlSaveFormatFile (file, doc, 1);
  if (res >= 0)
    res = 0;
  else
    error (errno, "cannot save setup to `%s'", file);
  xmlFreeDoc (doc);
  return res;
}

/* Load current setup from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
setup_do_load (xmlDocPtr doc, xmlNodePtr node)
{
  int res;
  
  res = 0;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE)
	continue;
      if (strcmp ((const char *)node->name, "modules") == 0)
	res = cereal_modules_load (doc, node);
      else if (strcmp ((const char *)node->name, "instances") == 0)
	res = me_load_setup (doc, node);
      else if (strcmp ((const char *)node->name, "connections") == 0)
	res = port_connections_load (doc, node);
      if (res != 0)
	break;
    }
  return res;
}

/* Load current setup from FILE;
   returns: 0 if OK, -1 on error (current setup is already destroyed) */
int
setup_load (const char *file)
{
  xmlDocPtr doc;
  xmlNodePtr node;
  int res;
  
  setup_destroy ();
  setup_init ();
  doc = xmlParseFile (file);
  if (doc == NULL)
    {
      error (errno, "cannot load setup from `%s'", file);
      res = -1;
      goto err;
    }
  /* XML document parsing is intentionally liberal, but consider the right to
     reject any document not conforming to the DTD reserved ;-) */
  node = xmlDocGetRootElement (doc);
  if (node->type != XML_ELEMENT_NODE
      || strcmp ((const char *)node->name, "cereal_setup") != 0)
    {
      error (0, "invalid root element in `%s'", file);
      res = -1;
      goto err_doc;
    }
  res = setup_do_load (doc, node);
 err_doc:
  xmlFreeDoc (doc);
 err:
  return res;
}

/* Save current state to FILE, UI state is saved by FN with DATA;
   returns: 0 if OK, -=1 on error */
int
state_save (const char *file,
	    void (*fn) (struct _xmlNode *node, intptr_t data), intptr_t data)
{
  int res;
  xmlDocPtr doc;
  xmlNodePtr node;

  doc = xmlNewDoc ((const xmlChar *)XML_DEFAULT_VERSION);
  doc->children = xmlNewDocNode (doc, NULL, (const xmlChar *)"cereal_state",
				 NULL);
  setup_do_save (xmlNewChild (doc->children, NULL,
			      (const xmlChar *)"cereal_setup", NULL));
  me_save_state (xmlNewChild (doc->children, NULL,
			      (const xmlChar *)"instances_state", NULL));
  node = xmlNewChild (doc->children, NULL, (const xmlChar *)"global_state",
		      NULL);
  schedule_save_state (xmlNewChild (node, NULL,
				    (const xmlChar *)"global_schedule", NULL));
  fn (xmlNewChild (doc->children, NULL, (const xmlChar *)"ui_state", NULL),
      data);
  res = xmlSaveFormatFile (file, doc, 1);
  if (res >= 0)
    res = 0;
  else
    error (errno, "cannot save setup to `%s'", file);
  xmlFreeDoc (doc);
  return res;
}  

/* Load current state from FILE (or create empty state if FILE is
   cereal_setup document), UI state is loaded by UI_FN with DATA,
   call SETUP_FN with DATA when setup is loaded;
   returns: 1 if loaded cereal_state, 0 if loaded cereal_setup, -1 on error
   (current state is already destroyed) */
int
state_load (const char *file, void (*setup_fn) (intptr_t data),
	    int (*ui_fn) (struct _xmlDoc *doc, struct _xmlNode *node,
			  intptr_t data), intptr_t data)
{
  xmlDocPtr doc;
  xmlNodePtr node;
  int res, have_setup;

  setup_destroy ();
  setup_init ();
  doc = xmlParseFile (file);
  if (doc == NULL)
    {
      error (errno, "cannot load state from `%s'", file);
      res = -1;
      goto err;
    }
  /* XML document parsing is intentionally liberal, but consider the right to
     reject any document not conforming to the DTD reserved ;-) */
  node = xmlDocGetRootElement (doc);
  if (node->type != XML_ELEMENT_NODE)
    goto err_root;
  if (strcmp ((const char *)node->name, "cereal_setup") == 0)
    {
      res = setup_do_load (doc, node);
      setup_fn (data);
      goto err_doc;
    }
  if (strcmp ((const char *)node->name, "cereal_state") != 0)
    goto err_root;
  res = 0;
  have_setup = 0;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE)
	continue;
      if (strcmp ((const char *)node->name, "cereal_setup") == 0)
	{
	  if (have_setup != 0)
	    {
	      error (0, "<cereal_setup> element appears twice in `%s'", file);
	      res = -1;
	      goto err_doc;
	    }
	  res = setup_do_load (doc, node);
	  setup_fn (data);
	  have_setup = 1;
	}
      else if (strcmp ((const char *)node->name, "instances_state") == 0)
	{
	  if (have_setup == 0)
	    goto err_no_setup;
	  res = me_load_state (doc, node);
	}
      else if (strcmp ((const char *)node->name, "global_state") == 0)
	{
	  xmlNodePtr n;

	  if (have_setup == 0)
	    goto err_no_setup;
	  for (n = node->children; n != NULL; n = n->next)
	    {
	      if (n->type != XML_ELEMENT_NODE)
		continue;
	      if (strcmp ((const char *)n->name, "global_schedule") == 0)
		res = schedule_load_state (doc, n);
	    }
	}
      else if (strcmp ((const char *)node->name, "ui_state") == 0)
	{
	  if (have_setup == 0)
	    goto err_no_setup;
	  res = ui_fn (doc, node, data);
	}
      if (res != 0)
	goto err_doc;
    }
  if (have_setup == 0)
    goto err_no_setup;
  res = 1;
  goto err_doc;

 err_root:
  error (0, "invalid root element in `%s'", file);
  res = -1;
  goto err_doc;

 err_no_setup:
  error (0, "<cereal_setup> is not the first element of <cereal_state> in "
	 "`%s'", file);
  res = -1;
 err_doc:
  xmlFreeDoc (doc);
 err:
  return res;
}

/* Called when setup is loaded;
   Do nothing */
void
generic_state_load_setup_fn (intptr_t data)
{
  (void)data;
}
