/* EDS engine / volume meter combination
   Copyright (C) 2002 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <klineedit.h>
#include <libxml/tree.h>

#include <cereal.h>
#include <lib.h>
#include <module.h>

#include "eds_endlg.h"

/* Instance state */
struct mis
{
  unsigned flags; /* See MI_* below */
  struct schedule_time time; /* Last time of input transition */
  double on_time, off_time;
  unsigned time_valid; /* 0, 1, 2 (2 = at least one full cycle) */
  double change_time; /* Time between output changes */
  unsigned timeout; /* Number of output changes without 0->1 transition */
  struct schedule_event *change_event;
  eds_enDlg *dlg;
};

enum
{
  MI_OUT	= 0x01, /* Current value of output bit */
  MI_IN		= 0x02, /* Last value of input bit */
  MI_ACTIVE	= 0x04, /* change_event is active */
};

/* Number of output changes without 0->1 transition to stop the engine */
enum { MI_TIMEOUT = 200 };

enum
{
  S1P_IN, 
  S1P_OUT,
  NUM_S1P,
};

/* Create a new module instance;
   returns: module instance pointer if OK, NULL on error */
static void *
eds_en_mi_new (void)
{
  struct mis *mi;

  mi = (struct mis *)xmalloc (sizeof (*mi));
  mi->flags = 0;
  mi->time_valid = 0;
  mi->timeout = 0;
  mi->change_event = schedule_new (mi);
  mi->dlg = new eds_enDlg ();
  mi->dlg->show ();
  return mi;
}

/* Delete MI_ (may be in any emulation state) */
static void
eds_en_mi_delete (void *mi_)
{
  struct mis *mi;

  mi = (struct mis *)mi_;
  schedule_delete (mi->change_event);
  delete mi->dlg;
  free (mi);
}

/* Save state of MI_ to NODE */
static void
eds_en_save_state (void *mi_, xmlNodePtr node)
{
  /* FIXME: implement */
  (void)mi_;
  (void)node;
}

/* Load state of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
eds_en_load_state (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  /* FIXME: implement */
  (void)mi_;
  (void)doc;
  (void)node;
  return 0;
}

/* Get number of ports of this space in MI_ */
static size_t
space_get_size (void *mi_)
{
  (void)mi_;
  return NUM_S1P;
}

/* Change output, ... */
static void
change_fn (void *mi_)
{
  struct mis *mi;

  mi = (struct mis *)mi_;
  mi->flags ^= MI_OUT;
  if (++mi->timeout != MI_TIMEOUT)
    schedule_add (mi->change_event, mi->change_time, change_fn);
  else
    {
      mi->flags &= ~MI_ACTIVE;
      mi->dlg->inputText->setText ("");
      mi->dlg->volumeText->setText ("");
      mi->dlg->frequencyText->setText ("");
    }
}

/* Write VAL to bit port in MI_ */
static void
space_write_fn (void *mi_, intptr_t data, int val)
{
  struct mis *mi;

  (void)data;
  mi = (struct mis *)mi_;
  if (val == 0)
    {
      if ((mi->flags & MI_IN) != 0)
	{
	  if (mi->time_valid != 0)
	    mi->on_time = schedule_difftime (&cereal_time, &mi->time);
	  mi->time = cereal_time;
	  if (mi->time_valid != 2)
	    mi->time_valid++;
	  mi->timeout = 0;
	  mi->flags &= ~MI_IN;
	}
    }
  else if ((mi->flags & MI_IN) == 0)
    {
      if (mi->time_valid != 0)
	mi->off_time = schedule_difftime (&cereal_time, &mi->time);
      if (mi->time_valid == 2)
	{
	  double ratio, volume, frequency;
	  
	  ratio = mi->on_time / (mi->on_time + mi->off_time);
	  mi->dlg->inputText->setText (QString::number (ratio * 100.0)
				       .append (" %"));
	  volume = 9.8 * ratio + 0.1;
	  mi->dlg->volumeText->setText (QString::number (volume));
	  frequency = 58 + (volume - 0.5) * ((580 - 58) / (5.0 - 0.5));
	  mi->dlg->frequencyText->setText (QString
					   ::number (frequency));
	  mi->change_time = 0.5 / frequency;
	  if ((mi->flags & MI_ACTIVE) == 0)
	    {
	      mi->flags |= MI_ACTIVE;
	      schedule_add (mi->change_event, mi->change_time, change_fn);
	    }
	}
      mi->time = cereal_time;
      if (mi->time_valid != 2)
	mi->time_valid++;
      mi->timeout = 0;
      mi->flags |= MI_IN;
    }
}

/* Save port_fn for PORT writing to DEST */
static void
space_get_write_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_S1P);
  if (port == S1P_IN)
    {
      dest->fn.write_1 = space_write_fn;
      dest->mi = (struct mis *)mi_;
    }
  else
    dest->fn.write_1 = NULL;
}

/* Read bit port value for emulation or display (may have no side effects);
   returns: port value */
static int
space_display_fn (void *mi_, intptr_t data)
{
  struct mis *mi;

  (void)data;
  mi = (struct mis *)mi_;
  return (mi->flags & MI_OUT) != 0;
}

/* Save port_fn for PORT reading to DEST */
static void
space_get_read_fn (void *mi_, size_t port, struct port_fn *dest)

{
  assert (port < NUM_S1P);
  if (port == S1P_OUT)
    {
      dest->fn.read_1 = space_display_fn;
      dest->mi = (struct mis *)mi_;
    }
  else
    dest->fn.read_1 = NULL;
}

/* Save port_fn for PORT displaying to DEST */
static void
space_get_display_fn (void *mi_, size_t port, struct port_fn *dest)

{
  assert (port < NUM_S1P);
  if (port == S1P_OUT)
    {
      dest->fn.display_1 = space_display_fn;
      dest->mi = (struct mis *)mi_;
    }
  else
    dest->fn.display_1 = NULL;
}

static const struct space eds_en_space =
{
  space_get_size,
  { space_get_read_fn, space_get_write_fn, space_get_display_fn, NULL, },
  { NULL,  NULL,  NULL,  NULL }
};

struct cereal_module eds_en_module =
{
  NULL, NULL, "eds_en", NULL, NULL, NULL, eds_en_mi_new, eds_en_mi_delete,
  generic_module_set_option, generic_module_get_option,
  generic_module_save_setup, generic_module_load_setup, eds_en_save_state,
  eds_en_load_state,
  { { &eds_en_space, 1, NULL }, { NULL, 0, NULL }, { NULL, 0, NULL} }
};

extern "C" struct cereal_module *
register_self (void)
{
  return cereal_module_register (&eds_en_module) == 0 ? &eds_en_module : NULL;
}
