/* EDS led (active low)
   Copyright (C) 2003, 2004 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <cassert>
#include <kinstance.h>
#include <klibloader.h>
#include <kled.h>
#include <klocale.h>
#include <kparts/plugin.h>
#include <libxml/tree.h>

#include <expr.h>
#include <me.h>
#include <module.h>
#include "../kde/kcmainwin.h"
#include "../kde/kcuistatehandler.h"
#include "../kde/kcwindow.h"
#include "../kde/kcwindowprovider.h"
#include "../kde/kde.h"

#include "eds_led.h"

static const char KCEDSLedXMLNs[] =
"http://www.fw.cz/mitr/cereal/modules/eds_led";

KCEDSLedViewWin::KCEDSLedViewWin (KCWindowProvider *provider,
				  KCMainWin *mainWin, struct me_entry *me,
				  QWidget *parent, const char *name)
  : KLed (parent, name), KCWindow (this, provider, me->name, mainWin), me_ (me)
{
  setWFlags (getWFlags () | Qt::WDestructiveClose);

  setCaption (i18n ("LED: %1").arg (me->name));

  QString m (me->name);
  value = expr_compile (QString ("[%1/bit/value]").arg (m), 0);
  if (value == NULL)
    {
      value = expr_compile ("0", 0);
      assert (value != NULL);
    }
  cerealUpdate ();
}

KCEDSLedViewWin::~KCEDSLedViewWin ()
{
  free (value);
}

/* Display (well, not) error message */
static void
silent_error_handler (const QString &)
{
}

/** Update view because emulation state has changed */
void
KCEDSLedViewWin::cerealUpdate ()
{
  intmax_t val;
  void (*prev_error_handler) (const QString &msg) = error_handler;
  error_handler = silent_error_handler;
  if (expr_exec (value, &val) != 0)
    goto err;
  setState (val == 0 ? KLed::On : KLed::Off);
 err:
  error_handler = prev_error_handler;
}

/** Save current UI state under uiState */
void
KCEDSLedViewWin::saveState (struct _xmlNode *uiState)
{
  xmlNodePtr node = KCUIStateHandler::saveWidget (this, "KCEDSLedViewWin",
						  uiState, NULL);
  xmlNsPtr ns = xmlNewNs (node, (const xmlChar *)KCEDSLedXMLNs,
			  (const xmlChar *)"eds_led");
  xmlSetNsProp (node, ns, (const xmlChar *)"me", (const xmlChar *)me_->name);
}

/** Load UI state from node of doc;
    returns: 0 if OK, -1 on error */
int
KCEDSLedViewWin::loadState (struct _xmlDoc *doc, struct _xmlNode *node)
{
  if (KCUIStateHandler::loadWidget (this, doc, node) != 0)
    return -1;
  return 0;
}

KCEDSLed::KCEDSLed (QObject *parent, const char *name)
  : KParts::Plugin (parent, name)
{
  bit_value_module = cereal_module_find ("bit_value");
  if (bit_value_module == NULL)
    return; /* No way. */

  if (!parent->inherits ("KCMainWin"))
    return; /* No way either. */
  mainWin = static_cast <KCMainWin *> (parent);

  mainWin->registerStateHandler (this);
  mainWin->registerProvider (this);
}

KCEDSLed::~KCEDSLed ()
{
}

/** Save current UI state under uiState */
void
KCEDSLed::saveState (xmlNodePtr)
{ /* Nothing */
}

/** Load UI state from uiState of doc;
    returns: 0 if OK, -1 on error */
int
KCEDSLed::loadState (xmlDocPtr doc, xmlNodePtr uiState)
{
  int res = 0;
  for (xmlNodePtr node = uiState->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE
	  || strcmp ((const char *)node->name, "ui_window") != 0)
	continue;
      xmlChar *str = xmlGetProp (node, (const xmlChar *)"class");
      if (str == NULL)
	continue;
      if (strcmp ((const char *)str, "KCEDSLedViewWin") == 0)
	{
	  xmlChar *name = xmlGetNsProp (node, (const xmlChar *)"me",
					(const xmlChar *)KCEDSLedXMLNs);
	  if (name == NULL)
	    {
	      res = -1;
	      goto err;
	    }
	  struct me_entry *me = me_get ((const char *)name);
	  xmlFree (name);
	  if (me == NULL)
	    {
	      res = -1;
	      goto err;
	    }
	  KCEDSLedViewWin *w = new KCEDSLedViewWin (this, mainWin, me,
						    mainWin->viewParent ());
	  res = w->loadState (doc, node);
	  if (res == 0)
	    w->show ();
	  else
	    delete w;
	}
    err:
      xmlFree (str);
      if (res != 0)
	{
	  error (0, "Invalid eds_led ui_window state");
	  break;
	}
    }
  return res;
}

/** @return list of names of windows that can be created by provideWindow () */
QStringList
KCEDSLed::providedWindowList ()
{
  QStringList res;
  for (const struct me_entry *me = me_list; me != NULL; me = me->next)
    {
      if (me->module == bit_value_module)
	res.append (me->name);
    }
  return res;
}

/** Create a window corresponding to member i of the list returned by
    providedWindowList () */
void
KCEDSLed::provideWindow (unsigned i)
{
  for (struct me_entry *me = me_list; me != NULL; me = me->next)
    {
      if (me->module == bit_value_module)
	{
	  if (i == 0)
	    {
	      KCEDSLedViewWin *w
		= new KCEDSLedViewWin (this, mainWin, me,
				       mainWin->viewParent ());
	      w->show ();
	      break;
	    }
	  i--;
	}
    }
}

KCEDSLedFactory::KCEDSLedFactory (QObject *parent, const char *name)
  : KLibFactory (parent, name)
{
  instance = new KInstance ("KCEDSLedFactory");
}

KCEDSLedFactory::~KCEDSLedFactory ()
{
  delete instance;
}

QObject *
KCEDSLedFactory::createObject (QObject *parent, const char *name, const char *,
			       const QStringList &)
{
  QObject *obj = new KCEDSLed (parent, name);
  emit objectCreated (obj);
  return obj;
}

KInstance *KCEDSLedFactory::instance = 0;

extern "C" void *
init_libkceds_led ()
{
  return new KCEDSLedFactory;
}

#include "eds_led.moc"
