/* A simple module providing a write-only bit which reports its changes
   Copyright (C) 2002, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>

#include <lib.h>
#include <module.h>

/* One bit_report instance state */
struct mi
{
  char *name; /* For reporting */
  unsigned flags; /* See MI_* below */
};

enum
{
  MI_VAL	= 0x01, /* Stored value */
  MI_KNOWN	= 0x02 /* Previous value is known, so we can report changes */
};

/* Default name */
static char default_name[] = "bit_report";

/* Create a new module instance;
   returns: module instance pointer if OK, NULL on error */
static void *
bit_report_mi_new (void)
{
  struct mi *mi;

  mi = xmalloc (sizeof (*mi));
  mi->flags = 0;
  mi->name = default_name;
  return mi;
}

/* Delete MI_ (may be in any emulation state) */
static void
bit_report_mi_delete (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  free (mi);
}

/* Set option NAME of MI_ to VALUE;
   returns: 0 if OK, -1 on error */
static int
bit_report_set_option (void *mi_, const char *name, const char *value)
{
  struct mi *mi;

  mi = mi_;
  if (strcmp (name, "name") == 0)
    {
      if (mi->name != default_name)
	free (mi->name);
      mi->name = xstrdup (value);
      return 0;
    }
  error (0, "unknown option `%s'", name);
  return -1;
}

/* Get value of option NAME of MI_;
   returns: string for free () or NULL on error */
static char *
bit_report_get_option (void *mi_, const char *name)
{
  struct mi *mi;
  
  mi = mi_;
  if (strcmp (name, "name") == 0)
    return xstrdup (mi->name);
  error (0, "unknown option `%s'", name);
  return NULL;
}

/* Our XML namespace */
static const xmlChar bit_report_xml_ns[] =
"http://www.fw.cz/mitr/cereal/modules/bit_report";

/* Save setup (configuration options) of MI_ to NODE */
static void
bit_report_save_setup (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;

  mi = mi_;
  ns = xmlNewNs (node, bit_report_xml_ns, (const xmlChar *)"br");
  xmlNewTextChild (node, ns, (const xmlChar *)"name",
		   (const xmlChar *)mi->name);
}

/* Load setup (configuration options) of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
bit_report_load_setup (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  return generic_module_load_setup_options (mi_, doc, node, &bit_report_module,
					    bit_report_xml_ns);
}

/* Save state of MI_ to NODE */
static void
bit_report_save_state (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;

  mi = mi_;
  ns = xmlNewNs (node, bit_report_xml_ns, (const xmlChar *)"bit_report");
  xmlNewTextChild (node, ns, (const xmlChar *)"value",
		   (const xmlChar *)((mi->flags & MI_KNOWN) == 0 ? "unknown"
				     : (mi->flags & MI_VAL) != 0 ? "1" : "0"));
}

/* Load state of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
bit_report_load_state (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  int res;

  mi = mi_;
  ns = xmlSearchNsByHref (doc, node, bit_report_xml_ns);
  res = 0;
  if (ns == NULL)
    goto err;
  for (node = node->children; node != NULL; node = node->next)
    {
      xmlChar *str;

      if (node->type != XML_ELEMENT_NODE || node->ns != ns)
	continue;
      if (strcmp ((const char *)node->name, "value") != 0)
	continue;
      str = xmlNodeListGetString (doc, node->children, 1);
      if (str == NULL)
	continue;
      if (strcmp ((const char *)str, "unknown") == 0)
	mi->flags &= ~MI_KNOWN;
      else
	{
	  unsigned long val;
	  
	  if (get_num ((const char *)str, 0, &val) != 0)
	    {
	      res = -1;
	      goto err;
	    }
	  if (val == 0)
	    mi->flags = (mi->flags & ~MI_VAL) | MI_KNOWN;
	  else if (val == 1)
	    mi->flags |= MI_VAL | MI_KNOWN;
	  else
	    {
	      error (0, "invalid <bit_report:value> value `%s'", str);
	      res = -1;
	      goto err;
	    }
	}
      xmlFree (str);
    }
 err:
  return res;
}

/* Get number of ports of this space in MI_ */
static size_t
space_get_size (void *mi_)
{
  (void)mi_;
  return 1;
}

/* Write VAL to bit port in MI_ */
static void
space_write_fn (void *mi_, intptr_t data, int val)
{
  struct mi *mi;
  unsigned flag;

  (void)data;
  mi = mi_;
  flag = val != 0 ? MI_VAL : 0;
  if ((mi->flags & MI_KNOWN) != 0 && (mi->flags & MI_VAL) != flag)
    printf ("%s: %d written\n", mi->name, val != 0);
  mi->flags = flag | MI_KNOWN;
}

/* Save port_fn for PORT writing to DEST */
static void
space_get_write_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port == 0);
  dest->fn.write_1 = space_write_fn;
  dest->mi = mi_;
}

/* Read bit port value for display (unlike port_1_read_t may have no side
   effects);
   returns: port value */
static int
space_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  (void)data;
  mi = mi_;
  return (mi->flags & MI_VAL) != 0;
}

/* Save port_fn for PORT displaying to DEST */
static void
space_get_display_fn (void *mi_, size_t port, struct port_fn *dest)

{
  assert (port == 0);
  dest->fn.display_1 = space_display_fn;
  dest->mi = mi_;
}

/* Modify bit port value by user */
static void
space_modify_fn (void *mi_, intptr_t data, int val)
{
  struct mi *mi;

  (void)data;
  mi = mi_;
  mi->flags = val != 0 ? MI_VAL | MI_KNOWN : MI_KNOWN;
}

/* Save port_fn for PORT modification to DEST */
static void
space_get_modify_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port == 0);
  dest->fn.modify_1 = space_modify_fn;
  dest->mi = mi_;
}

static const struct space bit_report_space =
{
  .get_size = space_get_size, .get_fn =
  {
    [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = space_get_write_fn,
    [PORT_TYPE_DISPLAY] = space_get_display_fn,
    [PORT_TYPE_MODIFY] = space_get_modify_fn
  },
  .set_fn =
  {
    [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
    [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
  }
};

struct cereal_module bit_report_module =
{
  .next = NULL, .pprev = NULL, .name = "bit_report", .compat_names = NULL,
  .xml = NULL, .handle = NULL, .mi_new = bit_report_mi_new,
  .mi_delete = bit_report_mi_delete, .set_option = bit_report_set_option,
  .get_option = bit_report_get_option, .save_setup = bit_report_save_setup,
  .load_setup = bit_report_load_setup, .save_state = bit_report_save_state,
  .load_state = bit_report_load_state, .w =
  {
    [WIDTH_1] =
    { .spaces = &bit_report_space, .num_spaces = 1, .xml_spaces = NULL },
    [WIDTH_8] = { .spaces = NULL, .num_spaces = 0, .xml_spaces = NULL },
    [WIDTH_16] = { .spaces = NULL, .num_spaces = 0, .xml_spaces = NULL }
  }
};
