/* A 8-bit wide latch, modelled after Czech MH74ALS373
   Copyright (C) 2002, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>

#include <cereal.h>
#include <lib.h>
#include <module.h>

enum
{
  WIDTH = 8
};

struct mi
{
  uint8_t val;
  struct port_fn read[WIDTH], write[WIDTH], e_read, oe_read;
};

enum
{
  S1_IN		= 0,
  S1_OUT	= WIDTH,
  S1_E		= 2 * WIDTH,
  S1_OE,
  NUM_S1
};

/* Create a new module instance;
   returns: module instance pointer if OK, NULL on error */
static void *
byte_latch_mi_new (void)
{
  struct mi *mi;
  struct port_fn *pf;

  mi = xmalloc (sizeof (*mi));
  mi->val = 0;
  for (pf = mi->read; pf < mi->read + ARRAY_SIZE (mi->read); pf++)
    pf->fn.read_1 = NULL;
  for (pf = mi->write; pf < mi->write + ARRAY_SIZE (mi->write); pf++)
    pf->fn.write_1 = NULL;
  mi->e_read.fn.read_1 = NULL;
  mi->oe_read.fn.read_1 = NULL;
  return mi;
}

/* Delete MI_ (may be in any emulation state) */
static void
byte_latch_mi_delete (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  free (mi);
}

/* Our XML namespace */
static const xmlChar byte_latch_xml_ns[] =
"http://www.fw.cz/mitr/cereal/modules/byte_latch";

/* Save state of MI_ to NODE */
static void
byte_latch_save_state (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  char buf[32];

  mi = mi_;
  ns = xmlNewNs (node, byte_latch_xml_ns, (const xmlChar *)"byte_latch");
  sprintf (buf, "0x%02X", mi->val);
  xmlNewTextChild (node, ns, (const xmlChar *)"value", (xmlChar *)buf);
}

/* Load state of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
byte_latch_load_state (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  int res;

  mi = mi_;
  ns = xmlSearchNsByHref (doc, node, byte_latch_xml_ns);
  res = 0;
  if (ns == NULL)
    goto err;
  for (node = node->children; node != NULL; node = node->next)
    {
      unsigned long val;

      if (node->type != XML_ELEMENT_NODE || node->ns != ns)
	continue;
      if (strcmp ((const char *)node->name, "value") == 0)
	{
	  res = get_xml_num (doc, node, 0x100, &val);
	  if (res != 0)
	    goto err;
	  mi->val = val;
	}
    }
 err:
  return res;
}

/* Get number of ports of this space in MI_ */
static size_t 
space_1_get_size (void *mi_)
{
  (void)mi_;
  return NUM_S1;
}

/* Read bit port value in MI_;
   returns: port value */
static int
space_1_read_fn (void *mi_, intptr_t data)
{
  struct mi *mi;
  
  assert ((size_t)data < WIDTH);
  mi = mi_;
  if (mi->oe_read.fn.read_1 != NULL
      && mi->oe_read.fn.read_1 (mi->oe_read.mi, mi->oe_read.data) != 0)
    return 1;
  return (mi->val & (1 << data)) != 0;
}

/* Save port_fn for PORT reading to DEST */
static void 
space_1_get_read_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_S1);
  if (port >= S1_OUT && port < S1_OUT + WIDTH)
    {
      dest->fn.read_1 = space_1_read_fn;
      dest->mi = mi_;
      dest->data = port - S1_OUT;
    }
  else
    dest->fn.read_1 = NULL;
}

/* Write VAL to bit port in MI_ */
static void
space_1_write_fn (void *mi_, intptr_t data, int val)
{
  struct mi *mi;

  assert ((size_t)data < NUM_S1 && (data < S1_OUT || data >= S1_OUT + WIDTH));
  mi = mi_;
  /* Done this way to suppress GCC warning when S1_IN == 0 */
  if ((size_t)(data - S1_IN) < WIDTH)
    {
      uint8_t mask;
      
      /* FIXME: what happens when both /OE and E are high? */
      if ((mi->oe_read.fn.read_1 != NULL
	   && mi->oe_read.fn.read_1 (mi->oe_read.mi, mi->oe_read.data) != 0)
	  || (mi->e_read.fn.read_1 != NULL
	      && mi->e_read.fn.read_1 (mi->e_read.mi, mi->e_read.data) == 0))
	return;
      data -= S1_IN;
      mask = 1 << data;
      if ((((mi->val & mask) != 0) ^ (val != 0)) != 0)
	{
	  const struct port_fn *pf;
	  
	  mi->val ^= mask;
	  pf = mi->write + data;
	  if (pf->fn.write_1 != NULL)
	    pf->fn.write_1 (pf->mi, pf->data, val);
	}
    }
  else
    {
      switch (data)
	{
	case S1_E:
	  if (val != 0)
	    {
	      size_t i;
	      uint8_t mask;

	      mask = 0x01;
	      for (i = 0; i < WIDTH; i++)
		{
		  const struct port_fn *pf;
 
		  pf = mi->read + i;
		  if (pf->fn.read_1 != NULL)
		    {
		      int v;

		      v = pf->fn.read_1 (pf->mi, pf->data);
		      if (((v != 0) ^ ((mi->val & mask) != 0)) != 0)
			{
			  mi->val ^= mask;
			  pf = mi->write + i;
			  if (pf->fn.write_1 != NULL)
			    pf->fn.write_1 (pf->mi, pf->data, v);
			}
		    }
		  mask <<= 1;
		}
	    }
	  break;

	case S1_OE:
	  {
	    size_t i;
	    uint8_t mask;
	    
	    mask = 0x01;
	    for (i = 0; i < WIDTH; i++)
	      {
		const struct port_fn *pf;
		
		pf = mi->write + i;
		if (pf->fn.write_1 != NULL && (mi->val & mask) == 0)
		  /* Write 1 when going to high-impedance, restore 0 when
		     going back */
		  pf->fn.write_1 (pf->mi, pf->data, val);
		mask <<= 1;
	      }
	    break;
	  }

	default:
	  assert ("BUG" == NULL);
	}
    }
}

/* Save port_fn for PORT writing to DEST */
static void 
space_1_get_write_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_S1);
  if (port >= S1_OUT && port < S1_OUT + WIDTH)
    dest->fn.write_1 = NULL;
  else
    {
      dest->fn.write_1 = space_1_write_fn;
      dest->mi = mi_;
      dest->data = port;
    }
}

/* Read bit port value for display (unlike port_1_read_t may have no side
   effects);
   returns: port value */
static int
space_1_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;
  
  assert ((size_t)data < WIDTH);
  mi = mi_;
  return (mi->val & (1 << data)) != 0;
}

/* Save port_fn for PORT displaying to DEST */
static void 
space_1_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_S1);
  if (port >= S1_OUT && port < S1_OUT + WIDTH)
    {
      dest->fn.display_1 = space_1_display_fn;
      dest->mi = mi_;
      dest->data = port - S1_OUT;
    }
  else
    dest->fn.display_1 = NULL;
}

/* Set peer read_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int 
space_1_set_read_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;

  assert (port < NUM_S1);
  mi = mi_;
  if ((size_t)(port - S1_IN) < WIDTH)
    pf = mi->read + port - S1_IN;
  else
    {
      switch (port)
	{
	case S1_E:
	  pf = &mi->e_read;
	  break;
	  
	case S1_OE:
	  pf = &mi->oe_read;
	  break;
	  
	default:
	  return ENOTSUP;
	}
    }
  if (pf->fn.read_1 != NULL && src->fn.read_1 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

/* Set peer write_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int 
space_1_set_write_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;

  assert (port < NUM_S1);
  mi = mi_;
  if (port >= S1_OUT && port < S1_OUT + WIDTH)
    {
      struct port_fn *pf;

      pf = mi->write + (port - S1_OUT);
      if (pf->fn.write_1 != NULL && src->fn.write_1 != NULL)
	return EBUSY;
      *pf = *src;
      return 0;
    }
  return ENOTSUP;
}

static const struct space byte_latch_space_1 =
{
  .get_size = space_1_get_size, .get_fn =
  {
    [PORT_TYPE_READ] = space_1_get_read_fn,
    [PORT_TYPE_WRITE] = space_1_get_write_fn,
    [PORT_TYPE_DISPLAY] = space_1_get_display_fn, [PORT_TYPE_MODIFY] = NULL
  },
  .set_fn =
  {
    [PORT_TYPE_READ] = space_1_set_read_fn,
    [PORT_TYPE_WRITE] = space_1_set_write_fn, [PORT_TYPE_DISPLAY] = NULL,
    [PORT_TYPE_MODIFY] = NULL
  }
};

struct cereal_module byte_latch_module =
{
  .next = NULL, .pprev = NULL, .name = "byte_latch", .compat_names = NULL,
  .xml = NULL, .handle = NULL, .mi_new = byte_latch_mi_new,
  .mi_delete = byte_latch_mi_delete, .set_option = generic_module_set_option,
  .get_option = generic_module_get_option,
  .save_setup = generic_module_save_setup,
  .load_setup = generic_module_load_setup, .save_state = byte_latch_save_state,
  .load_state = byte_latch_load_state, .w =
  {
    [WIDTH_1] =
    { .spaces = &byte_latch_space_1, .num_spaces = 1, .xml_spaces = NULL },
    [WIDTH_8] = { .spaces = NULL, .num_spaces = 0, .xml_spaces = NULL },
    [WIDTH_16] = { .spaces = NULL, .num_spaces = 0, .xml_spaces = NULL }
  }
};
