/* A simple module providing a byte that can be read or written
   Copyright (C) 2002, 2004, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>

#include <cereal.h>
#include <lib.h>
#include <module.h>

/* One byte_value instance state */
struct mi
{
  uint8_t val;
  struct port_fn write;
};

/* Create a new module instance;
   returns: module instance pointer if OK, NULL on error */
static void *
byte_value_mi_new (void)
{
  struct mi *mi;

  mi = xmalloc (sizeof (*mi));
  mi->val = 0;
  mi->write.fn.write_8 = NULL;
  return mi;
}

/* Delete MI_ (may be in any emulation state) */
static void
byte_value_mi_delete (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  free (mi);
}

/* Our XML namespaces */
static const xmlChar byte_value_xml_ns[] =
"http://www.fw.cz/mitr/cereal/modules/byte_value";
static const xmlChar byte_value_xml_old_ns[] =
"http://www.fw.cz/mitr/cereal/modules/byte_constant";

/* Save state of MI_ to NODE */
static void
byte_value_save_state (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  char buf[32];

  mi = mi_;
  ns = xmlNewNs (node, byte_value_xml_ns, (const xmlChar *)"byte_value");
  sprintf (buf, "0x%02X", mi->val);
  xmlNewTextChild (node, ns, (const xmlChar *)"value", (xmlChar *)buf);
}

/* Load state of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
byte_value_load_state (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  int res;

  mi = mi_;
  ns = xmlSearchNsByHref (doc, node, byte_value_xml_ns);
  if (ns == NULL)
    ns = xmlSearchNsByHref (doc, node, byte_value_xml_old_ns);
  res = 0;
  if (ns == NULL)
    goto err;
  for (node = node->children; node != NULL; node = node->next)
    {
      unsigned long val;

      if (node->type != XML_ELEMENT_NODE || node->ns != ns)
	continue;
      if (strcmp ((const char *)node->name, "value") != 0)
	continue;
      res = get_xml_num (doc, node, 0x100, &val);
      if (res != 0)
	goto err;
      mi->val = val;
    }
 err:
  return res;
}

/* Get number of ports of this space in MI_ */
static size_t
space_get_size (void *mi_)
{
  (void)mi_;
  return 1;
}

/* Read byte port value (for emulation or display; may have no side effects);
   returns: port value */
static uint8_t
space_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;
  
  (void)data;
  mi = mi_;
  return mi->val;
}

/* Save port_fn for PORT reading to DEST */
static void
space_get_read_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port == 0);
  dest->fn.read_8 = space_display_fn;
  dest->mi = mi_;
}

/* Save port_fn for PORT displaying to DEST */
static void
space_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port == 0);
  dest->fn.display_8 = space_display_fn;
  dest->mi = mi_;
}

/* Modify byte port value by user */
static void
space_modify_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  (void)data;
  mi = mi_;
  mi->val = val;
  if (mi->write.fn.write_8 != NULL)
    mi->write.fn.write_8 (mi->write.mi, mi->write.data, val);
}

/* Save port_fn for PORT writing to DEST */
static void
space_get_write_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port == 0);
  dest->fn.write_8 = space_modify_fn;
  dest->mi = mi_;
}

/* Save port_fn for PORT modification to DEST */
static void
space_get_modify_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port == 0);
  dest->fn.modify_8 = space_modify_fn;
  dest->mi = mi_;
}

/* Set peer write_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int
space_set_write_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;

  mi = mi_;
  assert (port == 0);
  if (mi->write.fn.write_8 != NULL && src->fn.read_8 != NULL)
    return EBUSY;
  mi->write = *src;
  return 0;
}

static const struct space byte_value_space =
{
  .get_size = space_get_size, .get_fn =
  {
    [PORT_TYPE_READ] = space_get_read_fn,
    [PORT_TYPE_WRITE] = space_get_write_fn,
    [PORT_TYPE_DISPLAY] = space_get_display_fn,
    [PORT_TYPE_MODIFY] = space_get_modify_fn,
  },
  .set_fn =
  {
    [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = space_set_write_fn,
    [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
  }
};

static const char *const byte_value_compat_names[] =
  { "byte_constant", "byte_sink", NULL };

struct cereal_module byte_value_module =
{
  .next = NULL, .pprev = NULL, .name = "byte_value",
  .compat_names = byte_value_compat_names, .xml = NULL, .handle = NULL,
  .mi_new = byte_value_mi_new, .mi_delete = byte_value_mi_delete,
  .set_option = generic_module_set_option,
  .get_option = generic_module_get_option,
  .save_setup = generic_module_save_setup,
  .load_setup = generic_module_load_setup,
  .save_state = byte_value_save_state, .load_state = byte_value_load_state,
  .w =
  {
    [WIDTH_1] = { .spaces = NULL, .num_spaces = 0, .xml_spaces = NULL },
    [WIDTH_8] =
    { .spaces = &byte_value_space, .num_spaces = 1, .xml_spaces = NULL },
    [WIDTH_16] = { .spaces = NULL, .num_spaces = 0, .xml_spaces = NULL }
  }
};
