/* A 2^n * 8 memory chip.  Most of the handling is guesswork.
   Copyright (C) 2002, 2004, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>

#include <cereal.h>
#include <lib.h>
#include <module.h>

enum
{
  DEFAULT_VALUE = 0x00, /* Value memory is filled with until modified */
  DATA_BITS	= 8,
  ADDR_BITS	= 16
};

/* Pin space addresses */
enum
{
  S1_DATA	= 0,
  S1_ADDR	= S1_DATA + DATA_BITS,
  S1_CE		= S1_ADDR + ADDR_BITS,
  S1_OE,
  S1_WE,
  NUM_S1
};

struct mi
{
  size_t data_size;
  uint8_t *data;
  uint8_t latch;
  struct port_fn data_read[DATA_BITS], data_write[DATA_BITS], addr[ADDR_BITS];
  struct port_fn ce, oe, we;
};

/* Create a new module instance;
   returns: module instance pointer if OK, NULL on error */
static void *
memory_mi_new (void)
{
  struct mi *mi;
  struct port_fn *pf;

  mi = xmalloc (sizeof (*mi));
  mi->data_size = 1;
  mi->data = xmalloc (mi->data_size);
  memset (mi->data, DEFAULT_VALUE, mi->data_size);
  mi->latch = 0xFF;
  for (pf = mi->data_read; pf < mi->data_read + ARRAY_SIZE (mi->data_read);
       pf++)
    pf->fn.read_1 = NULL;
  for (pf = mi->data_write; pf < mi->data_write + ARRAY_SIZE (mi->data_write);
       pf++)
    pf->fn.write_1 = NULL;
  for (pf = mi->addr; pf < mi->addr + ARRAY_SIZE (mi->addr); pf++)
    pf->fn.read_1 = NULL;
  mi->ce.fn.read_1 = NULL;
  mi->oe.fn.read_1 = NULL;
  mi->we.fn.read_1 = NULL;
  return mi;
}

/* Delete MI_ (may be in any emulation state) */
static void 
memory_mi_delete (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  free (mi->data); /* free () checks for NULL anyway */
  free (mi);
}

/* Set option NAME of MI_ to VALUE;
   returns: 0 if OK, -1 on error */
static int 
memory_set_option (void *mi_, const char *name, const char *value)
{
  struct mi *mi;

  mi = mi_;
  if (strcmp (name, "data_size") == 0)
    {
      unsigned long val;

      if (get_num (value, 0, &val) != 0)
	return -1;
      if (val == 0 || val > 0x10000 || (val & (val - 1)) != 0)
	{
	  error (0, "option `data_size' argument must be a power of two less "
		 "or equal to 65536");
	  return -1;
	}
      mi->data = xrealloc (mi->data, val);
      if (mi->data_size < val)
	memset (mi->data + mi->data_size, DEFAULT_VALUE, val - mi->data_size);
      mi->data_size = val;
      return 0;
    }
  if (strcmp (name, "load_hex") == 0)
    return load_hex_file (value, mi->data, mi->data_size);
  error (0, "unknown option `%s'", name);
  return -1;
}

/* Get value of option NAME of MI_;
   returns: string for free () or NULL on error */
static char *
memory_get_option (void *mi_, const char *name)
{
  struct mi *mi;

  mi = mi_;
  if (strcmp (name, "data_size") == 0)
    {
      char *res;

      assert (sizeof (mi->data_size) * CHAR_BIT <= 64);
      res = xmalloc (32); /* Overkill */
      sprintf (res, "%zu", mi->data_size);
      return res;
    }
  else if (strcmp (name, "load_hex") == 0)
    {
      error (0, "option `%s' can't be read", name);
      return NULL;
    }
  error (0, "unknown option `%s'", name);
  return NULL;
}

/* Our XML namespace */
static const xmlChar memory_xml_ns[] = 
"http://www.fw.cz/mitr/cereal/modules/memory";

/* Save setup (configuration options) of MI_ to NODE */
static void
memory_save_setup (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  char buf[32];

  mi = mi_;
  ns = xmlNewNs (node, memory_xml_ns, (const xmlChar *)"mem");
  assert (sizeof (mi->data_size) * CHAR_BIT <= 64);
  sprintf (buf, "%zu", mi->data_size);
  xmlNewTextChild (node, ns, (const xmlChar *)"data_size", (xmlChar *)buf);
}

/* Load setup (configuration options) of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
memory_load_setup (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  return generic_module_load_setup_options (mi_, doc, node, &memory_module,
					    memory_xml_ns);
}

/* Save state of MI_ to NODE */
static void
memory_save_state (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  char buf[32];

  mi = mi_;
  ns = xmlNewNs (node, memory_xml_ns, (const xmlChar *)"mem");
  sprintf (buf, "0x%02X", mi->latch);
  xmlNewTextChild (node, ns, (const xmlChar *)"latch", (xmlChar *)buf);
  save_hex_xml (node, ns, "data", mi->data, mi->data_size);
}

/* Load state of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
memory_load_state (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  int res;

  mi = mi_;
  ns = xmlSearchNsByHref (doc, node, memory_xml_ns);
  res = 0;
  if (ns == NULL)
    goto err;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE || node->ns != ns)
	continue;
      if (strcmp ((const char *)node->name, "latch") == 0)
	{
	  unsigned long val;

	  res = get_xml_num (doc, node, 0x100, &val);
	  if (res != 0)
	    goto err;
	  mi->latch = val;
	}
      else if (strcmp ((const char *)node->name, "data") == 0)
	{
	  res = load_hex_xml (doc, node, mi->data, mi->data_size);
	  if (res != 0)
	    goto err;
	}
    }
 err:
  return res;
}

/* Get number of ports of this space in MI_ */
static size_t 
pin_get_size (void *mi_)
{
  (void)mi_;
  return NUM_S1;
}

/* Read bit port value (for emulation or display, may have no side effects);
   returns: port value */
static int
pin_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  mi = mi_;
  assert ((size_t)data < DATA_BITS);
  return (mi->latch & (1 << data)) != 0;
}

/* Save port_fn for PORT reading to DEST */
static void 
pin_get_read_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_S1);
  if ((size_t)(port - S1_DATA) < DATA_BITS)
    {
      dest->fn.read_1 = pin_display_fn;
      dest->mi = mi_;
      dest->data = port - S1_DATA;
    }
  else
    dest->fn.read_1 = NULL;
}

/* Read memory address in MI */
static uint16_t
read_addr (struct mi *mi)
{
  const struct port_fn *pf;
  uint16_t val;
  unsigned long mask; /* Need to represent at least 65536 */
  int warn;
  
  val = 0;
  warn = 0;
  pf = mi->addr;
  for (mask = 0x01; mask < mi->data_size; mask <<= 1)
    {
      if (pf->fn.read_1 == NULL)
	{
	  val |= mask; /* Just guessing a bogus value */
	  warn = 1;
	}
      else if (pf->fn.read_1 (pf->mi, pf->data) != 0)
	val |= mask;
      pf++;
    }
  if (warn != 0)
    {
      error (0, "needed address lines not connected");
      emulation_result |= ER_WARNING;
    }
  return val;
}

/* Write VAL to bit port in MI_ */
static void
pin_write_fn (void *mi_, intptr_t data, int val)
{
  struct mi *mi;
  
  assert ((size_t)data < NUM_S1);
  mi = mi_;
  if ((data >= S1_DATA && data < S1_DATA + DATA_BITS)
      || (data >= S1_ADDR && data < S1_ADDR + ADDR_BITS))
    return; /* We will poll when necessary */
  switch (data)
    {
    case S1_CE: case S1_OE:
      if (val != 0)
	{
	  const struct port_fn *pf;

	  for (pf = mi->data_write;
	       pf < mi->data_write + ARRAY_SIZE (mi->data_write); pf++)
	    {
	      if (pf->fn.write_1 != NULL)
		pf->fn.write_1 (pf->mi, pf->data, 1);
	    }
	  mi->latch = 0xFF;
	}
      else if (data == S1_OE
	       && (mi->ce.fn.read_1 == NULL
		   || mi->ce.fn.read_1 (mi->ce.mi, mi->ce.data) == 0))
	{
	  const struct port_fn *pf;
	  uint8_t mask;
	  	  
	  mi->latch = mi->data[read_addr (mi)];
	  mask = 0x01;
	  for (pf = mi->data_write;
	       pf < mi->data_write + ARRAY_SIZE (mi->data_write); pf++)
	    {
	      if (pf->fn.write_1 != NULL)
		pf->fn.write_1 (pf->mi, pf->data, mi->latch & mask);
	      mask <<= 1;
	    }
	}
      break;

    case S1_WE:
      if (val == 0 && (mi->ce.fn.read_1 == NULL
		       || mi->ce.fn.read_1 (mi->ce.mi, mi->ce.data) == 0))
	{
	  const struct port_fn *pf;
	  uint8_t v;

	  v = 0;
	  pf = mi->data_read + ARRAY_SIZE (mi->data_read) - 1;
	  do
	    {
	      v <<= 1;
	      if (pf->fn.read_1 != NULL)
		v |= pf->fn.read_1 (pf->mi, pf->data) != 0;
	      else
		v |= 1; /* Just guessing a bogus value */
	    }
	  while (pf-- != mi->data_read);
	  mi->data[read_addr (mi)] = v;
	}
      break;

    default:
      assert ("BUG" == NULL);
      return;
    }
}

/* Save port_fn for PORT writing to DEST */
static void 
pin_get_write_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_S1);
  dest->fn.write_1 = pin_write_fn;
  dest->mi = mi_;
  dest->data = port;
}

/* Save port_fn for PORT displaying to DEST */
static void 
pin_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_S1);
  if ((size_t)(port - S1_DATA) < DATA_BITS)
    {
      dest->fn.read_1 = pin_display_fn;
      dest->mi = mi_;
      dest->data = port - S1_DATA;
    }
  else
    dest->fn.read_1 = NULL;
}

/* Set peer read_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int 
pin_set_read_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;

  assert (port < NUM_S1);
  mi = mi_;
  if ((size_t)(port - S1_DATA) < DATA_BITS)
    pf = mi->data_read + (port - S1_DATA);
  else if (port >= S1_ADDR && port < S1_ADDR + ADDR_BITS)
    pf = mi->addr + (port - S1_ADDR);
  else
    {
      switch (port)
	{
	case S1_CE:
	  pf = &mi->ce;
	  break;

	case S1_OE:
	  pf = &mi->oe;
	  break;
	  
	case S1_WE:
	  pf = &mi->we;
	  break;

	default:
	  assert ("BUG" == NULL);
	  return ENOTSUP;
	}
    }
  if (pf->fn.read_1 != NULL && src->fn.read_1 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

/* Set peer write_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int 
pin_set_write_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;

  assert (port < NUM_S1);
  mi = mi_;
  if ((size_t)(port - S1_DATA) < DATA_BITS)
    pf = mi->data_write + (port - S1_DATA);
  else
    return ENOTSUP;
  if (pf->fn.write_1 != NULL && src->fn.write_1 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

static const struct space pin_space =
{
  .get_size = pin_get_size, .get_fn =
  {
    [PORT_TYPE_READ] = pin_get_read_fn, [PORT_TYPE_WRITE] = pin_get_write_fn,
    [PORT_TYPE_DISPLAY] = pin_get_display_fn, [PORT_TYPE_MODIFY] = NULL
  },
  .set_fn =
  {
    [PORT_TYPE_READ] = pin_set_read_fn, [PORT_TYPE_WRITE] = pin_set_write_fn,
    [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
  }
};

/* Get number of ports of this space in MI_ */
static size_t 
data_get_size (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  return mi->data_size;
}

/* Read byte port value for display (unlike port_8_read_t may have no side
   effects);
   returns: port value */
static uint8_t 
data_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;
  
  mi = mi_;
  assert ((size_t)data < mi->data_size);
  return mi->data[data];
}

/* Save port_fn for PORT displaying to DEST */
static void
data_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < mi->data_size);
  dest->fn.display_8 = data_display_fn;
  dest->mi = mi;
  dest->data = port;
}

/* Modify byte port value by user */
static void
data_modify_fn (void *mi_, intptr_t data, uint8_t val)
{
  struct mi *mi;

  mi = mi_;
  assert ((size_t)data < mi->data_size);
  mi->data[data] = val;
}

/* Save port_fn for PORT modification to DEST */
static void
data_get_modify_fn (void *mi_, size_t port, struct port_fn *dest)
{
  struct mi *mi;

  mi = mi_;
  assert (port < mi->data_size);
  dest->fn.modify_8 = data_modify_fn;
  dest->mi = mi;
  dest->data = port;
}

static const struct space data_space =
{
  .get_size = data_get_size, .get_fn =
  {
    [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
    [PORT_TYPE_DISPLAY] = data_get_display_fn,
    [PORT_TYPE_MODIFY] = data_get_modify_fn
  },
  .set_fn =
  {
    [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = NULL,
    [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
  }
};

struct cereal_module memory_module =
{
  .next = NULL, .pprev = NULL, .name = "memory", .compat_names = NULL,
  .xml = NULL, .handle = NULL, .mi_new = memory_mi_new,
  .mi_delete = memory_mi_delete, .set_option = memory_set_option,
  .get_option = memory_get_option, .save_setup = memory_save_setup,
  .load_setup = memory_load_setup, .save_state = memory_save_state,
  .load_state = memory_load_state, .w =
  {
    [WIDTH_1] = { .spaces = &pin_space, .num_spaces = 1, .xml_spaces = NULL },
    [WIDTH_8] = { .spaces = &data_space, .num_spaces = 1, .xml_spaces = NULL },
    [WIDTH_16] = { .spaces = NULL, .num_spaces = 0, .xml_spaces = NULL }
  }
};
