/* A very simple UART modeled after the 8051 one
   Copyright (C) 2002, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>

#include <cereal.h>
#include <lib.h>
#include <module.h>

/* Default values */
enum
{
  DEFAULT_BAUD_RATE = 9600,
  DEFAULT_DATA_BITS = 8
};

/* Bit space addresses */
enum
{
  BIT_RXD,
  BIT_TXD,
  BIT_RX_DONE,
  BIT_TX_DONE,
  NUM_BIT
};

/* Data space addresses */
enum
{
  DATA_RX_DATA,
  DATA_TX_DATA,
  NUM_DATA
};

struct mi
{
  uint16_t rx_buf, tx_buf; /* Being sent / received */
  uint16_t rx_data; /* Last data received */
  unsigned rx_step, tx_step;
  unsigned rx_status, flags; /* See RX_*, MI_* */
  unsigned long baud_rate;
  unsigned data_bits;
  struct schedule_event *event;
  struct port_fn rxd_fn, txd_fn, rx_done_fn, tx_done_fn;
  struct port_fn rx_data_fn;
  double step_time; /* = 1 / (16 * baud_rate) */
};

/* Update also rx_status_text! */
enum
{
  RX_START_0, /* Looking for start bit, read 0 last time */
  RX_START_1, /* Looking for start bit, read 1 last time */
  /* Keep these three values consecutive! */
  RX_RECV_0,  /* Receiving data, seen only zeroes for this bit */
  RX_RECV_1,  /* Receiving data, seen single one for this bit */
  RX_RECV_2   /* Receiving data, seen two ones for this bit */
};

enum
{
  MI_TXD	= 0x01,
  MI_RX_DONE	= 0x02,
  MI_TX_DONE	= 0x04,
  MI_SENDING	= 0x08
};

/* Read current value of RXD in MI */
static int
read_rxd (struct mi *mi)
{
  if (mi->rxd_fn.fn.read_1 != NULL)
    return mi->rxd_fn.fn.read_1 (mi->rxd_fn.mi, mi->rxd_fn.data);
  else
    {
      error (0, "RXD pin not connected for reading");
      emulation_result |= ER_WARNING;
      return 1; /* Just guessing a bogus value */
    }
}

/* Emulate next UART event of MI */
static void
uart_event (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  if ((mi->flags & MI_SENDING) != 0 && mi->tx_step == 0)
    {
      mi->tx_buf >>= 1;
      if (mi->tx_buf == 0x01)
	{
	  mi->flags = (mi->flags | MI_TXD | MI_TX_DONE) & ~MI_SENDING;
	  if (mi->tx_done_fn.fn.write_1 != NULL)
	    mi->tx_done_fn.fn.write_1 (mi->tx_done_fn.mi, mi->tx_done_fn.data,
				       1);
	}
      else
	{
	  if ((mi->tx_buf & 0x01) == 0)
	    mi->flags &= ~MI_TXD;
	  else
	    mi->flags |= MI_TXD;
	}
      if (mi->txd_fn.fn.write_1 != NULL)
	mi->txd_fn.fn.write_1 (mi->txd_fn.mi, mi->txd_fn.data,
			       (mi->tx_buf & 0x01) != 0);
    }
  mi->tx_step = (mi->tx_step + 1) & 0x0F;
  switch (mi->rx_status)
    {
    case RX_START_0:
      if (read_rxd (mi) != 0)
	mi->rx_status = RX_START_1;
      break;

    case RX_START_1:
      if (read_rxd (mi) == 0)
	{
	  mi->rx_buf = (1 << (mi->data_bits + 1)) - 1;
	  mi->rx_step = 0;
	  mi->rx_status = RX_RECV_0;
	}
      break;

    case RX_RECV_0: case RX_RECV_1: case RX_RECV_2:
      {
	int cnt, now;

	if (mi->rx_step < 7 || mi->rx_step > 9)
	  break;
	assert (RX_RECV_1 == RX_RECV_0 + 1 && RX_RECV_2 == RX_RECV_1 + 1);
	cnt = mi->rx_status - RX_RECV_0;
	now = read_rxd (mi);
	if (now != 0)
	  cnt++;
	if (mi->rx_step != 9)
	  {
	    mi->rx_status = RX_RECV_0 + cnt;
	    break;
	  }
	if (cnt > 1) /* Received an one */
	  mi->rx_buf |= 1 << (mi->data_bits + 1);
	if (mi->rx_buf == (1 << (mi->data_bits + 2)) - 1)
	  /* Start bit was received as 1 */
	  mi->rx_status = now == 0 ? RX_START_0 : RX_START_1;
	else if ((mi->rx_buf & 0x01) != 0)
	  {
	    mi->rx_buf >>= 1;
	    mi->rx_status = RX_RECV_0;
	  }
	else /* All bits read */
	  {
	    if ((mi->flags & MI_RX_DONE) == 0 &&
		(mi->rx_buf & (1 << (mi->data_bits + 1))) != 0)
	      {
		mi->rx_data = (mi->rx_buf >> 1) & ((1 << mi->data_bits) - 1);
		if (mi->rx_data_fn.fn.write_16 != NULL)
		  mi->rx_data_fn.fn.write_16 (mi->rx_data_fn.mi,
					      mi->rx_data_fn.data,
					      mi->rx_data);
		mi->flags |= MI_RX_DONE;
		if (mi->rx_done_fn.fn.write_1 != NULL)
		  mi->rx_done_fn.fn.write_1 (mi->rx_done_fn.mi,
					     mi->rx_done_fn.data, 1);
	      }
	    mi->rx_status = now == 0 ? RX_START_0 : RX_START_1;
	  }
	break;
      }
    }
  mi->rx_step = (mi->rx_step + 1) & 0x0F;
  schedule_add (mi->event, mi->step_time, uart_event);
}

/* Create a new module instance;
   returns: module instance pointer if OK, NULL on error */
static void *
uart_mi_new (void)
{
  struct mi *mi;

  mi = xmalloc (sizeof (*mi));
  mi->rx_buf = 0;
  mi->tx_buf = 0;
  mi->rx_data = 0;
  mi->rx_step = 0;
  mi->tx_step = 0;
  mi->rx_status = RX_START_0; /* We want to see 1 and 0 */
  mi->flags = 0;
  mi->baud_rate = DEFAULT_BAUD_RATE;
  mi->data_bits = DEFAULT_DATA_BITS;
  mi->event = schedule_new (mi);
  mi->rxd_fn.fn.read_1 = NULL;
  mi->txd_fn.fn.write_1 = NULL;
  mi->rx_done_fn.fn.write_1 = NULL;
  mi->tx_done_fn.fn.write_1 = NULL;
  mi->rx_data_fn.fn.write_16 = NULL;
  mi->step_time = (1.0 / 16.0) / DEFAULT_BAUD_RATE;
  schedule_add (mi->event, 0, uart_event);
  return mi;
}

/* Delete MI_ (may be in any emulation state) */
static void
uart_mi_delete (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  schedule_delete (mi->event);
  free (mi);
}

/* Set option NAME of MI_ to VALUE;
   returns: 0 if OK, -1 on error */
static int
uart_set_option (void *mi_, const char *name, const char *value)
{
  struct mi *mi;

  mi = mi_;
  if (strcmp (name, "baud_rate") == 0)
    {
      unsigned long val;

      if (get_num (value, 0, &val) != 0)
	return -1;
      if (val == 0)
	{
	  error (0, "option `baud rate' argument cannot be 0");
	  return -1;
	}
      mi->baud_rate = val;
      mi->step_time = (1.0 / 16.0) / val;
      return 0;
    }
  if (strcmp (name, "data_bits") == 0)
    {
      unsigned long val;

      if (get_num (value, 0, &val) != 0)
	return -1;
      if (val == 0 || val > 13)
	{
	  error (0, "option `data_bits' argument must be non-zero and less or "
		 "equal to 13");
	  return -1;
	}
      mi->data_bits = val;
      return 0;
    }
  error (0, "unknown option `%s'", name);
  return -1;
}

/* Get value of option NAME of MI_;
   returns: string for free () or NULL on error */
static char *
uart_get_option (void *mi_, const char *name)
{
  struct mi *mi;

  mi = mi_;
  if (strcmp (name, "baud_rate") == 0)
    {
      char *res;

      assert (sizeof (mi->baud_rate) * CHAR_BIT <= 64);
      res = xmalloc (32); /* Overkill */
      sprintf (res, "%lu", mi->baud_rate);
      return res;
    }
  else if (strcmp (name, "data_bits") == 0)
    {
      char *res;

      assert (sizeof (mi->data_bits) * CHAR_BIT <= 64);
      res = xmalloc (32); /* Overkill */
      sprintf (res, "%u", mi->data_bits);
      return res;
    }
  error (0, "unknown option `%s'", name);
  return NULL;
}

/* Our XML namespace */
static const xmlChar uart_xml_ns[]
= "http://www.fw.cz/mitr/cereal/modules/uart";

/* Save setup (configuration options) of MI_ to NODE */
static void
uart_save_setup (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  char buf[32];

  mi = mi_;
  ns = xmlNewNs (node, uart_xml_ns, (const xmlChar *)"uart");
  assert (sizeof (mi->baud_rate) * CHAR_BIT <= 64);
  sprintf (buf, "%lu", mi->baud_rate);
  xmlNewTextChild (node, ns, (const xmlChar *)"baud_rate", (xmlChar *)buf);
  assert (sizeof (mi->data_bits) * CHAR_BIT <= 64);
  sprintf (buf, "%u", mi->data_bits);
  xmlNewTextChild (node, ns, (const xmlChar *)"data_bits", (xmlChar *)buf);
}

/* Load setup (configuration options) of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
uart_load_setup (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  return generic_module_load_setup_options (mi_, doc, node, &uart_module,
					    uart_xml_ns);
}

/* Values saved for rx_status */
static const char *const rx_status_text[] =
{
  [RX_START_0] = "RX_START_0", [RX_START_1] = "RX_START_1",
  [RX_RECV_0] = "RX_RECV_0", [RX_RECV_1] = "RX_RECV_1",
  [RX_RECV_2] = "RX_RECV_2"
};

/* Event handling function list */
static const struct schedule_fn_desc uart_event_list[] =
  { { .fn = uart_event, .name = "uart" }, { .fn = NULL, .name = NULL } };

/* Save state of MI_ to NODE */
static void
uart_save_state (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  char buf[32];

  mi = mi_;
  ns = xmlNewNs (node, uart_xml_ns, (const xmlChar *)"uart");
  sprintf (buf, "0x%04X", mi->rx_buf);
  xmlNewTextChild (node, ns, (const xmlChar *)"rx_buf", (xmlChar *)buf);
  sprintf (buf, "0x%04X", mi->tx_buf);
  xmlNewTextChild (node, ns, (const xmlChar *)"tx_buf", (xmlChar *)buf);
  sprintf (buf, "0x%04X", mi->rx_data);
  xmlNewTextChild (node, ns, (const xmlChar *)"rx_data", (xmlChar *)buf);
  assert ((mi->rx_step & ~0x0Fu) == 0);
  sprintf (buf, "%u", mi->rx_step);
  xmlNewTextChild (node, ns, (const xmlChar *)"rx_step", (xmlChar *)buf);
  assert ((mi->tx_step & ~0x0Fu) == 0);
  sprintf (buf, "%u", mi->tx_step);
  xmlNewTextChild (node, ns, (const xmlChar *)"tx_step", (xmlChar *)buf);
  xmlNewTextChild (node, ns, (const xmlChar *)"rx_status",
		   (const xmlChar *)rx_status_text[mi->rx_status]);
  sprintf (buf, "%d", (mi->flags & MI_TXD) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"txd", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_RX_DONE) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"rx_done", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_TX_DONE) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"tx_done", (xmlChar *)buf);
  sprintf (buf, "%d", (mi->flags & MI_SENDING) != 0);
  xmlNewTextChild (node, ns, (const xmlChar *)"sending", (xmlChar *)buf);
  schedule_save_event (mi->event, node, ns, "event", uart_event_list);
}

/* Load state of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
uart_load_state (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  int res;

  mi = mi_;
  ns = xmlSearchNsByHref (doc, node, uart_xml_ns);
  res = 0;
  if (ns == NULL)
    goto err;
  for (node = node->children; node != NULL; node = node->next)
    {
      if (node->type != XML_ELEMENT_NODE || node->ns != ns)
	continue;
#define GET_VAL(NAME, LIMIT)					\
      if (strcmp ((const char *)node->name, #NAME) == 0)	\
	do							\
	  {							\
	    unsigned long val;					\
								\
	    res = get_xml_num (doc, node, LIMIT, &val);		\
	    if (res != 0)					\
	      goto err;						\
	    mi->NAME = val;					\
	  }							\
	while (0)
#define GET_FLAG(NAME, BIT)				\
      if (strcmp ((const char *)node->name, NAME) == 0)	\
	do						\
	  {						\
	    unsigned long val;				\
							\
	    res = get_xml_num (doc, node, 2, &val);	\
	    if (res != 0)				\
	      goto err;					\
	    if (val == 0)				\
	      mi->flags &= ~(BIT);			\
	    else					\
	      mi->flags |= (BIT);			\
	  }						\
	while (0)
      GET_VAL (rx_buf, 0x10000);
      else GET_VAL (tx_buf, 0x10000);
      else GET_VAL (rx_data, 0x10000);
      else GET_VAL (rx_step, 0x10);
      else GET_VAL (tx_step, 0x10);
      else if (strcmp ((const char *)node->name, "rx_status") == 0)
	{
	  res = get_xml_enum (doc, node, rx_status_text,
			      ARRAY_SIZE (rx_status_text), &mi->rx_status);
	  if (res != 0)
	    goto err;
	}
      else GET_FLAG ("txd", MI_TXD);
      else GET_FLAG ("rx_done", MI_RX_DONE);
      else GET_FLAG ("tx_done", MI_TX_DONE);
      else GET_FLAG ("sending", MI_SENDING);
      else if (strcmp ((const char *)node->name, "event") == 0)
	{
	  res = schedule_load_event (mi->event, node, uart_event_list);
	  if (res != 0)
	    goto err;
	}
#undef GET_FLAG
#undef GET_VAL
    }
 err:
  return res;
}

/* Get number of ports of this space in MI_ */
static size_t
bit_get_size (void *mi_)
{
  (void)mi_;
  return NUM_BIT;
}

/* Read bit port value for emulation or display (may have no side effects);
   returns: port value */
static int
bit_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;
  unsigned mask;

  assert ((size_t)data < NUM_BIT);
  mi = mi_;
  switch (data)
    {
    case BIT_TXD:
      mask = MI_TXD;
      break;

    case BIT_RX_DONE:
      mask = MI_RX_DONE;
      break;

    case BIT_TX_DONE:
      mask = MI_TX_DONE;
      break;

    default:
      assert ("BUG" == NULL);
      return 0;
    }
  return (mi->flags & mask) != 0;
}

/* Save port_fn for PORT reading to DEST */
static void
bit_get_read_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_BIT);
  switch (port)
    {
    case BIT_RXD:
      dest->fn.read_1 = NULL;
      break;
      
    case BIT_TXD: case BIT_RX_DONE: case BIT_TX_DONE:
      dest->fn.read_1 = bit_display_fn;
      dest->mi = mi_;
      dest->data = port;
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Write VAL to bit port in MI_ */
static void
bit_write_fn (void *mi_, intptr_t data, int val)
{
  (void)mi_;
  (void)val;
  assert (data == BIT_RXD);
  return; /* We will poll */
}

/* Modify bit port value by user */
static void
bit_modify_fn (void *mi_, intptr_t data, int val)
{
  struct mi *mi;
  unsigned mask;

  assert ((size_t)data < NUM_BIT);
  mi = mi_;
  switch (data)
    {
    case BIT_RX_DONE:
      mask = MI_RX_DONE;
      break;

    case BIT_TX_DONE:
      mask = MI_TX_DONE;
      break;

    default:
      assert ("BUG" == NULL);
      return;
    }
  if (val == 0)
    mi->flags &= ~mask;
  else
    mi->flags |= mask;
}

/* Save port_fn for PORT writing to DEST */
static void
bit_get_write_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_BIT);
  switch (port)
    {
    case BIT_RXD:
      dest->fn.write_1 = bit_write_fn;
      dest->mi = mi_;
      dest->data = port;
      break;

    case BIT_TXD:
      dest->fn.write_1 = NULL;
      break;
      
    case BIT_RX_DONE: case BIT_TX_DONE:
      dest->fn.write_1 = bit_modify_fn;
      dest->mi = mi_;
      dest->data = port;
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Save port_fn for PORT displaying to DEST */
static void
bit_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_BIT);
  switch (port)
    {
    case BIT_RXD:
      dest->fn.read_1 = NULL;
      break;
      
    case BIT_TXD: case BIT_RX_DONE: case BIT_TX_DONE:
      dest->fn.read_1 = bit_display_fn;
      dest->mi = mi_;
      dest->data = port;
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Save port_fn for PORT modification to DEST */
static void
bit_get_modify_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_BIT);
  switch (port)
    {
    case BIT_RXD: case BIT_TXD:
      dest->fn.write_1 = NULL;
      break;
      
    case BIT_RX_DONE: case BIT_TX_DONE:
      dest->fn.write_1 = bit_modify_fn;
      dest->mi = mi_;
      dest->data = port;
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Set peer read_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int
bit_set_read_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;

  assert (port < NUM_BIT);
  mi = mi_;
  switch (port)
    {
    case BIT_RXD:
      pf = &mi->rxd_fn;
      break;

    case BIT_TXD: case BIT_RX_DONE: case BIT_TX_DONE:
      return ENOTSUP;

    default:
      assert ("BUG" == NULL);
      return ENOTSUP;
    }
  if (pf->fn.read_1 != NULL && src->fn.read_1 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

/* Set peer write_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int
bit_set_write_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;

  assert (port < NUM_BIT);
  mi = mi_;
  switch (port)
    {
    case BIT_RXD:
      return ENOTSUP;
      
    case BIT_TXD:
      pf = &mi->txd_fn;
      break;

    case BIT_RX_DONE:
      pf = &mi->rx_done_fn;
      break;

    case BIT_TX_DONE:
      pf = &mi->tx_done_fn;
      break;

    default:
      assert ("BUG" == NULL);
      return ENOTSUP;
    }
  if (pf->fn.write_1 != NULL && src->fn.write_1 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

static const struct space bit_space =
{
  .get_size = bit_get_size, .get_fn =
  {
    [PORT_TYPE_READ] = bit_get_read_fn, [PORT_TYPE_WRITE] = bit_get_write_fn,
    [PORT_TYPE_DISPLAY] = bit_get_display_fn,
    [PORT_TYPE_MODIFY] = bit_get_modify_fn
  },
  .set_fn =
  {
    [PORT_TYPE_READ] = bit_set_read_fn, [PORT_TYPE_WRITE] = bit_set_write_fn,
    [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
  }
};

/* Get number of ports of this space in MI_ */
static size_t
data_get_size (void *mi_)
{
  (void)mi_;

  return NUM_DATA;
}

/* Read 16b port value for emulation or display (may have no side effects);
   returns: port value */
static uint16_t
data_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;

  assert (data == DATA_RX_DATA);
  mi = mi_;
  return mi->rx_data;
}

/* Save port_fn for PORT reading to DEST */
static void
data_get_read_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_DATA);
  switch (port)
    {
    case DATA_RX_DATA:
      dest->fn.read_16 = data_display_fn;
      dest->mi = mi_;
      dest->data = port;
      break;

    case DATA_TX_DATA:
      dest->fn.read_16 = NULL;
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Write VAL to 16b port in MI_ */
static void
data_write_fn (void *mi_, intptr_t data, uint16_t val)
{
  struct mi *mi;

  assert (data == DATA_TX_DATA);
  mi = mi_;
  if (val >= (1 << mi->data_bits))
    {
      val &= (1 << mi->data_bits) - 1;
      error (0, "value %u cannot fit in %u bits", val, mi->data_bits);
      emulation_result |= ER_WARNING;
    }
  /* Bit 0 blank to eliminate a special case, bit 1 = 0 is the start bit. */
  mi->tx_buf = (1 << 2) << mi->data_bits | val << 2;
  mi->flags |= MI_SENDING;
}

/* Save port_fn for PORT writing to DEST */
static void
data_get_write_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_DATA);
  switch (port)
    {
    case DATA_RX_DATA:
      dest->fn.write_16 = NULL;
      break;

    case DATA_TX_DATA:
      dest->fn.write_16 = data_write_fn;
      dest->mi = mi_;
      dest->data = port;
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Save port_fn for PORT displaying to DEST */
static void
data_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port < NUM_DATA);
  switch (port)
    {
    case DATA_RX_DATA:
      dest->fn.display_16 = data_display_fn;
      dest->mi = mi_;
      dest->data = port;
      break;

    case DATA_TX_DATA:
      dest->fn.read_16 = NULL;
      break;

    default:
      assert ("BUG" == NULL);
    }
}

/* Set peer write_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int
data_set_write_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;
  struct port_fn *pf;

  assert (port < NUM_DATA);
  mi = mi_;
  switch (port)
    {
    case DATA_RX_DATA:
      pf = &mi->rx_data_fn;
      break;

    case DATA_TX_DATA:
      return ENOTSUP;
      
    default:
      assert ("BUG" == NULL);
      return ENOTSUP;
    }
  if (pf->fn.write_16 != NULL && src->fn.write_16 != NULL)
    return EBUSY;
  *pf = *src;
  return 0;
}

static const struct space data_space =
{
  .get_size = data_get_size, .get_fn =
  {
    [PORT_TYPE_READ] = data_get_read_fn, [PORT_TYPE_WRITE] = data_get_write_fn,
    [PORT_TYPE_DISPLAY] = data_get_display_fn, [PORT_TYPE_MODIFY] = NULL
  },
  .set_fn =
  {
    [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = data_set_write_fn,
    [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
  }
};

struct cereal_module uart_module =
{
  .next = NULL, .pprev = NULL, .name = "uart", .compat_names = NULL,
  .xml = NULL, .handle = NULL, .mi_new = uart_mi_new,
  .mi_delete = uart_mi_delete, .set_option = uart_set_option,
  .get_option = uart_get_option, .save_setup = uart_save_setup,
  .load_setup = uart_load_setup, .save_state = uart_save_state,
  .load_state = uart_load_state, .w =
  {
    [WIDTH_1] = { .spaces = &bit_space, .num_spaces = 1, .xml_spaces = NULL },
    [WIDTH_8] = { .spaces = NULL, .num_spaces = 0, .xml_spaces = NULL },
    [WIDTH_16] = { .spaces = &data_space, .num_spaces = 1, .xml_spaces = NULL }
  }
};
