/* Text-mode interactive interface to cereal
   Copyright (C) 2002, 2004, 2005 Miloslav Trmac <mitr@volny.cz>

   You can redistribute and/or modify this program under the terms of the
   GNU General Public License version 2 as published by the Free Software
   Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place Suite 330, Boston, MA 02111-1307, USA. */
#include <config.h>

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <libxml/tree.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <breakpoint.h>
#include <cereal.h>
#include <expr.h>
#include <lib.h>
#include <me.h>
#include <module.h>
#include <port.h>
#include <setup.h>

/* Report printf-like ERR, append message for ERR_NO if ERR_NO != 0;
   Each frontend defines its own implementation. */
void
error (int err_no, const char *err, ...)
{
  va_list ap;

  fprintf (stderr, "%s: ", program_name);
  va_start (ap, err);
  vfprintf (stderr, err, ap);
  va_end (ap);
  if (err_no == 0)
    fputc ('\n', stderr);
  else
    fprintf (stderr, ": %s\n", strerror (err_no));
}

/* Connect NUM strings at STRS, separate them by spaces;
   returns: connected string (for free ());
   Note that splitting strings on white space and then connecting them
   is not strictly right, we don't detect errors like '  ', but we didn't
   promise to hand-hold the user using cereal_text ;-) */
static char *
str_connect (const char **strs, size_t num_strs)
{
  size_t total_len;
  const char **str;
  char *res, *dest;

  total_len = num_strs; /* ' ' or '\0' after each string */
  for (str = strs; str < strs + num_strs; str++)
    total_len += strlen (*str);
  res = xmalloc (total_len);
  dest = res;
  for (str = strs; str < strs + num_strs - 1; str++)
    {
      size_t len;
      
      len = strlen (*str);
      memcpy (dest, *str, len);
      dest += len;
      *dest++ = ' ';
    }
  strcpy (dest, *str);
  return res;
}

/* Do command with ARGC, ARGV */
static void do_command (size_t argc, const char **argv);

/* Handle 'help' command */
static void
do_help (size_t argc, const char **argv)
{
  if (argc == 1 || strcmp (argv[1], "help") == 0
      || strcmp (argv[1], "--help") == 0)
    {
      printf ("help [command]\n"
	      "Displays help for running COMMAND.  Equivalent to `COMMAND "
	      "--help'.\n");
      return;
    }
  /* Modify and redispatch.  We are normally not allowed to modify argv, but
     this is OK, because repeating the original command is equivalent to
     repeating the modified one. */
  argv[0] = argv[1];
  argv[1] = "--help";
  do_command (argc, argv);
}

/* Mapping between ER_* bits and strings */
struct result_bit
{
  int bit;
  const char *string;
};

static const struct result_bit result_table[] =
{
  { .bit = ER_HALT, .string = "halt" }, { .bit = ER_ERROR, .string = "error" },
  { .bit = ER_WARNING, .string = "warning" },
  { .bit = ER_INSN, .string = "insn" },
  { .bit = ER_BREAKPOINT, .string = "breakpoint" }
};

/* Handle 'step' command */
static void
do_step (size_t argc, const char **argv)
{
  int res, mask;

  if (argc > 1 && strcmp (argv[1], "--help") == 0)
    {
      printf ("step [error | warning | insn | breakpoint]*\n"
	      "Starts emulation.  Optional parameters define conditions when "
	      "emulation is\n"
	      "stopped.  If there is no parameter, emulation stops after next "
	      "action (which\n"
	      "may be less than one instruction).\n");
      return;
    }
  mask = ER_HALT;
  while (*++argv != NULL)
    {
      const struct result_bit *rb;

      for (rb = result_table; rb < result_table + ARRAY_SIZE (result_table);
	   rb++)
	{
	  if (strcmp (*argv, rb->string) == 0 && rb->bit != ER_HALT)
	    {
	      mask |= rb->bit;
	      goto next_cond;
	    }
	}
      error (0, "unknown stop condition `%s'", *argv);
      return;
      
    next_cond:
      ;
    }
  if (!schedule_pending ())
    {
      error (0, "nothing to emulate");
      return;
    }
  do
    {
      res = schedule_run ();
      if (breakpoint_check () != NULL)
	res |= ER_BREAKPOINT;
      if (res != 0)
	{
	  const struct result_bit *rb;
	  
	  printf ("Step results: ");
	  for (rb = result_table;
	       rb < result_table + ARRAY_SIZE (result_table); rb++)
	    {
	      if ((res & rb->bit) != 0)
		printf ("%s ", rb->string);
	    }
	  putchar ('\n');
	}
    }
  while (mask != ER_HALT && (res & mask) == 0);
}

/* Handle 'bp_del' command */
static void
do_bp_del (size_t argc, const char **argv)
{
  struct breakpoint *bp;
  unsigned long id;

  (void)argv;
  if (argc != 2 || strcmp (argv[1], "--help") == 0)
    {
      printf ("bp_del id\n"
	      "Deletes breakpoint ID.\n");
      return;
    }
  if (get_num (argv[1], 10, &id) != 0)
    return;
  for (bp = breakpoint_list; bp != NULL; bp = bp->next)
    {
      if ((uintptr_t)bp->data == id)
	{
	  breakpoint_delete (bp);
	  return;
	}
    }
  error (0, "breakpoint %lu not found", id);
}

/* Handle 'bp_list' command */
static void
do_bp_list (size_t argc, const char **argv)
{
  struct breakpoint *bp;

  (void)argv;
  if (argc != 1) /* Also when strcmp (argv[1], "--help") == 0 */
    {
      printf ("bp_list\n"
	      "Lists defined breakpoints, mark with '+' if they are "
	      "currently triggered.\n");
      return;
    }
  bp = breakpoint_list;
  if (bp == NULL)
    {
      printf ("No breakpoints defined\n");
      return;
    }
  do
    {
      printf ("%4" PRIdPTR ". %c %s\n", bp->data,
	      breakpoint_eval (bp) != 0 ? '+' : ' ', bp->expr);
      bp = bp->next;
    }
  while (bp != NULL);
}

/* Handle 'breakpoint' command */
static void
do_breakpoint (size_t argc, const char **argv)
{
  static intptr_t last_id; /* = 0; */
  struct breakpoint *bp;
  char *expr;

  if (argc < 2 || strcmp (argv[1], "--help") == 0)
    {
      printf ("breakpoint expression\n"
	      "Makes new breakpoint that triggers when EXPRESSION is "
	      "nonzero.\n");
      return;
    }
  if (last_id + 1 <= 0)
    {
      /* Not really, but the user must be mad anyway */
      error (0, "too many breakpoints");
      return;
    }
  expr = str_connect (argv + 1, argc - 1);
  bp = breakpoint_new (expr);
  free (expr);
  if (bp == NULL)
    return;
  bp->data = ++last_id;
}

/* Handle 'connect_1' command */
static void
do_connect_1 (size_t argc, const char **argv)
{
  struct connect_entry skel;

  if (argc != 4 || strcmp (argv[1], "--help") == 0)
    {
      printf ("connect_1 {read | write | display | modify} port_1 port_2\n"
	      "Connects bit port PORT_1 of specified type to PORT_2.\n");
      return;
    }
  if (port_find (WIDTH_1, argv[2], &skel.pi1) != 1
      || port_find (WIDTH_1, argv[3], &skel.pi2) != 1)
    return;
  skel.type = port_type_find (argv[1]);
  if (skel.type < 0)
    {
      error (0, "unknown port type `%s'", argv[1]);
      return;
    }
  skel.width = WIDTH_1;
  (void)port_connect (&skel);
}

/* Handle 'print' command */
static void
do_print (size_t argc, const char **argv)
{
  char *expr;
  intmax_t val;

  if (argc < 2 || strcmp (argv[1], "--help") == 0)
    {
      printf ("print expression\n"
	      "Prints current value of EXPRESSION.\n");
      return;
    }
  expr = str_connect (argv + 1, argc - 1);
  if (expr_eval (expr, &val) == 0)
    printf ("0x%" PRIXMAX "\n", val);
  free (expr);
}

/* Handle 'set' command */
static void
do_set (size_t argc, const char **argv)
{
  char *dest, *src;
  const char **p;

  for (p = argv + 1; p < argv + argc; p++)
    {
      if (strcmp (*p, "=") == 0)
	break;
    }
  if (argc < 4 || p == argv + argc || p == argv + 1 || p == argv + argc - 1
      || strcmp (argv[1], "--help") == 0)
    {
      printf ("set DEST = SRC\n"
	      "Sets port referenced in DEST so that DEST == SRC.\n"
	      "Note that the `=' operator must be separated by space on both "
	      "sides.\n");
      return;
    }
  dest = str_connect (argv + 1, p - argv - 1);
  src = str_connect (p + 1, argc - (p - argv) - 1);
  (void)expr_modify (dest, src);
  free (src);
  free (dest);
}

/* Handle 'mod_delete' command */
static void
do_mod_delete (size_t argc, const char **argv)
{
  if (argc != 2 || strcmp (argv[1], "--help") == 0)
    {
      printf ("mod_delete name\n"
	      "Deletes module NAME\n");
      return;
    }
  (void)me_delete (argv[1]);
}
 
/* Handle 'mod_new' command */
static void
do_mod_new (size_t argc, const char **argv)
{
  const struct cereal_module *module;
  
  if (argc != 3 || strcmp (argv[1], "--help") == 0)
    {
      printf ("mod_new type name\n"
	      "Creates new module NAME of type TYPE\n");
      return;
    }
  module = cereal_module_find (argv[1]);
  if (module == NULL)
    {
      error (0, "unknown module type `%s'", argv[1]);
      return;
    }
  (void)me_new (module, argv[2]);
}

/* Handle 'mod_rename' command */
static void
do_mod_rename (size_t argc, const char **argv)
{
  if (argc != 3 || strcmp (argv[1], "--help") == 0)
    {
      printf ("mod_rename old new\n"
	      "Renames module OLD to NEW\n");
      return;
    }
  (void)me_rename (argv[1], argv[2]);
}

/* Handle 'option' command */
static void
do_option (size_t argc, const char **argv)
{
  const struct me_entry *me;
  
  if ((argc != 3 && argc != 4) || strcmp (argv[1], "--help") == 0)
    {
      printf ("option module option [value]\n"
	      "Sets option OPTION in MODULE to VALUE if VALUE argument is "
	      "present, displays\n"
	      "current value of option OPTION in MODULE otherwise.\n");
      return;
    }
  me = me_get (argv[1]);
  if (me == NULL)
    return;
  if (argc == 4)
    me->module->set_option (me->mi, argv[2], argv[3]);
  else
    {
      char *res;

      res = me->module->get_option (me->mi, argv[2]);
      if (res != NULL)
	{
	  printf ("%s\n", res);
	  free (res);
	}
    }
}
     
/* Handle 'setup_load' command */
static void
do_setup_load (size_t argc, const char **argv)
{
  if (argc != 2 || strcmp (argv[1], "--help") == 0)
    {
      printf ("setup_load filename\n"
	      "Loads setup from FILENAME.\n");
      return;
    }
  setup_load (argv[1]);
}
  
/* Handle 'setup_save' command */
static void
do_setup_save (size_t argc, const char **argv)
{
  if (argc != 2 || strcmp (argv[1], "--help") == 0)
    {
      printf ("setup_save filename\n"
	      "Saves setup to FILENAME.\n");
      return;
    }
  setup_save (argv[1]);
}
  
/** Load UI state from NODE of DOC;
    returns: 0 if OK, -1 on error */
static int
state_load_ui_callback (xmlDocPtr doc, xmlNodePtr node, intptr_t data)
{
  int res;
  
  (void)data;
  (void)doc;
  res = 0;
  for (node = node->children; node != NULL; node = node->next)
    {
      xmlChar *str;
      struct breakpoint *bp;
      
      if (node->type != XML_ELEMENT_NODE
	  || strcmp ((const char *)node->name, "ui_breakpoint") != 0)
	continue;
      str = xmlGetProp (node, (const xmlChar *)"expr");
      if (str == NULL)
	continue;
      bp = breakpoint_new ((const char *)str);
      xmlFree (str);
      if (bp == NULL)
	{
	  res = -1;
	  goto err;
	}
      str = xmlGetProp (node, (const xmlChar *)"id");
      if (str != NULL)
	{
	  intptr_t val;
	  int ret;
	  
	  ret = sscanf ((const char *)str, "%" SCNdPTR, &val);
	  xmlFree (str);
	  if (ret != 1)
	    {
	      breakpoint_delete (bp);
	      res = -1;
	      goto err;
	    }
	  bp->data = val;
	}
    err:
      ;
    }
  return 0;
}

/* Handle 'state_load' command */
static void
do_state_load (size_t argc, const char **argv)
{
  if (argc != 2 || strcmp (argv[1], "--help") == 0)
    {
      printf ("state_load filename\n"
	      "Loads state from FILENAME.\n");
      return;
    }
  state_load (argv[1], generic_state_load_setup_fn, state_load_ui_callback, 0);
}

/** Save UI state under NODE;
    returns: 0 if OK, -1 on error */
static void
state_save_callback (xmlNodePtr node, intptr_t data)
{
  const struct breakpoint *bp;
  
  (void)data;
  for (bp = breakpoint_list; bp != NULL; bp = bp->next)
    {
      char buf[32];
      xmlNodePtr n;
      
      n = xmlNewChild (node, NULL, (const xmlChar *)"ui_breakpoint", NULL);
      xmlSetProp (n, (const xmlChar *)"expr", (const xmlChar *)bp->expr);
      assert (sizeof (bp->data) * CHAR_BIT <= 64);
      sprintf (buf, "%" PRIdPTR, bp->data);
      xmlSetProp (n, (const xmlChar *)"id", (xmlChar *)buf);
    }
}

/* Handle 'state_save' command */
static void
do_state_save (size_t argc, const char **argv)
{
  if (argc != 2 || strcmp (argv[1], "--help") == 0)
    {
      printf ("state_save filename\n"
	      "Saves state to FILENAME.\n");
      return;
    }
  state_save (argv[1], state_save_callback, 0);
}
  
/* Do command with ARGC, ARGV */
static void
do_command (size_t argc, const char **argv)
{
  if (strcmp (argv[0], "bp_del") == 0)
    do_bp_del (argc, argv);
  else if (strcmp (argv[0], "bp_list") == 0)
    do_bp_list (argc, argv);
  else if (strcmp (argv[0], "breakpoint") == 0)
    do_breakpoint (argc, argv);
  else if (strcmp (argv[0], "connect_1") == 0)
    do_connect_1 (argc, argv);
  else if (strcmp (argv[0], "help") == 0)
    do_help (argc, argv);
  else if (strcmp (argv[0], "mod_delete") == 0)
    do_mod_delete (argc, argv);
  else if (strcmp (argv[0], "mod_new") == 0)
    do_mod_new (argc, argv);
  else if (strcmp (argv[0], "mod_rename") == 0)
    do_mod_rename (argc, argv);
  else if (strcmp (argv[0], "option") == 0)
    do_option (argc, argv);
  else if (strcmp (argv[0], "print") == 0)
    do_print (argc, argv);
  else if (strcmp (argv[0], "set") == 0)
    do_set (argc, argv);
  else if (strcmp (argv[0], "setup_load") == 0)
    do_setup_load (argc, argv);
  else if (strcmp (argv[0], "setup_save") == 0)
    do_setup_save (argc, argv);
  else if (strcmp (argv[0], "state_load") == 0)
    do_state_load (argc, argv);
  else if (strcmp (argv[0], "state_save") == 0)
    do_state_save (argc, argv);
  else if (strcmp (argv[0], "step") == 0)
    do_step (argc, argv);
  else
    error (0, "unknown command `%s'", argv[0]);
}

/* Read a command and execute it;
   returns: 0 at EOF, 1 otherwise */
inline static int
read_command (void)
{
  static char *buffer; /* = NULL; */
  static size_t *args; /* = NULL; */
  static size_t buffer_size /* = 0 */, args_size; /* = 0; */
  static const char **argv; /* = NULL; allocated (args_size + 1) entries */
  static size_t argc; /* = 0; */
  char *buf;
  size_t *argp;
  int c;
  
  c = getchar ();
  if (c == EOF)
    return 0;
  buf = buffer;
  argp = args;
  while (c != '\n')
    {
      while (c == ' ' || c == '\t')
	c = getchar ();
      if (c == '\n' || c == EOF)
	break;
      if (argp == args + args_size)
	{
	  args = xrealloc (args, (args_size + 16) * sizeof (*args));
	  argv = xrealloc (argv, (args_size + 16 + 1) * sizeof (*argv));
	  argp = args + args_size;
	  args_size += 16;
	}
      *argp++ = buf - buffer;
      while (c != ' ' && c != '\t' && c != '\n' && c != EOF)
	{
	  if (buf == buffer + buffer_size)
	    {
	      buffer = xrealloc (buffer,
				 (buffer_size + 128) * sizeof (*buffer));
	      buf = buffer + buffer_size;
	      buffer_size += 128;
	    }
	  *buf++ = c;
	  c = getchar ();
	}
      if (buf == buffer + buffer_size)
	{
	  buffer = xrealloc (buffer,
			     (buffer_size + 128) * sizeof (*buffer));
	  buf = buffer + buffer_size;
	  buffer_size += 128;
	}
      *buf++ = 0;
    }
  if (argp != args) /* Empty line means "repeat previous command" */
    {
      size_t *src;
      const char **dst;

      dst = argv;
      for (src = args; src != argp; src++)
	*dst++ = buffer + *src;
      *dst = 0;
      argc = dst - argv;
    }
  if (argv != NULL)
    do_command (argc, argv);
  return 1;
}

/* Set program_name from information in ARGV_0 */
inline static void
set_program_name (const char *argv_0)
{
  if (argv_0 == NULL)
    argv_0 = "cereal_text";
  program_name = xstrdup (argv_0);
}

/* Print help and exit */
inline static void attribute__ ((noreturn))
arg_help (void)     
{
  printf ("Usage: %s\n"
	  "Runs the text-mode intefrace to cereal\n"
	  "Report bugs to <" PACKAGE_BUGREPORT ">.\n", program_name);
  exit (EXIT_SUCCESS);
}

/* Print version and exit */
inline static void attribute__ ((noreturn))
arg_version (void)
{
  printf ("Cereal Emulates REALity " PACKAGE_VERSION "\n"
	  "Copyright (C) 2002 Miloslav Trmac <mitr@volny.cz>\n"
	  "You may redistribute copies of cereal under the terms of the GNU "
	  "General\n"
	  "Public License, version 2.  See the file COPYING in the source "
	  "tarball.\n"
	  "cereal comes with NO WARRANTY, to the extent permitted by law.\n");
  exit (EXIT_SUCCESS);
}

/* Print usage message and exit */
inline static void attribute__ ((noreturn))
arg_usage (void)
{
  error (0, "usage: %s", program_name);
  exit (EXIT_FAILURE);
}

int
main (int argc, char *argv[])
{
#ifndef SETVBUF_REVERSED  
  setvbuf (stdout, NULL, _IOLBF, BUFSIZ);
#else
  setvbuf (stdout, _IOLBF, NULL, BUFSIZ);
#endif
  set_program_name (argv[0]);
  if (argc != 1)
    {
      if (argc != 0)
	{
	  while (*++argv != NULL)
	    {
	      if (strcmp (*argv, "--help") == 0)
		arg_help ();
	      else if (strcmp (*argv, "--version") == 0)
		arg_version ();
	    }
	}
      arg_usage ();
    }
  setup_init ();
  while (read_command ())
    ;
  setup_destroy ();
  return EXIT_SUCCESS;
}
